/* LELUYA - GOROSTAS program block */
/* This function calculates the Psi variable with the approximation of
   a point source. */

#include "GOROSTAS.h"

int _GOROSTAS_Find_Psi (void){

  int error_flag;

  int lambda_max; /* wavelength with the most grid points */

  double New_Psi;

  double x,z, xsin, zcos;
  double d; /* d = sqrt(x^2 + z^2) */

  struct _triangle_structure *triangle;

  double step, stop_step_scale=1.0E-06, stop_step, step_new;
  double reduce = 0.8;

  int dust_Psi, dust_i; 

  /* interpolation variables */
  double x1,x2,x3,y1,y2,y3;
  double c1,c2,c3;
  double denominator;

  /* integral variables */
  double *f;    
  double energy;
  int lambda_k;


  /************************************************************************/
  /* first, we need to find the closest point to the source along the line
     of total optical depth (specified by the user) */

  /* find a grid with the most grid points  (let's call it grid_max)*/
  lambda_max = 0;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++)
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > 
	grid_per_lambda[lambda_max].TOTAL_NUMBER_OF_VERTICES) 
      lambda_max = lambda_k;
 
  /* start from the point _1_ */
  x = xsin = sin(Theta_Total_Value); 
  z = zcos = cos(Theta_Total_Value); 
  d = 1.0;

  /* get the initial step size and triangle which contains the point (x,z) */
  step = _ZELENJAK_size_limit(x,z)*_MIN_VERTEX_DISTANCE_;
  start_triangle_list=grid_per_lambda[lambda_max].start_triangle;
  error_flag =  _ZELENJAK_locate_point_in_grid(x,z,&triangle);
  if (triangle!=NULL) step *= -1.0;

  /* Here we play a game: if the point (x,z) is out of the dust, increase
     the distance from the star and repeat this procedure, but if the point
     has the dust then reduce the distance and repeat the procedure. When the
     the search turns direction, reduce the step by factor "reduce". This 
     process goes on as long as the step is bigger than "stop_step" and the
     point has dust. */

  stop_step = stop_step_scale * fabs(step);
  error_flag=FALSE;

  /*while (fabs(step) > stop_step || triangle==NULL){*/
  /* error_flag==FALSE includes triangle==NULL here: */ 
  while (fabs(step) > 1.0E-10 || error_flag==FALSE){
    d = d + step;
    x = d*xsin; 
    z = d*zcos;
    start_triangle_list=grid_per_lambda[lambda_max].start_triangle;
    error_flag =  _ZELENJAK_locate_point_in_grid(x,z,&triangle);
    if (triangle==NULL) error_flag = FALSE;
    else {
      /* check if this triangle is in the empty space: error_flag = FALSE */
      error_flag = FALSE;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
	  if (triangle->v1->T_dust[dust_i]>0.0 && 
	      triangle->v2->T_dust[dust_i]>0.0 &&
	      triangle->v3->T_dust[dust_i]>0.0) {error_flag = TRUE;break;}
    }
    if ((error_flag==FALSE && step<0.0) || (error_flag==TRUE && step>0.0)) {
      step_new =  _ZELENJAK_size_limit(x,z)*_MIN_VERTEX_DISTANCE_;
      if (stop_step_scale*step_new < stop_step) 
        {step= step_new; stop_step = stop_step_scale * step;}
      step = reduce * fabs(step);
      if (error_flag==TRUE) step *= -1.0;
    }
  }

  /* now we have the point and the triangle which it belongs to. Now find what
     dust type we are talking about */
  dust_Psi = -1;
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
      if (triangle->v1->T_dust[dust_i]>0.0 && 
	  triangle->v2->T_dust[dust_i]>0.0 && 
	  triangle->v3->T_dust[dust_i]>0.0) {dust_Psi=dust_i; break;}
  }
  if (dust_Psi<0) { /* damn it! what a stupid situation. Take anything. */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (triangle->v1->T_dust[dust_i]>0.0) {dust_Psi=dust_i; break;}
  }
  if (dust_Psi<0) /* argh!! You are evil! Die! */
    { return 1;}

  /* yahoo!! get this calculation, finally */
  f = dvector(0,lambda_Max-1);

  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){

    /* find the triangle for this particular wavelength at (x,z) */
    start_triangle_list=grid_per_lambda[lambda_k].start_triangle;
    error_flag =  _ZELENJAK_locate_point_in_grid(x,z,&triangle);

    /* Good. Prepare the interpolation coefficients. */
    x1 = triangle->v1->x;
    x2 = triangle->v2->x;
    x3 = triangle->v3->x;
    y1 = triangle->v1->z;
    y2 = triangle->v2->z;
    y3 = triangle->v3->z;

    denominator = (x2-x1)*(y3-y1) - (x3-x1)*(y2-y1);
    if (denominator == 0.0) return 1;
    c1 = 1 + ( (x-x1)*(y2-y3)+(z-y1)*(x3-x2) )/denominator;
    c2 = ( (x-x1)*(y3-y1)-(z-y1)*(x3-x1) )/denominator;
    c3 = ( (z-y1)*(x2-x1)-(x-x1)*(y2-y1) )/denominator;

    energy =
    c1*grid_per_lambda[lambda_k].Utot_new[triangle->v1->identification_number]
   +c2*grid_per_lambda[lambda_k].Utot_new[triangle->v2->identification_number]
   +c3*grid_per_lambda[lambda_k].Utot_new[triangle->v3->identification_number];
    f[lambda_k] = q_abs(x,z,lambda_k,dust_Psi,&error_flag)*energy;
  }

  New_Psi = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,f);

  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    f[lambda_k] = q_abs(x,z,lambda_k,dust_Psi,&error_flag)*
          _MATH_LIB_Planck(Dust_Sublim_Temp[dust_Psi],Lambda_Grid[lambda_k]);
  }
  New_Psi=New_Psi/ _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,f);

  energy = Dust_Sublim_Temp[0] * Dust_Sublim_Temp[0];
  energy /= (Dust_Sublim_Temp[dust_Psi]*Dust_Sublim_Temp[dust_Psi]);
  energy /= d;
  energy = energy*energy;

  New_Psi *= energy;

  printf("SEARCH FOR A NEW PSI:\n");
  printf("Psi point at x=%13.11E, z=%13.11E, distance d=%E\n",x,z,d);
  printf("dust=%d : Old Psi=%E, new Psi=%E\n",dust_Psi,Psi,New_Psi);

  Psi = New_Psi;

  free_dvector(f,0,lambda_Max-1);
  return 0;
}

