/* LELUYA - GOROSTAS program block */
/* finds temperature from the radiative equilibrium at a given point, 
   given dust component, and given energy spectrum */
/* Temperature is obtained by solving: 
    _GOROSTAS_Equilibrium(T) = Function1(T) - Function2 = 0
   where: 
    Function1(T) = T^4 Int(q_abs Planck(T) dlambda)
    Function2 = T_sub[1]^4 Int(q_abs U(Y) dlambda) / Y^2 / Psi

   This is solved by the Ridder's method: _MATH_LIB_Ridder_Root().

   SEE HTML PAGES ABOUT THIS FUNCTION!!

   Errors: 
   1 = error in bracketing of the solution: fixing of T1 and T2 didn't work
   2 = can not find the solution by _MATH_LIB_Ridder_Root 
*/

#include "GOROSTAS.h"

double _GOROSTAS_Find_Temperature (double x, double z, double T1, double T2, 
                                 int dust_component, double *U, int *err){
/* x,z       = cooridnates of the point 
   T1, T2  = bracketing temperatures in the search for solution
   dust_component = index of the dust component for which this calculation
             is performed
   U       = energy spectrum 
*/
   double Function2;  /* see definition above */
   double *f;         /* used as a help during integrations over lambda */
   int    counter;    /* used as a help in the loops */
   int    error;      /* error flag */
   double f1,f2;      /* _GOROSTAS_Equilibrium function at T1 and T2 */

   *err = 0;
   /* first check if we deal with the empty space */
   if (INPUT.DUST_DENSITY_TYPE[dust_component]==99999){
     f1 = _T99999_;
     return f1;
   }

   /* find the Function2 variable */
   f=dvector(0,lambda_Max-1);
   for (counter=0; counter<lambda_Max; counter++)
               f[counter]=q_abs(x,z,counter,dust_component,&error)*U[counter];
   Function2 = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,f);
   Function2 *= (Dust_Sublim_Temp[0]*Dust_Sublim_Temp[0]*
                 Dust_Sublim_Temp[0]*Dust_Sublim_Temp[0]);
   Function2 /= ((x*x+z*z)*Psi);
   free_dvector(f,0,lambda_Max-1);

   /* check if T1 and T2 are bracketing the solution */
   if (T1 == T2) {*err = 1; return 0.0;}
   f1 = _GOROSTAS_Equilibrium(T1, x, z, dust_component, Function2);
   if (f1 == 0.0) return T1;
   f2 = _GOROSTAS_Equilibrium(T2, x, z, dust_component, Function2);
   if (f2 == 0.0) return T2;
   /* try to fix the T1 and T2 temperatures 100 times, if necessary */
   counter = 0;
   while (f1*f2>0 && counter<100) {
     if (fabs(f1)<fabs(f2)){
       /* factor 1.6 is taken from 1D Dusty as a successful choice */
       T1 = T1 + 1.6*(T1-T2);
       f1 = _GOROSTAS_Equilibrium(T1, x, z, dust_component, Function2);
       if (f1 == 0.0) return T1;
     }
     else {
       /* factor 1.6 is taken from 1D Dusty as a successful choice */
       T2 = T2 + 1.6*(T2-T1);
       f2 = _GOROSTAS_Equilibrium(T2, x, z, dust_component, Function2);
       if (f2 == 0.0) return T2;
     }
     counter++;
   }
   if (counter==100) {*err=1; return 0.0;}
   
   /* finally, find the solution */
   f1 = _MATH_LIB_Ridder_Root(&_GOROSTAS_Equilibrium, x, z, dust_component,
                        Function2, T1, T2, _Temperature_Accuracy, &error);
   if (error) {*err=2; return 0.0;}
   return f1;

} /* end of the function _GOROSTAS_Find_Temperature */
