/* LELUYA - GOROSTAS program block */
/* This function calculates the intensity around a given point and saves
   that data into four files: stellar, dust emission, dust scattering, total */

/* NOTE: "Intensity" means flux, surface brightness. */ 

#include "GOROSTAS.h"

int _GOROSTAS_Intensity_Around_Point (double point_x, double point_z) {

  struct _3DVector rho, rho_unit; /* position of a point on a shell (and the
				     unit vector of this position) */
  struct _vertex_structure *rho_vertex; /* rho vector transformed into 
                         a vertex because of _GOROSTAS_create_integral_directions
                         function */
  struct _3DVector V0,ray_direction; /* direction of the intensity integral*/

  int angular_rays_counter; /* counter of the angluar rays at a point */

  double angular_flux_weight; /* weight in the angular integral */
  int lambda_max_N; /* wavelength with the largest number of vertices */
  int lambda_k;       /* counter over wavelengths */
  int err;            /* error flag */

  double *Istar, *Iemission, *Iscatter, *Itotal; /* flux components */
  double Fstar, Femission, Fscatter, Ftotal; /* total flux */

  double Point_Temperature; /* dust temperature at the point */
  int dust_i; /* dust type counter */ 
  double dummy,tau_lambda;

  double Fst=0.0, Fem=0.0,Fsc=0.0;

  /* output files */
  FILE *Istar_file,*Iemission_file,*Iscatter_file,*Itotal_file;
  char filename1[256], filename2[256], filename3[256], filename4[256];

  if (my_MPI_rank!=0) return 0;

  printf("....angular distribution of intensity requested for (%E,%E)\n",
         point_x,point_z);

  Istar = dvector(0,lambda_Max);
  Iemission = dvector(0,lambda_Max);
  Iscatter = dvector(0,lambda_Max);
  Itotal = dvector(0,lambda_Max);

  /* create rho_vertex */
  rho_vertex = (struct _vertex_structure *) malloc 
                                            (sizeof(struct _vertex_structure));
  rho_vertex->prev = rho_vertex->next = NULL;
  rho_vertex->identification_number=0;
  rho_vertex->v1 = rho_vertex->v2 = rho_vertex->v3 = rho_vertex->v4 =
                   rho_vertex->v5 = rho_vertex->v6 = NULL;
  rho_vertex->level = 0;
  rho_vertex->type  = 0;
  rho_vertex->x = point_x;
  rho_vertex->z = point_z;
  rho_vertex->T_dust = NULL;
  rho_vertex->T_error = NULL;
  rho_vertex->Eta = 0.0;
  rho_vertex->triangles = NULL;

  dummy = sqrt(point_x*point_x+point_z*point_z);
  rho_unit.x = point_x/dummy;
  rho_unit.y = 0.0;
  rho_unit.z = point_z/dummy;

  rho.x = point_x;
  rho.y = 0.0;
  rho.z = point_z;


  /* open the output file */ 
#ifdef _2DD_MPI_
  sprintf(filename1,"%s/LELUYA_Angular_Istar_at_%E_%E.dat",
          _LELUYA_MPIOUT_PATH,point_x,point_z);
  sprintf(filename2,"%s/LELUYA_Angular_Iemission_at_%E_%E.dat",
          _LELUYA_MPIOUT_PATH,point_x,point_z);
  sprintf(filename3,"%s/LELUYA_Angular_Iscatter_at_%E_%E.dat",
          _LELUYA_MPIOUT_PATH,point_x,point_z);
  sprintf(filename4,"%s/LELUYA_Angular_Itotal_at_%E_%E.dat",
          _LELUYA_MPIOUT_PATH,point_x,point_z);
#else
  sprintf(filename1,"%s/LELUYA_Angular_Istar_at_%E_%E.dat",
          _LELUYA_OUTPUT_PATH,point_x,point_z);
  sprintf(filename2,"%s/LELUYA_Angular_Iemission_at_%E_%E.dat",
          _LELUYA_OUTPUT_PATH,point_x,point_z);
  sprintf(filename3,"%s/LELUYA_Angular_Iscatter_at_%E_%E.dat",
          _LELUYA_OUTPUT_PATH,point_x,point_z);
  sprintf(filename4,"%s/LELUYA_Angular_Itotal_at_%E_%E.dat",
          _LELUYA_OUTPUT_PATH,point_x,point_z);
#endif

  Istar_file = fopen(filename1,"w");
  Iemission_file = fopen(filename2,"w");
  Iscatter_file = fopen(filename3,"w");
  Itotal_file = fopen(filename4,"w");

  /* write the header */ 
  /*  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
      fprintf(Istar_file,"%lf[%.3d] ",Lambda_Grid[lambda_k],lambda_k); */
  fprintf(Istar_file,"\n");

  fprintf(Iemission_file,"Angular distribution of intensity around the point: %E, %E\n", 
          point_x,point_z);
  fprintf(Iemission_file,"theta[deg] phi[deg] ");
  /*  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
      fprintf(Iemission_file,"%lf[%.3d] ",Lambda_Grid[lambda_k],lambda_k);*/
  fprintf(Iemission_file,"\n");

  fprintf(Iscatter_file,"Angular distribution of intensity around the point: %E, %E\n", 
          point_x,point_z);
  fprintf(Iscatter_file,"theta[deg] phi[deg] ");
  /*  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
      fprintf(Iscatter_file,"%lf[%.3d] ",Lambda_Grid[lambda_k],lambda_k); */
  fprintf(Iscatter_file,"\n");

  fprintf(Itotal_file,"Angular distribution of intensity around the point: %E, %E\n", 
          point_x,point_z);
  fprintf(Itotal_file,"theta[deg] phi[deg] ");
  /*  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
      fprintf(Itotal_file,"%lf[%.3d] ",Lambda_Grid[lambda_k],lambda_k); */
  fprintf(Itotal_file,"\n");

  /* find the stellar intensity */ 
  ray_direction.x = Source_Radius*rho_unit.x;
  ray_direction.y = 0.0;
  ray_direction.z = Source_Radius*rho_unit.z;

  if (DEBUG.SCREEN_MESSAGE >= 1)   printf("Working on stellar flux\n");

  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
      TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
    start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
      TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;

    tau_lambda = _ZELENJAK_tau_over_grid (rho,ray_direction,lambda_k,FALSE);
    Istar[lambda_k] = Finite_Source_Flux_Correction(rho.x,rho.z,tau_lambda,lambda_k,
            start_vertex_list,start_triangle_list)*
            Total_Source_Flux(lambda_k,rho.x,rho.z,&err);

    Itotal[lambda_k]=Istar[lambda_k]; 
    /*    fprintf(Istar_file,"%E  ",Istar[lambda_k]); */

    if(lambda_k==14) Fst=Istar[lambda_k];
  }

  Fstar = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Istar);

  fprintf(Istar_file,"%E\n",Fstar);
  fclose(Istar_file);

  /* find the diffuse radiation */
  if (DEBUG.SCREEN_MESSAGE >= 1)   printf("Working on diffuse flux\n");

  /* find the most detailed grid (the largest number of vertices) */
  lambda_max_N = 0; 
  err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > err) {
      lambda_max_N = lambda_k;
      err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    }

  /* find the angular rays at rho for integration and flux calculation */
  start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
  TOTAL_NUMBER_OF_VERTICES =
	grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
	   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
  start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
  TOTAL_NUMBER_OF_TRIANGLES = 
    grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

  /* initialize the 3D cloud of points */ 
  err = _GOROSTAS_create_3D_cloud ();
  _GOROSTAS_create_integral_directions(rho_vertex);


   /* plot 3D view of sphercal triangles */
   err=_VODENJAK_plot_angular_integral_rays(rho_vertex,
					    "integral_rays_VRML",_GOROSTAS_N_phi_grid);

  /* loop over angular integrals */
  for (angular_rays_counter=0;
       angular_rays_counter<total_number_of_spherical_triangles; 
       angular_rays_counter++) {

    ray_direction = integral_directions[angular_rays_counter].direction;
    /* factor 2 comes from the axial symmetry */
    angular_flux_weight = integral_directions[angular_rays_counter].weight*2.0/(4.0*PI);

    /* find theta and phi angle of this direction */ 
    V0 = ray_direction;
    dummy = -V0.x*rho_unit.x - V0.z*rho_unit.z;
    fprintf(Iemission_file,"%E ",acos(dummy)*180/PI);
    fprintf(Iscatter_file,"%E ",acos(dummy)*180/PI);
    fprintf(Itotal_file,"%E ",acos(dummy)*180/PI);
    if (1.0-dummy*dummy>0.0)
      dummy = V0.y/sqrt(1.0-dummy*dummy);
    else 
      dummy = 0.0;
    if (dummy>=1.000) dummy=PI/2.0;
    else {
      if (- rho_unit.z*V0.x + rho_unit.x*V0.z>0.0)
	dummy=asin(dummy);
      else dummy = PI - asin(dummy);
     }
    fprintf(Iemission_file,"%E ",dummy*180/PI);
    fprintf(Iscatter_file,"%E ",dummy*180/PI);
    fprintf(Itotal_file,"%E ",dummy*180/PI);

    /* calculate the intensities */ 
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
        TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
      start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
        TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;

      tau_lambda = _VILOZMAJ_Intensity_Integral(lambda_k,rho,
		       ray_direction,&Iemission[lambda_k],&Iscatter[lambda_k],FALSE);

      Itotal[lambda_k] += ((Iemission[lambda_k]+Iscatter[lambda_k])*angular_flux_weight);
      /* write into the files */
      /*fprintf(Iemission_file,"%E  ",Iemission[lambda_k]);
      fprintf(Iscatter_file,"%E  ",Iscatter[lambda_k]);
      fprintf(Itotal_file,"%E  ",Itotal[lambda_k]);*/

    if(lambda_k==14) Fsc += (Iscatter[lambda_k]*angular_flux_weight);
    if(lambda_k==14) Fem += (Iemission[lambda_k]*angular_flux_weight);

    } /* end of the lambda loop */

    Femission = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Iemission) *
                angular_flux_weight;
    Fscatter = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Iscatter) *
                angular_flux_weight;
    Ftotal = Femission + Fscatter; 

    fprintf(Iemission_file,"%E\n",Femission);
    fprintf(Iscatter_file,"%E\n", Fscatter);
    fprintf(Itotal_file,"%E\n",Ftotal);
    fflush(Iemission_file);
    fflush(Iscatter_file);
    fflush(Itotal_file);

    if (DEBUG.SCREEN_MESSAGE >= 1)
       printf("Iemission, Iscatter for %d/%d ray finished\n",
              angular_rays_counter,total_number_of_spherical_triangles);

  } /* end of the loop over angular integrals */

  /* destroy the angular rays */
  total_number_of_spherical_triangles=0;
  /* destry the 3D cloud of points */
  err=_GOROSTAS_destroy_3D_cloud ();

  printf("Ustar[14]=%E Uem[14]=%E Uscatter[14]=%E\n",
	 grid_per_lambda[14].Ustar[2021],grid_per_lambda[14].Uem[2021],
	 grid_per_lambda[14].Usca[2021]);
  printf("NEW Ustar=%E Uem=%E Uscatter=%E\n",Fst,Fem,Fsc);

  /* Find the temperature at that point*/
  for (dust_i=0; dust_i < Max_Dust_Type; dust_i++) {
    Point_Temperature = -1;
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
      if (Dust_Density(point_x,point_z,dust_i,NULL,&err)>0.0){
        Point_Temperature = _GOROSTAS_Find_Temperature (point_x,point_z,
					    0.5,2000.0,dust_i,Itotal,&err);
      }
    printf("DUST TYPE %d TEMPERATURE = %lf\n",dust_i,Point_Temperature);
  }

  fclose(Iemission_file);
  fclose(Iscatter_file);
  fclose(Itotal_file);

  /* free variables */
  free(rho_vertex);
  free_dvector(Istar,0,lambda_Max);
  free_dvector(Iemission,0,lambda_Max);
  free_dvector(Iscatter,0,lambda_Max);
  free_dvector(Itotal,0,lambda_Max);

  return 0;
}
