/* LELUYA - GOROSTAS program block */
/* This function calculates the radiation pressure and saves it into a
   file */

/***********************************************************************/
/* July 17, 2007: Radiation pressure calculation for different grains with */
/* optical properties specified in separate input files. */
/* STEP 1: Activate this line: */

/* #define _RADPRESS_  */

/* STEP 2: Create N files "Radpress_dust_001.dat", "Radpress_dust_002.dat", etc.*/
/* and put them into LELUYA output directory. File structure is the same as structure */
/* dust cross section files for the main LELUYA input: the first 3 lines are ignored */
/* (as comments), then three columns (lambda, sigma absorption, sigma scattering). */
/* STEP 3: Set N in this line: */

#define _RADPRESS_N 3

/* STEP 4: recompile  */
/* The result is radiation pressure based on specified dust_xxx.dat cross */
/* sections written in separate files LELUYA_Radiation_Pressure_xxx.dat */
/***********************************************************************/

#include "GOROSTAS.h"

int _GOROSTAS_Radiation_Pressure (void) {

  struct _vertex_structure   *loop_vertex; /* a grid point */

  struct _3DVector V0,ray_direction; /* direction of the intensity integral*/
  int angular_rays_counter; /* counter of the angluar rays at a point */
  double angular_flux_weight; /* weight in the angular integral */

  int lambda_max_N; /* wavelength with the largest number of vertices */
  int lambda_k;       /* counter over wavelengths */
  int err;            /* error flag */

  /* array of points for calculation */ 
  int count_x, count_z;
  double Step_x, Step_z; 
  double x_coord, z_coord;
  double temperature;
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _3DVector line_grid_direction; /* direction of the log-line grid */

  double *Istar, *Iemission, *Iscatter, *Itotal; /* flux components */
  int dust_i; /* dust type counter */ 
  double dummy,tau_lambda;

  /* pressure tensor: z-axis of the local coordinate system is along the radial
     direction - dimension [3][3][lambda_Max] */
  double ***Pressure_Tensor;
  /* pressure as if there were no dust at [0,0] point */
  double F_force_scale;
  /* pressure tensor unit vectors r,theta,phi */
  struct _3DVector unit_r,unit_theta,unit_phi;
  /* cosine matrix: matrix of cosines between pressure tensor vectors and 
     intensity vector */
  double n_r,n_theta,n_phi;
  /* radiation pressure force */
  double F_r, F_theta,F_phi=0.0, F_r0, F_theta0;

  /* output files */
  FILE *pressure_file;
  char filename[256];

  /* flag that shows if it is an array or a line */
  int flag_grid_type;

  /* for DEBUG.CONTINUE==406 */
  int vertexID;
  struct _vertex_structure *v;
  struct _triangle_structure *t;

#ifdef _2DD_MPI_
  int tag = 0;
  int receive_flag;
  MPI_Status Receive_Status, Status;
  int MPI_radius_i, MPI_theta_counter;
  int process_counter;
  int count_x_recv, count_z_recv;
#endif

#ifdef _RADPRESS_
  int _RADPRESS_count; 
  FILE *RADPRESS_file;
  double *RADPRESS_abs, *RADPRESS_sca;
  RADPRESS_abs=dvector (0,lambda_Max-1);
  RADPRESS_sca=dvector (0,lambda_Max-1);

  for (_RADPRESS_count=1; _RADPRESS_count<=_RADPRESS_N; _RADPRESS_count++) {
  
  /* read cross sections */
  #ifdef _2DD_MPI_
  sprintf(filename,"%s/Radpress_dust_%.3d.dat",_LELUYA_MPIRUN_PATH,_RADPRESS_count);
  #else
  sprintf(filename,"%s/Radpress_dust_%.3d.dat",_LELUYA_RUN_PATH,_RADPRESS_count);
  #endif
  
  /* read dust cross sections */
  RADPRESS_file = fopen(filename,"r");
  if (RADPRESS_file==NULL) {printf("ERROR: %s\n",filename);return 5;}
    /* read the first three lines of comments */
    err= _PATULJAK__Search_for_char (RADPRESS_file,"\n");
    err= _PATULJAK__Search_for_char (RADPRESS_file,"\n");
    err= _PATULJAK__Search_for_char (RADPRESS_file,"\n");
    if (err) return 6;
  /* read the cross sections and interpolate them on lambda grid */
  err = _PATULJAK__Read_Abs_Sca (RADPRESS_file,RADPRESS_abs,RADPRESS_sca);
    if (err==1) return 6;
    if (err==2) return 7;
    if (err==3) return 8;
  fclose(RADPRESS_file);

  printf("#%d: ... radiation pressure field for dust #%.3d\n",my_MPI_rank,_RADPRESS_count);
#else
  printf("#%d: ... radiation pressure field\n",my_MPI_rank);
#endif

  Istar = dvector(0,lambda_Max);
  Iemission = dvector(0,lambda_Max);
  Iscatter = dvector(0,lambda_Max);
  Itotal = dvector(0,lambda_Max);
  Pressure_Tensor=d3tensor(0,2,0,2,0,lambda_Max);
  /* create loop_vertex */
  loop_vertex = (struct _vertex_structure *) malloc 
                                            (sizeof(struct _vertex_structure));
  loop_vertex->prev = loop_vertex->next = NULL;
  loop_vertex->identification_number=0;
  loop_vertex->v1 = loop_vertex->v2 = loop_vertex->v3 = loop_vertex->v4 =
                   loop_vertex->v5 = loop_vertex->v6 = NULL;
  loop_vertex->level = 0;
  loop_vertex->type  = 0;
  loop_vertex->x = loop_vertex->z = 0.0;
  loop_vertex->T_dust = dvector(0,Max_Dust_Type-1); ;
  loop_vertex->T_error = NULL;
  loop_vertex->Eta = 0.0;
  loop_vertex->triangles = NULL;

  if (my_MPI_rank==0) {
    /* open the output file */ 
#ifdef _2DD_MPI_
  if (DEBUG.CONTINUE!=407) {
  #ifdef _RADPRESS_
    sprintf(filename,"%s/LELUYA_Radiation_Pressure_%.3d.dat",
            _LELUYA_MPIOUT_PATH,_RADPRESS_count);
  #else
    sprintf(filename,"%s/LELUYA_Radiation_Pressure.dat",
            _LELUYA_MPIOUT_PATH);
  #endif
    }
  else sprintf(filename,"%s/LELUYA_Flux.dat",_LELUYA_MPIOUT_PATH);
#else
  if (DEBUG.CONTINUE!=407) {
  #ifdef _RADPRESS_
    sprintf(filename,"%s/LELUYA_Radiation_Pressure_%.3d.dat",
            _LELUYA_OUTPUT_PATH,_RADPRESS_count);
  #else
    sprintf(filename,"%s/LELUYA_Radiation_Pressure.dat",
          _LELUYA_OUTPUT_PATH);
  #endif
  }
  else sprintf(filename,"%s/LELUYA_Flux.dat",_LELUYA_MPIOUT_PATH);
#endif
    pressure_file = fopen(filename,"w");

    if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
      /* fprint the basic info about the grid */
      fprintf(pressure_file,"X1  X2   Z1  Z2  Nx  Nz\n");
      fprintf(pressure_file,"%E  %E  %E  %E  %d  %d\n",X1_403,X2_403,Z1_403,Z2_403,
          Nx_403,Nz_403);
      fprintf(pressure_file,"   x     z    ");
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        if (DEBUG.CONTINUE==405) 
          fprintf(pressure_file,"Force_theta[%d]  Force_radial[%d]  F_theta_scaled[%d]  F_radial_scaled[%d]  ",dust_i,dust_i,dust_i,dust_i);
        else 
          fprintf(pressure_file,"Flux_theta[%d]  Flux_radial[%d]  F_theta_scaled[%d]  F_radial_scaled[%d]  ",dust_i,dust_i,dust_i,dust_i);       
        }
      fprintf(pressure_file,"\n");
    }
  }

  /* find the most detailed grid (the largest number of vertices) */
  lambda_max_N = 0; 
  err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > err) {
      lambda_max_N = lambda_k;
      err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    }

  if (my_MPI_rank==0)
  if (DEBUG.CONTINUE==406) {
    /* fprint the basic info about the grid */
    fprintf(pressure_file,"%d  = number of vertices\n",
            grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES);
    fprintf(pressure_file,"%d  = number of dust types\n",Max_Dust_Type);
    fprintf(pressure_file,"ID      x           z     ");
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      fprintf(pressure_file,"Force_theta[%d]  Force_radial[%d]  F_theta_scaled[%d]  F_radial_scaled[%d]  ",dust_i,dust_i,dust_i,dust_i);
    fprintf(pressure_file,"\n");
  }
  
  start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
  TOTAL_NUMBER_OF_VERTICES =
    grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
       grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
  start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
  TOTAL_NUMBER_OF_TRIANGLES = 
           grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;
  /* create 3D cloud of points (used for calculating integral directions)*/
  err = _GOROSTAS_create_3D_cloud ();

  /* decide where to look for the rad. pressure force */
  if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407) && Nx_403>0) flag_grid_type=1;
  else if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {flag_grid_type=2; Nx_403 *= -1;Nz_403=1;}
  else flag_grid_type=3;

  if (flag_grid_type==1) {
    /* find the steps in array */
    Step_x = (X2_403  - X1_403) / (Nx_403 - 1);
    Step_z = (Z2_403  - Z1_403) / (Nz_403 - 1);
  }
  else if (flag_grid_type==2) {
    /* find the scale factor for the log-line grid*/
    line_grid_direction.x = X2_403  - X1_403;
    line_grid_direction.y = 0.0;
    line_grid_direction.z = Z2_403  - Z1_403;
    dummy = sqrt(line_grid_direction.x*line_grid_direction.x+
                 line_grid_direction.z*line_grid_direction.z);
    line_grid_direction.x /= dummy;
    line_grid_direction.z /= dummy;
    Step_x = pow((dummy+1.0),(1.0/(Nx_403-1)));
  }
  else {
    Step_x = Step_z = 0.0;
  }

#ifdef _2DD_MPI_
  current_triangle=NULL;

  if (my_MPI_rank==0) {
    count_x=0; /* DEBUG.CONTINUE==406 uses count_x as the vertex counter */
    count_z=0;
    /* send the inital jobs */
    for (process_counter=1; process_counter<total_MPI_process; process_counter++) {
      if ((count_x<Nx_403 && count_z<Nz_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) ||
          (count_x<grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES
            && DEBUG.CONTINUE==406)) {
        MPI_Send(&count_x,1,MPI_INT,process_counter,tag,MPI_communicator);
        if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407))
          MPI_Send(&count_z,1,MPI_INT,process_counter,tag,MPI_communicator);
    count_x++;
    if (count_x>=Nx_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {count_z++;count_x=0;}
      } 
      else {
    err=-1;
        MPI_Send(&err,1,MPI_INT,process_counter,tag,MPI_communicator);
      }
    }
    /* receive soultions until all array points are done */
    count_x_recv=0;
    count_z_recv=0;
    while ((count_x_recv<Nx_403 && count_z_recv<Nz_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) ||
           (count_x_recv<grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES
            && DEBUG.CONTINUE==406)) {
      if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
        /* get x_coord */ 
        MPI_Recv(&x_coord,1,MPI_DOUBLE,MPI_ANY_SOURCE,tag,MPI_communicator,&Receive_Status);
        /* get z_coord */ 
        MPI_Recv(&z_coord,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
        vertexID = (int)((x_coord - X1_403)/Step_x+1.0E-10)*Nx_403 +
                   (int)((z_coord - Z1_403)/Step_z+1.0E-10);
        fprintf(pressure_file,"%d %E %E ",vertexID,x_coord,z_coord);
      }
      else if (DEBUG.CONTINUE==406) {
        /* get vetrtex ID */
        MPI_Recv(&vertexID,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,&Receive_Status);
        v = grid_per_lambda[lambda_max_N].start_vertex;
        while (v->identification_number<vertexID) v=v->next;
        fprintf(pressure_file,"%d  %13.11E  %13.11E  ",vertexID,v->x,v->z);
      }
      /* get the data */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    MPI_Recv(&F_theta,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,
                 &Status);
    MPI_Recv(&F_theta0,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,
                 &Status);
        /* get the force scale */
        if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) if (dust_i==0) if (vertexID==0)
          MPI_Recv(&F_force_scale,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,
                   &Status);
    MPI_Recv(&F_r,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,
                 &Status);
    MPI_Recv(&F_r0,1,MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,
                 &Status);
    fprintf(pressure_file,"%E  %E  %E  %E ",F_theta,F_r,F_theta0,F_r0);
      }
      fprintf(pressure_file,"\n");
      fflush(pressure_file);
      /* send the next job */
      if ((count_x<Nx_403 && count_z<Nz_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) ||
          (count_x<grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES
            && DEBUG.CONTINUE==406)) {
        MPI_Send(&count_x,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
        if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407))
          MPI_Send(&count_z,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
    count_x++;
    if (count_x>=Nx_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {count_z++;count_x=0;}
      } 
      else {
    err=-1;
        MPI_Send(&err,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
      }
      /* increase the number of recevied points */
      count_x_recv++;
      if (count_x_recv>=Nx_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {count_z_recv++;count_x_recv=0;}
    }
    /* print force scale */
    if (DEBUG.CONTINUE!=407)
    fprintf(pressure_file,"%E = scaled radial force on dust type=0 at (%E,%E)=[0,0] as if there were no dust\n",
            F_force_scale,X1_403,Z1_403);
    else 
    fprintf(pressure_file,"%E = scaled flux (flux*r^2) at (%E,%E)=[0,0] as if there were no dust\n",
            F_force_scale,X1_403,Z1_403);
            
  } /* end of my_MPI_rank==0 */
  else { /* nonzero process */
    /* receive the first ever signal from 0 */ 
    MPI_Recv(&count_x,1,MPI_INT,0,tag,MPI_communicator,&Status);
    while (count_x>=0) {
      /* get the z coordinate */
      if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407))
        MPI_Recv(&count_z,1,MPI_INT,0,tag,MPI_communicator,&Status);

      if (flag_grid_type==1) {
    x_coord = X1_403 + count_x*Step_x;
    z_coord = Z1_403 + count_z*Step_z;
      }
      else if (flag_grid_type==2){
    x_coord = X1_403;
    z_coord = Z1_403;
    if (count_x>0) {
      dummy = Step_x; err = count_x-1; while(err>0) {dummy *= Step_x; err--;}
      x_coord += ((dummy-1.0)*line_grid_direction.x);
      z_coord += ((dummy-1.0)*line_grid_direction.z);
    }
      }
      else {
        v = grid_per_lambda[lambda_max_N].start_vertex;
        while (v->identification_number<count_x) v=v->next;
        x_coord = v->x;
        z_coord = v->z;
      }
#else
  current_triangle=NULL;

  /* loop over the array or the log-grid line */
  if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {x_coord = X1_403; z_coord = Z1_403;}
  else {
    v = grid_per_lambda[lambda_max_N].start_vertex;
    x_coord = v->x;z_coord = v->z;
  }
  count_x=0;
  do {
    if (flag_grid_type==1) z_coord = Z1_403;
    count_z=0;
    do {
#endif

      loop_vertex->x = x_coord;
      loop_vertex->z = z_coord;
      printf("#%d: ... find force at %E, %E [%d,%d]\n",my_MPI_rank,
         loop_vertex->x, loop_vertex->z,count_x,count_z);
   fflush(stdout); 

      if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
        /* find the temperature of this point */ 
        start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
        TOTAL_NUMBER_OF_VERTICES =
      grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
        TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
       grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
        start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
        TOTAL_NUMBER_OF_TRIANGLES = 
           grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;
        /* ... locate the grid triangle */ 
        err=_ZELENJAK_locate_point_in_grid(x_coord,z_coord,&current_triangle);
      }

      /* if the triangle is NULL then we don't need to calculate anything */
      if (current_triangle==NULL && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
#ifdef _2DD_MPI_
          MPI_Send(&x_coord,1,MPI_DOUBLE,0,tag,MPI_communicator);
          MPI_Send(&z_coord,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
          if (count_x==0 && count_z==0) fprintf(pressure_file,"I CAN'T FIND THE FLUX SCALE\n");
          vertexID = (int)((x_coord - X1_403)/Step_x+1.0E-10)*Nx_403 + 
                     (int)((z_coord - Z1_403)/Step_z+1.0E-10);
      fprintf(pressure_file,"%d  %E  %E  ",vertexID,x_coord,z_coord);
#endif
          F_r = 0;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++){
#ifdef _2DD_MPI_
            MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
            MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
            MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
            MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
        fprintf(pressure_file,"%E  %E  %E  %E ",F_r,F_r,F_r,F_r);
#endif
      }
      }
      else {
      /* there is a dust at this point, so continue with the calculation */ 
    if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
          for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
            if (current_triangle==NULL) 
          loop_vertex->T_dust[dust_i]= 0.0;
           else if (current_triangle->v1->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v1->T_dust[dust_i] > 0.0 &&
                current_triangle->v2->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v2->T_dust[dust_i] > 0.0 &&
        current_triangle->v3->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v3->T_dust[dust_i] > 0.0)
         loop_vertex->T_dust[dust_i] = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
                                     current_triangle->v1->T_dust[dust_i],
             current_triangle->v2->x, current_triangle->v2->z,
                                 current_triangle->v2->T_dust[dust_i],
         current_triangle->v3->x, current_triangle->v3->z,
                                     current_triangle->v3->T_dust[dust_i],
             x_coord,z_coord);
            else loop_vertex->T_dust[dust_i] = 0.0;
          }
    }
    else if (DEBUG.CONTINUE==406) {
          loop_vertex->x = v->x;
          loop_vertex->z = v->z;
          for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
            loop_vertex->T_dust[dust_i] = v->T_dust[dust_i];
    }

   if (my_MPI_rank==0) {
     if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) 
       fprintf(pressure_file,"%E  %E  ",x_coord,z_coord);
     else 
       fprintf(pressure_file,"%d  %E  %E  ",count_x,x_coord,z_coord);
   }
   V0.x =  loop_vertex->x;
   V0.y=0.0;
   V0.z = loop_vertex->z;
   /* find the unit vectors */
   unit_r.x = loop_vertex->x;
   unit_r.y = 0.0;
   unit_r.z = loop_vertex->z;
   dummy = sqrt(unit_r.x*unit_r.x+unit_r.z*unit_r.z);
   unit_r.x /= dummy;
   unit_r.z /= dummy;

   unit_theta.x = unit_r.z;
   unit_theta.y = 0.0;
   unit_theta.z = - unit_r.x;

   unit_phi.x = 0.0;
   unit_phi.y = 1.0;
   unit_phi.z = 0.0;

   /* find the stellar radiation pressure */
   ray_direction.x = Source_Radius*unit_r.x;
   ray_direction.y = 0.0;
   ray_direction.z = Source_Radius*unit_r.z;

   for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
     start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
       TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
     start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
       TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;

     tau_lambda = _ZELENJAK_tau_over_grid (V0,ray_direction,lambda_k,FALSE);
     /* find the stellar contribution ("dummy" variable is the tangental correction) */ 
     dummy = 0.0;
     Istar[lambda_k] = Finite_Source_Flux_Correction_with_direction(V0.x,V0.z,
          tau_lambda,lambda_k,unit_r,unit_theta,
              start_vertex_list,start_triangle_list,&dummy)* 
              Total_Source_Flux(lambda_k,V0.x,V0.z,&err);
     /* set the pressure variable */
     Pressure_Tensor[0][0][lambda_k] = 0.0;
     Pressure_Tensor[0][1][lambda_k] = 0.0;
     Pressure_Tensor[0][2][lambda_k] = 0.0;
     Pressure_Tensor[1][0][lambda_k] = 0.0;
     Pressure_Tensor[1][1][lambda_k] = 0.0;
     Pressure_Tensor[1][2][lambda_k] = 0.0;
     Pressure_Tensor[2][0][lambda_k] = 0.0;
     Pressure_Tensor[2][1][lambda_k] = 
           Total_Source_Flux(lambda_k,V0.x,V0.z,&err)*dummy; /* tangental */
     Pressure_Tensor[2][2][lambda_k] = Istar[lambda_k]; /* radial */
   }

   /* find the diffuse radiation */

   /* find the angular rays at loop_vertex for angular integration  */
   start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
   TOTAL_NUMBER_OF_VERTICES =
    grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
   TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
       grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
   start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
   TOTAL_NUMBER_OF_TRIANGLES = 
           grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

   _GOROSTAS_create_integral_directions(loop_vertex);
   /* loop over angular integrals */
   for (angular_rays_counter=0;
        angular_rays_counter<total_number_of_spherical_triangles; 
        angular_rays_counter++) {

     /* printf("RAY #%d/%d\n",angular_rays_counter,total_number_of_spherical_triangles);
        fflush(stdout);  */

     ray_direction = integral_directions[angular_rays_counter].direction;
     /* factor 2 comes from the axial symmetry */
     angular_flux_weight = integral_directions[angular_rays_counter].weight*2.0/(4*PI);
     /* find cosines matrix */
     n_r = -(ray_direction.x*unit_r.x + ray_direction.z*unit_r.z);
     n_phi=0;
     n_theta = -(ray_direction.x*unit_theta.x + ray_direction.z*unit_theta.z);

     /* calculate the intensities */ 
     for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
       start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
         TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
       start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
         TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;

       tau_lambda = _VILOZMAJ_Intensity_Integral(lambda_k,V0,
               ray_direction,&Iemission[lambda_k],&Iscatter[lambda_k],FALSE);
       /* radiation pressure force */
       Itotal[lambda_k] = (Iemission[lambda_k]+Iscatter[lambda_k])*angular_flux_weight;
       Pressure_Tensor[0][0][lambda_k] += (Itotal[lambda_k]*n_theta);
       Pressure_Tensor[1][1][lambda_k] += (Itotal[lambda_k]*n_r);
     } /* end of the lambda loop */

   } /* end of the loop over angular integrals */
   /* destroy the angular rays */
   total_number_of_spherical_triangles=0;

#ifdef _2DD_MPI_
   /* this will trigger the receive mode in the zero process */ 
   if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
     MPI_Send(&x_coord,1,MPI_DOUBLE,0,tag,MPI_communicator);
     MPI_Send(&z_coord,1,MPI_DOUBLE,0,tag,MPI_communicator);
   }
   else if (DEBUG.CONTINUE==406) {
     MPI_Send(&count_x,1,MPI_INT,0,tag,MPI_communicator);
   }
#endif
   if (DEBUG.CONTINUE!=407) {
   /* ------ calculate radiation pressure ----- */

   /* find the force for each dust type */
   for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {

     /* this is dust_i component of Eta */ 
     if ((Max_Dust_Temp<=0.0 && loop_vertex->T_dust[dust_i]<= 
                                Dust_Sublim_Temp[dust_i]*(1.0+_SUBLIMATION_TOLERANCE_) 
          && loop_vertex->T_dust[dust_i]>0.0 )
         ||
     (Max_Dust_Temp>0.0 && loop_vertex->T_dust[dust_i]<=Max_Dust_Temp
          && loop_vertex->T_dust[dust_i]>0.0)
        ) {
       F_r = Dust_Density(V0.x,V0.z,dust_i,NULL,&err) * Tau_Total / Eta_Denominator;
       F_r = F_r / (V0.x*V0.x+V0.z*V0.z);
     }
     else F_r = -1;

     if (F_r>-2.0) {
       /* tangental force at each wavelength */
       #ifdef _RADPRESS_
       for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
         Itotal[lambda_k] = (RADPRESS_abs[lambda_k] + RADPRESS_sca[lambda_k]) *
                            F_r*(Pressure_Tensor[2][1][lambda_k]+
                                 Pressure_Tensor[0][0][lambda_k]);
       #else
       for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
         Itotal[lambda_k] = (Sigma_Dust_Abs[dust_i][lambda_k] + 
                                Sigma_Dust_Sca[dust_i][lambda_k]) *
                            F_r*(Pressure_Tensor[2][1][lambda_k]+
                                 Pressure_Tensor[0][0][lambda_k]);
       #endif
       F_theta = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Itotal);
       F_theta0 = F_theta/F_r;
       if (F_r<0.0) F_theta = 0.0;

#ifdef _2DD_MPI_
       MPI_Send(&F_theta,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_theta0,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
       fprintf(pressure_file,"%E  ",F_theta);
#endif

       /* radial force at each wavelength */
       for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
       #ifdef _RADPRESS_
         Itotal[lambda_k] = (RADPRESS_abs[lambda_k] + RADPRESS_sca[lambda_k]) *
                            F_r*(Pressure_Tensor[2][2][lambda_k]+
                                 Pressure_Tensor[1][1][lambda_k]);       
       #else
         Itotal[lambda_k] = (Sigma_Dust_Abs[dust_i][lambda_k] + 
                                Sigma_Dust_Sca[dust_i][lambda_k]) *
                            F_r*(Pressure_Tensor[2][2][lambda_k]+
                                 Pressure_Tensor[1][1][lambda_k]);
       #endif
       dummy = F_r;
       F_r = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Itotal);
       F_r0 = F_r/dummy;
       if (dummy<0.0) F_r = 0.0;

       /* find the force scale */
       if (DEBUG.CONTINUE==405) if (count_x==0 && count_z==0 && dust_i==0) {
         for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
       #ifdef _RADPRESS_
           Istar[lambda_k] = (RADPRESS_abs[lambda_k] + 
                              RADPRESS_sca[lambda_k]) * Istar[lambda_k];
       #else
           Istar[lambda_k] = (Sigma_Dust_Abs[dust_i][lambda_k] +
                              Sigma_Dust_Sca[dust_i][lambda_k]) * Istar[lambda_k];
       #endif
         F_force_scale = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Istar);
#ifdef _2DD_MPI_
         MPI_Send(&F_force_scale,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
         fprintf(pressure_file,"%E = force scale\n");
#endif
       }

#ifdef _2DD_MPI_
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_r0,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
       fprintf(pressure_file,"%E  %E  %E  ",F_r,F_theta0,F_r0);
#endif
     }
     else { 
       F_r = 0.0;
#ifdef _2DD_MPI_
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
       fprintf(pressure_file,"%E  %E  %E  %E  ",F_r,F_r,F_r,F_r);
#endif
     }
   }} /* end of radiation pressure calculation */ 
   else {
     /* ------ calculate flux ----- */

     dummy = 1.0 / (V0.x*V0.x+V0.z*V0.z);
     /* tangental flux at each wavelength */
     for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
         Itotal[lambda_k] = dummy * (Pressure_Tensor[2][1][lambda_k]+
                                 Pressure_Tensor[0][0][lambda_k]);
     F_theta = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Itotal);
     F_theta0 = F_theta/dummy;

#ifdef _2DD_MPI_
       MPI_Send(&F_theta,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_theta0,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
       fprintf(pressure_file,"%E  ",F_theta);
#endif

     /* radial flux at each wavelength */
     for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
         Itotal[lambda_k] = dummy*(Pressure_Tensor[2][2][lambda_k]+
                                 Pressure_Tensor[1][1][lambda_k]);
     F_r = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Itotal);
     F_r0 = F_r/dummy;

     /* find the flux scale */
     if (count_x==0 && count_z==0) {
       F_force_scale = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Istar);
#ifdef _2DD_MPI_
         MPI_Send(&F_force_scale,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
         fprintf(pressure_file,"%E = flux scale\n");
#endif
       }

#ifdef _2DD_MPI_
       MPI_Send(&F_r,1,MPI_DOUBLE,0,tag,MPI_communicator);
       MPI_Send(&F_r0,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
       fprintf(pressure_file,"%E  %E  %E  ",F_r,F_theta0,F_r0);
#endif
   
   } /* end of flux calculation */

      } /* end of: if (current_triangle==NULL) */ 

#ifdef _2DD_MPI_
      MPI_Recv(&count_x,1,MPI_INT,0,tag,MPI_communicator,&Status);
    }
  }
#else
   fprintf(pressure_file,"\n");
   fflush(pressure_file);

      if ((DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407)) {
        /* go to the next array point */ 
        if (flag_grid_type==1) z_coord += Step_z;
        count_z++;
      }

    } while (count_z <Nz_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407));
    count_x++;
    if (flag_grid_type==1) x_coord += Step_x;
    else if (flag_grid_type==2){
      dummy = Step_x; err = count_x-1; while(err>0) {dummy *= Step_x; err--;}
      x_coord = X1_403+(dummy-1.0)*line_grid_direction.x;
      z_coord = Z1_403+(dummy-1.0)*line_grid_direction.z;
    }
    else {
      v=v->next;
      x_coord = v->x;
      z_coord = v->z;
    }
  } while ((count_x<Nx_403 && (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407))  || 
           (count_x<grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES 
            && DEBUG.CONTINUE==406));
#endif

  if (my_MPI_rank==0) {
    if (DEBUG.CONTINUE==406) {
      /* print triangles */
      fprintf(pressure_file,"%d = number of triangles\n",
              grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES);
      t = grid_per_lambda[lambda_max_N].start_triangle;
      do {
        fprintf(pressure_file,"%d  %d  %d\n",
                t->v1->identification_number,
                t->v2->identification_number,
                t->v3->identification_number);
        t=t->next;
      } while (t!=grid_per_lambda[lambda_max_N].start_triangle);
    }
    fclose(pressure_file);
  }

  /* destry the 3D cloud of points (used for calculating integral directions) */
  err=_GOROSTAS_destroy_3D_cloud ();

  /* free variables */
  free_dvector(Istar,0,lambda_Max);
  free_dvector(Iemission,0,lambda_Max);
  free_dvector(Iscatter,0,lambda_Max);
  free_dvector(Itotal,0,lambda_Max);
  free_d3tensor(Pressure_Tensor,0,2,0,2,0,lambda_Max);
  free_dvector(loop_vertex->T_dust,0,Max_Dust_Type-1);
  free(loop_vertex);
#ifdef _RADPRESS_
  } /* end of  _RADPRESS_count loop */
  free_dvector(RADPRESS_abs,0,lambda_Max-1);
  free_dvector(RADPRESS_sca,0,lambda_Max-1);
#endif

#ifdef _2DD_MPI_
  MPI_Barrier(MPI_communicator);
#endif

  return 0;
}
