/* 2D DUSTY - two dimensional radiative transfer in astrophysical dust */
/* with a central energy source.  */
/* January, 2000. VERSION: Gamma 0.0 */

/**** environment variables as macros ******************/
/* used in: _VILOZMAJ_Radiative_Transfer_Solver  */
#ifndef LEL_RADTRANS_N /* Max. number of the matrix grid vertices */
#define LEL_RADTRANS_N 0
#endif
#ifndef LEL_RADTRANS_L /* Max. number of wavelengths to solve */
#define LEL_RADTRANS_L 0
#endif
/***********************************************************************/
/******************* The Main Code Block : PERUN ***********************/
/***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "_install_parameters.c"
#include "_global_constants.h"
#include "_global_structures.h"
#include "LEPAMARA/LEPAMARA.h" 
#include "PESJANEK/PESJANEK.h"
#include "PATULJAK/PATULJAK.h"
#include "ZELENJAK/ZELENJAK.h"
#include "PERUNIKA/PERUNIKA.h"
#include "VILOZMAJ/VILOZMAJ.h"

int    MPI_argc;
char **MPI_argv;
int my_MPI_rank;       /* rank of my process */
int total_MPI_process; /* total number of processes */
#ifdef _2DD_MPI_
#include <mpi.h>
MPI_Comm MPI_communicator;
#endif

/***********************************************************************/
/******************* Code constants (#define) **************************/
 

/***********************************************************************/
/******************* Code variable structures **************************/


/***********************************************************************/
/*********************** Algorithm flags *******************************/
/***** DEBUGGER FLAGS: */

#include "_debugger_flags.h"
struct _Debugger_Flags DEBUG;

/***** INPUT DATA FLAGS: */
#include "_data_calc_output_flags.h"
struct _Input_Data_Flags INPUT;

/***** CALCULATION FLAGS: */

/***** OUTPUT FLAGS: */

/***********************************************************************/
/******************* Algorithm global constants ************************/

/***** some constants **********/
double SQRT3; /* sqrt(3) */
double SQRT3_2; /* sqrt(3)/2 */ 
double PI_SQRT3; /* PI/SQRT3 */

/*** lambda grid variables ***/
double *Lambda_Grid;       /* vector with the wavelenghts at the 
                              grid points */
int    lambda_Max=0;       /* total number of grid points: index of
                              the last wavelength in the grid */
/*** total optical depth variables ***/
double Tau_Total;          /* total optical depth toward the central
                              energy source at azimuthal angle 
                              Theta_Total_Value and wavelength 
                              Lambda_Total_Value. This optical depth is
                              used for scaling of all other optical
                              depths used in the program */
double Theta_Total_Value;  /* azimuthal angle which defines the line
                              of total optical depth Tau_Total*/
int    Theta_Total;        /* integer representation of Theta_Total_Value
                              in the angular grid of the central energy
                              source */
double Tau_Min_Grid=0.1;   /* optical depth below which we do not calculate
			      the radiative transfer matrix, but rather use
                              the approximate solution for optically thin
			      dust */
double Lambda_Min_Grid=6.0;/* the shortest wavelength for which we can
                               use the approximate solution of optically thin
                               dust and avoid calculation of radiative transfer
                               matrix (used in combination with Tau_Min_Grid)*/

double Grid_Tolerance_Tau_Factor=0.02; /* if we have already built 
			      a grid at an optical depth tau, then we will use
			      the same grid for the optical depths between
                              "tau*Grid_Tolerance_Tau_Factor" and "tau"
			      (this also means that Grid_Tolerance_Tau_Factor
			      has to be smaller than 1)*/
double Lambda_Total_Value; /* wavelength at which the total optical depth
                              Tau_Total is specified */
int    Lambda_Total;       /* integer representation of Lambda_Total_Value
                              in Lambda_Grid */
double *Source_Flux;       /* the central 'point' source flux
                              (this is the 'point source approximation' if
                              the source has a finite size) */
double Rout;               /* Outer radius of the dust envelope */
double Source_Radius;      /* The scaled radius of the source */
double Source_Teff;        /* The effective source temperature */
/*** dust properties *********/
int    Max_Dust_Type=0;    /* Total number of different dust type in 
                              the multigrain mixture */
char **Dust_Component_Name;/* names of the dust components */
double *Dust_Sublim_Temp;  /* sublimation temperatures */
double Max_Dust_Temp;      /* maximum dust temperature (it can be higher than
                              the sublimation temperature) */
double **Sigma_Dust_Abs;   /* dust absorption cross sections */
double **Sigma_Dust_Sca;   /* dust scattering cross sections */
                           /* Sigma_Dust_Abs and Sigma_Dust_Sca have two 
                              indeces: one represents different dust types 
                              from 0 to Max_Dust_Type-1, and another shows 
                              the wavelength. For example, Sigma_Dust_Abs[i][j]
                              gives the absorption cross section for the dust 
                              type 'i' at wavelength Lambda_Grid[j] */
struct density_parameters *Dust_Density_Param;
                           /* list of lists to all necessary parameters
                              required for calculation of the dust density
                              profiles for each of the dust components.
                              (see _data_calc_output_flags.h for details about 
                               the dust density models)*/
double Eta_Denominator;     /* denominator in the Eta function */
struct _density_grid_structure Density_2D_Grid; /* density grid for
                              a non-analytic dust density */
/*** radiative transfer *********/
double Psi;               /* 'Psi' variable (see equations for definition) */

/* here is the list of integral rays. It has to be created for each grid point
   separately. The ray is integral_directions[].direction.  */
struct _spherical_triangle_structure 
                   integral_directions[MAX_NUMBER_OF_SPHERICAL_TRIANGLES+4];
int total_number_of_spherical_triangles=0; /* total number of rays */ 

/*** triangular grid variables *********/
struct _vertex_structure *start_vertex_list; /* this is the grid !!
                              that is a pointer to the grid vertices
                              which form the grid. */
struct _triangle_structure *start_triangle_list; /* when the vertices
                              are connected into the mesh, they create 
                              triangles. This triangles are the mesh cells. */
struct _grid_per_lambda_structure *grid_per_lambda; /* list of grids,
			      one grid at each wavelength */
struct _grid_per_lambda_structure *old_grid_per_lambda; /* these are the
                           grids from previous iteration */
struct _vertex_structure *start_old_vertex_list; /* vertex list from a previous
                              iteration */
struct _triangle_structure *start_old_triangle_list; /* triangle list form 
                              a previous iteration */
struct _quadrangle_structure *start_quadrangle_list; /* these quadrangles are
                              corrections of the sublimation surfaces. They
                              are created in _PERUNIKA_sublimation_extension
                              and destroyed in _PERUNIKA_main */
int _INITIAL_GRID_ARRAY_r = 300;   /* total number of the radial grid steps */
int _INITIAL_GRID_ARRAY_theta = 80;/* total number of the angular grid steps
                                        in the array of intital approximation.
                                        is used for generation of the initial
                                        grid. */
double *initial_radial_array;/* radial grid for initial_temperature_array */
double ***initial_energy_array;/* array of the initial energy density and ... */
double ***initial_temperature_array; /* initial dust temperature. This array is
                          used for generation of the initial triangular grid */
int START_LEVEL=4; /* the grid cell with their level number smaller or
                           equal then START_LEVEL will be created for sure */
double CELL_SIZE_LIMIT=0.05;/* the smallest geometrical size od a grid cell
                                 (at r=1) */
double ETA_RATIO_LIMIT=1.0E30;/* ratio between eta values at two grid points
                                  has to be bigger than ETA_RATIO_LIMIT or
                                  1/ETA_RATIO_LIMIT  */
int TOTAL_NUMBER_OF_VERTICES = 0; /* total number of the grid vrtices */
int TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 0; /* total number of the grid
                                 vertices which are not EMPTY_SPACE */
int TOTAL_NUMBER_OF_TRIANGLES = 0;/* total number of the grid triangles */
double epsilon_GRID_ERROR=1.1; /* tolerance in the optical depth criterium
                                 for a new grid point */
struct _3DVector basic_icosahedron[12]; /* used for angluar integral rays
                                  (see: _GOROSTAS_create_integral_directions)*/
int T_error_neighbors;  /* 0=no adjacent vertexes are involved in checking 
                                  their error for skipping rad.transfer calculation */
double T_error_tolerance = 0.0; /* if INPUT.SKIP_POINT is 1 or 2 then this 
                                   number shows the temperature error below 
                                   which the radiative transfer is not required*/
double R_allowed_range_low, R_allowed_range_high; /* if INPUT.SKIP_POINT_R is
                                   3 or 4 then this is the range of radii for
                                   radiative transfer: low<=R<=high */
double T_allowed_range_low, T_allowed_range_high; /* if INPUT.SKIP_POINT_T is
                                   5 or 6 then this is the range of temperatures for
                                   radiative transfer: low<=T<=high */
int STEP_FUNCTION_HAZARD_AVOIDANCE=FALSE; /* a sudden step in the dust density can
                                    produce a grid point during the grid generation
                                   but indicate this point as outside the dust later
                                   on due to small errors in the floating point
                                   arithmetics. This flag shows when the dust
                                  density has to take care of this in
                                  _ZELENJAK_Dust_Density.c */
/* see _GOROSTAS_create_integral_direction subroutinefor this: */ 
double _GOROSTAS_cos_phi[_GOROSTAS_N_phi_grid+1],
       _GOROSTAS_sin_phi[_GOROSTAS_N_phi_grid+1];

/***** Luminosity/Flux calculation ************/
double Output_Lambda_Image = -1.0; /* the wavelength at which we need an image
                                 (comes with the flag -L ) */
double Output_Theta_Image = -1.0; /* the inclination angle at which we need
                                 an image (comes with the flag -A) */
/*** Image plotting variables (output) ********/
int PLOT_IMAGE_SIZE_X = 400; /* size of the plotted image (the GIF image will */
int PLOT_IMAGE_SIZE_Y = 400; /* be lager. This is just the part with the data)*/

/*** iterations *******************************/
int TOTAL_NUMBER_OF_ITERATIONS = 1; /* the total number of iterations:
				       default=1 or specified by -n flag */
int GLOBAL_ITERATION_COUNTER;
int TOTAL_NUMBER_OF_GRID_BINS = 1; /* the total number of steps within
		     one iteration step based on the vertex temperature:
				       default=1 or specified by -k flag */
int GLOBAL_BIN_COUNTER;
int INITIAL_BIN_COUNTER = -1; /* set by -K flag, shows from
                     which temperature or radial bin to start the first iteration */
double *Temperature_Bins; /* if TOTAL_NUMBER_OF_GRID_BINS>1 then
                    _GOROSTAS_Find_Temperature_Bins (flag -k) creates these bins 
                       according to the distribution of vertex temperatures */
double *Radial_Bins; /* if TOTAL_NUMBER_OF_GRID_BINS>1 then
                       _GOROSTAS_Find_Radial_Bins (flag -j) creates these bins 
                       according to the distribution of vertex radii */
int Bin_Type; /* method of bin selection: 0=none, 1=temperature, 2=radii */
int MAX_LAMBDA_PER_PROCESS = 5; /* maximum number of rad.trans. matrices 
                                   calculated simultaneously (flag -M)*/

/* variables for -403 flag: output temperature array */
int TEMPERATURE_ARRAY_TYPE; /* flag showing: 1=Cartesian 2=Polar coordinates
                                      3=Polar with log(r)*/
double X1_403, X2_403, Z1_403,Z2_403; /* the temperature box*/
int Nx_403, Nz_403; /* grid resolution */ 
/* variables for -404 flag: */
double X_404=1.0, Z_404=0.0;

/* random generator seed */
long seed;

#if (LEL_RADTRANS_N && LEL_RADTRANS_L)
  /* variables for _VILOZMAJ_Radiative_Transfer_Solver  */
  double Rad_Trans_Matrix[LEL_RADTRANS_L][LEL_RADTRANS_N][LEL_RADTRANS_N];
  double Rad_Trans_Emission[LEL_RADTRANS_L][LEL_RADTRANS_N]; 
#endif

/***********************************************************************/
/********************** Body of the PERUN block *************************/
main(int argc, char *argv[]) 
{
  char Root_Directory[256]; /* directory from which the code has been started*/
  FILE *GLOBAL_FILE;        /* pointer to the main (global) input file */
  char Main_Dusty_File[256];/* the main (global) input file */
  FILE *DUMP_FILE;          /* the flag -s 202 dump file */
  char Dump_File_name[256];
  int  err = 0;             /* error flag used for run-time checking */
  int  err_location;        /* error flag used for location of the error */
  int  message = 0;         /* debugger flag used for checking of the code */
  int  message_location;    /* debugger flag used for location of the message*/

  double theta_view; /* viewing angle (direction of the observer) */
  int TOTAL_NUMBER_OF_IMAGE_CELLS; /* number of cells in an 2D image */
  struct _image_cell_structure *Image; /* 2D image array */
  struct _image_cell_structure *Image_cell;
  int theta_counter; 
  int lambda_k, lambda_kk; /* lambda grid counters */
  int dust_i; /* dust component counter */ 
  struct density_parameters *temp_density_parameter;

#ifdef _2DD_MPI_
  int tag = 0;
  MPI_Status Status;
#endif

 /* un-comment this if stdout has to be redirected */
 /* stdout = fopen("STDOUT_LELUYA.txt","w"); */
 
#if (LEL_RADTRANS_N)
      printf("WARNING: LEL_RADTRANS_N=%d\n",LEL_RADTRANS_N);
#endif
#if (LEL_RADTRANS_L)
      printf("WARNING: LEL_RADTRANS_L=%d\n",LEL_RADTRANS_L);
#endif

  /* some global constants */
  SQRT3 = sqrt(3.0);
  SQRT3_2 = SQRT3/2.0;
  PI_SQRT3 = PI/SQRT3;
#ifdef _2DD_MPI_
    sprintf(Main_Dusty_File,"%s/Leluya.inp",_LELUYA_MPIRUN_PATH);
#else
    sprintf(Main_Dusty_File,"%s/Leluya.inp",_LELUYA_RUN_PATH);
#endif


  /* put default values into the flags */
  DEBUG.SCREEN_MESSAGE = 0;
  DEBUG.STOP = 0;
  DEBUG.CONTINUE = 0;
  DEBUG.NEW_POINT_X = -1.0;
  DEBUG.NEW_POINT_Z = -1.0;
  DEBUG.ISOTHERM_BENDING = -1.0;
  DEBUG.MAX_TRIANGLE_ANGLE = -1.0;
  DEBUG.AZIMUTHAL_GRID = -1;
  DEBUG.AZIMUTHAL_GRID1 = 5.0;
  DEBUG.DENSITY_UPDATE = -1;
  DEBUG.DENSITY_UPDATE_M = -1;
  DEBUG.DENSITY_UPDATE_R = -1;
  DEBUG.INPUT_DENSITY_UPDATE = -1;
  DEBUG.INPUT_DENSITY_UPDATE_M = -1;
  DEBUG.INPUT_DENSITY_UPDATE_R = -1;
  TOTAL_NUMBER_OF_ITERATIONS = 1;
  Bin_Type = 0;
  INPUT.SKIP_POINT = FALSE;
  INPUT.SKIP_POINT_R = FALSE;
  INPUT.SKIP_POINT_T = FALSE;
  T_error_tolerance = 0.0;
  T_error_neighbors = 0;
  R_allowed_range_low = 0.0;
  R_allowed_range_high = 1E30; /* just a large number */
  T_allowed_range_low = 0.0;
  T_allowed_range_high = 1E6; /* just a large number */
  INPUT.TAU_LIMIT = -1.0;
  INPUT.TAU_SURFACE_SCALE = -1.0;
  INPUT.LAMBDA_EMISSION_ONLY = -1.0;
  CELL_SIZE_LIMIT=0.05;
  TOTAL_NUMBER_OF_GRID_BINS = 1;
  INITIAL_BIN_COUNTER = -1;
  MAX_LAMBDA_PER_PROCESS = 5;
  INPUT.DEEPLY_EMBEDDED = FALSE;
  INPUT.POINT_SOURCE_LIMIT=0.05;
  INPUT.TEMPERATURE_ACCELERATE_min_err = -1;
  INPUT.TEMPERATURE_ACCELERATE_scale = 1;
  Grid_Tolerance_Tau_Factor=0.02;
  Tau_Min_Grid=0.1;
  Max_Dust_Temp = 1000000.0; /* just a large number, to make sure it is activated */

  Density_2D_Grid.lambda = -1;
  Density_2D_Grid.dust = -1;
  Density_2D_Grid.TOTAL_NUMBER_OF_VERTICES=-1;
  Density_2D_Grid.dust_types = NULL;
  Density_2D_Grid.density = NULL;
  Density_2D_Grid.triangle = NULL;

  initial_radial_array = NULL;
  initial_energy_array = NULL;
  initial_temperature_array = NULL;

  /* pre-calculate cos(phi) and sin(phi) for 
     _GOROSTAS_create_integral_direction subroutine */
  for (theta_counter=0; theta_counter<=_GOROSTAS_N_phi_grid;theta_counter++){ 
    _GOROSTAS_cos_phi[theta_counter] = cos(theta_counter*PI/_GOROSTAS_N_phi_grid);
    _GOROSTAS_sin_phi[theta_counter] = sin(theta_counter*PI/_GOROSTAS_N_phi_grid);
  }

  /* initialize the basic_icosahedron */
  basic_icosahedron[0].y = sqrt(2.0/(5.0+sqrt(5.0)));
  basic_icosahedron[0].x = basic_icosahedron[0].y/sqrt(3.0);
  basic_icosahedron[0].z = sqrt(1.0-4.0*basic_icosahedron[0].y*basic_icosahedron[0].y/3.0);

  basic_icosahedron[1].x = basic_icosahedron[0].x;
  basic_icosahedron[1].y = - basic_icosahedron[0].y;
  basic_icosahedron[1].z = basic_icosahedron[0].z;

  basic_icosahedron[2].x = - basic_icosahedron[0].x;
  basic_icosahedron[2].y = basic_icosahedron[0].y;
  basic_icosahedron[2].z = - basic_icosahedron[0].z;

  basic_icosahedron[3].x = - basic_icosahedron[0].x;
  basic_icosahedron[3].y = - basic_icosahedron[0].y;
  basic_icosahedron[3].z = - basic_icosahedron[0].z;

  basic_icosahedron[4].x = (sqrt(5.0)+1.0)*basic_icosahedron[0].y/sqrt(3.0);
  basic_icosahedron[4].y = 0.0;
  basic_icosahedron[4].z = (sqrt(5.0)-1.0)*basic_icosahedron[0].y*0.5/sqrt(3.0);

  basic_icosahedron[5].x = basic_icosahedron[0].y * 2.0/sqrt(3.0);
  basic_icosahedron[5].y = 0.0;
  basic_icosahedron[5].z = - basic_icosahedron[0].z;

  basic_icosahedron[6].x = - basic_icosahedron[5].x;
  basic_icosahedron[6].y = 0.0;
  basic_icosahedron[6].z = basic_icosahedron[0].z;

  basic_icosahedron[7].x = - basic_icosahedron[4].x;
  basic_icosahedron[7].y = 0.0;
  basic_icosahedron[7].z = - basic_icosahedron[4].z;

  basic_icosahedron[8].x = 0.5*basic_icosahedron[4].x;
  basic_icosahedron[8].y = basic_icosahedron[8].x*sqrt(3.0);
  basic_icosahedron[8].z = - basic_icosahedron[4].z;

  basic_icosahedron[9].x = basic_icosahedron[8].x;
  basic_icosahedron[9].y = - basic_icosahedron[8].y;
  basic_icosahedron[9].z = - basic_icosahedron[4].z;

  basic_icosahedron[10].x = - basic_icosahedron[8].x;
  basic_icosahedron[10].y = basic_icosahedron[8].y;
  basic_icosahedron[10].z = basic_icosahedron[4].z;

  basic_icosahedron[11].x = - basic_icosahedron[8].x;
  basic_icosahedron[11].y = - basic_icosahedron[8].y;
  basic_icosahedron[11].z = basic_icosahedron[4].z;

#ifndef _2DD_MPI_
  /***** Analyze the input flags and find the name of global input file 
         (this function will fix the flags in DEBUG) */
  err=_LEPAMARA__Analyze_Input_Flags(argc, argv, Root_Directory, Main_Dusty_File);
  if (err) {
    err_location = 1;
    _PESJANEK__Error_Message (err_location,err);
    exit(1);
  }

  /* stop here if that's what you want */
  if (DEBUG.STOP==1) {
    err_location = 2;
    err = 1;
    _PESJANEK__Error_Message (err_location,err);
    exit(1);
  }

  my_MPI_rank=0; total_MPI_process=1;

/*
  DEBUG.CONTINUE = 203;
  DEBUG.SCREEN_MESSAGE = 1;
  DEBUG.STOP=302;
  TOTAL_NUMBER_OF_ITERATIONS = 1;
  INPUT.SKIP_POINT = 0; T_error_tolerance=0.0;
  INPUT.TAU_LIMIT = 100.0;
  TOTAL_NUMBER_OF_GRID_BINS = 1;
  Bin_Type = 0;
  INITIAL_BIN_COUNTER = -1;
  INPUT.TAU_SURFACE_SCALE = -1.0;
  INPUT.DEEPLY_EMBEDDED = FALSE;
  INPUT.POINT_SOURCE_LIMIT=0.05;
*/

#else 
  /* Start parallel implementation */
  /* MPI_argc=argc; MPI_argv=argv; */
  /* argc=0; argv=NULL; */

  /*** initial grid generation */
  /*
  DEBUG.STOP=202;
  DEBUG.SCREEN_MESSAGE = 1;
  DEBUG.AZIMUTHAL_GRID = 1;
  DEBUG.AZIMUTHAL_GRID1 = 1.0;
  Grid_Tolerance_Tau_Factor=0.0;
  Tau_Min_Grid=0.0;
  INPUT.POINT_SOURCE_LIMIT=0.0;
  */

/* a typical run */
/*
  DEBUG.CONTINUE = 202;
  DEBUG.SCREEN_MESSAGE = 1;
  DEBUG.STOP=302;
  TOTAL_NUMBER_OF_ITERATIONS = 10;
  INPUT.SKIP_POINT = 0; T_error_tolerance=0.003;
  INPUT.TAU_LIMIT = 40.0;
  TOTAL_NUMBER_OF_GRID_BINS = -1;
  Bin_Type = 2;
  INITIAL_BIN_COUNTER = -1;
  MAX_LAMBDA_PER_PROCESS = 93;
  INPUT.LAMBDA_EMISSION_ONLY = 10.0;
  INPUT.TAU_SURFACE_SCALE = -1.0;
  INPUT.DEEPLY_EMBEDDED = FALSE;
  INPUT.POINT_SOURCE_LIMIT=0.0;
*/

/* luminosity conservation */
/*
  DEBUG.CONTINUE = 4;
  DEBUG.SCREEN_MESSAGE = 1;
*/

/* images and SED */
/*
  DEBUG.CONTINUE = 401;
  Output_Theta_Image = 89.999;
  DEBUG.SCREEN_MESSAGE = 1;
*/

/* rad. pressure  */
/*
  DEBUG.CONTINUE = 405;
  DEBUG.SCREEN_MESSAGE = 1;
  X1_403=0.0; X2_403=2.0; Z1_403=0.0; Z2_403=2.0;
  Nx_403=40; Nz_403=40;
*/



  MPI_Init(&argc, &argv);
  MPI_communicator = MPI_COMM_WORLD;
  MPI_Comm_rank(MPI_communicator, &my_MPI_rank);
  MPI_Comm_size(MPI_communicator, &total_MPI_process);

  /* un-comment this if stdout has to be redirected */
  sprintf(Dump_File_name,"STDOUT_LELUYA_%.3d.txt",my_MPI_rank);
   stdout = fopen(Dump_File_name,"w");
   fflush(stdout); 

  err=_LEPAMARA__Analyze_Input_Flags(argc, argv, Root_Directory, Main_Dusty_File);
  if (err) {
    err_location = 1;
    _PESJANEK__Error_Message (err_location,err);
    exit(1);
  }

#endif

  if (INPUT.DEEPLY_EMBEDDED && INPUT.TAU_LIMIT<=0.0)
    INPUT.TAU_LIMIT = 20.0;
  if (INPUT.DEEPLY_EMBEDDED && INPUT.SKIP_POINT != 2) {
    INPUT.SKIP_POINT = 2; T_error_tolerance = 0.003; }

  /* initialize the random generator */
  seed = -6546145; /* put whatever integer number here... */
  seed = seed*(my_MPI_rank+1);
  ran2(&seed);


  if (INPUT.TAU_LIMIT>0.0  && DEBUG.SCREEN_MESSAGE >= 2)
    printf("INPUT TAU LIMIT = %lf\n",INPUT.TAU_LIMIT);

  /* start here with the iteration loop */
  if (TOTAL_NUMBER_OF_GRID_BINS<=0) TOTAL_NUMBER_OF_GRID_BINS=1;
  GLOBAL_ITERATION_COUNTER=0;
  do {

  if (TOTAL_NUMBER_OF_ITERATIONS>1 && my_MPI_rank==0) {
    printf("##############################################\n");
    printf("########## iteration %.2d #####################\n",
	   GLOBAL_ITERATION_COUNTER);
  }
  if (TOTAL_NUMBER_OF_ITERATIONS>1) DEBUG.STOP=302;

  /* start here with the temperature-bin loop */
  GLOBAL_BIN_COUNTER=0;
  /* check if -K flag was activated: */ 
  if (INITIAL_BIN_COUNTER>=0 && GLOBAL_ITERATION_COUNTER==0)
     GLOBAL_BIN_COUNTER=INITIAL_BIN_COUNTER;
  do {

  if (TOTAL_NUMBER_OF_GRID_BINS>1 && my_MPI_rank==0) {
    printf("##############################################\n");
    if (Bin_Type==1) 
      printf("########## temperature bin %.2d #####################\n",
             GLOBAL_BIN_COUNTER);
    if (Bin_Type==2)
      printf("########## radial bin %.2d #####################\n",
             GLOBAL_BIN_COUNTER);
  }
  if (TOTAL_NUMBER_OF_GRID_BINS>1)  DEBUG.STOP=302;

  /* continue according to DEBUG.CONTINUE flag */ 
  switch (DEBUG.CONTINUE) {

  case 0:
  case 1:
  case 100:
    /*------------------------ INPUT ---------------------------*/
    /***** open the global input file */
    GLOBAL_FILE=fopen(Main_Dusty_File,"rt");
    if (GLOBAL_FILE==NULL) {
      err_location = 1;
      err = 5;
      _PESJANEK__Error_Message (err_location,err);
      printf("global input file: %s\n",Main_Dusty_File);
#ifdef _2DD_MPI_
  MPI_Finalize();
#endif
      exit(1);
    }

    /* debugger messages if required by DEBUG.SCREEN_MESSAGE */
    if (DEBUG.SCREEN_MESSAGE >= 1 ) {
      message_location = 1;
      message = 1;
      _PESJANEK__Debugger_Message (message_location,message);
      printf("    Root Directory   : %s\n",Root_Directory);
      printf("    Global Input File: %s\n\n",Main_Dusty_File);
    }

    /***** read the input data and transform them into the useful form for 
           the grid generator and radiative transfer solver */
     err=_PATULJAK__Read_Input_Data (GLOBAL_FILE,Main_Dusty_File);
    /* check a possible error */
    if (err) {
      /* errors are handled in _PATULJAK__Read_Input_Data */
#ifdef _2DD_MPI_
      MPI_Finalize();
#endif
      exit(1);
    }

    /***** close the global input file ******/
    fclose(GLOBAL_FILE);

    /* create the initial temperature grid (if the grid does not exist)
       (this corresponds to status = 0 or status = 1 in PERUNIKA_main) */
    if (DEBUG.CONTINUE<=2 || DEBUG.CONTINUE == 100){
      initial_energy_array=d3tensor(0,_INITIAL_GRID_ARRAY_r-1,
                                    0,_INITIAL_GRID_ARRAY_theta-1,
                                    0,lambda_Max-1);
      initial_temperature_array=d3tensor(0,Max_Dust_Type-1,
                                         0,_INITIAL_GRID_ARRAY_r-1,
                                         0,_INITIAL_GRID_ARRAY_theta-1);
      initial_radial_array = dvector(0,_INITIAL_GRID_ARRAY_r-1);
      err = _PERUNIKA_initialize_initial_arrays(0);
      if (err!=0){
#ifdef _2DD_MPI_
        MPI_Finalize();
#endif
        exit(1);
      } 
    }
    err = _LEPAMARA_Find_Initial_Eta_Denominator();
    if (err) {
#ifdef _2DD_MPI_
      MPI_Finalize();
#endif
      exit(1);
    }

    /* stop here if that's what you want */
    if (DEBUG.STOP==2 || DEBUG.STOP==200) {
      _PESJANEK_Dump_Data_STOP2 ();
      err_location = 2;
      err = 2;
      _PESJANEK__Error_Message (err_location,err);
#ifdef _2DD_MPI_
      MPI_Finalize();
#endif
      exit(1);
    }


  case 2:
  case 200:
  case 201:
  case 202:
  case 203:
  case 204:
  case 205:
  case 206:
  case 207:
  case 208:
  case 209:
  case 210:

#ifdef _2DD_MPI_
    if (DEBUG.CONTINUE !=100){ 
      /*if (my_MPI_rank>0) MPI_Recv(&(DEBUG.CONTINUE),1,MPI_INT,my_MPI_rank-1,
	tag,MPI_communicator,&Status);*/
      if (DEBUG.SCREEN_MESSAGE >= 2 ) {
        printf("process#%.3d: read the input data files\n",my_MPI_rank);
      }
    }
#endif

    if (DEBUG.SCREEN_MESSAGE >= 1 )
      printf("process#%.3d: read the input data\n",my_MPI_rank);

    /*------------------- INITIAL GRID ----------------------------*/
    /* read the dumped data */
    if (DEBUG.CONTINUE !=100)
    if (DEBUG.CONTINUE == 2 || DEBUG.CONTINUE == 200 || DEBUG.CONTINUE == 201
        || DEBUG.CONTINUE == 202 || DEBUG.CONTINUE == 203 
        || DEBUG.CONTINUE == 204 || DEBUG.CONTINUE == 205
        || DEBUG.CONTINUE == 206 || DEBUG.CONTINUE == 207
        || DEBUG.CONTINUE == 208 || DEBUG.CONTINUE == 209
	|| DEBUG.CONTINUE == 210) {
#ifdef _2DD_MPI_
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP2.2dd",_LELUYA_MPIRUN_PATH);
  if ( (TOTAL_NUMBER_OF_ITERATIONS>1 && GLOBAL_ITERATION_COUNTER>0) ||
       (TOTAL_NUMBER_OF_GRID_BINS>1 && GLOBAL_BIN_COUNTER>0) ) {
    lambda_k = GLOBAL_BIN_COUNTER - 1;
    if (lambda_k<0) lambda_k=TOTAL_NUMBER_OF_GRID_BINS - 1;
    lambda_kk = GLOBAL_ITERATION_COUNTER;
    if (GLOBAL_BIN_COUNTER==0) lambda_kk=GLOBAL_ITERATION_COUNTER-1;
    if (lambda_kk<0) lambda_kk=0;
    sprintf(Dump_File_name,"%s/LELUYA.dump.STOP3_%.2d_%.2d.2dd",
            _LELUYA_MPIOUT_PATH,lambda_k,lambda_kk);
  }
#else
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP2.2dd",_LELUYA_RUN_PATH);
  if ( (TOTAL_NUMBER_OF_ITERATIONS>1 && GLOBAL_ITERATION_COUNTER>0) ||
    (TOTAL_NUMBER_OF_GRID_BINS>1 && GLOBAL_BIN_COUNTER>0) )
    sprintf(Dump_File_name,"%s/LELUYA.dump.STOP3.2dd",_LELUYA_OUTPUT_PATH);
#endif
      if (DEBUG.SCREEN_MESSAGE >= 2 ) printf("FILE=%s\n",Dump_File_name);
      DUMP_FILE = fopen (Dump_File_name,"r");
      if (DUMP_FILE==NULL) {
        printf("---- WARNING!! Can not opet the dump file:\n");
        printf("               %s\n",Dump_File_name);
	err = 1;
      }
      else err = _LEPAMARA_Read_Data_CONTINUE2(DUMP_FILE);
      fclose (DUMP_FILE);
      if (err) {
#ifdef _2DD_MPI_
	MPI_Finalize();
#endif
	exit (1);
      }
    }

    fflush(stdout);
 
    /* create the initial grid */
    if (DEBUG.CONTINUE <= 2 || DEBUG.CONTINUE == 200 || DEBUG.CONTINUE == 100){
      err = _PERUNIKA_main(0);
    }
    if (DEBUG.CONTINUE == 201) {
      err = _PERUNIKA_main(1);
    }

    if (DEBUG.CONTINUE == 202 || DEBUG.CONTINUE == 203 || 
        DEBUG.CONTINUE == 204 || DEBUG.CONTINUE == 205 ||
        DEBUG.CONTINUE == 206 || DEBUG.CONTINUE == 207 ||
        DEBUG.CONTINUE == 208 || DEBUG.CONTINUE == 209 ||
	DEBUG.CONTINUE == 210) {
#ifdef _2DD_MPI_
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_MPIRUN_PATH);
  if ( (TOTAL_NUMBER_OF_ITERATIONS>1 && GLOBAL_ITERATION_COUNTER>0) ||
       (TOTAL_NUMBER_OF_GRID_BINS>1 && GLOBAL_BIN_COUNTER>0) ) {
    lambda_k = GLOBAL_BIN_COUNTER - 1;
    if (lambda_k<0) lambda_k=TOTAL_NUMBER_OF_GRID_BINS - 1;
    lambda_kk = GLOBAL_ITERATION_COUNTER;
    if (GLOBAL_BIN_COUNTER==0) lambda_kk=GLOBAL_ITERATION_COUNTER-1;
    if (lambda_kk<0) lambda_kk=0;
    sprintf(Dump_File_name,"%s/LELUYA.dump.STOP302_%.2d_%.2d.2dd",
            _LELUYA_MPIOUT_PATH,lambda_k,lambda_kk);
  }
#else
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_RUN_PATH);
  if ( (TOTAL_NUMBER_OF_ITERATIONS>1 && GLOBAL_ITERATION_COUNTER>0) ||
    (TOTAL_NUMBER_OF_GRID_BINS>1 && GLOBAL_BIN_COUNTER>0) )
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP302.2dd",_LELUYA_OUTPUT_PATH);
#endif

      if (DEBUG.SCREEN_MESSAGE >= 2 ) printf("FILE=%s\n",Dump_File_name);
      DUMP_FILE = fopen (Dump_File_name,"r");
      if (DUMP_FILE==NULL) {
        printf("---- WARNING!! Can not opet the dump file:\n");
        printf("               %s\n",Dump_File_name);
	err = 1;
      }
      else err = _LEPAMARA_Read_Data_CONTINUE202 (DUMP_FILE);
      fclose (DUMP_FILE);
    }
    if (err) {
#ifdef _2DD_MPI_
      MPI_Barrier(MPI_communicator);
      MPI_Finalize();
#endif
      exit(1);
    }

#ifdef _2DD_MPI_
    /*if (my_MPI_rank<total_MPI_process-1) 
    MPI_Send(&(DEBUG.CONTINUE),1,MPI_INT,my_MPI_rank+1,tag,MPI_communicator);*/
  MPI_Barrier(MPI_communicator);
#endif
    /* if (DEBUG.CONTINUE == 203) then check if we need sublimation surface
       correction */
    if ((DEBUG.CONTINUE == 203 &&  GLOBAL_ITERATION_COUNTER>0 &&
	 GLOBAL_BIN_COUNTER==0) || DEBUG.CONTINUE == 204) {
      if (DEBUG.CONTINUE == 2044)  _PERUNIKA_remove_hot_dust(2);
      err = _PERUNIKA_main(2);
    }

    if (DEBUG.CONTINUE == 2055) {
      _PERUNIKA_remove_hot_dust(2);
    }

    /* set ln(density) at pixels */
    if (DEBUG.INPUT_DENSITY_UPDATE<0 && DEBUG.DENSITY_UPDATE>=0) 
      _GOROSTAS_set_logarithmic_density_at_pixels(0);

    /* add a point to the grids */
    if (DEBUG.CONTINUE == 206) {
      if(DEBUG.NEW_POINT_X>=0.0) 
	err = _PERUNIKA_add_point_to_grids(DEBUG.NEW_POINT_X,DEBUG.NEW_POINT_Z);
      else {
        theta_view = -DEBUG.NEW_POINT_X*RAD1;
	err = _PERUNIKA_add_points_along_radial_line(theta_view,DEBUG.NEW_POINT_Z);
      }
    }
    /* split triangles in the grids */
    if (DEBUG.CONTINUE == 207) {
      if(DEBUG.NEW_POINT_X>=0.0) 
        err=_PERUNIKA_split_triangle_at_given_point(DEBUG.NEW_POINT_X,DEBUG.NEW_POINT_Z);
      else {
        theta_view = -DEBUG.NEW_POINT_X*RAD1;
	err = _PERUNIKA_add_points_along_radial_line(theta_view,DEBUG.NEW_POINT_Z);
      }
    }
    /* refine grid to make smooth triangles */
    if (DEBUG.CONTINUE == 208) {
      err = _PERUNIKA_smooth_grid_temperature();
    }
    if (DEBUG.CONTINUE == 209) {
      err = _PERUNIKA_smooth_grid_cells_shape();
    }
    if (DEBUG.CONTINUE == 210) {
      theta_view = DEBUG.MAX_TRIANGLE_ANGLE*RAD1;
      err = _PERUNIKA_add_points_along_radial_line(theta_view,0);
    }

  case 3:
  case 300:

    /* find grid bins for the loop */
    if (    (TOTAL_NUMBER_OF_GRID_BINS>1 && 
             GLOBAL_BIN_COUNTER==0)
         || (GLOBAL_ITERATION_COUNTER==0 &&
             GLOBAL_BIN_COUNTER==INITIAL_BIN_COUNTER)
       )  {
      if (Bin_Type==1) _GOROSTAS_Find_Temperature_Bins ();
      if (Bin_Type==2) _GOROSTAS_Find_Radial_Bins ();
    }

    if (DEBUG.CONTINUE == 202 || DEBUG.CONTINUE == 203) err = _VILOZMAJ_main(); 
    /******************* create some visualization output ********************/
  if (my_MPI_rank==0 && (DEBUG.CONTINUE == 205)) {
    /* find the max TOTAL_NUMBER_OF_VERTICES */
    lambda_kk = 0;
    for (lambda_k=1; lambda_k<lambda_Max; lambda_k++)
      if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > 
          grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES)
        lambda_kk = lambda_k;

    start_vertex_list = grid_per_lambda[lambda_kk].start_vertex;
    start_triangle_list = grid_per_lambda[lambda_kk].start_triangle;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
#ifdef _2DD_MPI_
    sprintf(Dump_File_name,
 "%s/2D_dusty.Temperature_Error.dust=%.2d.max_at_lambda=%.3d.process=%.2d",
	   _LELUYA_MPIOUT_PATH,dust_i,lambda_kk,my_MPI_rank);
#else
    sprintf(Dump_File_name,
 "%s/2D_dusty.Temperature_Error.dust=%.2d.max_at_lambda=%.3d.process=%.2d",
	   _LELUYA_OUTPUT_PATH,dust_i,lambda_kk,my_MPI_rank);
#endif
       _VODENJAK_plot_vrml(Dump_File_name,3,dust_i,lambda_kk); 
    }

    sprintf(Dump_File_name,"2DD_interactive_grid.update.max_at_lambda=%.3d",lambda_kk);
    err = _VODENJAK_plot_interactive_vrml(Dump_File_name,0,lambda_kk); 
    err = _VODENJAK_plot_PovRay(Dump_File_name,0,lambda_kk);
#ifdef _2DD_MPI_
    sprintf(Dump_File_name,"%s/2DD_Dust_surface.max_at_lambda=%.3d",_LELUYA_MPIOUT_PATH,lambda_kk);
#else
    sprintf(Dump_File_name,"%s/2DD_Dust_surface.max_at_lambda=%.3d",_LELUYA_OUTPUT_PATH,lambda_kk);
#endif

     err = _VODENJAK_plot_dust_edge_vrml(Dump_File_name,lambda_kk); 

    /* find the min TOTAL_NUMBER_OF_VERTICES */
    lambda_kk = 0;
    for (lambda_k=1; lambda_k<lambda_Max; lambda_k++)
      if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES < 
          grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES)
        lambda_kk = lambda_k;

    start_vertex_list = grid_per_lambda[lambda_kk].start_vertex;
    start_triangle_list = grid_per_lambda[lambda_kk].start_triangle;
sprintf(Dump_File_name,"2DD_interactive_grid.update.min_at_lambda=%.3d",lambda_kk);
     err = _VODENJAK_plot_interactive_vrml(Dump_File_name,0,lambda_kk); 

lambda_kk = 38;
start_vertex_list = grid_per_lambda[lambda_kk].start_vertex;
    start_triangle_list = grid_per_lambda[lambda_kk].start_triangle;
sprintf(Dump_File_name,"2DD_interactive_grid.update.min_at_lambda=%.3d",lambda_kk);
     err = _VODENJAK_plot_interactive_vrml(Dump_File_name,0,lambda_kk);

  }

    /******************* END OF: create some visualization output *************/

    /* stop here: flag -s 302 */
    if (DEBUG.STOP==302) {
      /* dump the grids in case of the master process */
      if (my_MPI_rank == 0) {
        _PESJANEK_Dump_Data_STOP2 ();
#ifdef _2DD_MPI_
   sprintf(Dump_File_name,"%s/LELUYA.dump.STOP302_%.2d_%.2d.2dd",
          _LELUYA_MPIOUT_PATH,GLOBAL_BIN_COUNTER,
          GLOBAL_ITERATION_COUNTER);
#else
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP302.2dd",_LELUYA_OUTPUT_PATH);
#endif
        DUMP_FILE = fopen (Dump_File_name,"w");
        if (DUMP_FILE==NULL) {
          printf("---- WARNING!! Can not opet the dump file:\n");
          printf("               %s\n",Dump_File_name);
        }
        else {
          _PESJANEK_Dump_Data_STOP202(DUMP_FILE);
        }
        fclose(DUMP_FILE); 
      }
#ifdef _2DD_MPI_
      MPI_Barrier(MPI_communicator);
#endif
      if (GLOBAL_ITERATION_COUNTER == TOTAL_NUMBER_OF_ITERATIONS-1) {
	printf("process#%.3d: WARNING!! The program is stopped\n",my_MPI_rank);
	printf("              due to the user's command line flag: -s 302\n");
      }
      break; 
    }

  case 4:
  case 400:
  case 401:
  case 402:
  case 403:
  case 404:
  case 405:
  case 406:
  case 407:

    if (DEBUG.CONTINUE == 4 || DEBUG.CONTINUE == 400 || DEBUG.CONTINUE == 401 ||
        DEBUG.CONTINUE == 402 || DEBUG.CONTINUE == 403 || DEBUG.CONTINUE == 404 ||
        DEBUG.CONTINUE == 405 || DEBUG.CONTINUE == 406 || DEBUG.CONTINUE == 407){
#ifdef _2DD_MPI_
    sprintf(Dump_File_name,"%s/LELUYA.dump.STOP2.2dd",_LELUYA_MPIRUN_PATH);
#else
    sprintf(Dump_File_name,"%s/LELUYA.dump.STOP2.2dd",_LELUYA_RUN_PATH);
#endif
      if (DEBUG.SCREEN_MESSAGE >= 1 ) printf("FILE=%s\n",Dump_File_name);
      DUMP_FILE = fopen (Dump_File_name,"r");
      if (DUMP_FILE==NULL) {
        printf("---- WARNING!! Can not opet the dump file:\n");
        printf("               %s\n",Dump_File_name);
	err = 1;
      }
      else err = _LEPAMARA_Read_Data_CONTINUE2(DUMP_FILE);
      fclose (DUMP_FILE);
      if (err) {
	exit (1);
      }
#ifdef _2DD_MPI_
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_MPIRUN_PATH);
#else
  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_RUN_PATH);
#endif
      if (DEBUG.SCREEN_MESSAGE >= 1 ) printf("FILE=%s\n",Dump_File_name);
      DUMP_FILE = fopen (Dump_File_name,"r");
      if (DUMP_FILE==NULL) {
        printf("---- WARNING!! Can not opet the dump file:\n");
        printf("               %s\n",Dump_File_name);
	err = 1;
      }
      else err = _LEPAMARA_Read_Data_CONTINUE202 (DUMP_FILE);
      fclose (DUMP_FILE);
      if (err) {
	exit(1);
      }
    }

    if (DEBUG.CONTINUE == 4 || DEBUG.CONTINUE == 400) {
      /* calculate the luminosity */
      err = _VILOZMAJ_Luminosity_shells ();
    }

    if (DEBUG.CONTINUE == 401) {
      /* calculate the images */
      theta_counter=0;
      while (theta_counter<=18) {

        if (Output_Theta_Image<0.0)
          theta_view = theta_counter*5.0*RAD1;
        else 
          theta_view = Output_Theta_Image*RAD1;

        TOTAL_NUMBER_OF_IMAGE_CELLS=0;
        Image = NULL;
        printf("  Inclination angle = %lf deg\n",theta_view/RAD1);
        err = _VILOZMAJ_Output_Image_Flux(theta_view,&TOTAL_NUMBER_OF_IMAGE_CELLS,
				    &Image);
        if (err==0) {
          /* destroy the image */
          Image_cell = Image;
          do {
	    if (Image_cell->tau!=NULL)
              free_dvector(Image_cell->tau,0,lambda_Max-1);
	    if (Image_cell->Intensity_em!=NULL)
	      free_dvector(Image_cell->Intensity_em,0,lambda_Max-1);
	    if (Image_cell->Intensity_sca!=NULL)
	      free_dvector(Image_cell->Intensity_sca,0,lambda_Max-1);
	    if (Image_cell->Intensity_star!=NULL)
	      free_dvector(Image_cell->Intensity_star,0,lambda_Max-1);
	    if (Image_cell->next == Image_cell) Image=NULL;
	    else {
	      Image = Image->next;
	      Image->prev = Image_cell->prev;
	      Image_cell->prev->next = Image; 
	    }
	    free(Image_cell);
	    TOTAL_NUMBER_OF_IMAGE_CELLS--;

	    Image_cell = Image;
	  } while(Image_cell!=NULL);
        } /* end of if (err==0) */

        if (Output_Theta_Image<0.0) theta_counter++;
        else theta_counter = 19;
      }
    }

    if (DEBUG.CONTINUE == 402) {
      /* calculate the radial profiles */
      err = _VILOZMAJ_Radial_Profiles ();
    }

    if (DEBUG.CONTINUE == 403) {
      /* create the output temperature array files */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	err = _VILOZMAJ_Temperature_Array (dust_i);
    }

    if (DEBUG.CONTINUE == 404) {
      /* calculate the angular flux (brigntess) distribution 
         around (X_404,Z_404) */
	err =_GOROSTAS_Intensity_Around_Point(X_404,Z_404);
    }

   if (DEBUG.CONTINUE == 405 || DEBUG.CONTINUE == 406 || DEBUG.CONTINUE == 407) {
     /* calculate the radiation pressure force */
     err = _GOROSTAS_Radiation_Pressure();
   }

  default :
    break;
  }

    /****** destroy the allocated variables ******/ 

    /* remove the density grid if it exists */
    if (DEBUG.DENSITY_UPDATE>=0)
      _GOROSTAS_set_logarithmic_density_at_pixels(-1);

    /* remove the vertex and triangle grids: */ 
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      TOTAL_NUMBER_OF_VERTICES =
                        grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
      free_dvector(grid_per_lambda[lambda_k].Ustar,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(grid_per_lambda[lambda_k].Usca,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(grid_per_lambda[lambda_k].Uem,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(grid_per_lambda[lambda_k].Utot_new,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
      start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
      if (start_vertex_list!=NULL) {
        for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) 
	  if (grid_per_lambda[lambda_kk].start_vertex==start_vertex_list)
	    {
	      grid_per_lambda[lambda_kk].start_vertex=NULL;
	      grid_per_lambda[lambda_kk].start_triangle=NULL;
	    }
        _PERUNIKA_destroy_vertex_triangle_list(start_vertex_list,
					   start_triangle_list);
      }
    }
    free_grid_per_lambda_vector(grid_per_lambda,0,lambda_Max-1);

    /* remove the dust density parameters: */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_density_parameter=Dust_Density_Param[dust_i].next;
      while(temp_density_parameter!=NULL) {
	Dust_Density_Param[dust_i].next=temp_density_parameter->next;
	free(temp_density_parameter);
        temp_density_parameter=Dust_Density_Param[dust_i].next;
      }
    }
    free_density_parameters_vector(Dust_Density_Param,0,Max_Dust_Type-1);
   
    /* remove the dust sigmas, names, and sublimation temperatures: */
    free_dmatrix(Sigma_Dust_Abs,0,Max_Dust_Type-1,0,lambda_Max-1);
    free_dmatrix(Sigma_Dust_Sca,0,Max_Dust_Type-1,0,lambda_Max-1);
    free_dvector(Dust_Sublim_Temp,0,Max_Dust_Type-1);
    free_cmatrix(Dust_Component_Name,0,Max_Dust_Type-1,
                              0,_DUST_COMPONENT_NAME_SIZE_-1);

    /* remove the source flux and the lambda grid: */
    free_dvector(Source_Flux,0,lambda_Max-1);
    free_dvector(Lambda_Grid,0,lambda_Max-1);
    lambda_Max = 0; Max_Dust_Type = 0;

    /* if the iteration loop is active, save the current status */
    if (TOTAL_NUMBER_OF_ITERATIONS > 1 && my_MPI_rank == 0)
#ifdef _2DD_MPI_
      sprintf(Dump_File_name,"%s/iteration=%.2d.2dd",
	    _LELUYA_MPIOUT_PATH,GLOBAL_ITERATION_COUNTER);
#else
      sprintf(Dump_File_name,"%s/iteration=%.2d.2dd",
	    _LELUYA_OUTPUT_PATH,GLOBAL_ITERATION_COUNTER);
#endif
    if (TOTAL_NUMBER_OF_GRID_BINS > 1 && my_MPI_rank == 0)
#ifdef _2DD_MPI_
      sprintf(Dump_File_name,"%s/iteration=%.2d_Tbin=%.2d.2dd",
	    _LELUYA_MPIOUT_PATH,GLOBAL_ITERATION_COUNTER,
            GLOBAL_BIN_COUNTER);
#else
      sprintf(Dump_File_name,"%s/iteration=%.2d_Tbin=%.2d.2dd",
	    _LELUYA_OUTPUT_PATH,GLOBAL_ITERATION_COUNTER,
            GLOBAL_BIN_COUNTER);
#endif
    if ((TOTAL_NUMBER_OF_ITERATIONS > 1 ||
	TOTAL_NUMBER_OF_GRID_BINS > 1) && my_MPI_rank == 0) {
      DUMP_FILE = fopen (Dump_File_name,"w");
      fprintf(DUMP_FILE,"iteration #%.2d temperature bin=%.2d\n",
              GLOBAL_ITERATION_COUNTER,GLOBAL_BIN_COUNTER);
      if (TOTAL_NUMBER_OF_GRID_BINS > 1) {
	if (Bin_Type==1)
fprintf(DUMP_FILE,"##Rad.Trans. for vertices between T<=%lfK and >%lfK\n",
	       Temperature_Bins[GLOBAL_BIN_COUNTER],
	       Temperature_Bins[GLOBAL_BIN_COUNTER+1]);
	if (Bin_Type==2)
fprintf(DUMP_FILE,"##Rad.Trans. for vertices between R<=%lf and >%lf\n",
	       Radial_Bins[GLOBAL_BIN_COUNTER+1],
	       Radial_Bins[GLOBAL_BIN_COUNTER]);
      }
      fprintf(DUMP_FILE,"  Psi=%E\n",Psi);
      fprintf(DUMP_FILE,"  Eta denominator=%E\n",Eta_Denominator);
      fclose( DUMP_FILE );
    }

#ifdef _2DD_MPI_
    if (TOTAL_NUMBER_OF_ITERATIONS>1 || TOTAL_NUMBER_OF_GRID_BINS>1)
      MPI_Barrier(MPI_communicator);
    /* remove the previous iteration files*/
    if (my_MPI_rank==0) 
	if ( (TOTAL_NUMBER_OF_ITERATIONS>1 && GLOBAL_ITERATION_COUNTER>0) ||
	     (TOTAL_NUMBER_OF_GRID_BINS>1 && GLOBAL_BIN_COUNTER>0) ) {
	  lambda_k = GLOBAL_BIN_COUNTER - 1;
	  if (lambda_k<0) lambda_k=TOTAL_NUMBER_OF_GRID_BINS - 1;
	  lambda_kk = GLOBAL_ITERATION_COUNTER;
	  if (GLOBAL_BIN_COUNTER==0) lambda_kk=GLOBAL_ITERATION_COUNTER-1;
	  if (lambda_kk<0) lambda_kk=0;
	  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP3_%.2d_%.2d.2dd",
		  _LELUYA_MPIOUT_PATH,lambda_k,lambda_kk);
	  /* remove(Dump_File_name); */
	  sprintf(Dump_File_name,"%s/LELUYA.dump.STOP302_%.2d_%.2d.2dd",
		  _LELUYA_MPIOUT_PATH,lambda_k,lambda_kk);
	  /* remove(Dump_File_name); */
	}
#endif

    /* end of the temperature-bin loop */ 
    GLOBAL_BIN_COUNTER++;
    /* destroy the grid bins */
    if(TOTAL_NUMBER_OF_GRID_BINS>1 && 
       GLOBAL_BIN_COUNTER==TOTAL_NUMBER_OF_GRID_BINS) {
      if (Bin_Type==1) free_dvector(Temperature_Bins,0,TOTAL_NUMBER_OF_GRID_BINS);
      if (Bin_Type==2) free_dvector(Radial_Bins,0,TOTAL_NUMBER_OF_GRID_BINS);
    }
  } while (GLOBAL_BIN_COUNTER<TOTAL_NUMBER_OF_GRID_BINS);


    /* end of the iteration loop */ 
    GLOBAL_ITERATION_COUNTER++;
  } while (GLOBAL_ITERATION_COUNTER<TOTAL_NUMBER_OF_ITERATIONS);

  /* END THE PROGRAM */ 
  err_location = 1;
  err = 6;
      /* 
     _PESJANEK__Error_Message (err_location,err);
      */
  printf ("process#%.3d: END OF THE PERUN BLOCK: EXIT\n",my_MPI_rank);
#ifdef _2DD_MPI_
  MPI_Barrier(MPI_communicator);
  MPI_Finalize();
#endif

} /* end of the PERUN block */
