/* LELUYA - LEPAMARA program block */
/* Functions for analysis of the input flags and fixing DEBUGGING flags */

/* Here is the list of all flags used by LELUYA: 
   -d n: debug : run-time messages on the screen about the program status 
         ("n" shows the level of debug messages)
   -s n: stop : stop the running code at position specified by the number "n"
   -c n: continue: continue the calculation at position specified by 
         the number "n"
   -c 202: Radiative transfer, no grid updates
   -c 203: Radiative transfer, with grid updates
   -c 204: No radiative transfer, only grid updates (use it with -s 302)
   -c 205: No radiative transfer, no grid updates (used only for visualization)
   -c 206 x z : Add new grid point at (x,z) at all wavelengths
   -c 206 a p : (IMPORTANT: a<0) Add new points along the radial line of
                angle |a|deg, if at least one side of the encountered triangle
                is bigger than p*size_limit
   -c 207 x z : Split the triangle at (x,z) at all wavelengths
   -c 207 a p : (IMPORTANT: a<0) Split triangles along the radial line of
                angle |a|deg, if at least one side of the encountered triangle
                is bigger than p*size_limit
   -c 208 f   : use isotherms to smooth the grid, make new vertices if the local
                isotherms bends more than f degrees
   -c 209 f   : find triangles which have one if its angles > f and split them
   -c 210 a f1 f2: cuts through the grid along the radial line of angle a from
                radius f1 to f2 
   -c 402 : radial (or angular) temperature scan. It comes together with -A f
         and -L f. If -A f<0 than it is an angular scan at radius f.
   -c 403 n x1 x2 z1 z2 Nx Nz : Creates an output file with the temperatures
         within the box [x1,z1][x2,z2] with [Nx,Nz] resolution
         n = 1 Cartesian grid, n=2 Polar grid, n=3 Polar rid with log(r)
   -c 404 x z : calculate angular distribution of flux around the point (x,z)
         and save this into three files: star, dust emission, dust scatter, and
         total flux
   -c 405 x1 x2 z1 z2 Nx Nz: calculates the radiation pressure force on a
         rectangular grid [Nx,Nz] over the box [x1,z1][x2,z2]
   -c 405 x1 x2 z1 z2 N : (IMPORTANT: N<0) calculates the radiation pressure 
         force on a log-grid with N points along the radial line from (x1,z1) 
         to (x2,z2)  
   -c 406: calculates the radiation pressure force at each grid point
   -c 407 x1 x2 z1 z2 Nx Nz: calculates the flux (see -c 405)
   -c 407 x1 x2 z1 z2 N : (IMPORTANT: N<0) calculates the flux (see -c 405)
   -L f: image at f: create a theoretical image at the wavelength f 
         (works with the flag -c 401)
   -A f: image at f: create a theoretical image at the inclination angle f 
         (works with the flag -c 401)
   -n n: number of iterations (works only with -c 202 or -c 203, it creates
         -s 302)
   -t n f: if the temperature error at a grid point, or n levels around it, 
         is less than a given value f, this point does not need radiative
         transfer calculation (this does not apply to the first iteration
   -T n f: the same as -t, except that this applies to all iterations
   -E f: the rad.trans. integrals stop when the optical depth gets larger
         than f. The flag is off when f<0.0  
   -G f: this specifies CELL_SIZE_LIMIT=f, and it works ONLY IF -c flag is 
         less than 2!! Default CELL_SIZE_LIMIT=0.05
   -j n: split the radiative transfer job (one iteration) into n substeps
         according to the vertex distance to (0,0)
   -J n: shows from which radial bin to start the first iteration
   -k n: split the radiative transfer job (one iteration) into n substeps
         according to the vertex temperatures
   -K n: shows from which temperature or radial bin to start the first iteration
   -M n: maximum number of rad.trans. matrices calculated simultaneously 
   -O f: for wavelengths longer than f we do not invert 
         the scattering matrix, thus only dust thermal emission exists.
   -S f: activates surface grid recognition. f controls the density of this 
         surface grid. 
   -W  : activates skipping deeply embedded vertexes. It requires -E and -T to 
         be active. If they are not then use the default values: -E 20.0 -T 0.003
   -h f: if the apparent angular radius of the source at some point is smaller
         than f then we use the point source approximation
         (if f<0.0 then we copy the stellar intensity form previous iteration)
   -b f1 f2: acceleration method where the new temperature is: 
         Tnew=Told(1+f2*DeltaT) for |DeltaT|>=f1 where DeltaT is the temperature 
         error.
   -Z n: Azimuthal triangular grid, where grid points are located on concentric 
         rings around the star. n=0 is the default griding method.
         existing Z: 1,2,3 
   -g f1 f2: control over the grid copies based on the optical depth. The same 
         grid1 is going to be used for grid2 if: 
         1st) (tau1<f2)     2nd) (tau1>tau2*f1 && tau1>f2) 
   -v    Density update flag:
   -v 0     = memorize ln(density) grid
   -v 1 M R = use vertical hydrostatic equilibrium for updating the flare disk
              (M is the stellar mass, R is the stellar radius, both in the solar units)
              (works in combination with -Z flag (azimuthal grid) and Flared Disk 
               density profile (density type = 8) )
   -m f: maximum dust temperature (it can be higher than the sublimation temperature)
   -r f1 f2 : calculate radiative transfer for point between f1 and f2 in radius
              (do not apply this rule in the first iteration)
   -R f1 f2 : the same as -r, except that it is applied also on the first iteration
   -x f1 f2 : calculate radiative transfer for point between f1 and f2 in temperature
              (do not apply this rule in the first iteration)
   -X f1 f2 : the same as -x, except that it is applied also on the first iteration
*/


#include <stdio.h>

#include "LEPAMARA.h"

int _LEPAMARA__Analyze_Input_Flags (int argc, char *argv[], char Root_Directory[],
                                 char Main_Dusty_File[]){

  int NAME_EXIST = 0;   /* is the name picked up? */
  int argc_i;           /* counter instead of argc */
  int flag_i;           /* counter of flags within argv[] */
  char default_name[] = "Leluya.inp";

  /* put default values into the flags */
  DEBUG.SCREEN_MESSAGE = 0;
  DEBUG.STOP = 0;
  DEBUG.CONTINUE = 0;
  DEBUG.NEW_POINT_X = -1.0;
  DEBUG.NEW_POINT_Z = -1.0;
  DEBUG.ADD_POINTS_R1 = -1.0;
  DEBUG.ADD_POINTS_R2 = -1.0;
  DEBUG.ISOTHERM_BENDING = -1.0;
  DEBUG.MAX_TRIANGLE_ANGLE = -1.0;
  DEBUG.AZIMUTHAL_GRID = -1;
  DEBUG.DENSITY_UPDATE = -1;
  DEBUG.DENSITY_UPDATE_M = -1;
  DEBUG.DENSITY_UPDATE_R = -1;
  TOTAL_NUMBER_OF_ITERATIONS = 1;
  INPUT.SKIP_POINT = FALSE;
  INPUT.SKIP_POINT_R = FALSE;
  INPUT.SKIP_POINT_T = FALSE;
  INPUT.TAU_LIMIT = -1.0;
  INPUT.TAU_SURFACE_SCALE = -1.0;
  INPUT.LAMBDA_EMISSION_ONLY = -1.0;
  CELL_SIZE_LIMIT=0.05;
  TOTAL_NUMBER_OF_GRID_BINS = 1;
  INITIAL_BIN_COUNTER = -1;
  Bin_Type = 0; 
  MAX_LAMBDA_PER_PROCESS = 5;
  INPUT.DEEPLY_EMBEDDED = FALSE;
  INPUT.POINT_SOURCE_LIMIT=0.05;
  INPUT.TEMPERATURE_ACCELERATE_min_err = -1;
  INPUT.TEMPERATURE_ACCELERATE_scale = 1;
  Grid_Tolerance_Tau_Factor=0.02;
  Tau_Min_Grid=0.1;
  Max_Dust_Temp=1000000.0; /* just a large number, to make sure it is activated */

  /* directory from which the code has been started */
  sprintf(Root_Directory,"%s",argv[0]);

  /* loop over the command line arguments */
  /* if they don't exist, use the default name */
  argc--;
  if (!argc) {
#ifdef _2DD_MPI_
    sprintf(Main_Dusty_File,"%s/%s",_LELUYA_MPIRUN_PATH,default_name);
#else
    sprintf(Main_Dusty_File,"%s/%s",_LELUYA_RUN_PATH,default_name);
#endif
    return 0;
  }
  /* loop over command line arguments */
  argc_i = 1;
  while (argc_i <= argc ) {
    /* if this string is not an input flag, then it is the file name */
    if (argv[argc_i][0]!='-'){
       if (NAME_EXIST) return 2;
#ifdef _2DD_MPI_
       sprintf(Main_Dusty_File,"%s/%s",_LELUYA_MPIRUN_PATH,argv[argc_i]);
#else
       sprintf(Main_Dusty_File,"%s/%s",_LELUYA_RUN_PATH,argv[argc_i]);
#endif
       NAME_EXIST = 1;
    }
    else {
      /* the current string should be a list of flags */
      if (argv[argc_i][1]=='\0') return 1;
      flag_i=1;
      while (argv[argc_i][flag_i]!='\0') {
        switch (argv[argc_i][flag_i]) {

      /* option -d : debug : run-time messages on the screen about 
                                 the program status */  
      case 'd':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (DEBUG.SCREEN_MESSAGE!=0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&DEBUG.SCREEN_MESSAGE)) return 1;
            if (DEBUG.SCREEN_MESSAGE < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option -s n: stop : stop the running code at
                                 position specified by the number "n" */
      case 's':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (DEBUG.STOP != 0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&DEBUG.STOP)) return 1;
            if (DEBUG.STOP < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

          /* option -c n: continue: continue the calculation at 
                                 position specified by the number "n" */
      case 'c':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (DEBUG.CONTINUE != 0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&DEBUG.CONTINUE)) return 1;
            if (DEBUG.CONTINUE < 2) return 1;
            /* if -c 206, 207 then read additional variables */
            if (DEBUG.CONTINUE==206 || DEBUG.CONTINUE==207) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.NEW_POINT_X))
        {DEBUG.CONTINUE=0;return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.NEW_POINT_Z))
        {DEBUG.CONTINUE=0;return 1;}
          if (DEBUG.NEW_POINT_Z<=0)
                {printf("\nERROR with -c 206 or -c 207 flag!!\n");return 1;}
        }
            /* if -c 208 then read additional variables */
            if (DEBUG.CONTINUE==208) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.ISOTHERM_BENDING))
        {DEBUG.CONTINUE=0;return 1;}
          if (DEBUG.ISOTHERM_BENDING<=0)
                {printf("\nERROR with -c 208 flag!!\n");return 1;}
        }
        /* if -c 209 then read additional variables */
        if (DEBUG.CONTINUE==209) {
           argc_i++;
           if (!sscanf(argv[argc_i],"%lf",&DEBUG.MAX_TRIANGLE_ANGLE))
             {DEBUG.CONTINUE=0;return 1;}
           if (DEBUG.MAX_TRIANGLE_ANGLE<=90 || DEBUG.MAX_TRIANGLE_ANGLE>=180)
           {printf("\nERROR with -c 209 flag!!\n");return 1;}
        }
        /* if -c 210 then read additional variables */
        if (DEBUG.CONTINUE==210) {
           argc_i++;
           if (!sscanf(argv[argc_i],"%lf",&DEBUG.MAX_TRIANGLE_ANGLE))
             {DEBUG.CONTINUE=0;return 1;}
           if (DEBUG.MAX_TRIANGLE_ANGLE<0 || DEBUG.MAX_TRIANGLE_ANGLE>90)
             {printf("\nERROR with -c 210 flag!!\n");return 1;}
               argc_i++;
               if (!sscanf(argv[argc_i],"%lf",&DEBUG.ADD_POINTS_R1))
             {DEBUG.CONTINUE=0;return 1;}
           if (DEBUG.ADD_POINTS_R1<=0.0)
             {printf("\nERROR with -c 210 flag!!\n");return 1;}
               argc_i++;
               if (!sscanf(argv[argc_i],"%lf",&DEBUG.ADD_POINTS_R2))
             {DEBUG.CONTINUE=0;return 1;}
           if (DEBUG.ADD_POINTS_R2<=0.0 ||
                   DEBUG.ADD_POINTS_R2<=DEBUG.ADD_POINTS_R1) 
             {printf("\nERROR 2 with -c 210 flag!!\n");return 1;}
        }
        /* if -c 403 then read additional variables */
        if (DEBUG.CONTINUE==403) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%d",&TEMPERATURE_ARRAY_TYPE))
        {DEBUG.CONTINUE=0;return 1;}
          if (TEMPERATURE_ARRAY_TYPE!=1 && TEMPERATURE_ARRAY_TYPE!=2 &&
                  TEMPERATURE_ARRAY_TYPE!=3 && TEMPERATURE_ARRAY_TYPE!=4)
                 {printf("\nThis type of temperature array is not supported!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&X1_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (X1_403<0){printf("\nERROR with -c 403 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&X2_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (X2_403<=X1_403)
                {printf("\nERROR with -c 403 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&Z1_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Z1_403<0){printf("\nERROR with -c 403 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&Z2_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Z2_403<=Z1_403)
                {printf("\nERROR with -c 403 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%d",&Nx_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Nx_403<0){printf("\nERROR with -c 403 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%d",&Nz_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Nz_403<0){printf("\nERROR with -c 403 flag!!\n");return 1;}
        }
        /* if -c 404 then read additional variables */
        if (DEBUG.CONTINUE==404) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&X_404))
        {DEBUG.CONTINUE=0;return 1;}
          if (X_404<0.0){printf("\nERROR with -c 404 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&Z_404))
        {DEBUG.CONTINUE=0;return 1;}
          if (Z_404<0.0){printf("\nERROR with -c 404 flag!!\n");return 1;}
        }
        /* if -c 405 or -c 407 then read additional variables */
        if (DEBUG.CONTINUE==405 || DEBUG.CONTINUE==407) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&X1_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (X1_403<0){printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&X2_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (X2_403<=X1_403)
                {printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&Z1_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Z1_403<0){printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&Z2_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Z2_403<=Z1_403)
                {printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
              argc_i++;
              if (!sscanf(argv[argc_i],"%d",&Nx_403))
        {DEBUG.CONTINUE=0;return 1;}
          if (Nx_403>-2 && Nx_403<2)
                {printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
          if (Nx_403>0) {
                argc_i++;
                if (!sscanf(argv[argc_i],"%d",&Nz_403))
              {DEBUG.CONTINUE=0;return 1;}
            if (Nz_403<0){printf("\nERROR with -c 405 or -c 407 flag!!\n");return 1;}
          }
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -L f: image at f: create a theoretical image at 
                            the wavelength f (works with the flag -c 401) */
          case 'L':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (Output_Lambda_Image>0.0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&Output_Lambda_Image)) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -A f: image at f: create a theoretical image at
                      the inclination angle f (works with the flag -c 401) */
          case 'A':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (Output_Theta_Image>=0.0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&Output_Theta_Image)) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -n n: number of iterations 
               (works only with -c 202 or -c 203, it creates -s 302) */ 
          case 'n':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (TOTAL_NUMBER_OF_ITERATIONS > 1) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&TOTAL_NUMBER_OF_ITERATIONS)) 
               return 1;
        if (TOTAL_NUMBER_OF_ITERATIONS < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -t n: skip the radiative transfer calculation at a point
         (this does not apply to the first iteration) */
          case 't':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&T_error_neighbors)) 
               return 1;
        if (T_error_neighbors < 0 || T_error_neighbors>1) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_error_tolerance)) 
               return 1;
        if (T_error_tolerance <= 0.0) return 1;
            INPUT.SKIP_POINT = 1; /* TRUE1 */
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -T n: skip the radiative transfer calculation at a point
         (this applies to all iterations) */
          case 'T':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&T_error_neighbors)) 
               return 1;
        if (T_error_neighbors < 0 || T_error_neighbors>1) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_error_tolerance)) 
               return 1;
        if (T_error_tolerance <= 0.0) return 1;
            INPUT.SKIP_POINT = 2; /* TRUE2 */
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -r n: skip the radiative transfer calculation at a point
         (this does not apply to the first iteration) */
          case 'r':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT_R != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&R_allowed_range_low)) 
               return 1;
        if (R_allowed_range_low < 0.0) return 1;
            INPUT.SKIP_POINT_R = 3; /* TRUE3 */
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&R_allowed_range_high)) 
               return 1;
        if (R_allowed_range_high <= R_allowed_range_low) {
              printf("ERROR: -r flag: lower radius is larger than the higher radius\n");
              return 1;
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -R n: skip the radiative transfer calculation at a point
         (this applies to all iterations) */
          case 'R':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT_R != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&R_allowed_range_low)) 
               return 1;
        if (R_allowed_range_low < 0.0) return 1;
            INPUT.SKIP_POINT_R = 4; /* TRUE4 */
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&R_allowed_range_high)) 
               return 1;
        if (R_allowed_range_high <= R_allowed_range_low) {
              printf("ERROR: -R flag: lower radius is larger than the higher radius\n");
              return 1;
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -x n: skip the radiative transfer calculation at a point
         (this does not apply to the first iteration) */
          case 'x':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT_T != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_allowed_range_low)) 
               return 1;
        if (T_allowed_range_low < 0.0) return 1;
            INPUT.SKIP_POINT_T = 5; /* TRUE5 */
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_allowed_range_high)) 
               return 1;
        if (T_allowed_range_high <= T_allowed_range_low) {
              printf("ERROR: -x flag: lower radius is larger than the higher radius\n");
              return 1;
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* option  -X n: skip the radiative transfer calculation at a point
         (this applies to all iterations) */
          case 'X':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.SKIP_POINT_T != FALSE) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_allowed_range_low)) 
               return 1;
        if (T_allowed_range_low < 0.0) return 1;
            INPUT.SKIP_POINT_T = 6; /* TRUE6 */
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&T_allowed_range_high)) 
               return 1;
        if (T_allowed_range_high <= T_allowed_range_low) {
              printf("ERROR: -X flag: lower temperature is larger than the higher temperature\n");
              return 1;
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -E f: the rad.trans. integrals stop when the optical 
            depth gets larger than f. The flag is off when f<0.0 */
          case 'E':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INPUT.TAU_LIMIT>0.0) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.TAU_LIMIT)) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -G f: this specifies CELL_SIZE_LIMIT=f, and it works 
            ONLY IF -c flag is less than 2!! Default CELL_SIZE_LIMIT=0.05 */
          case 'G':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&CELL_SIZE_LIMIT)) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -j n: split the radiative transfer job (one iteration) 
                       into n substeps according to the vertex distance to (0,0)
           (works only with -c 202 or -c 203, it creates -s 302) */ 
          case 'j':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (TOTAL_NUMBER_OF_GRID_BINS > 1) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&TOTAL_NUMBER_OF_GRID_BINS)) 
               return 1;
        if (TOTAL_NUMBER_OF_GRID_BINS < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
        Bin_Type = 2; 
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -J n:  shows from which radial bin to start 
                   the first iteration(n has to be <= that n in -j flag) */
          case 'J':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INITIAL_BIN_COUNTER > 1) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&INITIAL_BIN_COUNTER)) 
               return 1;
        if (INITIAL_BIN_COUNTER < 1) return 1;
            INITIAL_BIN_COUNTER -= 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
        Bin_Type = 2; 
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -k n: split the radiative transfer job (one iteration) 
           into n substeps according to the vertex temperatures
           (works only with -c 202 or -c 203, it creates -s 302) */ 
          case 'k':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (TOTAL_NUMBER_OF_GRID_BINS > 1) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&TOTAL_NUMBER_OF_GRID_BINS)) 
               return 1;
        if (TOTAL_NUMBER_OF_GRID_BINS < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
        Bin_Type = 1; 
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -K n: shows from which temperature bin to start 
           the first iteration (n has to be <= that n in -K flag) */
          case 'K':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            if (INITIAL_BIN_COUNTER > 1) return 4;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&INITIAL_BIN_COUNTER)) 
               return 1;
        if (INITIAL_BIN_COUNTER < 1) return 1;
            INITIAL_BIN_COUNTER -= 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
        /* Bin_Type = 1; */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -M n: maximum number of rad.trans. matrices calculated 
                        simultaneously */
          case 'M':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&MAX_LAMBDA_PER_PROCESS)) 
               return 1;
        if (MAX_LAMBDA_PER_PROCESS < 1) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -O f: for wavelengths longer than LAMBDA_EMISSION_ONLY=f 
                        do not invet the scattering matrix, thus only dust 
                        thermal emission exists. */
          case 'O':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.LAMBDA_EMISSION_ONLY))
               return 1;
            if (INPUT.LAMBDA_EMISSION_ONLY <= 0.0) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -S f: activates surface grid recognition. f controls 
            the density of this surface grid. f should be > 1.0 */
          case 'S':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.TAU_SURFACE_SCALE)) return 1;
        if (INPUT.TAU_SURFACE_SCALE>0.0 && INPUT.TAU_SURFACE_SCALE<=1.0) {
          printf("INPUT FLAG ERROR: -S requires flag number > 1.0\n");
          return 1; 
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

      /* activates skipping deeply embedded vertices. It requries -E and -T to 
         be active. If they are not then use the default values: -E 20.0 -T 0.003 */
          case 'W':
            if (INPUT.DEEPLY_EMBEDDED != FALSE) return 4;
            INPUT.DEEPLY_EMBEDDED = TRUE;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -h f:if the apparent angular radius of the source at some 
            point is smaller than f then we use the point source approximation */
          case 'h':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.POINT_SOURCE_LIMIT)) return 1;
        if (INPUT.POINT_SOURCE_LIMIT<0.0) {
          printf("FLAG -h MESSAGE: use old stellar energy densities\n");
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

          /* option -b f1 f2: acceleration method where the new temperature is: 
          Tnew=Told(1+f2*DeltaT) for DeltaT>=f1 where DeltaT is the temperature 
          error.   */
          case 'b':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.TEMPERATURE_ACCELERATE_min_err)) 
              {printf("-b flag error!!\n");return 1;}
        if (INPUT.TEMPERATURE_ACCELERATE_min_err<0.0) {
          printf("INPUT FLAG ERROR: -b requires the first flag >= 0.0\n");
          return 1; 
        }
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&INPUT.TEMPERATURE_ACCELERATE_scale)) 
          {printf("-b flag error!!\n");return 1;}
        if (INPUT.TEMPERATURE_ACCELERATE_scale<=1.0){
              printf("INPUT FLAG ERROR: -b requires the second flag > 1.0\n");
              return 1;
            }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -Z n:Azimuthal triangular grid, where grid points are located on 
        concentric rings around the star. n=0 is the default griding method. */
          case 'Z':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&DEBUG.AZIMUTHAL_GRID)) return 1;
        if (DEBUG.AZIMUTHAL_GRID!=0 && DEBUG.AZIMUTHAL_GRID!=1 && DEBUG.AZIMUTHAL_GRID!=2
                 && DEBUG.AZIMUTHAL_GRID!=3) {
          printf("INPUT FLAG ERROR: -Z accepts only methods 0, 1 or 2\n");
          return 1; 
        }
            if (DEBUG.AZIMUTHAL_GRID==1 || DEBUG.AZIMUTHAL_GRID==2 || DEBUG.AZIMUTHAL_GRID==3) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.AZIMUTHAL_GRID1))
        {printf("ERROR WITH -Z 1 or 2\n");return 1;}
              if (DEBUG.AZIMUTHAL_GRID1<=0.0)
        {printf("ERROR WITH -Z 1 or 2\n");return 1;}
              if (DEBUG.AZIMUTHAL_GRID==3) {
        if (DEBUG.AZIMUTHAL_GRID1 >=90.0)
              {printf("ERROR WITH -Z 3\n");return 1;}
                DEBUG.AZIMUTHAL_GRID1 *= RAD1;
          }
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -g f1 f2: control over the grid copies based on the optical 
            depth. The same grid1 is going to be used for grid2 if: 
              1st) (tau1<f2)     2nd) (tau1>tau2*f1 && tau1>f2) */
          case 'g':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&Grid_Tolerance_Tau_Factor)) return 1;
        if (Grid_Tolerance_Tau_Factor<0.0 || Grid_Tolerance_Tau_Factor>=1.0) {
          printf("INPUT FLAG ERROR: -g \n");
          return 1; 
        }
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&Tau_Min_Grid)) return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

         /* option -v n f1 f2:  Density update flag: n=0 for ln(density), n=1
                                for the vertical hydrostatic equilibrium */
          case 'v':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%d",&DEBUG.DENSITY_UPDATE)) return 1;
        if (DEBUG.DENSITY_UPDATE<0 || DEBUG.DENSITY_UPDATE>1) {
          printf("INPUT FLAG ERROR: -v \n");
          return 1; 
        }
            if (DEBUG.DENSITY_UPDATE == 1) {
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.DENSITY_UPDATE_M)) {
                printf("Error with -v 1 M\n");
                return 1;
              }
              argc_i++;
              if (!sscanf(argv[argc_i],"%lf",&DEBUG.DENSITY_UPDATE_R)) {
                printf("Error with -v 1 M R\n");
                return 1;
              }
              if (DEBUG.DENSITY_UPDATE_M<=0.0 || DEBUG.DENSITY_UPDATE_R<=0.0){
                printf("Error with -v 1 M R (M and R should be > 0.0)\n");
                return 1;
              }
        }
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;

        /* option -m f: maximum dust temperature (it can be higher than 
                    the sublimation temperature) */
          case 'm':
            if (argv[argc_i][flag_i+1]!='\0') return 1;
            if (argc_i==argc) return 1;
            argc_i++;
            if (!sscanf(argv[argc_i],"%lf",&Max_Dust_Temp)) 
               return 1;
            /* prepare the flag_i variable to stop this loop over flag_i */
            flag_i = 0;
            while (argv[argc_i][flag_i]!='\0') flag_i++;
            flag_i--;
            break;


      default:
            return 3;
            /* this 'break' command is unreachable: 
            break;*/
    } /* end of the switch command*/
        flag_i++;
      }/* end of the loop over flags within the current string argv[argc_i] */
    }/* end of checking of the string argv[argc_i] */

    argc_i++;
  }/*end of the loop over command line arguments */

  if (DEBUG.STOP > 0 && DEBUG.STOP<=DEBUG.CONTINUE ) {
    return 7;
  }

  /* check the -n flag */ 
  if (TOTAL_NUMBER_OF_ITERATIONS > 1 && DEBUG.CONTINUE!=202 && 
      DEBUG.CONTINUE!=203) {
    printf("INPUT FLAG ERROR: -n flag active, but -c is not 202 or 203\n");
    return 1;
  }
  if (TOTAL_NUMBER_OF_ITERATIONS > 1) DEBUG.STOP=302;

  /* check the -k flag */
  if (TOTAL_NUMBER_OF_GRID_BINS > 1 && DEBUG.CONTINUE!=202 && 
      DEBUG.CONTINUE!=203) {
    printf("INPUT FLAG ERROR: -k flag active, but -c is not 202 or 203\n");
    return 1;
  }
  if (TOTAL_NUMBER_OF_GRID_BINS > 1) DEBUG.STOP=302;

  /* check the radial profile -c 402 */
  if (DEBUG.CONTINUE==402 && Output_Lambda_Image<=0.0) {
    printf("##############\n");
    printf("ERROR: Specify the output wavelength -L for this radial profile\n");
    printf("##############\n");
    return 1;
  }

  /* check the -K flag */
  if (INITIAL_BIN_COUNTER>=TOTAL_NUMBER_OF_GRID_BINS) {
      printf("##############\n");
      printf("WARNING!!! -K flag indicates a nonexistant bin iteration!\n");
      INITIAL_BIN_COUNTER = TOTAL_NUMBER_OF_GRID_BINS-1;
      printf("           I set this iteration to %d!\n",
             INITIAL_BIN_COUNTER);
      printf("##############\n");
  }

  /* if the file name is not specified, used the default name */
  if (NAME_EXIST==0) {
#ifdef _2DD_MPI_
    sprintf(Main_Dusty_File,"%s/%s",_LELUYA_MPIRUN_PATH,default_name);
#else
    sprintf(Main_Dusty_File,"%s/%s",_LELUYA_RUN_PATH,default_name);
#endif
  }

  /* check if CELL_SIZE_LIMIT is correctly specifed */
  if (CELL_SIZE_LIMIT!=0.05 && DEBUG.CONTINUE!=0 && DEBUG.CONTINUE!=1 
      && DEBUG.CONTINUE!=100) {
      printf("##############\n");
      printf("WARNING!!! -G flag does not work because of -c flag!\n");
      printf("##############\n");
    }
  printf("You use CELL_SIZE_LIMIT=%lf\n",CELL_SIZE_LIMIT);

  if (INPUT.DEEPLY_EMBEDDED && INPUT.TAU_LIMIT<=0.0) {
    printf("##############\n");
    printf("WARNING!!! -W flag activated without -E !!\n");
    printf("USING DEFAULT: -E 20.0\n");
    printf("##############\n");
    INPUT.TAU_LIMIT = 20.0;
  }

  if (INPUT.DEEPLY_EMBEDDED && INPUT.SKIP_POINT != 2) {
    printf("##############\n");
    printf("WARNING!!! -W flag activated without -T !!\n");
    printf("USING DEFAULT: -T 0.003\n");
    printf("##############\n");
    INPUT.SKIP_POINT = 2;
    T_error_tolerance = 0.003;
  }

  /* check if the density update is possible */
  if (DEBUG.DENSITY_UPDATE == 1) {
    if (DEBUG.AZIMUTHAL_GRID<0) 
     if (DEBUG.CONTINUE!=202 && DEBUG.CONTINUE!=203 && DEBUG.CONTINUE!=204 &&
         DEBUG.CONTINUE!=205){
      printf("##############\n");
      printf("ERROR!!! -v flag activated without -Z (azimuthal grid) !!\n");
      printf("##############\n");
      return 1;
    }
  }

  /* End of the function "_LEPAMARA__Analyze_Input_Flags" */
  return 0;
}
