/* LELUYA - LEPAMARA program block */
/* This program calculates the denominator of the eta function (see equations
   for definition). This denominator has integrals over the dust densities
   along the total optical depth's line. The problem with this is that we
   don't know the locations of the dust sublimation surfaces. Therefore this
   function has to use some approximations to give the initial guess for 
   the denominator.
   The approximations used in the program are:
   - optically thin dust
   - a point source
   - the diffuse radiation is neglected
   With these approximations we can find the positions of the sublimation 
   surfaces (called "sublimation radii" in the text and program below).

   SEE THE HTML PAGES ABOUT THIS FUNCTION !!
*/

#include "LEPAMARA.h"

/* functions used ONLY by _LEPAMARA_Find_Initial_Eta_Denominator */
int _LEPAMARA_Find_Psi (void);
/**************************************************************/

int _LEPAMARA_Find_Initial_Eta_Denominator (void) {
  double x_1,z_1;          /* coordinates of the point _1_ */
  double x,z;              /* coordinates of a point on the line of the
                              total optical depth */
  int N_step = 500;         /* number of the grid points between two adjacent
                              sublimation radial distances. It is fixed for
                              now, but if you find some problems with this
                              function, move it to the list of input
                              variables */
  int FLAG_INCREASE_N_step;/* flag which shows if N_step has to be increased */
  int N_tot;               /* total number of the radial grid points on the
                              line of total optical depth (without
                              the point _1_). This number is equal to 
                              N_step*Max_Dust_Type */
  double *Sublimation_Radius;/* list of the sublimation distances on the line
                              of total optical depth (it as an array from 0
                              to Max_Dust_Type-1 ) */
  double **Temperature_Profile;/* temperature radial profiles (for each
                              dust component separately). If the dust should
                              not exist at some point, the temperature value
                              is -1.0 */
  double *Interpolated_Temperature;/* this is used at the end of each 
                              iteration, when the temperature has to be
                              interpolated over the new grid points.
                              When the interpolation is finished, these
                              interpolated temperatures are copied to 
                              Temperature_Profile[][] */
  int degeneracy;          /* takes care about the grid generation when
                              two or more sublimation radii are the same */
  int *sublimation_grid_position; /* gives the position of the sublimation
                              grid points within Radial_Grid_New.
                              The position of the i-th sublimation radius is: 
                              sublimation_grid_position[i]*N_step */
  double *Radial_Grid;     /* list of the grid points on the line of total 
                              optical depth */
  double *Radial_Grid_New; /* a new radial grid at the end of each iteration.
                              After the temperature interpolation is done,
                              these points are moved into Radial_Grid  */
  double *Radial_Optical_Depth;/* optical depth at some grid point. This
                              variable is NOT an array over the grid points,
                              it goes only over the wavelengths. It is 
                              accumulating the optical depth from one point
                              to another in a loop over the grid points */
  double d_i;              /* grid factor: gives a logarithmic spacing
                              of the radial grid R(i+1)=d_i*R(i) */
  double **Radial_Density; /* dust density on the line of the total optical
                              depth (dust density profiles). It is depends 
                              on the sublimation distances */
  double Radius_Convergence;/* this variable is the largest value among the
                              percentage differences between the new and old 
                              sublimation distances. It is used for checking
                              if the iteration loop should stop */
  double *U_source;        /* central energy source = the point source flux
                              Source_Flux[] multiplied by e^(- optical depth),
                              where the optical depth is 
                              Radial_Optical_Depth */
  double density_step;     /* density difference between two adjacent 
                              radial grid points; linear (trapezoid) 
                              approximation */
  double Current_Sublimation_Radius;/* it is used in a search for a new
                              sublimation distance, and then compared with
                              the old value to calculate the error */
  int lambda_k;            /* dummy variable: used whenever we need a counter
                              over the lambda grid (from 0 to lambda_Max-1 ) */
  int dust_i;              /* dummy variable: used whenever we need a counter
                              over the dust components
                              (from 0 to Max_Dust_Type-1 ) */
  int radial_j;            /* dummy variable: used whenever we need a counter
                              over the radial grid on the line of the total
                              optical depth (from 0 to N_tot ) */
  int angular_n;
  int iteration_counter;   /* counts the iteration steps */
  int err_location;        /* error location flag */
  int err;                 /* error flag */
  double dummy1;           /* no special meaning */
  int dust_j;              /* used in sorting of the dust components */
  int radial_i;            /* used in interpolation of the temperature over 
                              the new grid points */
  struct _vertex_structure *point; /* used for the final Eta_Denominator
			      calculation with _PERUNIKA_distance_to_edge */

  do { /* GLOBAL LOOP FOR BASED ON N_step */ 
  FLAG_INCREASE_N_step = FALSE;

  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("------------------------------------\n");
    printf("FUNCTION: _LEPAMARA_Find_Initial_Eta_Denominator\n");
  }

  /* create the arrays needed in the program */
  Sublimation_Radius = dvector(0,Max_Dust_Type-1);
  N_tot = N_step * Max_Dust_Type;
  Temperature_Profile = dmatrix(0,Max_Dust_Type-1,0,N_tot);
  /* NOTE: Temperature_Profile goes from 0 to N_tot because there are 
     all together N_tot+1 points when the points _1_ and Rout are included.
     The same for Radial_Grid */
  Interpolated_Temperature = dvector(0,N_tot);
  Radial_Grid = dvector(0,N_tot);
  Radial_Grid_New = dvector(0,N_tot);
  sublimation_grid_position = ivector(0,Max_Dust_Type-1);
  Radial_Optical_Depth = dvector(0,lambda_Max-1);
  Radial_Density = dmatrix(0,Max_Dust_Type-1,0,N_tot);
  U_source = dvector(0,lambda_Max-1);

  /* find the coordinates of the point _1_ */
  x_1 = 1.0 * sin(Theta_Total_Value);  /* X-cooridnate */
  z_1 = 1.0 * cos(Theta_Total_Value);  /* Z-cooridnate */

  /* find Psi (see equations for definition) */
  err =  _LEPAMARA_Find_Psi();
  if (err) return 1;
  if (DEBUG.SCREEN_MESSAGE >= 2 ) {
    printf(" PSI value (BEFORE the 'eta denominator' calculation)=%lf\n",Psi);
  }


  /* initialize the sublimation radii assuming the temperature profiles
     T(y)=T_0 * y^(-0.5) */
  Sublimation_Radius[0] = 1.0;
  if (Max_Dust_Type>1) 
    for (dust_i=1; dust_i<Max_Dust_Type; dust_i++) {
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        Sublimation_Radius[dust_i]=(Dust_Sublim_Temp[0]*Dust_Sublim_Temp[0])/
                           (Dust_Sublim_Temp[dust_i]*Dust_Sublim_Temp[dust_i]);
      else 
	Sublimation_Radius[dust_i]=Rout;
    }

  if (DEBUG.SCREEN_MESSAGE >= 2 ) {
    printf("The initial configuration of the sublimation distances:\n");
    printf("DUST COMPONENT : DISTANCE\n");
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      printf("%s : %lf\n",Dust_Component_Name[dust_i],
                          Sublimation_Radius[dust_i]);
    }
    printf("------------\n");
  }


  /* create the radial grid */
  Radial_Grid[0] = 1.0;
  degeneracy = 1;
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    if (dust_i+1 == Max_Dust_Type) 
      d_i = exp(log(Rout/Sublimation_Radius[dust_i-degeneracy+1])/
                                                          (degeneracy*N_step));
    else
      d_i = exp(log(Sublimation_Radius[dust_i+1]/
                 Sublimation_Radius[dust_i-degeneracy+1])/(degeneracy*N_step));
    if (dust_i+1<Max_Dust_Type && 
        Sublimation_Radius[dust_i]==Sublimation_Radius[dust_i+1]) 
      degeneracy++;
    else {
      Radial_Grid[(1-degeneracy+dust_i)*N_step+1]=
                                   d_i*Sublimation_Radius[dust_i-degeneracy+1];
      for (radial_j=2; radial_j<=degeneracy*N_step; radial_j++)
        Radial_Grid[(1-degeneracy+dust_i)*N_step+radial_j] = 
                     d_i*Radial_Grid[(1-degeneracy+dust_i)*N_step+radial_j-1];
      degeneracy=1;
    }
  }
  /* just to be sure that the last point is exactly at the outer edge */
  Radial_Grid[N_tot] = Rout;

  /* initialize the temperature profiles (a simple 1/y^0.5 profile) */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    for (radial_j=0; radial_j<=N_tot; radial_j++) {
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
        Temperature_Profile[dust_i][radial_j] = Dust_Sublim_Temp[0] /
                                              sqrt(Radial_Grid[radial_j]);
        if (Temperature_Profile[dust_i][radial_j] > Dust_Sublim_Temp[dust_i])
          /* the dust component dust_i can not exist at this temperature */
          Temperature_Profile[dust_i][radial_j] = -1.0;
      }
      else 
	Temperature_Profile[dust_i][radial_j] = _T99999_; 
    }
  
  
  if (DEBUG.SCREEN_MESSAGE >= 3 ) {
    printf("\nThe initial temperature radial profiles\n");
    printf("-----------------------------------------\n");
    printf("radial distance |");
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                 printf(" %s |",Dust_Component_Name[dust_i]);
    printf("\n");
    for (radial_j=0; radial_j<=N_tot; radial_j++) {
      printf("%lf  ",Radial_Grid[radial_j]);
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        printf("%lf  ",Temperature_Profile[dust_i][radial_j]);
      }
      printf("\n");
    }
    printf("-----------------------------------------\n");
  }

  /* here starts the iteration loop: it is iteration over the positions
     of the sublimation distances Sublimation_Radius[]. The loop goes on as 
     long as at least one sublimation distance don't converge with the 
     precision of _SUBLIMATION_RADIUS_ACCURACY_ */
  iteration_counter = 1;
  do {

    /* update Psi (see equations for definition) */
    err =  _LEPAMARA_Find_Psi();
    if (err) return 1;
    if (DEBUG.SCREEN_MESSAGE >= 2 ) {
      printf("\n------------\n");
      printf("PSI value (iteration #%d)=%lf\n",iteration_counter,Psi);
    }

    /* calculate the Eta function's denominator based on current temperature
       profiles */
    Eta_Denominator = 0.0;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      /* find the dust density profiles */
      for (radial_j=0; radial_j<=N_tot; radial_j++) {
        if (Temperature_Profile[dust_i][radial_j]<=0.0) 
                               Radial_Density[dust_i][radial_j]=0.0;
        else {
          x = Radial_Grid[radial_j] * x_1;
          z = Radial_Grid[radial_j] * z_1;
          Radial_Density[dust_i][radial_j]=Dust_Density(x,z,dust_i,NULL,&err);
	  if ( Radial_Density[dust_i][radial_j]<=0.0)
                               Temperature_Profile[dust_i][radial_j]=-1.0;
	}
      }
      /* integrate the dust density profiles and add them to the eta
         denominator */
      Eta_Denominator = Eta_Denominator +
                        (Sigma_Dust_Abs[dust_i][Lambda_Total] +
                         Sigma_Dust_Sca[dust_i][Lambda_Total]) *
         _MATH_LIB_simpson(N_tot+1,0,N_tot,Radial_Grid,Radial_Density[dust_i]);
      if (DEBUG.SCREEN_MESSAGE >= 3 ) {
        printf("--- Dust component:%s \n    updated ETA_DENOMINATOR += %E\n",
               Dust_Component_Name[dust_i],Eta_Denominator);
        printf("Sigma_Dust_Abs+Sigma_Dust_Sca=%E\n",
                Sigma_Dust_Abs[dust_i][Lambda_Total] +
                Sigma_Dust_Sca[dust_i][Lambda_Total]);
      }
    }
    else 
      for (radial_j=0; radial_j<=N_tot; radial_j++)
	  Radial_Density[dust_i][radial_j]=_Density99999_;

    if (DEBUG.SCREEN_MESSAGE >=2) 
      printf("\nIteration # %d starts with Eta_Denominator = %E\n",
             iteration_counter,Eta_Denominator);
    if (DEBUG.SCREEN_MESSAGE >= 3 ) {
      printf("\nThe dust density radial profiles:\n");
      printf("ITERATION # = %d\n",iteration_counter);
      printf("-----------------------------------------\n");
      printf("radial distance |");
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                 printf(" %s |",Dust_Component_Name[dust_i]);
      printf("\n");
      for (radial_j=0; radial_j<=N_tot; radial_j++) {
        printf("%lf  ",Radial_Grid[radial_j]);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          printf("%12.10lf  ",Radial_Density[dust_i][radial_j]);
        }
        printf("\n");
      }
      printf("-----------------------------------------\n");
    }

    /* set all optical depth to zero (this is at the point _1_) */
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      Radial_Optical_Depth[lambda_k] = 0.0;
      U_source[lambda_k] = Source_Flux[lambda_k];
    }

    /* calculate the dust temperature at the point _1_ for all dust
       components */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 && 
        Radial_Density[dust_i][0]>0) {
      Temperature_Profile[dust_i][0]=_GOROSTAS_Find_Temperature (x_1,z_1,
                       0.9*Dust_Sublim_Temp[0],1.1*Dust_Sublim_Temp[0],
                       dust_i,U_source,&err);
      if (err>0) {
        err_location=8; err+=2;
        _PESJANEK__Error_Message (err_location,err);
        printf("(iteration step #%d, dust component #%d)\n",
               iteration_counter,dust_i);
        printf("(radial distance = point _1_)\n");
        return 1;
      }
      /* check if the dust temperatures are larger then sublimation */
      if (Temperature_Profile[dust_i][0]/Dust_Sublim_Temp[dust_i]-1.0
                                                   > _SUBLIMATION_TOLERANCE_) 
                                  Temperature_Profile[dust_i][0] = -1.0;
    }
    else 
      Temperature_Profile[dust_i][0] = -1.0;

    /* go over all radial grid points and update the dust temperatures */
    /* NOTE: the loop starts from 1 because we already have checked 
       the point _1_ */
    for (radial_j=1; radial_j<=N_tot; radial_j++) {
      /* find the optical depth at the radial point Radial_Grid[radial_j] */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        if (Radial_Density[dust_i][radial_j] > 0.0 &&
                                   Radial_Density[dust_i][radial_j-1] > 0.0) {
          density_step = 0.5*(Radial_Grid[radial_j] - Radial_Grid[radial_j-1])*
                             (Radial_Density[dust_i][radial_j] +
                              Radial_Density[dust_i][radial_j-1]);
          for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
            Radial_Optical_Depth[lambda_k] += (Tau_Total*density_step*
                             (Sigma_Dust_Abs[dust_i][lambda_k] +
                              Sigma_Dust_Sca[dust_i][lambda_k])/
                              Eta_Denominator);
	}
      /* update the energy spectrum from the central source because the 
         optical depth has changed */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
        U_source[lambda_k] *= exp(-Radial_Optical_Depth[lambda_k]);
      /* find new dust temperatures of all dust components */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 && 
          Temperature_Profile[dust_i][radial_j]>0.0){
        x = Radial_Grid[radial_j]*x_1;
        z = Radial_Grid[radial_j]*z_1;
        Temperature_Profile[dust_i][radial_j]=
                       _GOROSTAS_Find_Temperature (x,z,
                       0.9*Temperature_Profile[dust_i][radial_j],
                       1.1*Temperature_Profile[dust_i][radial_j],
                       dust_i,U_source,&err);
        if (err>0) {
          err_location=8; err+=2;
          _PESJANEK__Error_Message (err_location,err);
          printf("(iteration step #%d, dust component #%d)\n",
                 iteration_counter,dust_i);
          printf("(radial distance = %lf)\n",Radial_Grid[radial_j]);
          return 1;
        }
        /* check if the new temperature is less then the sublimation 
           temperature + error tolerance */
        if (Temperature_Profile[dust_i][radial_j]/Dust_Sublim_Temp[dust_i]-1.0
                                                    > _SUBLIMATION_TOLERANCE_) 
         Temperature_Profile[dust_i][radial_j] = -1.0;
        /* check if the new temperature is acceptable, but higher then 
           the sublimation temperature */
        else 
          if (Temperature_Profile[dust_i][radial_j]>Dust_Sublim_Temp[dust_i])
            Temperature_Profile[dust_i][radial_j] = Dust_Sublim_Temp[dust_i];
      } /* end of the loop over the dust components during the calculation 
           of the new dust temperatures */
    } /* end of the loop over the radial grid points */

    /* print the updated temperatures */
    if (DEBUG.SCREEN_MESSAGE >= 3 ) {
      printf("\nThe temperature radial profiles:\n");
      printf("ITERATION # %d\n", iteration_counter);
      printf("-----------------------------------------\n");
      printf("radial distance |");
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                 printf(" %s |",Dust_Component_Name[dust_i]);
      printf("\n");
      for (radial_j=0; radial_j<=N_tot; radial_j++) {
        printf("%lf  ",Radial_Grid[radial_j]);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          printf("%lf  ",Temperature_Profile[dust_i][radial_j]);
        }
        printf("\n");
      }
      printf("-----------------------------------------\n");
    }

    /* update the list of sublimation radii */
    if (DEBUG.SCREEN_MESSAGE >= 2 ) printf("\nSublimation distances:\n");
    Radius_Convergence = 0.0;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      /* search for the sublimation distance from the temperature profile */
      /* (search for the first point which has temperature > 0.0 ) */
      Current_Sublimation_Radius = 1.0;
      radial_j=0;
      while (Temperature_Profile[dust_i][radial_j]<=0.0 && radial_j<=N_tot) {
        radial_j++;
        Current_Sublimation_Radius = Radial_Grid[radial_j];
      }
      /* the sublimation radius is usually not exactly at the position of an 
         old grid point and therefore it has to be interpolated */
      if (radial_j >= N_tot) {
        Current_Sublimation_Radius = Radial_Grid[N_tot];
	FLAG_INCREASE_N_step = TRUE;
	break;
      }
      else if (Radial_Density[dust_i][(radial_j==0?0:radial_j-1)]>0.0)
      if (radial_j<N_tot) {
        if (DEBUG.SCREEN_MESSAGE >= 3){
          printf("interpolation between T1=%EK, y1=%E, and T2=%EK, y2=%E ",
                 Temperature_Profile[dust_i][radial_j], Radial_Grid[radial_j],
             Temperature_Profile[dust_i][radial_j+1], Radial_Grid[radial_j+1]);
          printf("at T_sublim=%EK\n",Dust_Sublim_Temp[dust_i]);
	}
        err=_MATH_LIB_log_interpolation(Temperature_Profile[dust_i][radial_j],
                Radial_Grid[radial_j],Temperature_Profile[dust_i][radial_j+1],
                Radial_Grid[radial_j+1],Dust_Sublim_Temp[dust_i],
                &Current_Sublimation_Radius);
        if (Current_Sublimation_Radius<1.0) Current_Sublimation_Radius=1.0;
      }
      if (DEBUG.SCREEN_MESSAGE >= 2 ) {
        printf("Dust name: %s \n",Dust_Component_Name[dust_i]);
        printf("OLD distance: %12.10lf    NEW distance:%12.10lf  ",
                Sublimation_Radius[dust_i], Current_Sublimation_Radius);
      }
      /* find the percentage difference between the new and old distance */
      dummy1 = fabs(Current_Sublimation_Radius/Sublimation_Radius[dust_i]-1.0);
      if (Current_Sublimation_Radius==1.0 &&
          Temperature_Profile[dust_i][0]/Dust_Sublim_Temp[dust_i]
          < 1.0 - _SUBLIMATION_TOLERANCE_)
         dummy1 = 1.0;
      if (DEBUG.SCREEN_MESSAGE >= 2 ) printf("Diff=%12.10lf\n",dummy1);
      if (Radius_Convergence < dummy1) Radius_Convergence = dummy1;
      Sublimation_Radius[dust_i] = Current_Sublimation_Radius;
    }

    /* stop the whole thing if N_step has to be increased */
    if (FLAG_INCREASE_N_step) break;

    /* sort the dust components according to their sublimation distances */
    /* bubble sort method */
    dust_i = 0;
    while (dust_i < Max_Dust_Type) {
      dust_j = dust_i + 1;
      while (dust_j < Max_Dust_Type) {
        if (Sublimation_Radius[dust_i] > Sublimation_Radius[dust_j] ||
            (Sublimation_Radius[dust_i] == Sublimation_Radius[dust_j] &&
             Temperature_Profile[dust_i][0]>Temperature_Profile[dust_j][0] &&
             Sublimation_Radius[dust_i] == 1.0)
           ) {
          if (DEBUG.SCREEN_MESSAGE >= 2 ) 
            printf("DUST EXCHANGE DETECTED: <%s> <---> <%s>\n",
                   Dust_Component_Name[dust_i],Dust_Component_Name[dust_j]);
          err = _ZELENJAK_exchange_components(dust_i,dust_j);
	  /* exchange dust temperature grid if it exists */
	  if (initial_temperature_array!=NULL) {
            for (angular_n=0; angular_n<_INITIAL_GRID_ARRAY_theta; angular_n++)
              for (radial_j=0; radial_j<_INITIAL_GRID_ARRAY_r; radial_j++) {
                dummy1 = initial_temperature_array[dust_i][radial_j][angular_n];
                initial_temperature_array[dust_i][radial_j][angular_n] = 
		  initial_temperature_array[dust_j][radial_j][angular_n];
                initial_temperature_array[dust_j][radial_j][angular_n] = dummy1;
                dummy1 = initial_energy_array[dust_i][radial_j][angular_n];
                initial_energy_array[dust_i][radial_j][angular_n] =
                  initial_energy_array[dust_j][radial_j][angular_n];
		initial_energy_array[dust_j][radial_j][angular_n] = dummy1;
	      }
	  }
          /* exchange sublimation distances */
          dummy1 = Sublimation_Radius[dust_i];
          Sublimation_Radius[dust_i] = Sublimation_Radius[dust_j];
          Sublimation_Radius[dust_j] = dummy1;
          /* exchange temperature profiles */
          for (radial_j=0; radial_j<=N_tot; radial_j++) {
            dummy1 = Temperature_Profile[dust_i][radial_j];
            Temperature_Profile[dust_i][radial_j] = 
                                        Temperature_Profile[dust_j][radial_j];
            Temperature_Profile[dust_j][radial_j] = dummy1;
	  }
	}
        dust_j++;
      }
      dust_i++;
    }

    /* calculate a new grid due to the updated sublimation distances */
    Radial_Grid_New[0] = 1.0;
    degeneracy = 1;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (dust_i==0) sublimation_grid_position[0] = 0;
      else {
        if (Sublimation_Radius[dust_i] != Sublimation_Radius[dust_i-1])
          sublimation_grid_position[dust_i]=dust_i;
        else 
         sublimation_grid_position[dust_i]=sublimation_grid_position[dust_i-1];
      }
      if (dust_i+1 == Max_Dust_Type) 
        d_i = exp(log(Rout/Sublimation_Radius[dust_i-degeneracy+1])/
                                                          (degeneracy*N_step));
      else
        d_i = exp(log(Sublimation_Radius[dust_i+1]/
                 Sublimation_Radius[dust_i-degeneracy+1])/(degeneracy*N_step));
      if (dust_i+1<Max_Dust_Type && 
          Sublimation_Radius[dust_i]==Sublimation_Radius[dust_i+1]) 
        degeneracy++;
      else {
        if (Tau_Total>10.0) d_i = 1.0 + (d_i-1.0)/Tau_Total; /* increase resolution */
        Radial_Grid_New[(1-degeneracy+dust_i)*N_step+1] = 
                                   d_i*Sublimation_Radius[dust_i-degeneracy+1];
        for (radial_j=2; radial_j<=degeneracy*N_step; radial_j++)
          Radial_Grid_New[(1-degeneracy+dust_i)*N_step+radial_j] = 
                 d_i*Radial_Grid_New[(1-degeneracy+dust_i)*N_step+radial_j-1];
        degeneracy=1;
      }
    }
    /* just to be sure that the last point is exactly at the outer edge */
    Radial_Grid_New[N_tot] = Rout;

    /* interpolate the temperature profiles over the new grid */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      /* the point _1_ is fixed and it doesn't need interpolation */
      Interpolated_Temperature[0] = Temperature_Profile[dust_i][0];
      /* go over the new grid points */
      radial_i=0;
      for (radial_j=1; radial_j<=N_tot; radial_j++) {
        /*sublimation point*/
        if (radial_j==sublimation_grid_position[dust_i]*N_step)
          Interpolated_Temperature[radial_j]=Dust_Sublim_Temp[dust_i];
        /*overheated dust*/
        else if (radial_j<sublimation_grid_position[dust_i]*N_step)
          Interpolated_Temperature[radial_j]= -1.0;
        else {   /* points where we need interpolation */
          /* find the first point in the old grid which has larger distance
             from the center and where the dust exist */
          while (( Radial_Grid[radial_i]<=Radial_Grid_New[radial_j] ||
                   Temperature_Profile[dust_i][radial_i]<=0.0 ) &&
                 radial_i<=N_tot) radial_i++;
          /* be careful, maybe we work with the last point on the old grid */
          if (radial_i>=N_tot){ /* use the last two points for interpolation */
            if (Radial_Grid[N_tot] != Radial_Grid[N_tot-1] )
              err = _MATH_LIB_log_interpolation
               (Radial_Grid[N_tot-1],Temperature_Profile[dust_i][N_tot-1],
                Radial_Grid[N_tot],Temperature_Profile[dust_i][N_tot],
                Radial_Grid_New[radial_j],&Interpolated_Temperature[radial_j]);
            else Interpolated_Temperature[radial_j] = 
                                            Temperature_Profile[dust_i][N_tot];
	  } 
	  else { /* OK, it is not the last point in the old grid, but ... */
            if (Temperature_Profile[dust_i][radial_i-1] <=0.0)
              /* the point at which the dust starts in the old grid */
              err = _MATH_LIB_log_interpolation
              (Radial_Grid_New[sublimation_grid_position[dust_i]*N_step],
               Dust_Sublim_Temp[dust_i],
               Radial_Grid[radial_i],Temperature_Profile[dust_i][radial_i],
              Radial_Grid_New[radial_j],&(Interpolated_Temperature[radial_j]));
            else 
              /* the points radial_i and radial_i-1 are bracketing the 
                 new point radial_j */
              err = _MATH_LIB_log_interpolation
              (Radial_Grid[radial_i-1],Temperature_Profile[dust_i][radial_i-1],
               Radial_Grid[radial_i],Temperature_Profile[dust_i][radial_i],
              Radial_Grid_New[radial_j],&(Interpolated_Temperature[radial_j]));
	  } /* end: if (radial_i < N_tot) */
        } /* end: if (radial_j > dust_i*N_step) */
      } /* end of the loop over radial_j */
      /* move the new temperature profile into Temperature_Profile[dust_i][] */
      for (radial_j=0; radial_j<=N_tot; radial_j++)
        Temperature_Profile[dust_i][radial_j] = 
                                            Interpolated_Temperature[radial_j];
    }

    /* move the grid points into Radial_Grid variable */
    for (radial_j=0; radial_j<=N_tot; radial_j++) 
      Radial_Grid[radial_j] = Radial_Grid_New[radial_j];

    /* print the new interpolated temperature profiles */
    if (DEBUG.SCREEN_MESSAGE >= 3 ) {
      printf("\nThe INTERPOLATED temperature radial profiles:\n");
      printf("ITERATION # %d\n", iteration_counter);
      printf("-----------------------------------------\n");
      printf("radial distance |");
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                 printf(" %s |",Dust_Component_Name[dust_i]);
      printf("\n");
      for (radial_j=0; radial_j<=N_tot; radial_j++) {
        printf("%lf  ",Radial_Grid[radial_j]);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          printf("%lf  ",Temperature_Profile[dust_i][radial_j]);
        }
        printf("\n");
      }
      printf("-----------------------------------------\n");
    }

    /* scale the cross sections with the dust at point _1_ */
    err = _ZELENJAK_Scale_Cross_Sections ();

    if (DEBUG.SCREEN_MESSAGE >= 1) 
      printf("##### Iteration #%d has finished\n",iteration_counter);

    /* the end of the iteration loop */
    iteration_counter++;

    /* do not go into infinite loop */
    if (iteration_counter>4*Max_Dust_Type) {FLAG_INCREASE_N_step=TRUE; break;}

  } while (Radius_Convergence > _SUBLIMATION_RADIUS_ACCURACY_);

  if (!FLAG_INCREASE_N_step) {
    /* calculate the Eta function's denominator based on the final temperature
       profiles */
    Eta_Denominator = 0.0;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
        /* find the dust density profiles */
        for (radial_j=0; radial_j<=N_tot; radial_j++) {
        if (Temperature_Profile[dust_i][radial_j]<=0.0) 
                             Radial_Density[dust_i][radial_j]=0.0;
        else {
          x = Radial_Grid[radial_j] * x_1;
          z = Radial_Grid[radial_j] * z_1;
          Radial_Density[dust_i][radial_j]=Dust_Density(x,z,dust_i,NULL,&err);
	  if ( Radial_Density[dust_i][radial_j]<=0.0)
                               Temperature_Profile[dust_i][radial_j]=-1.0;
        }
      }
      /* integrate the dust density profiles and add them to the eta
         denominator */
      Eta_Denominator = Eta_Denominator +
                      (Sigma_Dust_Abs[dust_i][Lambda_Total] +
                       Sigma_Dust_Sca[dust_i][Lambda_Total]) *
         _MATH_LIB_simpson(N_tot+1,0,N_tot,Radial_Grid,Radial_Density[dust_i]);
      }
    /* make Eta_Denominator more precise (when Tau_Total is large this 
       error can be very big) */
    point = (struct _vertex_structure *) malloc (sizeof(struct _vertex_structure));
    point->x = sin(Theta_Total_Value);
    point->z = cos(Theta_Total_Value);
    dummy1 = _PERUNIKA_distance_to_edge(point,Rout*sin(Theta_Total_Value),
                                      Rout*cos(Theta_Total_Value),Lambda_Total,
                                      -1,&d_i,0);
    Eta_Denominator *= (dummy1/Tau_Total); /* correction */
    free(point);
  
    /* find Psi (see equations for definition) */
    err =  _LEPAMARA_Find_Psi();
    if (err) return 1;
    if (DEBUG.SCREEN_MESSAGE >= 2 ) {
      printf("\n------------\n");
      printf("PSI value (AFTER the 'eta denominator' calculation)=%lf\n",Psi);
    }

    if (DEBUG.SCREEN_MESSAGE >= 1 ) {
      printf("\ntauold=%lf The initial guess of the eta function's denominator = %lf\n",dummy1,Eta_Denominator);
    }
  }

  /* destroy the  arrays used in the program */
  free_dvector(Sublimation_Radius,0,Max_Dust_Type-1);
  free_dmatrix(Temperature_Profile,0,Max_Dust_Type-1,0,N_tot);
  free_dvector(Interpolated_Temperature,0,N_tot);
  free_dvector(Radial_Grid,0,N_tot);
  free_dvector(Radial_Grid_New,0,N_tot);
  free_ivector(sublimation_grid_position,0,Max_Dust_Type-1);
  free_dvector(Radial_Optical_Depth,0,lambda_Max-1);
  free_dmatrix(Radial_Density,0,Max_Dust_Type-1,0,N_tot);
  free_dvector(U_source,0,lambda_Max-1);

  if (FLAG_INCREASE_N_step) {
    N_step *= 2;
    printf("#### Search starts from the beginning: N_step is increased to %d\n",N_step);
  }

  } while (FLAG_INCREASE_N_step); /* END OF THE LOOP BASED ON N_step */

  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("Dust ordering:\n");
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      printf("      %d: %s\n",dust_i,Dust_Component_Name[dust_i]);
  }


  return 0;
} /* end of the function _LEPAMARA_Find_Initial_Eta_Denominator */

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

/* calculate the Psi value */

int _LEPAMARA_Find_Psi (void){
  double *f;              
  double x_1,z_1;          /* coordinates of the point _1_ */
  int lambda_k;
  int err_location, err;

  f = dvector(0,lambda_Max-1);

  /* find the coordinates of the point _1_ */
  x_1 = 1.0 * sin(Theta_Total_Value);  /* X-cooridnate */
  z_1 = 1.0 * cos(Theta_Total_Value);  /* Z-cooridnate */

  /* find Psi (see equations for definition) */
  /* NOTE: it would be better if this is moved to some other place, but
           for now, it can stay here */

  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
    f[lambda_k] = q_abs(x_1,z_1,lambda_k,0,&err)*Source_Flux[lambda_k];
  if (err>0) {
    err_location=8; err=2;
    _PESJANEK__Error_Message (err_location,err);
    printf("(coordiantes:  x=%lf, z=%lf = point _1_)\n",x_1,z_1);
    printf("(wavelength #%d: %E\n",lambda_k,Lambda_Grid[lambda_k]);
    return 1;
  }
  Psi = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,f);
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
    f[lambda_k] = q_abs(x_1,z_1,lambda_k,0,&err)*
                  _MATH_LIB_Planck(Dust_Sublim_Temp[0],Lambda_Grid[lambda_k]);
  Psi = Psi / _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,f);

  free_dvector(f,0,lambda_Max-1);
  return 0;
}
