#define TINY 1.0E-20

#include "MATH_LIB.h"

/* decomposition of a dmatrix into LU form. Used in combination with 'lubksb' 
   to solve linear equations or invert a dmatrix */
/* This is a parallel version of ludcmp, written by Dejan Vinkovic, 2002 */

void _MATH_LIB_ludcmp_parallel(double **a, int n, int *indx, double *d)
{
  int i,imax,j,k;
  double big,dum,sum,temp;
  double *vv;
#ifdef _2DD_MPI_
  int Nmax,ni,nj,bij,bj;
  int tag = 0;
  int receive_flag;
  MPI_Status Receive_Status, Status;
#endif

time_t sek;
struct tm *t;
char str[30];
time (&sek);
t = localtime(&sek);
strcpy(str,ctime(&sek));
if (my_MPI_rank==0)
  printf("Local time in LUDCMP_parallel: %s\n",str);

  vv=dvector(1,n);
  *d=1.0;
 
#ifdef _2DD_MPI_
  if (my_MPI_rank==0)
    /* check for singularity!! */
    for (i=1;i<=n;i++) {
      big=0.0;
      for (j=1;j<=n;j++)
	if ((temp=fabs(a[i][j])) > big) big=temp;
      if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
    }

  /* find the total number of columns 
     (it has to be a multiple of total_MPI_process) */
  if (n%total_MPI_process != 0)
    Nmax = n + total_MPI_process - n%total_MPI_process;
  else
    Nmax = n;

  /* the main loop: loop over the matrix columns */
  for (j=1;j<=Nmax;j++) {
    /* calculate the column for this process */
    if ((j-1)%total_MPI_process == my_MPI_rank) {

      /* get the initial broadcasted data from smaller processes *
      for (bij=2;bij<=my_MPI_rank+1;bij++) 
	for (bj=1;bj<=my_MPI_rank;bj++) {
	  nj=((int)((j-1)/total_MPI_process))*total_MPI_process + bj;
	  ni=bij-bj;
          if (nj<=n){
            MPI_Bcast (&(a[ni][nj]),1,MPI_DOUBLE,(bj-1),MPI_communicator);
	  }
	  } */ 

      /* loop over the rows */
      for (i=1-my_MPI_rank;i<=n+total_MPI_process-1-my_MPI_rank;i++) {
	if (j<=n && i>0 && i<=n) {
	  /* this column and row exist */ 
	  if (i<=j) {
	    sum=a[i][j];
	    for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
	    a[i][j]=sum;
	    if (i==j) if (a[j][j] == 0.0) a[j][j]=TINY;
	  }
	  else {
	    sum=a[i][j];
	    for (k=1;k<j;k++) sum -= a[i][k]*a[k][j];
	    a[i][j]=sum/a[j][j];
	  }
	}
	/* broadcast the data */
	for (bj=0;bj<total_MPI_process;bj++) {
	  ni = i+my_MPI_rank-bj;
	  nj=((int)((j-1)/total_MPI_process))*total_MPI_process + bj + 1;
	  if (ni>0 && ni<=n && nj<=n) {
	    MPI_Bcast (&(a[ni][nj]),1,MPI_DOUBLE,bj,MPI_communicator);
	  }
	} 
      }

      /* get the final row data from larger processes *
      for (bij=n+2+my_MPI_rank;bij<=n+total_MPI_process;bij++) 
	for (bj=2+my_MPI_rank;bj<=total_MPI_process;bj++) {
	  nj=((int)((j-1)/total_MPI_process))*total_MPI_process + bj;
	  ni=bij-bj;
	  if (nj<=n){
            MPI_Bcast (&(a[ni][nj]),1,MPI_DOUBLE,(bj-1),MPI_communicator);
	  }
	  }*/ 

    } /* end of the main loop over columns */ 
    if (j<=n)  indx[j]=j;
  }

#else
	/* If this does not use MPI then it is the same as ordinary 
           ludcmp subroutine */ 

        for (i=1;i<=n;i++) {
                big=0.0;
                for (j=1;j<=n;j++)
                        if ((temp=fabs(a[i][j])) > big) big=temp;
                if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
                vv[i]=1.0/big;
        }
        for (j=1;j<=n;j++) {
                for (i=1;i<j;i++) {
                        sum=a[i][j];
                        for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                }
                big=0.0;
                for (i=j;i<=n;i++) {
                        sum=a[i][j];
                        for (k=1;k<j;k++)
                                sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                        if ( (dum=vv[i]*fabs(sum)) >= big) {
                                big=dum;
                                imax=i;
                        }
                }
                if (j != imax) {
                        for (k=1;k<=n;k++) {
                                dum=a[imax][k];
                                a[imax][k]=a[j][k];
                                a[j][k]=dum;
                        }
                        *d = -(*d);
                        vv[imax]=vv[j];
                }
                indx[j]=imax;
                if (a[j][j] == 0.0) a[j][j]=TINY;
                if (j != n) {
                        dum=1.0/(a[j][j]);
                        for (i=j+1;i<=n;i++) a[i][j] *= dum;
                }
        }
#endif

  free_dvector(vv,1,n);
}










