/* LELUYA - PATULJAK program block */
/* this function reads the parameters required by a dust density model.
   The parameters for a dust component 'dust_counter' are stored in the 
   variable 'Dust_Density_Param[dust_counter]', and they describe the model
   'INPUT.DUST_DENSITY_TYPE[dust_counter]'. They are specified in the following
   way:
   1st parameter:
      Dust_Density_Param[dust_counter].param
   2nd parameter:
      Dust_Density_Param[dust_counter].next->param
   3rd parameter: 
      Dust_Density_Param[dust_counter].next->next->param
   etc...

   In this function we have to read these parameters from the input file,
   and memorize them. Don't forget to allocate a memory for them!! 
   Example: if you work with the 2nd parameter and you need a third one:

   Dust_Density_Param[dust_counter].next->next = (struct density_parameters *) 
                                 malloc (sizeof(struct density_parameters));

   Also, if you are introducing a new model, write a short description
   in _data_calc_output_flags.h, where you can find the list of existing 
   models, and give a name to your model (up to 50 characters):

     sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"New Model Name");

   */

#include "PATULJAK.h"

int _PATULJAK__Read_Density_Type (FILE *GLOBAL_FILE,int dust_counter){
  int err;             /* error flag used during input operations */


  /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==99999) {
    /* this is empty space, there is no input parameters for it*/
    /* This empty space "starts" at: */ 
    Dust_Density_Param[dust_counter].param = 0.1;
    Dust_Density_Param[dust_counter].next = NULL;
    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Empty Space");
    /* and has a constant density equal to _Density99999_ */
    return 0;
  }
  /***********************************************************************/
  /***********************************************************************/  

  /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==1) {
    /* Welcome to Trinity model ! 
       This model is used by Dejan Vinkovic for testing of the code
       and it doesn't have a fixed structure. DO NOT USE THIS MODEL !! */
   
    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Trinity Model"); 

     /* this is a flared disk + wall with the density:
         Rho = Rho0 * (x^2+y^2)^(-p/2)*exp(-h0*z^2/(x^2+y^2)^m)
       and the surface z=1.5*h0*(x^2+y^2)^(m/2)

       parameters:  
         p  = Dust_Density_Param[dust_counter].param
         m  = Dust_Density_Param[dust_counter].next->param
         h0 = Dust_Density_Param[dust_counter].next->next->param
         Rho0 = Dust_Density_Param[dust_counter].next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read m */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    /* read h0 */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
      Dust_Density_Param[dust_counter].next->next->next->param <=0.0 ) return 1;

    Dust_Density_Param[dust_counter].next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==1) */
  /***********************************************************************/
  /***********************************************************************/  

  /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==2) {
    /* Welcome to Smooth_Sphere model! */
   
    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Smooth Sphere Model"); 

    /* density \propto r^(-p)
       where: r=a*sqrt(x*x+z*z)
       parameters: a    = Dust_Density_Param[dust_counter].param
       parameters: p    = Dust_Density_Param[dust_counter].next->param
    */

    /* read 'a' */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 || Dust_Density_Param[dust_counter].param<=0.0) return 1;

    /* read 'p' */
    Dust_Density_Param[dust_counter].next = (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    if (Dust_Density_Param[dust_counter].next==NULL) return 1;
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].next->param);
    if (err!=1) return 1;

    Dust_Density_Param[dust_counter].next->next = NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==2) */
  /***********************************************************************/
  /***********************************************************************/  

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==3) {
    /* Welcome to Ellipsoid model !
       Introduced by Zeljko Ivezic & Dejan Vinkovic
                                   Princeton, September 21, 2001
    */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Ellipsoid Model");

    /* dust density = 1 / (X*X + e*Z*Z)^(n/2) 
       parameters: n    = Dust_Density_Param[dust_counter].param
                   e    = Dust_Density_Param[dust_counter].next->param
    */

    /* read 'n' */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1) return 1;

    /* read 'e' */
    Dust_Density_Param[dust_counter].next = (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    if (Dust_Density_Param[dust_counter].next==NULL) return 1;
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].next->param);
    if (err!=1 || Dust_Density_Param[dust_counter].next->param <= 0.0) return 1;

    Dust_Density_Param[dust_counter].next->next = NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==3) */
  /***********************************************************************/
  /***********************************************************************/


  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==4) {
    /* Welcome to the flared disk model with sharp surface */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Flared Disk with Sharp Surface");
    /* this is a flared disk with the density:
         Rho = Rho0 * (x^2+z^2)^(-p/2)*exp(-h0*z^2/(x^2+z^2)^m)
       and the surface z=1.5*h0*(x^2+z^2)^(m/2)

       parameters:  
         p  = Dust_Density_Param[dust_counter].param
         m  = Dust_Density_Param[dust_counter].next->param
         h0 = Dust_Density_Param[dust_counter].next->next->param
         Rho0 = Dust_Density_Param[dust_counter].next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read m */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    /* read h0 */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
      Dust_Density_Param[dust_counter].next->next->next->param <=0.0 ) return 1;

    Dust_Density_Param[dust_counter].next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==4) */
  /***********************************************************************/
  /***********************************************************************/

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==5) {
    /* A spherical shell without a cone: "Cone model"   */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Cone Model");
    /*
       if Theta_cone>=0 then the dust exists for Theta>=Theta_cone
       else the dust exists for Theta<Theta_cone

       dust density = a*rho^(-p)

       parameters:  
       cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       a     = Dust_Density_Param[dust_counter].next->param
       -p    = Dust_Density_Param[dust_counter].next->next->param

       input parameters: "Theta_cone" (in deg, NOT radians), "a", and "p"
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read Theta_cone */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].param>=0.0)
      Dust_Density_Param[dust_counter].param = 
                              cos(RAD1*Dust_Density_Param[dust_counter].param);
    else
      Dust_Density_Param[dust_counter].param = 
                              -cos(RAD1*Dust_Density_Param[dust_counter].param);
    /* read "a" */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 || Dust_Density_Param[dust_counter].next->param <=0.0 ) return 1;
    /* read p */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;

    Dust_Density_Param[dust_counter].next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==5) */
  /***********************************************************************/
  /***********************************************************************/
 
  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==6) {
    /* A spherical shell without a cone: "Finite Cone model"   */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Finite Cone Model");
    /*
       if Theta_cone<0 then the dust exists for Theta<Theta_cone AND 
       radius < Rcone, else it exists outside this region

       dust density = a*rho^(-p)

       parameters:  
       cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       a     = Dust_Density_Param[dust_counter].next->param
       -p    = Dust_Density_Param[dust_counter].next->next->param
       Rcone = Dust_Density_Param[dust_counter].next->next->next->param

       input parameters: "Theta_cone" (in deg, NOT radians), "a", "p", "Rcone"
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read Theta_cone */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].param>=0.0)
      Dust_Density_Param[dust_counter].param = 
                              cos(RAD1*Dust_Density_Param[dust_counter].param);
    else
      Dust_Density_Param[dust_counter].param = 
                              -cos(RAD1*Dust_Density_Param[dust_counter].param);
    /* read "a" */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 || Dust_Density_Param[dust_counter].next->param <=0.0 ) return 1;
    /* read p */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* Rcone */
    Dust_Density_Param[dust_counter].next->next->next = 
      (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
        Dust_Density_Param[dust_counter].next->next->next->param<=0.0) return 1;

    Dust_Density_Param[dust_counter].next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==6) */
  /***********************************************************************/
  /***********************************************************************/

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==7) {
    /* A simple disk model defined with the angle of cutoff cone and
       outer radius   */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Simple Disk");
    /*
       the dust exists for Theta>=Theta_cone and R<Rdisk

       dust density = a*rho^(-p)

       parameters:  
       cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       a     = Dust_Density_Param[dust_counter].next->param
       -p    = Dust_Density_Param[dust_counter].next->next->param
       Rdisk = Dust_Density_Param[dust_counter].next->next->next->param
       Rin = Dust_Density_Param[dust_counter].next->next->next->next->param

       input parameters: "Theta_cone" (in deg, NOT radians),"a","p","Rdisk"
       NOTE: Rin is zero by default. If Rdisk<0 then Rdisk=|Rdisk| and Rin is
             specified in the input file
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read Theta_cone */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].param<0.0)
      {printf("INPUT ERROR in model 'Simple Disk': Theta_Disk<=0\n");return 1;}
    Dust_Density_Param[dust_counter].param = 
                              cos(RAD1*Dust_Density_Param[dust_counter].param);
    /* read "a" */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 || Dust_Density_Param[dust_counter].next->param <=0.0 ) return 1;
    /* read p */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rdisk */
    Dust_Density_Param[dust_counter].next->next->next= 
        (struct density_parameters *)malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1) return 1;

    Dust_Density_Param[dust_counter].next->next->next->next= 
        (struct density_parameters *)malloc (sizeof(struct density_parameters));
    if (Dust_Density_Param[dust_counter].next->next->next->param < 0.0) {
      /* read if Rdisk < 0.0 */
      Dust_Density_Param[dust_counter].next->next->next->param *= -1.0;
      err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
      if (err) return 1;
      err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->next->next->param));
      if (err!=1) return 1;
    }
    else {
      Dust_Density_Param[dust_counter].next->next->next->next->param = 0.0;
    }

    Dust_Density_Param[dust_counter].next->next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==7) */
  /***********************************************************************/
  /***********************************************************************/

 
  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==8) {
    /* Welcome to the flared disk model */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Flared Disk");
    /* this is a flared disk with the density:
         Rho = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))
        (NO DISK SURFACE!)

       parameters:  
         p  = Dust_Density_Param[dust_counter].param
         m  = Dust_Density_Param[dust_counter].next->param
         h0 = Dust_Density_Param[dust_counter].next->next->param
         Rho0 = Dust_Density_Param[dust_counter].next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read m */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    /* read h0 */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
      Dust_Density_Param[dust_counter].next->next->next->param <=0.0 ) return 1;

    Dust_Density_Param[dust_counter].next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==8) */
  /***********************************************************************/
  /***********************************************************************/

 
  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==9) {
    /* Welcome to the flat disk model */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Flat Disk");
    /* this is a flat disk with the density:
         Rho = Rho0 * x^(-p) for Xmin<x<Xmax and z<Zmax

       parameters:  
         p  = Dust_Density_Param[dust_counter].param
         Xmin = Dust_Density_Param[dust_counter].next->param
         Xmax = Dust_Density_Param[dust_counter].next->next->param
         Zmax = Dust_Density_Param[dust_counter].next->next->next->param
         Rho0 = Dust_Density_Param[dust_counter].next->next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read Xmin */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    if (Dust_Density_Param[dust_counter].next->param < 1.0)
      Dust_Density_Param[dust_counter].next->param = 1.0;
    /* read Xmax */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->param > Rout)
      Dust_Density_Param[dust_counter].next->next->param = Rout;
    /* read Zmax */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->param));
    if (err!=1) return 1;


    Dust_Density_Param[dust_counter].next->next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==9) */
  /***********************************************************************/
  /***********************************************************************/

 
  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==10) {
    /* Welcome to the flared disk + halo model */
    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Flared Disk + Spherical Halo");

    /* this is a combination of flared disk and halo with the density:

     Halo = halo dust density profile = Rho1 * r^(-s)
     Disk = disk dust density profile = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))

     if (Halo>=Disk) Rho = Halo
     else Rho = Disk

       parameters:  
       p   = Dust_Density_Param[dust_counter].param
       m   = Dust_Density_Param[dust_counter].next->param
       h0  = Dust_Density_Param[dust_counter].next->next->param
       Rho0= Dust_Density_Param[dust_counter].next->next->next->param
       s   = Dust_Density_Param[dust_counter].next->next->next->next->param
       Rho1= Dust_Density_Param[dust_counter].next->next->next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read m */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    /* read h0 */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
      Dust_Density_Param[dust_counter].next->next->next->param <=0.0 ) return 1;
    /* read s */
    Dust_Density_Param[dust_counter].next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->param));
    if (err!=1) return 1;
    /* read Rho1 */
    Dust_Density_Param[dust_counter].next->next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->param));
    if (err!=1 || 
        (Dust_Density_Param[dust_counter].next->next->next->next->next->param) <= 0.0) 
       return 1;

    /* close the list */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==10) */
  /***********************************************************************/
  /***********************************************************************/

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==11) {
    /* Welcome to the flared disk + flared halo model */
    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "Flared Disk + Flared Halo");
    /* this is a combination of flared disk and flared halo with the density:

     Halo = halo dust density profile = Rho1 * x^(-s)*exp(-h1*z^2/x^(2k))
     Disk = disk dust density profile = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))

     if (Halo>=Disk) Rho = Halo
     else Rho = Disk

       parameters:  
       p   = Dust_Density_Param[dust_counter].param
       m   = Dust_Density_Param[dust_counter].next->param
       h0  = Dust_Density_Param[dust_counter].next->next->param
       Rho0= Dust_Density_Param[dust_counter].next->next->next->param
       s   = Dust_Density_Param[dust_counter].next->next->next->next->param
       k   = Dust_Density_Param[dust_counter].next->next->next->next->next->param
       h1  = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param
       Rho1= Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    /* read p */ 
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;
    /* read m */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;
    /* read h0 */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
	       &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    /* read Rho0 */
    Dust_Density_Param[dust_counter].next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1 || 
      Dust_Density_Param[dust_counter].next->next->next->param <=0.0 ) return 1;
    /* read s */
    Dust_Density_Param[dust_counter].next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->param));
    if (err!=1) return 1;
    /* read k */
    Dust_Density_Param[dust_counter].next->next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->param));
    if (err!=1 ) return 1;
    /* read h1 */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->next->param));
    if (err!=1 ) return 1;
    /* read Rho1 */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next->next= 
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param));
    if (err!=1 || 
        (Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param) <= 0.0) 
       return 1;

    /* close the list */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==11) */
  /***********************************************************************/
  /***********************************************************************/

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==102) {
    /* Welcome to the Bonn2001-CIT3 model: CIT3_Jan17_01 */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],
             "CIT3_Jan17_01");

    /* radial density is prop. to a/r^0.5  when theta<theta0
                                  1/r^2.0  when theta>theta0 */
    /* parameters:  theta0 = Dust_Density_Param[dust_counter].param
                    a = Dust_Density_Param[dust_counter].next->param
    */

    /* read theta0 */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 || Dust_Density_Param[dust_counter].param <=0.0 || 
        Dust_Density_Param[dust_counter].param >= 90.) return 1;
    Dust_Density_Param[dust_counter].param = 
                            cos(Dust_Density_Param[dust_counter].param*PI/180.);
    /* read a */ 
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].next->param);
    if (err!=1 || Dust_Density_Param[dust_counter].next->param <= 0.0) return 1;
    Dust_Density_Param[dust_counter].next->next = NULL;
    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==102) */

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==103) {
    /* Welcome to the step-flat disk model */

    sprintf(INPUT.DUST_DENSITY_TYPE_NAME[dust_counter],"Step-Flat Disk");

    /* A flat disk with a step in density along the x-axis:

         Rho = Rho1 * x^(-p1) for Xmin<x<Xmid and z<Zmax
         Rho = Rho2 * x^(-p2) for Xmid<x<Xmax and z<Zmax

       parameters:
         p1  = Dust_Density_Param[dust_counter].param
         p2  = Dust_Density_Param[dust_counter].next->param
         Xmin = Dust_Density_Param[dust_counter].next->next->param
         Xmid = Dust_Density_Param[dust_counter].next->next->next->param
         Xmax = Dust_Density_Param[dust_counter].next->next->next->next->param
         Zmax = Dust_Density_Param[dust_counter].next->next->next->next->next->param
         Rho1 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param
         Rho2 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param
    */

    /* read the parameters */
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;

    /* read p1 */
    err=fscanf(GLOBAL_FILE,"%lf",&Dust_Density_Param[dust_counter].param);
    if (err!=1 ) return 1;

    /* read p2 */
    Dust_Density_Param[dust_counter].next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",&(Dust_Density_Param[dust_counter].next->param));
    if (err!=1 ) return 1;

    /* read Xmin */
    Dust_Density_Param[dust_counter].next->next= (struct density_parameters *)
                                  malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->param < 1.0)
      Dust_Density_Param[dust_counter].next->next->param = 1.0;

    /* read Xmid */
    Dust_Density_Param[dust_counter].next->next->next=
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->param <= 
        Dust_Density_Param[dust_counter].next->next->param) return 1;

    /* read Xmax */
    Dust_Density_Param[dust_counter].next->next->next->next=
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->param < 
        Dust_Density_Param[dust_counter].next->next->next->param) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->param > Rout)
      Dust_Density_Param[dust_counter].next->next->next->next->param = Rout;
    
    /* read Zmax */
    Dust_Density_Param[dust_counter].next->next->next->next->next=
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->next->param <= 0.0)  return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->next->param > Rout)
      Dust_Density_Param[dust_counter].next->next->next->next->next->param = Rout;

    /* read Rho1 */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next=
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->next->next->param<=0.0) return 1;

    /* read Rho2 */
    Dust_Density_Param[dust_counter].next->next->next->next->next->next->next=
       (struct density_parameters *) malloc (sizeof(struct density_parameters));
    err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
    if (err) return 1;
    err=fscanf(GLOBAL_FILE,"%lf",
               &(Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param));
    if (err!=1) return 1;
    if (Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param<=0.0) return 1;


    Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->next=NULL;

    /* return OK flag */
    return 0;
  } /* end of (INPUT.DUST_DENSITY_TYPE[dust_counter]==102) */
  /***********************************************************************/
  /***********************************************************************/

  return 1; /* this line should not be reached!! */
} /* end of _PATULJAK__Read_Density_Type */

