/* LELUYA - PATULJAK program block */
/* This function reads the total optical depth specified for one
   line toward the central source.
   More details about the required variables Tau_Total, Theta_Total_Value,
   Lambda_Total_Value find in PATULJAK.h. */

#include "PATULJAK.h"

int _PATULJAK__Read_Tau_Total (FILE *GLOBAL_FILE) {
  int  lambda_counter;  /* used for the loop over wavelengths in the 
                           lambda grid */
  double *Temp_Lambda_Grid; /* temporary lambda grid in the case when
                           a new wavelength has to included */
  int  err;             /* error flag used for detection of a problem 
                              in reading of GLOBAL_FILE */

  /***** read Tau_Total *****************************************/
  /* locate the position of input data in the global input file */
  err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
  if (err) return 1;

  /* read the input value */
  err = fscanf(GLOBAL_FILE,"%lf",&Tau_Total);
  if (err!=1) return 2;

  /***** read the azimuthal angle *******************************/
  /* locate the position of input data in the global input file */
  err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
  if (err) return 3;

  /* read the input value */
  err = fscanf(GLOBAL_FILE,"%lf",&Theta_Total_Value);
  if (err!=1) return 4;

  /* check if it is in acceptable range */
  if (Theta_Total_Value < 0 || Theta_Total_Value > 90) return 5;

  /***** include this angle into the angular grid of the central
         source flux ********************************************/
  /***** NOTE: THIS OPTION DON'T EXIST FOR NOW !!!! *************/
  Theta_Total = 0;

  /***** read the wavelength ************************************/
  /* locate the position of input data in the global input file */
  err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
  if (err) return 6;

  /* read the input value */
  err = fscanf(GLOBAL_FILE,"%lf",&Lambda_Total_Value);
  if (err!=1) return 7;

  /* check if it is in the range of accaptable wavelengths */
  if (Lambda_Total_Value < Lambda_Grid[0] ||
      Lambda_Total_Value > Lambda_Grid[lambda_Max-1])
    return 8;

  /***** include this wavelength into the lambda grid **********/
  /* check if it is already in the grid */
  lambda_counter = 0;
  while (Lambda_Grid[lambda_counter] < Lambda_Total_Value)
    lambda_counter++;
  if (Lambda_Grid[lambda_counter] == Lambda_Total_Value){
    Lambda_Total = lambda_counter;
  }
  else {
    /* nope! It's not in the grid. We have to include it. */
    /* debuger messages if required by DEBUG.SCREEN_MESSAGE */
    if (DEBUG.SCREEN_MESSAGE >= 1 ) {
      printf("  the lambda grid has one additional point at %lf\n",
                                                      Lambda_Total_Value);
    }
    /* create a temporary lambda grid with one grid point more then now */
    Temp_Lambda_Grid=dvector(0,lambda_Max);
    /* copy the lambda grid into the temporary grid, and add the additional
       total optical depth wavelength */
    lambda_counter = 0;
    while (Lambda_Grid[lambda_counter] < Lambda_Total_Value) {
      Temp_Lambda_Grid[lambda_counter]=Lambda_Grid[lambda_counter];
      lambda_counter++;
    }
    Temp_Lambda_Grid[lambda_counter]=Lambda_Total_Value;
    lambda_Max++; /* increase the total number of lambda grid points */
    Lambda_Total = lambda_counter; /* memorize the position of 
                                      Lambda_Total_Value in the lambda grid */
    lambda_counter++;
    
    while (lambda_counter < lambda_Max ) {
      Temp_Lambda_Grid[lambda_counter]=Lambda_Grid[lambda_counter-1];
      lambda_counter++;
    }
    free_dvector(Lambda_Grid,0,lambda_Max-2);
    Lambda_Grid=Temp_Lambda_Grid; /* temporary grid becomes
                                             the new lambda grid */
  } /* end of (Lambda_Grid[lambda_counter] != Lambda_Total_Value)*/

  /* debuger messages if required by DEBUG.SCREEN_MESSAGE */
  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("  total optical depth = %lf\n",Tau_Total);
    printf("        at wavelength = %lf microns and\n",Lambda_Total_Value);
    printf("      azimuthal angle = %lf deg\n",Theta_Total_Value);
    printf("  position of this wavelength in the lambda grid = %d\n",
           Lambda_Total);
    printf("  (wavelength before =");
    if (Lambda_Total>0) printf("%lf ",Lambda_Grid[Lambda_Total-1]);
    else printf(" none ");
    printf(" and after =");
    if (Lambda_Total<lambda_Max-1) 
      printf("%lf)\n",Lambda_Grid[Lambda_Total+1]);
    else printf(" none)\n");
  }

  /* read the outer radius (scaled by the inner sublimation distance) */
  err = _PATULJAK__Search_for_char (GLOBAL_FILE,"=");
  if (err) return 9;
  err = fscanf(GLOBAL_FILE,"%lf",&Rout);
  if (err!=1) return 9;
  /* debuger messages if required by DEBUG.SCREEN_MESSAGE */
  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("  **** Outer Radius = %lf\n",Rout);
  }

  return 0; /* everything is OK */
} /* end of _PATULJAK__Read_Tau_Total () */
