/* LELUYA - PERUNIKA program block */
/* This part of the LELUYA code is generating the computational grid */

/* This function is performing the analysis of a grid cell */
/* The cell can be split into smaller ones or left unchanged, depends on 
   the optical depth position of the cell in the dust cloud */
/* see HTML manual for details */

#include "PERUNIKA.h"

/* calculation of the minimal allowed optical depth between two points */
double _PERUNIKA_minimal_tau_line (struct _vertex_structure *pointA,
                               struct _vertex_structure *pointB);

int _PERUNIKA_analyze_triangle (struct _triangle_structure *parent_triangle,
                            int lambda_computational_grid,int status){
  /*
  parent_triangle = triangle which is going to be analyzed for possible
                    splitting into smaller ones
  lambda_computational_grid = wavelength at which we need to create 
                    the computational grid
  status          = if status = 0 or 1, the temperature and energy density 
                    values will be taken from the initial approximation arrays
                    if status = 2, then they will be taken from the grid
                    built in the previous iteration
  The notaion used for the parent_triangle:
  UP triangle:                DOWN triangle:
                v2                          side2
               /  \                     v3--------v2
        side2 /    \ side1               \        /
             /      \                     \      / 
            /        \               side3 \    / side1
           v3---------v1                    \  /
               side3                         v1
  */

  short int UP, DOWN;     /* 1 (TRUE) or 0 (FALSE) = orientation of 
                             parent_triangle */
  double ZETA = 2.0;      /* logarithmic scale in the optical depth criterium
                            for a new grid point */
  int err_location,err,error; /* error flags */
  int dust_i;             /* counter over the dust components */
  /* the geometrical size of the parent_triangle sides: */
  double size_side;
  /* optical depth between the vertices of the parent_triangle: */
  double tau_side1;       /* optical depth size of the side1 */
  double tau_side2;       /* optical depth size of the side2 */
  double tau_side3;       /* optical depth size of the side3 */
  /* a triangle's side will not split if its optical depth size is larger 
     then the minimal optical depth: */
  double min_tau_side1;   /* minimal optical depth size for the side1 */
  double min_tau_side2;   /* minimal optical depth size for the side2 */
  double min_tau_side3;   /* minimal optical depth size for the side3 */
  /* flags which show if the triangle's sides can split or not: */
  short int split_side1 = 0;
  short int split_side2 = 0;
  short int split_side3 = 0;
  /* flags which show if a triangle's side is crossing over a dust edge */
  short int split_side1_edge = 0;
  short int split_side2_edge = 0;
  short int split_side3_edge = 0;
  /* an edge point, used for search for the distance toward the edge
     of the dust cloud */
  double edge_point_x, edge_point_z;/* coordinates */
  double edge_point_distance;      /* distance from a point to the edge */
  double edge_point_tau;           /* optical depth from a point to the edge */
  /* coordinates of the triangle's centroid */
  double centroid_x, centroid_y;
  /* size limit at (centroid_x, centroid_y) */
  double Size_Limit_at_centorid;
  /* new grid points created on the sides of the parent_triangle */
  struct _vertex_structure *new_vertex1; /* on the side 1 */
  struct _vertex_structure *new_vertex2; /* on the side 2 */
  struct _vertex_structure *new_vertex3; /* on the side 3 */
  /* new grid triangles created from the new grid vertecies */
  struct _triangle_structure *triangle1; /* points (v1,new_v1,new_v3) */
  struct _triangle_structure *triangle2; /* points (new_v1,v2,new_v2) */
  struct _triangle_structure *triangle3; /* points (new_v3,new_v2,v3) */
  struct _triangle_structure *triangle4; /* points (new_v1,new_v2,new_v3) */
  /* teperature error from _PERUNIKA_temperature interpolation */
  double temp_error;
  /* variables without a special purpose */
  double dummy; 

  /* check first if the total number of vertices with dust is too big */
  if (TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES >=
                              LIMIT_ON_THE_TOTAL_NUMBER_OF_VERTICES) return 1;

  /* find the size of the triangle's sides */
  size_side=_MATH_LIB_distance(parent_triangle->v1->x,parent_triangle->v1->z,
                               parent_triangle->v2->x,parent_triangle->v2->z);
  /* find the centroid */
  centroid_x = (parent_triangle->v1->x+parent_triangle->v2->x+
	        parent_triangle->v3->x)/3.0;
  centroid_y = (parent_triangle->v1->z+parent_triangle->v2->z+
	        parent_triangle->v3->z)/3.0;

  /* size limit at this location */
  Size_Limit_at_centorid = _ZELENJAK_size_limit(centroid_x,centroid_y);

  /* is it UP or DOWN triangle? */
  if (parent_triangle->v2->z > parent_triangle->v1->z &&
      parent_triangle->v2->z > parent_triangle->v3->z &&
	parent_triangle->v2->x < parent_triangle->v1->x ) {UP=1; DOWN=0;}
  else {UP=0; DOWN=1;}

  /* in the first part of this function, we have to find out if the
     parent_triangle has to be divided into smaller triangles */
  if (parent_triangle->v1->x <= 0.0 &&
      parent_triangle->v2->x <= 0.0 &&
      parent_triangle->v3->x <= 0.0) {
    split_side1 = 0; split_side2 = 0;split_side3 = 0;
  }
  /* the first few levels of the triangular gird are created for sure */
  else if (parent_triangle->level <= START_LEVEL) {
    split_side1 = 1; split_side2 = 1;split_side3 = 1;
    /*if (DEBUG.SCREEN_MESSAGE >= 1 ) 
      printf("... level = %d\n",parent_triangle->level);*/
  }
  /******************************************************************/
  /******************************************************************/
  /********************* Ohnaka model ******************************
  else if (parent_triangle->level <= START_LEVEL+2  && 
           (parent_triangle->v1->T_dust[0]>0.0 || 
            parent_triangle->v2->T_dust[0]>0.0 || 
            parent_triangle->v3->T_dust[0]>0.0)) { 
    split_side1 = 1; split_side2 = 1; split_side3 = 1;
  }
  else if ( (centroid_x*centroid_x+centroid_y*centroid_y>100) &&
            (  (parent_triangle->v1->x/parent_triangle->v1->z>5.00 && 
                parent_triangle->v1->x/parent_triangle->v1->z<7.115) ||
               (parent_triangle->v2->x/parent_triangle->v2->z>5.00 && 
                parent_triangle->v2->x/parent_triangle->v2->z<7.115) ||
               (parent_triangle->v3->x/parent_triangle->v3->z>5.00 && 
	        parent_triangle->v3->x/parent_triangle->v3->z<7.115) ||
               (centroid_x/centroid_y>5.00 && centroid_x/centroid_y<7.115)
            ) &&
           // (size_side > Size_Limit_at_centorid) &&
              parent_triangle->level <= 6
	  ) {
    split_side1 = 1; split_side2 = 1; split_side3 = 1;
    Size_Limit_at_centorid *= 0.2;
  }
  else if ( (centroid_x*centroid_x+centroid_y*centroid_y>100) &&
            (centroid_x*centroid_x+centroid_y*centroid_y<360000) &&
            (  (parent_triangle->v1->x/parent_triangle->v1->z>5.00 &&
                parent_triangle->v1->x/parent_triangle->v1->z<6.3) ||
               (parent_triangle->v2->x/parent_triangle->v2->z>5.00 &&
                parent_triangle->v2->x/parent_triangle->v2->z<6.3) ||
               (parent_triangle->v3->x/parent_triangle->v3->z>5.00 &&
                parent_triangle->v3->x/parent_triangle->v3->z<6.3) ||
               (centroid_x/centroid_y>5.00 && centroid_x/centroid_y<6.3)
            ) &&
           // (size_side > Size_Limit_at_centorid) &&
              parent_triangle->level <= 8
          ) {
    split_side1 = 1; split_side2 = 1; split_side3 = 1;
    Size_Limit_at_centorid *= 0.05;
  }
  ******************************************************************/
  /******************************************************************/
  else {

    min_tau_side1 = _PERUNIKA_minimal_tau_line (parent_triangle->v1,
                                            parent_triangle->v2);
    min_tau_side2 = _PERUNIKA_minimal_tau_line (parent_triangle->v2,
                                            parent_triangle->v3);
    min_tau_side3 = _PERUNIKA_minimal_tau_line (parent_triangle->v3,
                                            parent_triangle->v1);
    /* split_side is an argument in the function _PERUNIKA_optical_depth_distance:
       split_side=1 means that the side is going over a dust edge, or over
       very steep slope of Eta */
    tau_side1=_PERUNIKA_optical_depth_distance 
                       (parent_triangle->v1->x,parent_triangle->v1->z,
                        parent_triangle->v2->x,parent_triangle->v2->z,
	  		lambda_computational_grid, status, &split_side1_edge);

    tau_side2=_PERUNIKA_optical_depth_distance 
                       (parent_triangle->v2->x,parent_triangle->v2->z,
		        parent_triangle->v3->x,parent_triangle->v3->z,
                        lambda_computational_grid,status, &split_side2_edge);

    tau_side3=_PERUNIKA_optical_depth_distance 
                       (parent_triangle->v3->x,parent_triangle->v3->z,
			parent_triangle->v1->x,parent_triangle->v1->z,
                        lambda_computational_grid,status, &split_side3_edge);

    /* check if the side1 needs to be divided */
    if (tau_side1 > min_tau_side1) {
      if (parent_triangle->v1->Eta <= 0.0 || parent_triangle->v2->Eta <= 0.0)
        split_side1 = 1;
      else if (parent_triangle->v1->Eta/parent_triangle->v2->Eta
               > ETA_RATIO_LIMIT  ||
               parent_triangle->v2->Eta/parent_triangle->v1->Eta
               > ETA_RATIO_LIMIT )
        split_side1 = 1;
      else {
	/* vertices v1 and v2 should be inside of the Rout circle */
        /* we have to find the optical depth between the points and the
           edge of the dust cloud. The dust cloud can reach the edge of
           the Rout disk, but no further than that. Therefore we can integrate 
           the eta values from a point toward the edge as long as the dust
           exists. When we reach eta=0 or r=Rout, we can stop and use this
           optical depth and distance to find if the triangle's side has to
           be divided into two smaller lengths */

	/* from the point v1 toward the edge point */ 
	dummy = parent_triangle->v1->x*(parent_triangle->v1->x - 
		parent_triangle->v2->x) +  parent_triangle->v1->z*
	        (parent_triangle->v1->z - parent_triangle->v2->z);
	edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                (Rout*Rout-parent_triangle->v1->x*parent_triangle->v1->x
                -parent_triangle->v1->z*parent_triangle->v1->z))
                -dummy)/size_side;
        edge_point_x = parent_triangle->v1->x+edge_point_distance*
                (parent_triangle->v1->x - parent_triangle->v2->x)/size_side;
        edge_point_z = parent_triangle->v1->z+edge_point_distance*
                (parent_triangle->v1->z - parent_triangle->v2->z)/size_side;
        edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v1,
		edge_point_x, edge_point_z,lambda_computational_grid,1,
                &edge_point_distance,status);
        if (edge_point_distance==0.0) split_side1 = 1;
	else 
          /*if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_)*/
            if (tau_side1 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	      split_side1 = 1;

	/* from the point v2 toward the edge point */ 
	if (split_side1 == 0 ){
	  dummy = parent_triangle->v2->x*(parent_triangle->v2->x - 
	  	  parent_triangle->v1->x) +  parent_triangle->v2->z*
	          (parent_triangle->v2->z - parent_triangle->v1->z);
	  edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                  (Rout*Rout-parent_triangle->v2->x*parent_triangle->v2->x
                  -parent_triangle->v2->z*parent_triangle->v2->z))
                  -dummy)/size_side;
          edge_point_x = parent_triangle->v2->x+edge_point_distance*
                  (parent_triangle->v2->x - parent_triangle->v1->x)/size_side;
          edge_point_z = parent_triangle->v2->z+edge_point_distance*
                  (parent_triangle->v2->z - parent_triangle->v1->z)/size_side;
          edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v2,
		  edge_point_x, edge_point_z,lambda_computational_grid,1,
                  &edge_point_distance,status);
          if (edge_point_distance==0.0) split_side1 = 1;
          else 
            /*if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_)*/
              if (tau_side1 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	        split_side1 = 1;
	}
      } /* end of the optical depth distance criterium */
    } /* end of the checking of the side1 */

    /* check if the side2 needs to be divided */
    if (tau_side2 > min_tau_side2) {
      if (parent_triangle->v2->Eta <= 0.0 || parent_triangle->v3->Eta <= 0.0)
        split_side2 = 1;
      else if (parent_triangle->v2->Eta/parent_triangle->v3->Eta
               > ETA_RATIO_LIMIT  ||
               parent_triangle->v3->Eta/parent_triangle->v2->Eta
               > ETA_RATIO_LIMIT )
        split_side2 = 1;
      else {
	/* vertices v2 and v3 should be inside of the Rout circle */
        /* we have to find the optical depth between the points and the
           edge of the dust cloud. The dust cloud can reach the edge of
           the Rout disk, but no further than that. Therefore we can integrate 
           the eta values from a point toward the edge as long as the dust
           exists. When we reach eta=0 or r=Rout, we can stop and use this
           optical depth and distance to find if the triangle's side has to
           be divided into two smaller lengths */

	/* from the point v2 toward the edge point */ 
	dummy = parent_triangle->v2->x*(parent_triangle->v2->x - 
		parent_triangle->v3->x) +  parent_triangle->v2->z*
	        (parent_triangle->v2->z - parent_triangle->v3->z);
	edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                (Rout*Rout-parent_triangle->v2->x*parent_triangle->v2->x
                -parent_triangle->v2->z*parent_triangle->v2->z))
                -dummy)/size_side;
        edge_point_x = parent_triangle->v2->x+edge_point_distance*
                (parent_triangle->v2->x - parent_triangle->v3->x)/size_side;
        edge_point_z = parent_triangle->v2->z+edge_point_distance*
                (parent_triangle->v2->z - parent_triangle->v3->z)/size_side;
        edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v2,
		edge_point_x, edge_point_z,lambda_computational_grid,1,
                &edge_point_distance,status);
        if (edge_point_distance==0.0) split_side2 = 1;
        else 
	 /* if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_) */
            if (tau_side2 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	      split_side2 = 1;

	/* from the point v3 toward the edge point */ 
	if (split_side2 == 0 ){
	  dummy = parent_triangle->v3->x*(parent_triangle->v3->x - 
	  	  parent_triangle->v2->x) +  parent_triangle->v3->z*
	          (parent_triangle->v3->z - parent_triangle->v2->z);
	  edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                  (Rout*Rout-parent_triangle->v3->x*parent_triangle->v3->x
                  -parent_triangle->v3->z*parent_triangle->v3->z))
                  -dummy)/size_side;
          edge_point_x = parent_triangle->v3->x+edge_point_distance*
                  (parent_triangle->v3->x - parent_triangle->v2->x)/size_side;
          edge_point_z = parent_triangle->v3->z+edge_point_distance*
                  (parent_triangle->v3->z - parent_triangle->v2->z)/size_side;
          edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v3,
		  edge_point_x, edge_point_z,lambda_computational_grid,1,
                  &edge_point_distance,status);
          if (edge_point_distance==0.0) split_side2 = 1;
          else  
	   /* if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_) */
              if (tau_side2 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	        split_side2 = 1;
	}
      } /* end of the optical depth distance criterium */
    } /* end of the checking of the side2 */

    /* check if the side3 needs to be divided */
    if (tau_side3 > min_tau_side3) {
      if (parent_triangle->v3->Eta <= 0.0 || parent_triangle->v1->Eta <= 0.0)
        split_side3 = 1;
      else if (parent_triangle->v3->Eta/parent_triangle->v1->Eta
               > ETA_RATIO_LIMIT  ||
               parent_triangle->v1->Eta/parent_triangle->v3->Eta
               > ETA_RATIO_LIMIT )
        split_side3 = 1;
      else {
	/* vertices v3 and v1 should be inside of the Rout circle */
        /* we have to find the optical depth between the points and the
           edge of the dust cloud. The dust cloud can reach the edge of
           the Rout disk, but no further than that. Therefore we can integrate 
           the eta values from a point toward the edge as long as the dust
           exists. When we reach eta=0 or r=Rout, we can stop and use this
           optical depth and distance to find if the triangle's side has to
           be divided into two smaller lengths */

	/* from the point v3 toward the edge point */ 
	dummy = parent_triangle->v3->x*(parent_triangle->v3->x - 
		parent_triangle->v1->x) +  parent_triangle->v3->z*
	        (parent_triangle->v3->z - parent_triangle->v1->z);
	edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                (Rout*Rout-parent_triangle->v3->x*parent_triangle->v3->x
                -parent_triangle->v3->z*parent_triangle->v3->z))
                -dummy)/size_side;
        edge_point_x = parent_triangle->v3->x+edge_point_distance*
                (parent_triangle->v3->x - parent_triangle->v1->x)/size_side;
        edge_point_z = parent_triangle->v3->z+edge_point_distance*
                (parent_triangle->v3->z - parent_triangle->v1->z)/size_side;
        edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v3,
		edge_point_x, edge_point_z,lambda_computational_grid,1,
                &edge_point_distance,status);
        if (edge_point_distance==0.0) split_side3 = 1;
        else  
	 /* if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_) */
            if (tau_side3 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	      split_side3 = 1;

	/* from the point v1 toward the edge point */ 
	if (split_side3 == 0 ){
	  dummy = parent_triangle->v1->x*(parent_triangle->v1->x - 
	  	  parent_triangle->v3->x) +  parent_triangle->v1->z*
	          (parent_triangle->v1->z - parent_triangle->v3->z);
	  edge_point_distance = (sqrt(dummy*dummy+size_side*size_side*
                  (Rout*Rout-parent_triangle->v1->x*parent_triangle->v1->x
                  -parent_triangle->v1->z*parent_triangle->v1->z))
                  -dummy)/size_side;
          edge_point_x = parent_triangle->v1->x+edge_point_distance*
                  (parent_triangle->v1->x - parent_triangle->v3->x)/size_side;
          edge_point_z = parent_triangle->v1->z+edge_point_distance*
                  (parent_triangle->v1->z - parent_triangle->v3->z)/size_side;
          edge_point_tau = _PERUNIKA_distance_to_edge(parent_triangle->v1,
		  edge_point_x, edge_point_z,lambda_computational_grid,1,
                  &edge_point_distance,status);
          if (edge_point_distance==0.0) split_side3 = 1;
          else  
	    /*if (edge_point_tau < _OPTICAL_DEPTH_LIMIT_) */
              if (tau_side3 > epsilon_GRID_ERROR*edge_point_tau*
		 (ZETA*edge_point_distance/(edge_point_distance+size_side)-1))
	        split_side3 = 1;
	}
      } /* end of the optical depth distance criterium */
    } /* end of the checking of the side3 */

    /* if the sides still do not get split, use the last chance: maybe they 
       are crossing over a dust edge */
    if (size_side > Size_Limit_at_centorid) {
      if (split_side1==0 && split_side1_edge==1) split_side1=1;
      if (split_side2==0 && split_side2_edge==1) split_side2=1;
      if (split_side3==0 && split_side3_edge==1) split_side3=1;
    }
  } /* end of (parent_triangle->level > START_LEVEL) */


  /* split the triangle if at least one side of the triangle needs to
     be divided and if the triangle is bigger then the minimal geometrical
     size allowed */
  if ( (split_side1==1 || split_side2==1 || split_side3==1) &&
       size_side > Size_Limit_at_centorid ) {

    /* find a new grid point between v1 and v2 */
    /* ... but check first if this point already exists */
    if ( (DOWN && parent_triangle->v1->v2 != parent_triangle->v2 ) ||
         ( UP  && parent_triangle->v1->v3 != parent_triangle->v2 ) ){
      /* vertices v1 and v2 are not connected, which implies that there is
         at least one point between them. The point in the middle of the side1
         should be there, too. */
      new_vertex1 = parent_triangle->v1;
      while (new_vertex1->level != parent_triangle->level+1 &&
             new_vertex1 != NULL) {
        if (DOWN) new_vertex1 = new_vertex1->v2;
        if ( UP ) new_vertex1 = new_vertex1->v3;
      }
      if (new_vertex1 == NULL ) {
	err_location = 9;
        err = 5;
        _PESJANEK__Error_Message (err_location,err);
        printf("grid level =%d, triangle's side 1\n",parent_triangle->level);
        return 2;
      }
    }
    else {
      /* vertices v1 and v2 are connected and we need to create the point
         in the middle of the side1 */
      new_vertex1 = _PERUNIKA_create_vertex (&error);
      new_vertex1->v1=new_vertex1->v2=new_vertex1->v3=new_vertex1->v4=
	new_vertex1->v5=new_vertex1->v6=NULL;
      if (DOWN) {
        new_vertex1->v2 = parent_triangle->v2;
        parent_triangle->v2->v5 = new_vertex1;
        new_vertex1->v5 = parent_triangle->v1;
        parent_triangle->v1->v2 = new_vertex1;
      }
      if ( UP ) {
        new_vertex1->v3 = parent_triangle->v2;
        parent_triangle->v2->v6 = new_vertex1;
        new_vertex1->v6 = parent_triangle->v1;
        parent_triangle->v1->v3 = new_vertex1;
      }
      new_vertex1->level = parent_triangle->level+1;
      new_vertex1->x = 0.5*(parent_triangle->v1->x+parent_triangle->v2->x);
      new_vertex1->z = 0.5*(parent_triangle->v1->z+parent_triangle->v2->z);
      /* check if the new point is out of the computational domain */
      /*if (new_vertex1->x<0 || new_vertex1->z<0 || 
          new_vertex1->x*new_vertex1->x + new_vertex1->z*new_vertex1->z
          > Rout*Rout)
	new_vertex1->Eta=0.0;  yes, the point is out the domain  
      else */{ 
	/* the point is inside of the domain, find Eta now */
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
	  temp_error=1.0;
          new_vertex1->T_dust[dust_i]=_PERUNIKA_temperature(new_vertex1->x,
                                    new_vertex1->z,dust_i,status,&temp_error);
	  if (status>1) new_vertex1->T_error[dust_i]=temp_error;
	}
        new_vertex1->Eta=Eta(new_vertex1->x,new_vertex1->z,
                        lambda_computational_grid,new_vertex1->T_dust,&error);
      }
      /* set the type of this point: 1=real point, 3=fake point */
      if (new_vertex1->Eta > 0.0 && new_vertex1->x >= 0.0) {
	new_vertex1->type = 1;
	for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
              new_vertex1->T_dust[dust_i]>0.0)
	    {TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES++;break;}
      }
      else {
        new_vertex1->type = 3;
      }      
    } /* end of the search for the point between v1 and v2 */

    /* find a new grid point between v2 and v3 */
    /* ... but check first if this point already exists */
    if ( (DOWN && parent_triangle->v2->v4 != parent_triangle->v3 ) ||
         ( UP  && parent_triangle->v2->v5 != parent_triangle->v3 ) ){
      /* vertices v2 and v3 are not connected, which implies that there is
         at least one point between them. The point in the middle of the side2
         should be there, too. */
      new_vertex2 = parent_triangle->v2;
      while (new_vertex2->level != parent_triangle->level+1 &&
             new_vertex2 != NULL) {
        if (DOWN) new_vertex2 = new_vertex2->v4;
        if ( UP ) new_vertex2 = new_vertex2->v5;
      }
      if (new_vertex2 == NULL ) {
	err_location = 9;
        err = 5;
        _PESJANEK__Error_Message (err_location,err);
        printf("grid level =%d, triangle's side 2\n",parent_triangle->level);
        return 2;
      }
    }
    else {
      /* vertices v2 and v3 are connected and we need to create the point
         in the middle of the side2 */
      new_vertex2 = _PERUNIKA_create_vertex (&error);
      new_vertex2->v1=new_vertex2->v2=new_vertex2->v3=new_vertex2->v4=
	new_vertex2->v5=new_vertex2->v6=NULL;
      if (DOWN) {
        new_vertex2->v4 = parent_triangle->v3;
        parent_triangle->v3->v1 = new_vertex2;
        new_vertex2->v1 = parent_triangle->v2;
        parent_triangle->v2->v4 = new_vertex2;
      }
      if ( UP ) {
        new_vertex2->v5 = parent_triangle->v3;
        parent_triangle->v3->v2 = new_vertex2;
        new_vertex2->v2 = parent_triangle->v2;
        parent_triangle->v2->v5 = new_vertex2;
      }
      new_vertex2->level = parent_triangle->level+1;
      new_vertex2->x = 0.5*(parent_triangle->v2->x+parent_triangle->v3->x);
      new_vertex2->z = 0.5*(parent_triangle->v2->z+parent_triangle->v3->z);
      /* check if the new point is out of the computational domain */
      /*if (new_vertex2->x<0 || new_vertex2->z<0 || 
          new_vertex2->x*new_vertex2->x + new_vertex2->z*new_vertex2->z
          > Rout*Rout)
	new_vertex2->Eta=0.0;  yes, the point is out of the domain  
      else */{
	/* the point is inside of the domain, find Eta now */
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) { 
	  temp_error=1.0; 
          new_vertex2->T_dust[dust_i]=_PERUNIKA_temperature(new_vertex2->x,
                                   new_vertex2->z,dust_i,status,&temp_error);
	  if (status>1) new_vertex2->T_error[dust_i]=temp_error;
	}
        new_vertex2->Eta=Eta(new_vertex2->x,new_vertex2->z,
                        lambda_computational_grid,new_vertex2->T_dust,&error);
      }
      /* set the type of this point: 1=real point, 3=fake point */
      if (new_vertex2->Eta > 0.0 && new_vertex2->x >= 0.0) {
	new_vertex2->type = 1;
	for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
              new_vertex2->T_dust[dust_i]>0.0)
	    {TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES++;break;}
      }
      else {
        new_vertex2->type = 3;
      }      
    } /* end of the search for the point between v2 and v3 */

    /* find a new grid point between v3 and v1 */
    /* ... but check first if this point already exists */
    if ( (DOWN && parent_triangle->v3->v6 != parent_triangle->v1 ) ||
         ( UP  && parent_triangle->v3->v1 != parent_triangle->v1 ) ){
      /* vertices v3 and v1 are not connected, which implies that there is
         at least one point between them. The point in the middle of the side3
         should be there, too. */
      new_vertex3 = parent_triangle->v3;
      while (new_vertex3->level != parent_triangle->level+1 &&
             new_vertex3 != NULL) {
        if (DOWN) new_vertex3 = new_vertex3->v6;
        if ( UP ) new_vertex3 = new_vertex3->v1;
      }
      if (new_vertex3 == NULL ) {
	err_location = 9;
        err = 5;
        _PESJANEK__Error_Message (err_location,err);
        printf("grid level =%d, triangle's side 3\n",parent_triangle->level);
        return 2;
      }
    }
    else {
      /* vertices v3 and v1 are connected and we need to create the point
         in the middle of the side3 */
      new_vertex3 = _PERUNIKA_create_vertex (&error);
      new_vertex3->v1=new_vertex3->v2=new_vertex3->v3=new_vertex3->v4=
	new_vertex3->v5=new_vertex3->v6=NULL;
      if (DOWN) {
        new_vertex3->v6 = parent_triangle->v1;
        parent_triangle->v1->v3 = new_vertex3;
        new_vertex3->v3 = parent_triangle->v3;
        parent_triangle->v3->v6 = new_vertex3;
      }
      if ( UP ) {
        new_vertex3->v1 = parent_triangle->v1;
        parent_triangle->v1->v4 = new_vertex3;
        new_vertex3->v4 = parent_triangle->v3;
        parent_triangle->v3->v1 = new_vertex3;
      }
      new_vertex3->level = parent_triangle->level+1;
      new_vertex3->x = 0.5*(parent_triangle->v3->x+parent_triangle->v1->x);
      new_vertex3->z = 0.5*(parent_triangle->v3->z+parent_triangle->v1->z);
      /* check if the new point is out of the computational domain */
      /*if (new_vertex3->x<0 || new_vertex3->z<0 || 
          new_vertex3->x*new_vertex3->x + new_vertex3->z*new_vertex3->z
          > Rout*Rout)
	new_vertex3->Eta=0.0;  yes, the point is out of the domain 
      else */{
	/* the point is inside of the domain, find Eta now */
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
	  temp_error=1.0;
          new_vertex3->T_dust[dust_i]=_PERUNIKA_temperature(new_vertex3->x,
                                    new_vertex3->z,dust_i,status,&temp_error);
	  if (status>1) new_vertex3->T_error[dust_i]=temp_error;
	}
        new_vertex3->Eta=Eta(new_vertex3->x,new_vertex3->z,
                        lambda_computational_grid,new_vertex3->T_dust,&error);
      }
      /* set the type of this point: 1=real point, 3=fake point */
      if (new_vertex3->Eta > 0.0 && new_vertex3->x >= 0.0) {
	new_vertex3->type = 1;
	for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
              new_vertex3->T_dust[dust_i]>0.0)
	    {TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES++;break;}
      }
      else {
        new_vertex3->type = 3;
      }      
    } /* end of the search for the point between v3 and v1 */

    /* connect the new points between themselves */
    if (DOWN) {
      new_vertex1->v3 = new_vertex2;
      new_vertex2->v6 = new_vertex1;
      new_vertex2->v5 = new_vertex3;
      new_vertex3->v2 = new_vertex2;
      new_vertex3->v1 = new_vertex1;
      new_vertex1->v4 = new_vertex3;
    }
    if (UP) {
      new_vertex1->v4 = new_vertex2;
      new_vertex2->v1 = new_vertex1;
      new_vertex2->v6 = new_vertex3;
      new_vertex3->v3 = new_vertex2;
      new_vertex3->v2 = new_vertex1;
      new_vertex1->v5 = new_vertex3;
    }

    /* now, when we have new points, create the new triangles */
    /* connect the points (v1,new_v1,new_v3) */
    triangle1=_PERUNIKA_create_triangle(&error);
    triangle1->level = parent_triangle->level+1;
    triangle1->v1 = parent_triangle->v1;
    triangle1->v2 = new_vertex1;
    triangle1->v3 = new_vertex3;
    /* connect the points (new_v1,v2,new_v2) */
    triangle2=_PERUNIKA_create_triangle(&error);
    triangle2->level = parent_triangle->level+1;
    triangle2->v1 = new_vertex1;
    triangle2->v2 = parent_triangle->v2;
    triangle2->v3 = new_vertex2;
    /* connect the points (new_v3,new_v2,v3) */
    triangle3=_PERUNIKA_create_triangle(&error);
    triangle3->level = parent_triangle->level+1;
    triangle3->v1 = new_vertex3;
    triangle3->v2 = new_vertex2;
    triangle3->v3 = parent_triangle->v3;
    /* connect the points (new_v3,new_v2,new_v1) */
    triangle4=_PERUNIKA_create_triangle(&error);
    triangle4->level = parent_triangle->level+1;
    if (DOWN) { triangle4->v1 = new_vertex1;
                triangle4->v2 = new_vertex2;
                triangle4->v3 = new_vertex3; }
    if ( UP ) { triangle4->v1 = new_vertex3;
                triangle4->v2 = new_vertex1;
                triangle4->v3 = new_vertex2; }

    /* destroy the parent_triangle */
    error = _PERUNIKA_destroy_triangle (parent_triangle);

    /* analyze the new triangles */
    error= _PERUNIKA_analyze_triangle (triangle1,lambda_computational_grid,status);
    if (error) return error;

    error= _PERUNIKA_analyze_triangle (triangle2,lambda_computational_grid,status);
    if (error) return error;

    error= _PERUNIKA_analyze_triangle (triangle3,lambda_computational_grid,status);
    if (error) return error;

    error= _PERUNIKA_analyze_triangle (triangle4,lambda_computational_grid,status);
    if (error) return error;



  } /* end of the creation of the new triangles */
  else { 
    /* what to do when all three sides should not be split? */
    /* destroy this triangle if it has all vertices out of the dust and/or
       computational domain (that is, if all vetices are fake, type=3)*/
    if (parent_triangle->v1->type == 3 &&
        parent_triangle->v2->type == 3 &&
        parent_triangle->v3->type == 3 )
      error = _PERUNIKA_destroy_triangle (parent_triangle);
 if (start_triangle_list==NULL)
   printf("KAJ JE TO? (destroy fake triangle)\n");

    /* .. or print a message because this triangle will be part of the grid */
    /* else if (DEBUG.SCREEN_MESSAGE >= 3 ) {
      printf("--new triangle: point1=(%E,%E) point2=(%E,%E) point3=(%E,%E)\n",
	     parent_triangle->v1->x,parent_triangle->v1->z,
             parent_triangle->v2->x,parent_triangle->v2->z,
             parent_triangle->v3->x,parent_triangle->v3->z);
	     }*/
  }  

  return 0;
} /* end of  _PERUNIKA_analyze_triangle  */

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/* calculation of the minimal allowed optical depth between two points */
double _PERUNIKA_minimal_tau_line (struct _vertex_structure *pointA,
                               struct _vertex_structure *pointB){
  /* the exponential radius dependence is used:
     _PERUNIKA_minimal_tau_line = tau_min_0 * EXP( r * tau_min_1)
  the following numbers will give:
    _PERUNIKA_minimal_tau_line(r=1)    = 0.01
    _PERUNIKA_minimal_tau_line(r=Rout) = 0.1  

  = ln(0.1)/(1-Rout)
  double tau_min_1 = -2.302585093/(1.0-Rout); 
  double tau_min_0 = 0.01 * exp(-tau_min_1);
  */

  /* linear dependence */
  double tau_min_1 = (0.1 - 0.01) / (Rout - 1.0);
  double tau_min_0 = 0.01 - tau_min_1;

  double r_A, r_B, min_r;
  double minimal_tau;

  r_A = sqrt(pointA->x*pointA->x + pointA->z*pointA->z);
  r_B = sqrt(pointB->x*pointB->x + pointB->z*pointB->z);
  if (r_A < r_B) min_r = r_A;
  else min_r = r_B;
  /*
  minimal_tau = tau_min_0 * exp( min_r * tau_min_1); 
  */
 
  minimal_tau = tau_min_0 +  min_r * tau_min_1;

  return minimal_tau;
}
