/* LELUYA - PERUNIKA program block */
/* This part of the LELUYA code is generating the computational grid */

/* This function creates an azimuthal triangular grid. This grid has vertexes
   placed on concentric rings around the star. 
NOTE: This grid _requires_ the "empty space" dust type to be active.
*/

#include "PERUNIKA.h"

/************************************************************************/
/* subroutine: find next rho or theta */
double _PERUNIKA_Find_next_rho_or_theta (struct _vertex_structure *V0, 
       double *Tau_back, double *Tau_front, double *D_back, double *D_front,
       int geometry_flag, double dL_limit,int *type_flag,
       int lambda_computational_grid,int status);

/************************************************************************/
/* subroutine: interconnect rings A and B */
int  _PERUNIKA_create_azimuthal_grid_between_A_B (
    struct _vertex_structure *Va_start, struct _vertex_structure *Va_end, 
    struct _vertex_structure *Vb_start, struct _vertex_structure *Vb_end, 
    int lambda_computational_grid,int status); 

/************************************************************************/
/* subroutine: temperature estimates */
int  _PERUNIKA_create_azimuthal_grid_T_estimate (
    struct _vertex_structure *V_current,  
    int lambda_computational_grid,int status); 

/************************************************************************/

int _PERUNIKA_create_azimuthal_grid (int lambda_computational_grid,int status){

  /*
  lambda_computational_grid = wavelength at which we need to create
                    the computational grid
  status          = if status = 0 or 1, the temperature and energy density
                    values will be taken from the initial approximation arrays
                    if status = 2, then they will be taken from the grid
                    built in the previous iteration

  Outline of this subroutine:
                 |
        Vb_start o
                 |\      This sketch shows two rings: ring A and ring B. 
        Va_start o \     These rings start with the vertexes Va_start and
                 |\ \    Vb_start, and end with Va_end and Vb_end.
                 | \ \     The grid is created when vertexes on these rings 
                 |__o_o__    are interconnected to form triangles.
                  Va_end
                      Vb_end

    - create Va_start at (rho=1, theta=0)
    - do {
    | - find next theta and create Va_end at (rho=1, theta)
    - } while (theta < 0.5*PI)
    - do {
    | - find next rho 
    | - create Vb_start at (rho, theta=0)
    | - do {
    | | - find next theta and create Vb_end at (rho, theta)
    | - } while (theta < 0.5*PI)
    | - create triangles, that is grid, between ring A and ring B
    | - Va_start = Vb_start; Va_end = Vb_end;
    - } while (rho < Rout)
  */

  struct _vertex_structure *Va_start, *Vb_start, *Va_end, *Vb_end;

  /* polar coordinates */
  double rho, rho_start, theta, rho_old, dr, dL_limit;
  /* variables needed for finding the next rho or theta */
  double Tau_back, Tau_front, D_back, D_front;
  double Tau_back_rho, Tau_front_rho, D_back_rho, D_front_rho;
  struct _vertex_structure *vertex_rho_search;
  struct _vertex_structure *vertex_theta_search;
  int type_flag, dust_edge;

  int dust_i;
  double temp_error;
  int err;

  /*--------------------------------------------------------*/
  /* create the first ring at rho=1 */
  Va_start = _PERUNIKA_create_vertex (&err);
  Va_start->level= 667; /* it prevents close points removal*/
  Va_start->type = 1;
  /* find the initial rho */
  rho_start = 1.00000000005;
  Va_start->x = rho_start*sin(Theta_Total_Value);
  Va_start->z = rho_start*cos(Theta_Total_Value);
  dust_edge = TRUE;
  do {
    rho_start = (rho_start - 1.0)*2.0 + 1.0;

    Va_start->x = rho_start*sin(Theta_Total_Value);
    Va_start->z = rho_start*cos(Theta_Total_Value);
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_error=1.0;
      Va_start->T_dust[dust_i]=_PERUNIKA_temperature(Va_start->x,Va_start->z,
                                              dust_i,status,&temp_error);
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
        if (status<2) 
          if (Dust_Density(Va_start->x,Va_start->z,dust_i,NULL,&err)<=0.0)
            Va_start->T_dust[dust_i]=-1;
        if (Va_start->T_dust[dust_i]>0.0) dust_edge = FALSE;
      }
    }
  } while (dust_edge);
  rho_old = rho_start;
  rho = rho_start;

  /* now set this point to be at (0,1) */
  Va_start->x = 0.0;
  Va_start->z = rho;

  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    temp_error=1.0;
    Va_start->T_dust[dust_i]=_PERUNIKA_temperature(Va_start->x,Va_start->z,
                                              dust_i,status,&temp_error);
    if (status>1) Va_start->T_error[dust_i]=temp_error;
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      if (status<2) 
        if (Dust_Density(Va_start->x,Va_start->z,dust_i,NULL,&err)<=0.0)
          Va_start->T_dust[dust_i]=-1;
      if (Va_start->T_dust[dust_i]>0.0) Va_start->type = 2;
    }
  }
  Va_start->Eta=Eta(Va_start->x,Va_start->z,
                    lambda_computational_grid,Va_start->T_dust,&err);
  /****/

  /* vertex used only for searching for theta */
  vertex_theta_search = (struct _vertex_structure *) malloc (sizeof(struct _vertex_structure));
  vertex_theta_search->x = Va_start->x;
  vertex_theta_search->z = Va_start->z;
  vertex_theta_search->T_dust = dvector(0,Max_Dust_Type-1);
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    vertex_theta_search->T_dust[dust_i]=Va_start->T_dust[dust_i];
  }
  vertex_theta_search->Eta=Va_start->Eta;
  /***/
  /* vertex used only for searching for rho */
  vertex_rho_search = (struct _vertex_structure *) malloc (sizeof(struct _vertex_structure));
  if (DEBUG.AZIMUTHAL_GRID==3) {
    vertex_rho_search->x = 0.0;
    vertex_rho_search->z = rho;
  }
  else {
    vertex_rho_search->x = rho*sin(Theta_Total_Value);
    vertex_rho_search->z = rho*cos(Theta_Total_Value);
  }
  vertex_rho_search->T_dust = dvector(0,Max_Dust_Type-1);
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    temp_error=1.0;
    vertex_rho_search->T_dust[dust_i]=_PERUNIKA_temperature(vertex_rho_search->x,
                       vertex_rho_search->z,dust_i,status,&temp_error);
    if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
      if (Dust_Density(vertex_rho_search->x,vertex_rho_search->z,dust_i,NULL,&err)<=0.0)
        vertex_rho_search->T_dust[dust_i]=-1;

  }
  vertex_rho_search->Eta=Eta(vertex_rho_search->x,vertex_rho_search->z,
                      lambda_computational_grid,vertex_rho_search->T_dust,&err);
  /***/
  if (DEBUG.AZIMUTHAL_GRID==2) {
    /* find the initial dL_limit - for that we need dr */
    Tau_back_rho=Tau_front_rho=D_back_rho=D_front_rho=0.0;
    rho_old = _PERUNIKA_Find_next_rho_or_theta (vertex_rho_search,&Tau_back_rho,
                    &Tau_front_rho,&D_back_rho,&D_front_rho,1,0.0,&type_flag,
                    lambda_computational_grid,status);
    dr = rho_old - rho;
    rho_old = rho;
    dL_limit = rho*2.0*acos(1.0 - dr/rho);
  }
  else 
    dL_limit = 0.0;

  theta = 0.0;
  Tau_back=Tau_front=D_back=D_front=0.0;
  do {
    /* find next theta angle */ 
    theta = _PERUNIKA_Find_next_rho_or_theta (vertex_theta_search,&Tau_back,&Tau_front,
                               &D_back,&D_front,2,dL_limit,&type_flag,
                               lambda_computational_grid,status);
    /* check for the R~1 zone */
    if (DEBUG.AZIMUTHAL_GRID==3) 
      if (theta>DEBUG.AZIMUTHAL_GRID1) theta = DEBUG.AZIMUTHAL_GRID1;
    /* create new vertex */
    Va_end = _PERUNIKA_create_vertex (&err);
    Va_end->level= 667; /* it prevents close points removal*/
    Va_end->type = 1;
    Va_end->x = rho*sin(theta);
    Va_end->z = rho*cos(theta);
    if (theta==0.5*PI) {Va_end->x = rho; Va_end->z=0.0;} 
    if (DEBUG.AZIMUTHAL_GRID==3 && vertex_rho_search->x!=0.0) {
      if (theta>DEBUG.AZIMUTHAL_GRID1) {
        Va_end->x = rho;
        if (theta==0.5*PI) {Va_end->x = rho; Va_end->z=0.0;} 
        else Va_end->z = rho/tan(theta);
      }
      else {
        Va_end->x /= sin(DEBUG.AZIMUTHAL_GRID1);
        Va_end->z /= sin(DEBUG.AZIMUTHAL_GRID1);
      }
    }

    printf("V[%lf,%lf] rho=%12.10lf theta=%lf t_b=%lf t_f=%lf D_b=%lf D_f=%lf\n",
       Va_end->x,Va_end->z,rho,theta/RAD1,Tau_back,Tau_front,D_back,D_front);
    dust_edge = FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_error=1.0;
      Va_end->T_dust[dust_i]=_PERUNIKA_temperature(Va_end->x,Va_end->z,
                                            dust_i,status,&temp_error);
      if (status>1) Va_end->T_error[dust_i]=temp_error;

      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
        if (status<2) 
          if (Dust_Density(Va_end->x,Va_end->z,dust_i,NULL,&err)<=0.0)
            Va_end->T_dust[dust_i]=-1;
        if (Va_end->T_dust[dust_i]>0.0) Va_end->type = 2;
      }
      if (Va_end->T_dust[dust_i]*Va_end->prev->T_dust[dust_i]<0.0) {
	printf("edge for #%d, ",dust_i);
	dust_edge = TRUE;
      }
    }
    if (dust_edge) printf(" at H=%13.11E\n",Va_end->z);

    Va_end->Eta=Eta(Va_end->x,Va_end->z,
                    lambda_computational_grid,Va_end->T_dust,&err);

    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
          Va_end->T_dust[dust_i]>0.0)
        if (Tau_back<2.0 || Tau_front<2.0) 
          {Va_end->level= 668; break;} /* no compromise on
                                          griding method close to the surface */

    /****/
    vertex_theta_search->x = Va_end->x;
    vertex_theta_search->z = Va_end->z;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      vertex_theta_search->T_dust[dust_i]=Va_end->T_dust[dust_i];
    }
    vertex_theta_search->Eta=Va_end->Eta;
    /***/

    /* exit if this is R~1 zone */
    if (DEBUG.AZIMUTHAL_GRID==3) 
      if (theta==DEBUG.AZIMUTHAL_GRID1) theta = PI;

  } while (theta < 0.5*PI); 

  /*--------------------------------------------------------*/
  /* create other (concentric) rings all the way up to Rout */
  
  Tau_back_rho=Tau_front_rho=D_back_rho=D_front_rho=0.0;
  do {

    /* find next rho */ 
    rho = _PERUNIKA_Find_next_rho_or_theta (vertex_rho_search,&Tau_back_rho,
                    &Tau_front_rho,&D_back_rho,&D_front_rho,1,0.0,&type_flag,
                    lambda_computational_grid,status);

    if (DEBUG.AZIMUTHAL_GRID==3 &&
        /* check if we need to exit the R~1 zone */
        vertex_rho_search->x == 0.0 && 
        rho*sin(DEBUG.AZIMUTHAL_GRID1)>=1.0) {
      
      rho = rho_start;
      rho_old = rho_old*sin(DEBUG.AZIMUTHAL_GRID1);
      vertex_rho_search->x = rho;
      vertex_rho_search->z = 0.0;
      dr = rho - rho_old;
      Tau_back_rho=Tau_front_rho=D_back_rho=D_front_rho=0.0;
    }
    else {
      dr = rho - rho_old;
      if (vertex_rho_search->x == 0.0) vertex_rho_search->z = rho;
      else {
        vertex_rho_search->x = rho*sin(Theta_Total_Value);
        vertex_rho_search->z = rho*cos(Theta_Total_Value);
      }
    }

    dL_limit = rho*2.0*acos(1.0 - dr/rho);
    printf("new rho=%15.13E t_b=%E t_f=%E D_b=%E D_f=%lf DL=%lf deg\n",rho,Tau_back_rho,
           Tau_front_rho,D_back_rho,D_front_rho,dL_limit/rho/RAD1);
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_error=1.0;
      vertex_rho_search->T_dust[dust_i]=_PERUNIKA_temperature(vertex_rho_search->x,
                       vertex_rho_search->z,dust_i,status,&temp_error);
      if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
        if (Dust_Density(vertex_rho_search->x,vertex_rho_search->z,dust_i,NULL,&err)<=0.0)
          vertex_rho_search->T_dust[dust_i]=-1;

    }
    vertex_rho_search->Eta=Eta(vertex_rho_search->x,vertex_rho_search->z,
                      lambda_computational_grid,vertex_rho_search->T_dust,&err);
    /***/

    /* create the start of B ring */
    Vb_start = _PERUNIKA_create_vertex (&err);
    Vb_start->level= 667; /* it prevents close points removal*/
    Vb_start->type = 1;
    Vb_start->x = 0.0;
    Vb_start->z = rho;
    if (DEBUG.AZIMUTHAL_GRID==3) {
      if (rho>Rout*sin(DEBUG.AZIMUTHAL_GRID1)+1E-5) {
        Vb_start->x = rho;
	Vb_start->z = sqrt(Rout*Rout - rho*rho);
	theta = atan(Vb_start->x/Vb_start->z);
      }
      else {
	if (vertex_rho_search->x!=0.0)
          Vb_start->z /=  sin(DEBUG.AZIMUTHAL_GRID1); 
	theta = 0.0;  
      }
    }

    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_error=1.0;
      Vb_start->T_dust[dust_i]=_PERUNIKA_temperature(Vb_start->x,Vb_start->z,
                                            dust_i,status,&temp_error);
      if (status>1) Vb_start->T_error[dust_i]=temp_error;
      if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
        if (Dust_Density(Vb_start->x,Vb_start->z,dust_i,NULL,&err)<=0.0)
          Vb_start->T_dust[dust_i]=-1;
    }
    Vb_start->Eta=Eta(Vb_start->x,Vb_start->z,
                      lambda_computational_grid,Vb_start->T_dust,&err);

    if (DEBUG.AZIMUTHAL_GRID==3)
      if (rho>Rout*sin(DEBUG.AZIMUTHAL_GRID1)+1E-5) {
	/* special situation: the vertical edge */
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          temp_error=1.0;
          Vb_start->T_dust[dust_i]=_PERUNIKA_temperature(Vb_start->x,0.95*Vb_start->z,
                                            dust_i,status,&temp_error);
          if (status>1) Vb_start->T_error[dust_i]=temp_error;
          if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
          if (Dust_Density(Vb_start->x,0.95*Vb_start->z,dust_i,NULL,&err)<=0.0)
              Vb_start->T_dust[dust_i]=-1;
        }
        Vb_start->Eta=Eta(Vb_start->x,0.95*Vb_start->z,
                      lambda_computational_grid,Vb_start->T_dust,&err);
      }
    /****/
    vertex_theta_search->x = Vb_start->x;
    vertex_theta_search->z = Vb_start->z;
    if (DEBUG.AZIMUTHAL_GRID==3)
      if (rho>Rout*sin(DEBUG.AZIMUTHAL_GRID1)+1E-5) {
	if (vertex_theta_search->z>0.01) vertex_theta_search->z -= 0.01;
        else vertex_theta_search->z *=0.99;

      }
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      vertex_theta_search->T_dust[dust_i]=Vb_start->T_dust[dust_i];
    }
    vertex_theta_search->Eta=Vb_start->Eta;
    /***/

    Tau_back=Tau_front=D_back=D_front=0.0;
    if (DEBUG.AZIMUTHAL_GRID==3 && rho==Rout) {
      Vb_end = Vb_start;
    }
    else do {
      /**** outer edge is not a circle: *****/
      if (rho>0.995*Rout && status==2)
        if ((DEBUG.AZIMUTHAL_GRID==3 && theta<DEBUG.AZIMUTHAL_GRID1) ||
             DEBUG.AZIMUTHAL_GRID!=3) {
        vertex_theta_search->x = 0.995*Rout*sin(theta);
        vertex_theta_search->z = 0.995*Rout*cos(theta);
      }
      /**********************/

      /* find next theta angle */ 
      theta = _PERUNIKA_Find_next_rho_or_theta (vertex_theta_search,&Tau_back,&Tau_front,
                               &D_back,&D_front,2,dL_limit,&type_flag,
                               lambda_computational_grid,status);
      /* check for the R~1 zone */
      if (DEBUG.AZIMUTHAL_GRID==3) 
        if (theta>DEBUG.AZIMUTHAL_GRID1 && vertex_rho_search->x == 0.0 ) 
           theta = DEBUG.AZIMUTHAL_GRID1;
      /* create new vertex */
      Vb_end = _PERUNIKA_create_vertex (&err);
      Vb_end->level= 667; /* it prevents close points removal*/
      Vb_end->type = type_flag;
      Vb_end->x = rho*sin(theta);
      Vb_end->z = rho*cos(theta);
      if (theta==0.5*PI) {Vb_end->x = rho; Vb_end->z=0.0;} 
      if (DEBUG.AZIMUTHAL_GRID==3 && vertex_rho_search->x!=0.0){
        if (theta>DEBUG.AZIMUTHAL_GRID1) {
          Vb_end->x = rho;
          if (theta==0.5*PI) {Vb_end->x = rho; Vb_end->z=0.0;} 
          else Vb_end->z = rho/tan(theta);
        }
        else {
          Vb_end->x /= sin(DEBUG.AZIMUTHAL_GRID1);
          Vb_end->z /= sin(DEBUG.AZIMUTHAL_GRID1);
        }
      }

      if (rho==Rout || (DEBUG.AZIMUTHAL_GRID==3 && rho>0.9*Rout))
        printf("V[%lf,%lf] rho=%lf theta=%lf t_b=%lf t_f=%lf D_b=%lf D_f=%lf\n",
               Vb_end->x,Vb_end->z,rho,theta/RAD1,Tau_back,Tau_front,D_back,D_front);
      dust_edge = FALSE;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        temp_error=1.0;
        Vb_end->T_dust[dust_i]=_PERUNIKA_temperature(Vb_end->x,Vb_end->z,
                                            dust_i,status,&temp_error);
	/**** outer edge is not a circle: *****/
      if (sqrt(Vb_end->x*Vb_end->x+Vb_end->z*Vb_end->z)>0.995*Rout && status==2)
        if ((DEBUG.AZIMUTHAL_GRID==3 && theta<DEBUG.AZIMUTHAL_GRID1) ||
             DEBUG.AZIMUTHAL_GRID!=3) 
          Vb_end->T_dust[dust_i]=_PERUNIKA_temperature(0.995*Rout*sin(theta),0.995*Rout*cos(theta), 
                                                   dust_i,status,&temp_error);
	/**********************/
        if (status>1) Vb_end->T_error[dust_i]=temp_error;
        if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
          if (Dust_Density(Vb_end->x,Vb_end->z,dust_i,NULL,&err)<=0.0)
            Vb_end->T_dust[dust_i]=-1;

        if (Vb_end->T_dust[dust_i]*Vb_end->prev->T_dust[dust_i]<0.0) {
          printf("edge for #%d, ",dust_i);
          dust_edge = TRUE;
	}
      }
      if (dust_edge) {
/*
t=NULL;t_start=start_triangle_list;
start_triangle_list=old_grid_per_lambda[0].start_triangle;
_ZELENJAK_locate_point_in_grid(Vb_end->x,Vb_end->z,&t);
printf(" at [%d,%d,%d]",t->v1->identification_number,
               t->v2->identification_number,t->v3->identification_number);
start_triangle_list=t_start;
*/
        printf(" at H=%13.11E \n",Vb_end->z);
printf("<%13.11lf, %13.11lf>, \n",Vb_end->x,Vb_end->z);
      }

      Vb_end->Eta=Eta(Vb_end->x,Vb_end->z,
                      lambda_computational_grid,Vb_end->T_dust,&err);
      /**** outer edge is not a circle: *****/
      if (rho>0.995*Rout && status==2)
        if ((DEBUG.AZIMUTHAL_GRID==3 && theta<DEBUG.AZIMUTHAL_GRID1) ||
             DEBUG.AZIMUTHAL_GRID!=3) 
        Vb_end->Eta=Eta(0.995*Rout*sin(theta),0.995*Rout*cos(theta),
                      lambda_computational_grid,Vb_end->T_dust,&err);
      /**********************/
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
        if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
            Vb_end->T_dust[dust_i]>0.0)
          if ((Tau_back<2000.0 || Tau_front<2000.0))
            {Vb_end->level= 668; break;} /* no compromise on
                                            griding method close to the surface */

      /****************************************************************/
      /****************************************************************/
      /********** diffusion approximation for the disk interior *******/
      if (DEBUG.AZIMUTHAL_GRID==-3) 
	if (Tau_back_rho>50.0 && Tau_front_rho>50.) 
        /* if (Vb_end->x>1.000865 && Vb_end->x<1.0072 && Vb_end->z/Vb_end->x < 0.001) */
	  if (Vb_end->T_dust[0]>0.0 && Vb_end->prev->T_dust[0]>0.0) {
  
	    _PERUNIKA_create_azimuthal_grid_T_estimate (Vb_end,
			    lambda_computational_grid,status);
      }
      /****************************************************************/
      /****************************************************************/

      /****/
      vertex_theta_search->x = Vb_end->x;
      vertex_theta_search->z = Vb_end->z;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        vertex_theta_search->T_dust[dust_i]=Vb_end->T_dust[dust_i];
      }
      vertex_theta_search->Eta=Vb_end->Eta;
      /***/

      /* exit for the R~1 zone */
      if (DEBUG.AZIMUTHAL_GRID==3) 
        if (vertex_rho_search->x == 0.0 && theta==DEBUG.AZIMUTHAL_GRID1)
          theta=PI;

    } while (theta < 0.5*PI); 

    /* connect rings A and B */
    err=_PERUNIKA_create_azimuthal_grid_between_A_B(Va_start,Va_end,Vb_start,Vb_end,
                                            lambda_computational_grid,status);
    if (err==1) Vb_start = Va_start; /* breaching occurred */

    /* prepare for the next step */
    Va_start = Vb_start;
    Va_end = Vb_end;
    if (DEBUG.AZIMUTHAL_GRID==3)
      if (rho_old<Rout*sin(DEBUG.AZIMUTHAL_GRID1-1E-5) && 
          rho    >Rout*sin(DEBUG.AZIMUTHAL_GRID1-1E-5)) {
	while (Va_start->x/Va_start->z < tan(DEBUG.AZIMUTHAL_GRID1)-1E-5 )
          Va_start=Va_start->next;
    }
    rho_old = rho;

  } while (rho<Rout);

  /* clean variables */
  free_dvector(vertex_rho_search->T_dust,0,Max_Dust_Type-1);
  free(vertex_rho_search);
  free_dvector(vertex_theta_search->T_dust,0,Max_Dust_Type-1);
  free(vertex_theta_search);

  /* set the last ring to be level=667 */
  vertex_theta_search = Va_start;
  while (vertex_theta_search!=Va_end) {
     vertex_theta_search->level=667;
     vertex_theta_search = vertex_theta_search->next;
  }

  return 0;

} /* end of subroutine */

/************************************************************************/
/************************************************************************/
/* subroutine: interconnect rings A and B */
int  _PERUNIKA_create_azimuthal_grid_between_A_B (
    struct _vertex_structure *Va_start, struct _vertex_structure *Va_end, 
    struct _vertex_structure *Vb_start, struct _vertex_structure *Vb_end, 
    int lambda_computational_grid,int status){ 

  /* running vertexes on the rings */
  struct _vertex_structure *Va1, *Va2, *Vb1, *Vb2, *Vsearch; 
  /* flag showing what triangle to create: one connecting (Va1,Va2,Vb1) 
     called trainA, or (Va1,Vb2,Vb1) called trianB */
  int flag_A, flag_B, oldflag_A, oldflag_B, front_dust;
  /* the angle at Va2 and Vb2 */
  double angleA, angleB,radiusA,radiusB,tan_Theta0,temp_error; 
  /* triangle */
  struct _triangle_structure *trian;
  /* "breaching" vertexes: in case we want to remove a part of the B ring */
  struct _vertex_structure *breachA, *breachB;
  int error,dust_i;
  /* variables for keeping the inventory of rings' dust properties */
  int *dust_existA, *dust_existB, TstepA, TstepB;

  dust_existA = ivector(0,Max_Dust_Type);
  dust_existB = ivector(0,Max_Dust_Type);
  dust_existA[Max_Dust_Type] = FALSE;
  dust_existB[Max_Dust_Type] = FALSE;
  tan_Theta0 = tan(DEBUG.AZIMUTHAL_GRID1);
  breachA = breachB = NULL;

  /* scan the rings for existing dust types */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    dust_existA[dust_i] = FALSE;
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      Va1 = Va_start;
      while (Va1!=Va_end) {
        if (Va1->T_dust[dust_i]>0.0) {dust_existA[dust_i] = TRUE; break;}
        Va1 = Va1->next;
      }
    }
  }
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    dust_existB[dust_i] = FALSE;
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      Vb1 = Vb_start;
      while (Vb1!=Vb_end) {
        if (Vb1->T_dust[dust_i]>0.0) {dust_existB[dust_i] = TRUE; break;}
        Vb1 = Vb1->next;
      }
    }
  }

  /* if one dust type exists in one ring, but in the other, then mark those
     vertexes as type=2 */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      if (dust_existA[dust_i] && (!dust_existB[dust_i])){
	/* ring A has edge vertexes */
        Va1 = Va_start;
        while (Va1!=Va_end) {
          if (Va1->T_dust[dust_i]>0.0) Va1->type = 2;
          Va1 = Va1->next;
        }
      }
      else if (dust_existB[dust_i] && (!dust_existA[dust_i])){
	/* ring B has edge vertexes */
        Vb1 = Vb_start;
        while (Vb1!=Vb_end) {
          if (Vb1->T_dust[dust_i]>0.0) Vb1->type = 2;
          Vb1 = Vb1->next;
        }
      }
    }
  /* find dust steps on a ring itself */
  Va1 = Va_start;
  while (Va1!=Va_end) {
    Va2 = Va1->next;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
        if (Va1->T_dust[dust_i]>0.0 && Va2->T_dust[dust_i]<0.0) 
          {Va1->type = 2;break;}
        if (Va1->T_dust[dust_i]<0.0 && Va2->T_dust[dust_i]>0.0) 
          {Va2->type = 2;break;}
      }
    Va1 = Va2;
  }
  Vb1 = Vb_start;
  while (Vb1!=Vb_end) {
    Vb2 = Vb1->next;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
        if (Vb1->T_dust[dust_i]>0.0 && Vb2->T_dust[dust_i]<0.0) 
          {Vb1->type = 2;break;}
        if (Vb1->T_dust[dust_i]<0.0 && Vb2->T_dust[dust_i]>0.0) 
          {Vb2->type = 2;break;}
      }
    Vb1 = Vb2;
  }

  /* start connecting ring A with ring B */
  Va1 = Va_start; Va2 = Va1->next;
  Vb1 = Vb_start; Vb2 = Vb1->next;
  radiusA=sqrt(Va1->x*Va1->x+Va1->z*Va1->z);
  radiusB=sqrt(Vb1->x*Vb1->x+Vb1->z*Vb1->z);

  if (DEBUG.AZIMUTHAL_GRID==3 && radiusB<0.7*Rout && 
      radiusB*sin(DEBUG.AZIMUTHAL_GRID1)>1.0) 
  /* check if the A and B rings are too close to each other */
    if (radiusB-radiusA< 1.0E-3) {

      /* remove a part of the B ring */
      do {
	Vb1->prev->next = Vb2;
	Vb2->prev = Vb1->prev;
	_PERUNIKA_destroy_vertex(Vb1);
	Vb1 = Vb2;
	Vb2 = Vb2->next;
      } while (Vb2->x < Vb2->z*tan_Theta0+1E-7);
      /* "scroll" the A ring */
      do {
	Va1 = Va2;
	Va2 = Va2->next;
      } while (Va2->x < Va2->z*tan_Theta0-1E-7);
      /* we want only convex rings */
      temp_error = _MATH_LIB_nu_value(Vb1->x,Vb1->z,Va1->x,Va1->z,Va2->x,Va2->z);
      if (temp_error > 1.0 || temp_error<= 0.0) {
	Vb1->x = Vb2->x;
	Vb1->z = Vb1->x / tan_Theta0;
      }
      /* memorize the contact points for breaching between the rings */
      breachA = Va1;
      breachB = Vb1;
      /* triangle */
      trian=_PERUNIKA_create_triangle(&error);
      trian->level = 1;
      trian->v1 = Va1;
      trian->v2 = Va2;
      trian->v3 = Vb1;
      /* prepare for connecting the rings */
      Va1 = Va2;
      Va2 = Va2->next;
    }
  
  /* start connecting ring A with ring B */
  oldflag_A = FALSE; oldflag_B=FALSE;
  while (Va1!=Va_end || Vb1!=Vb_end) {
    flag_A = FALSE; flag_B=FALSE;
    if (Va1==Va_end) flag_B=TRUE;
    else if (Vb1==Vb_end) flag_A=TRUE;
    else {
      /* check for dust steps */
      TstepA=TstepB=Max_Dust_Type;
      if (DEBUG.AZIMUTHAL_GRID==3) temp_error = tan_Theta0;
      else temp_error = 0.0;
      if (Va1->x >= Va1->z*temp_error && Vb1->x >= Vb1->z*temp_error)
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
        if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
          if (Va2->T_dust[dust_i]*Vb1->T_dust[dust_i] < 0.0) TstepA=dust_i;
          else TstepA=Max_Dust_Type;
          if (Vb2->T_dust[dust_i]*Va1->T_dust[dust_i] < 0.0) TstepB=dust_i;
          else TstepB=Max_Dust_Type;
	  if (TstepA<Max_Dust_Type || TstepB<Max_Dust_Type) break;
	}
      /* check for large ring curvatures */
      angleA = _MATH_LIB_nu_value(Va2->x,Va2->z,Vb1->x,Vb1->z,Va1->x,Va1->z);
      angleB = _MATH_LIB_nu_value(Vb2->x,Vb2->z,Vb1->x,Vb1->z,Va1->x,Va1->z);

      if (angleA <=0.0 && angleB>0.0) flag_B=TRUE;
      else if (angleB <=0.0 && angleA>0.0) flag_A=TRUE;
      else if (angleA <=0.0 && angleB <=0.0) {
        /* create an intermediate vertex to fix this problem */
        Vb2 = _PERUNIKA_create_vertex (&error);
        Vb2->level= 667; /* it prevents close points removal*/
        Vb2->type = 1;
        radiusA=sqrt(Va1->x*Va1->x+Va1->z*Va1->z);
        radiusB=sqrt(Vb1->x*Vb1->x+Vb1->z*Vb1->z);
        Vb2->x = radiusB*Va1->x/radiusA;
        Vb2->z = radiusB*Va1->z/radiusA;

	if (DEBUG.AZIMUTHAL_GRID==3) {
	  /* check if this is a point on DEBUG.AZIMUTHAL_GRID1 */
	  if (Vb1->x <= Vb1->z*tan_Theta0 && 
              Vb1->next->x > Vb1->next->z*tan_Theta0) {
            Vb2->x = Vb1->next->x;
            Vb2->z = Vb1->next->x / tan_Theta0;
	  }
	}

        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          temp_error=1.0;
          Vb2->T_dust[dust_i]=_PERUNIKA_temperature(Vb2->x,Vb2->z,
                                                dust_i,status,&temp_error);
          if (status<2 && INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
            if (Dust_Density(Vb2->x,Vb2->z,dust_i,NULL,&error)<=0.0)
              Vb2->T_dust[dust_i]=-1;
          if (status>1) Vb2->T_error[dust_i]=temp_error;
          if (Vb2->T_dust[dust_i]*Vb1->T_dust[dust_i]<0.0 ||
              Vb2->T_dust[dust_i]*Vb1->next->T_dust[dust_i]<0.0 ||
              Vb2->T_dust[dust_i]*Va1->T_dust[dust_i]<0.0)
            Vb2->type = 2;
        }
        Vb2->Eta=Eta(Vb2->x,Vb2->z,lambda_computational_grid,Vb2->T_dust,&error);
        /* relocate this vertex in the vertex list */
        Vb2->prev->next = Vb2->next;
        Vb2->next->prev = Vb2->prev;
        Vb2->next = Vb1->next;
        Vb1->next->prev = Vb2;
        Vb2->prev = Vb1;
        Vb1->next = Vb2;
        /* activate triangle creation */
        flag_B=TRUE;
      }
      /* crossing a dust step */
      else if (TstepA<Max_Dust_Type && TstepB==Max_Dust_Type &&
               Va1->T_dust[TstepA]>0.0) {
	if (Va2->T_dust[TstepA]<0.0) {
	  /* I don't know how to connect these vertexes, thus I use the default */
          if (angleA < angleB) flag_A = TRUE;
          else flag_B=TRUE;
	}
	else flag_B=TRUE;
      }
      else if (TstepB<Max_Dust_Type && TstepA==Max_Dust_Type &&
               Vb1->T_dust[TstepB]>0.0) {
        if (Vb2->T_dust[TstepB]<0.0) {
	  /* I don't know how to connect these vertexes, thus I use the default */
          if (angleA < angleB) flag_A = TRUE;
          else flag_B=TRUE;
	}
	else flag_A=TRUE;
      }
      else if (TstepA<Max_Dust_Type && TstepB<Max_Dust_Type && 
               dust_existA[TstepA] && dust_existB[TstepA] &&
               Vb1->T_dust[TstepA]>0.0 && Vb2->T_dust[TstepA]>0.0) {
	/* check if the dust surface is coming up */
        front_dust = FALSE;
        Vsearch = Va2;
        while (Vsearch!=Va_end) {
          Vsearch = Vsearch->next;
          if (Vsearch->T_dust[TstepA]>0.0) {front_dust = TRUE; break;}
	}
        if (front_dust) flag_A=TRUE;
        else {
	  /* default */
          if (angleA < angleB) flag_A = TRUE;
          else flag_B=TRUE;
	}
      }
       else if (TstepA<Max_Dust_Type && TstepB<Max_Dust_Type && 
               dust_existA[TstepA] && dust_existB[TstepA] &&
               Va1->T_dust[TstepA]>0.0 && Va2->T_dust[TstepA]>0.0) {
	/* check if the dust surface is coming up */
        front_dust = FALSE;
        Vsearch = Vb2;
        while (Vsearch!=Vb_end) {
          Vsearch = Vsearch->next;
          if (Vsearch->T_dust[TstepA]>0.0) {front_dust = TRUE; break;}
	}
        if (front_dust) flag_B=TRUE;
        else {
	  /* default */
          if (angleA < angleB) flag_A = TRUE;
          else flag_B=TRUE;
	}
      }
      /* vertexes very close to the dust surface */
      else if (Va2->level==668 && Vb2->level==668) {
	if ((oldflag_A && oldflag_B) || (!oldflag_A && !oldflag_B)) {
          if (angleA < angleB) flag_A = TRUE;
          else flag_B=TRUE;
	}
	else {
          flag_A = 1 - oldflag_A;
          flag_B = 1 - oldflag_B;
	}
      }
      /* nothing special about the vertexes */
      else if (angleA < angleB) flag_A = TRUE;
      else flag_B=TRUE;
    }
    /* create triangle */
    trian=_PERUNIKA_create_triangle(&error);
    trian->level = 1;
    trian->v1 = Va1;
    trian->v3 = Vb1;
    if (flag_A) {
      trian->v2 = Va2;
      Va2->level=667;
      Va1 = Va2;
      if (Va1!=Va_end) Va2 = Va1->next;
    }
    else {
      trian->v2 = Vb2;
      Vb1 = Vb2;
      if (Vb1!=Vb_end) Vb2 = Vb1->next;
    }

    oldflag_A = flag_A; 
    oldflag_B = flag_B;

    /* exit if the ring A is not in the z=0 plane and Vb1 gets below
       the ring A */
    if (Va1==Va_end) if (Va_end->z!=0.0 && Vb1->z<Va_end->z) Vb1=Vb_end;

  } /* end of while (Va1!=Va_end && Vb1!=Vb_end) */

  free_ivector(dust_existA,0,Max_Dust_Type);
  free_ivector(dust_existB,0,Max_Dust_Type);

  /* check if we need to re-route the rings in case of breaching */
  if (breachA!=NULL) {
    Va_start->prev->next = breachA->next;
    breachA->next->prev = Va_start->prev;
    Va_end->next = Va_start;
    Va_start->prev = Va_end;
    breachA->next = breachB;
    breachB->prev = breachA;
    return 1; /* signal for breaching */
  }

  return 0;
}

/************************************************************************/
/************************************************************************/
/* subroutine: find next rho or theta */
double _PERUNIKA_Find_next_rho_or_theta (struct _vertex_structure *V0, 
       double *Tau_back, double *Tau_front, double *D_back, double *D_front,
       int geometry_flag, double dL_limit, int *type_flag,
       int lambda_computational_grid,int status){

  /* rho i theta of the current point in the search */
  double rho, theta, rho_new, theta_new, tan_Theta0;
  /* points during vertex search */
  struct _3DVector r_old, r_new;
  /* increments */ 
  double tau_old, tau_new, L_old, L_new, dL; 
  int create_vertex;
  /* eta */
  double *temperature,*temperature_old;
  double Eta_old, Eta_new;
  double temp_error;
  int dust_i;

  double ZETA = 2.0;      /* logarithmic scale in the optical depth criterium
                            for a new grid point */
  int dust_edge, steep_temperature;
  int err;

  temperature = dvector(0,Max_Dust_Type-1);
  temperature_old = dvector(0,Max_Dust_Type-1);
  tan_Theta0 = tan(DEBUG.AZIMUTHAL_GRID1);

  rho = sqrt(V0->x*V0->x + V0->z*V0->z);
  if (V0->z==0.0) theta = PI/2.0;
  else theta = atan(V0->x/V0->z); 

  /**** outer edge is not a circle: *****/
  if (geometry_flag==2 && rho>0.995*Rout && status==2) 
    if ((DEBUG.AZIMUTHAL_GRID==3 && theta<DEBUG.AZIMUTHAL_GRID1) ||
         DEBUG.AZIMUTHAL_GRID!=3) {
    rho = 0.995*Rout;
    /* V0->x = 0.995*Rout*sin(theta); */
    /* V0->z = 0.995*Rout*cos(theta); */
  }
  /**************************************/

  rho_new = rho; theta_new = theta;

  if ((*Tau_back)==0.0 || (*Tau_front)==0.0) {

    /* find optical depth distances */
    if (geometry_flag==1){
      (*Tau_back)= _PERUNIKA_distance_to_edge(V0,0.001*sin(theta),0.001*cos(theta),
                                          lambda_computational_grid,1,D_back,status);
      (*Tau_front)=_PERUNIKA_distance_to_edge(V0,1.1*Rout*sin(theta),1.1*Rout*cos(theta),
                                          lambda_computational_grid,1,D_front,status);
    }

    if (geometry_flag==2){
      if (DEBUG.AZIMUTHAL_GRID==3 && theta>=DEBUG.AZIMUTHAL_GRID1) {
        (*Tau_back)= _PERUNIKA_distance_to_edge(V0,V0->x,Rout,
                                            lambda_computational_grid,1,D_back,status);
        (*Tau_front)=_PERUNIKA_distance_to_edge(V0,V0->x,-Rout,
                                            lambda_computational_grid,1,D_front,status);
      }
      else {
        (*Tau_back)= _PERUNIKA_distance_to_edge(V0,rho,0.0,lambda_computational_grid,
                                            2,D_back,status);
        (*Tau_front)=_PERUNIKA_distance_to_edge(V0,rho,PI,lambda_computational_grid,
                                            2,D_front,status);
      }
    }

  }

  r_old.x = V0->x; r_old.z = V0->z;
  /**** outer edge is not a circle: *****/
  if (geometry_flag==2 && rho>=0.995*Rout && status==2) 
    if ((DEBUG.AZIMUTHAL_GRID==3 && theta<DEBUG.AZIMUTHAL_GRID1) ||
         DEBUG.AZIMUTHAL_GRID!=3) {
    r_old.x = 0.995*Rout*sin(theta);
    r_old.z = 0.995*Rout*cos(theta);
  }
  /**************************************/
  tau_old = 0.0; tau_new = 0.0; L_old = 0.0; 
  /* Eta_old = V0->Eta; */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)  {
    /* temperature_old[dust_i] = V0->T_dust[dust_i]; */
    temp_error=1.0;
    temperature_old[dust_i]=
        _PERUNIKA_temperature(r_old.x,r_old.z,dust_i,status,&temp_error);
    if (status<2) if (Dust_Density(r_old.x,r_old.z,dust_i,NULL,&err)<=0.0)
          temperature_old[dust_i]=-1;
    /* special treatment of the disk+halo model 
    if (geometry_flag==2 && INPUT.DUST_DENSITY_TYPE[dust_i]==10 &&
        theta<DEBUG.AZIMUTHAL_GRID1)
      temperature_old[dust_i]=-1; */
  }
  Eta_old=Eta(r_old.x,r_old.z,
                      lambda_computational_grid,temperature_old,&err);
  dL = 0.0001*_ZELENJAK_size_limit(r_old.x,r_old.z); /*??????*/
  create_vertex = FALSE;
  
  /* start the search for the next opitmal vertex location */
  do {

    /* new point */
    if (geometry_flag==1){
      r_new.x = r_old.x + dL*sin(Theta_Total_Value);
      r_new.z = r_old.z + dL*cos(Theta_Total_Value);
      rho_new = sqrt(r_new.x*r_new.x + r_new.z*r_new.z);
    }
    if (geometry_flag==2){
      r_new.x = rho*sin(theta+dL/rho);
      r_new.z = rho*cos(theta+dL/rho);
      if (r_new.z==0.0) theta_new = PI/2.0;
      else theta_new = atan(r_new.x/r_new.z); 
      if (DEBUG.AZIMUTHAL_GRID==3) {
	if (theta_new>DEBUG.AZIMUTHAL_GRID1 && theta<DEBUG.AZIMUTHAL_GRID1-1E-20) {
          theta_new=DEBUG.AZIMUTHAL_GRID1+1E-10;
          r_new.x = rho*sin(theta_new);
          r_new.z = rho*cos(theta_new);
          dL = rho*(theta_new - theta);
	}
        else if (theta>DEBUG.AZIMUTHAL_GRID1) {
	  r_new.x = r_old.x;
          if (theta+dL/rho>=PI/2.0) r_new.z =0.0;
          else r_new.z = r_new.x/tan(theta+dL/rho);
          rho_new = sqrt(r_new.x*r_new.x+r_new.z*r_new.z);
          if (r_new.z<=0.0) theta_new = PI/2.0;
          else theta_new = atan(r_new.x/r_new.z); 
	}
      }
    }
    L_new = L_old+dL;
    if (DEBUG.AZIMUTHAL_GRID==3 && rho_new>Rout*sin(DEBUG.AZIMUTHAL_GRID1)+1.0E-10 && 
        rho<Rout*sin(DEBUG.AZIMUTHAL_GRID1)-1.0E-10 && geometry_flag==1) {
      /* exactly on the point of crossing Rout corner */
      rho_new=Rout*sin(DEBUG.AZIMUTHAL_GRID1);
      printf("--- R point exactly at Rout*sin(DEBUG.AZIMUTHAL_GRID1)\n");
      create_vertex = TRUE;
      dust_edge = FALSE; 
      break;
    }
    else if ((geometry_flag==1 && rho_new>Rout) ||
             (geometry_flag==2 && r_new.z<0.0)) {
      create_vertex = TRUE;
      if (rho_new>Rout) rho_new=Rout;
      if (r_new.z<0.0) theta_new=0.5*PI;
      dust_edge = FALSE; 
      break; /* the new vertex is exactly on the domain edge */
    }

    /* new eta */
    dust_edge = FALSE; 
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
      temp_error=1.0;
      temperature[dust_i]=
        _PERUNIKA_temperature(r_new.x,r_new.z,dust_i,status,&temp_error);
      if (status<2) if (Dust_Density(r_new.x,r_new.z,dust_i,NULL,&err)<=0.0)
          temperature[dust_i]=-1;
      /* special treatment of the disk+halo model 
      if (geometry_flag==2 && INPUT.DUST_DENSITY_TYPE[dust_i]==10 &&
          V0->x<V0->z*tan_Theta0 && r_old.x<r_old.z*tan_Theta0) {
        if (r_new.x<r_new.z*tan_Theta0)
          temperature[dust_i]=-1;
	else
	  temperature_old[dust_i]=temperature[dust_i];
      }*/
      if (temperature[dust_i]*temperature_old[dust_i]<0.0) dust_edge=TRUE;
    }
    Eta_new=Eta(r_new.x,r_new.z,
                      lambda_computational_grid,temperature,&err);

    /********** check if the temperature gradient is very steep *****/
    steep_temperature=FALSE;
    if (!dust_edge && geometry_flag==2) {
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 && temperature[dust_i]>100.0)
	  if (temperature[dust_i]<0.9*V0->T_dust[dust_i]) 
             {steep_temperature=TRUE; break;}
    }
    /***************************************************/

    if (dust_edge) {
      /* refine the grid */
      if (dL > 1.0E-7) {
        dL /= 4.0;
        create_vertex = FALSE;
        continue;
      }
      else {
        create_vertex = TRUE;
        if (geometry_flag==-1) {
          /* scenario for an extremely resolved  edge on the radial grid */
          /* NOT ACTIVATED !!! */
          if (r_old.x==V0->x && r_old.z==V0->z ){
            (*Tau_back)=0.0; 
            if (Eta_old==0.0) (*Tau_front)=0.0;
	  }
	  else {
            r_new.x=r_old.x; r_new.z=r_old.z; dL=0.0; 
            (*Tau_front)=0.0;
            if (geometry_flag==1){rho_new = rho;}
            if (geometry_flag==2){theta_new = theta;}
	  }
	}

	if (geometry_flag==1) {
          /* printf("EDGE DETECTED: R=%14.12lf [%14.12E,%14.12E] theta=%lf",
	     sqrt(r_new.x*r_new.x+r_new.z*r_new.z),r_new.x,r_new.z,theta_new*180.0/PI);*/
          printf("EDGE DETECTED: R=%14.12lf ",rho_new);
          for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
            if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
              if (temperature[dust_i]*temperature_old[dust_i]<0.0) printf("#%d ",dust_i);
          printf("\n");
	}

	if (geometry_flag==1 || geometry_flag==2) {
          /* scenario for an edge on the azimuthal (AND RADIAL) grid */
          if (Eta_new == 0.0) {
            r_new.x=r_old.x; r_new.z=r_old.z; dL=0.0; 
            (*Tau_front)=0.0;
            printf(" --- edge moved back to (%14.12E,%14.12E)\n",r_new.x,r_new.z);
	  }
	  else {(*Tau_back)=0.0; if (Eta_old==0.0) (*Tau_front)=0.0;}
	}
        break;
      }
      /***************************************************************************/
    } else if (steep_temperature && 0) {
      create_vertex = TRUE;break;
      /***************************************************************************/
    } else if (DEBUG.AZIMUTHAL_GRID==3 && theta_new>DEBUG.AZIMUTHAL_GRID1 && 
               theta<DEBUG.AZIMUTHAL_GRID1 && geometry_flag==2) {
      /* force a new vertex at the turn point*/
      /* (update Tau_front because it goes vertically down now) */
      (*Tau_front)=0.0;(*D_front)=0.0;
      create_vertex = TRUE;break;
      /***************************************************************************/
 /* } else if (DEBUG.AZIMUTHAL_GRID==3 && theta_new>1.569574596 && 
               theta<1.569574596 && geometry_flag==2) {
      (*Tau_front)=0.0;(*D_front)=0.0;
      create_vertex = TRUE;break; */
      /* this is a point very close to the z=0 plane */
      /***************************************************************************/
    } else if (geometry_flag==2 && L_new/rho_new>5.0*RAD1 && (Eta_new==0.0 || 
               DEBUG.AZIMUTHAL_GRID==3)) {
      /* force a new point after 5deg in empty space*/
      create_vertex = TRUE;break;
      /***************************************************************************/
    } else if (geometry_flag==2 && DEBUG.AZIMUTHAL_GRID==1 &&
               L_new/rho_new>DEBUG.AZIMUTHAL_GRID1*RAD1 && Eta_new>0.0) {
      /* force a new point after DEBUG.AZIMUTHAL_GRID1 deg in non-empty space*/
      create_vertex = TRUE;break;
      /***************************************************************************/
    }  if (geometry_flag==2 && DEBUG.AZIMUTHAL_GRID==2 &&
               L_new/rho_new>3.0*RAD1 && Eta_new>0.0) {
      /* force a new point after 3.0 deg in non-empty space when DEBUG.AZIMUTHAL_GRID=2*/
      create_vertex = TRUE;break;
      /************************** 80deg Ohnaka disk *****************************
    } else if (geometry_flag==2 && r_new.x/r_new.z>5.67128182 
               && r_new.x/r_new.z<6.313751515 &&
               L_new/rho_new>0.1*RAD1 && Eta_new>0.0) {
      create_vertex = TRUE;break;
      **************************************************************************/
    } else if (geometry_flag==1 && rho_new>1.25*(rho_new-L_new)) {
      /* force a new point after 1.25*old_rho */
      create_vertex = TRUE;break;
      /***************************************************************************/
    } else if (Eta_new==0.0 && (*Tau_front)==0.0) {
      /* continue with the search if empty space */
      create_vertex = FALSE;
      tau_new=tau_old;
      /***************************************************************************/
      /**************** halo radii ***********************************************
    }  if (geometry_flag==1 && rho_new>1.007 && rho<=1.007) {
        create_vertex = TRUE;break;
    }  if (geometry_flag==1 && rho_new>3.146 && rho<=3.146) {
        create_vertex = TRUE;break;*/
    }  if (geometry_flag==1 && rho_new>3.05 && rho<=3.05) {
        create_vertex = TRUE;break;
      /***************************************************************************/
      /**************** step-flat disk *******************************************
    }  if (geometry_flag==1 && rho_new>2.0 && rho<=2.0) {
        rho_new = 2.0000001;
        (*Tau_front)=0.0; (*Tau_back)=0.0; tau_new=0.0;
        create_vertex = TRUE;break;
      ***************************************************************************/
      /**************** enforcing angular grid every 0.35deg *********************/
    } else if (geometry_flag==2 && DEBUG.AZIMUTHAL_GRID==3 && (*Tau_back)>1.0 &&
	       L_new/rho_new>0.5*RAD1) {
      create_vertex = TRUE;
      /***************************************************************************/
    } else {
      /* check if the distance is optimal for a new vertex */
      temp_error=0.5*dL*(Eta_old+Eta_new);
      if (tau_old == 0.0 && temp_error>0.1) {
        /* decrease the initial step */
        dL /= 4.0;
        create_vertex = FALSE;
        continue;
      }
      else if (tau_old>0.0 && temp_error>4*tau_old){
        dL /= 2.7;/* step can not be larger than 4x the current tau */
        create_vertex = FALSE;
        continue;
      }
      else if (temp_error<0.01*tau_old  && (tau_old<0.05 || tau_old>0.2) ) dL *=2;

      tau_new = tau_old + temp_error;

      /* varius possibilites for choosing vertex creation */
      if (geometry_flag==2 && INPUT.DUST_DENSITY_TYPE[0]==10 &&
          V0->x<V0->z*tan_Theta0 && r_new.x<r_new.z*tan_Theta0) {
	/* special treatment of the disk+halo model */
        tau_new = tau_old = 0.0;
        dL_limit = 5.0*RAD1*rho;
	(*Tau_back) = (*Tau_front) = 0.0;
      }

      if (geometry_flag==2 && (*Tau_back)<1.0 && tau_new>0.2) {
	/* enforce tau<=0.2 on the surface (important for scattering) */
        create_vertex = TRUE;break;
      }
      create_vertex = FALSE;
      if ((*Tau_back)<0.1) {
	if (tau_new>0.1005) {
          dL/=1.8; /* locate the tau=0.1 surface precisely */
          create_vertex = FALSE;
          continue;
        }
        if (tau_new>0.1) {create_vertex = TRUE;break;}
      }
      else if (DEBUG.AZIMUTHAL_GRID==2 && geometry_flag==2 && L_new>dL_limit &&
               L_new>DEBUG.AZIMUTHAL_GRID1*RAD1) {
        create_vertex = TRUE;break;
      }
      else if (temp_error<0.01) dL *=2;
      else if (geometry_flag==2 &&
               tau_new >= (*Tau_back)*((2.0*ZETA)*(*D_back)/((*D_back)+L_new)-1.0)) /* 200% increase on ZETA */
        {create_vertex = TRUE;break;}
      else if (geometry_flag==1 &&
               tau_new >= (*Tau_back)*((1.7*ZETA)*(*D_back)/((*D_back)+L_new)-1.0)) /* 170% increase on ZETA */
        {create_vertex = TRUE;break;}


      /* increased grid resolution on the "dark" side of dust*/
      if ((*Tau_front)-tau_new>0.1) {
        if (DEBUG.AZIMUTHAL_GRID!=3) {
          if (tau_new>=((*Tau_front)-tau_new)*
	      ((10.0*ZETA)*((*D_front)-L_new)/(*D_front)-1.0))/* 1000% increase on ZETA */
            {create_vertex = TRUE;break;} 
        }
	else {
          if (tau_new>=((*Tau_front)-tau_new)*
	      ((1.7*ZETA)*((*D_front)-L_new)/(*D_front)-1.0))/* 170% increase on ZETA */
            {create_vertex = TRUE;break;} 
        }
      }


    }

    /* prepare variables for the next step */
    rho = rho_new;
    theta = theta_new;
    r_old.x=r_new.x;
    r_old.z=r_new.z;
    L_old = L_new;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      temperature_old[dust_i] = temperature[dust_i];
    Eta_old = Eta_new;
    tau_old = tau_new;

  } while (!create_vertex);

  if (DEBUG.AZIMUTHAL_GRID==3 && DEBUG.AZIMUTHAL_GRID1 - theta_new > 0 &&
      DEBUG.AZIMUTHAL_GRID1 - theta_new < 1E-8) 
    theta_new = DEBUG.AZIMUTHAL_GRID1; /* this is a very special point on the first
                                  ring: on the edge of dust and at the turning point */

  if (!dust_edge) {
    (*Tau_back) += tau_new;
    (*Tau_front)-= tau_new;
  }
  if ((*Tau_front)<0.0) (*Tau_front)=0.0;
  (*D_back) += L_new;
  (*D_front)-= L_new;
  if ((*D_front)<0.0) {(*D_front)=0.0;(*Tau_front)=0.0;}

  if ((*D_front)<0.5*Rout && geometry_flag==1 ) {
    /* find optical depth distances */
    theta = theta_new;
    r_old.x=V0->x; r_old.z=V0->z;
    V0->x=r_new.x; V0->z=r_new.z;
    (*Tau_front)=_PERUNIKA_distance_to_edge(V0,1.1*Rout*sin(theta),1.1*Rout*cos(theta),
                                        lambda_computational_grid,1,D_front,status);
    V0->x=r_old.x; V0->z=r_old.z;
  }

  (*type_flag) = 1;
  if (dust_edge) (*type_flag) = 2;

  free_dvector(temperature,0,Max_Dust_Type-1);
  free_dvector(temperature_old,0,Max_Dust_Type-1);

  if (geometry_flag==1) return rho_new;
  if (geometry_flag==2) return theta_new;

  return 1;
}


/************************************************************************/
/************************************************************************/
/* subroutine: temperature estimates */
int  _PERUNIKA_create_azimuthal_grid_T_estimate (
    struct _vertex_structure *V_current,  
    int lambda_computational_grid,int status){

  double lambda_diffuse, tau_diffuse_back, tau_diffuse_front, T_diffuse;
  int lambda_diffuse_int;
  int TOTAL_NUMBER_OF_VERTICES_diffuse, TOTAL_NUMBER_OF_TRIANGLES_diffuse;
  struct _3DVector startpoint_diffuse, endpoint_diffuse;
  struct _vertex_structure *v_diffuse;
  struct _triangle_structure *t_diffuse;
  struct _triangle_structure *t, *t_start;
  double Fdown_bol, Fup_bol, J_bol, *J, *Fdown, *Fup;
  int dust_i, lambda_k,lambda_max_N,  angular_rays_counter;
  struct _3DVector ray_direction; /* direction of the intensity integral*/
  double angular_weight, angular_flux_weight; /* weight in the angular integral */
  double rho, dummy;
  int flag_plus_minus; /* flag which shows inward and outward flux at a point*/
  double Intensity_sca, Intensity_em; /*intensities calculated along an angular
				        integral ray */


  TOTAL_NUMBER_OF_VERTICES_diffuse = TOTAL_NUMBER_OF_VERTICES;
  v_diffuse = start_vertex_list;
  TOTAL_NUMBER_OF_TRIANGLES_diffuse = TOTAL_NUMBER_OF_TRIANGLES;
  t_diffuse = start_triangle_list;

  startpoint_diffuse.x = V_current->prev->x;
  startpoint_diffuse.y = 0.0;
  startpoint_diffuse.z = V_current->prev->z;

  /* find radial tauV */
  rho = sqrt( V_current->prev->x*V_current->prev->x + 
              V_current->prev->z*V_current->prev->z);
  endpoint_diffuse.x = V_current->prev->x / rho;
  endpoint_diffuse.y = 0.0;
  endpoint_diffuse.z = V_current->prev->z / rho;
  lambda_diffuse = 0.55;
  for (lambda_diffuse_int=0; lambda_diffuse_int<lambda_Max; 
	      lambda_diffuse_int++) 
         if (Lambda_Grid[lambda_diffuse_int]>=lambda_diffuse) break;

  start_vertex_list = old_grid_per_lambda[lambda_diffuse_int].start_vertex;
    TOTAL_NUMBER_OF_VERTICES=old_grid_per_lambda[lambda_diffuse_int].TOTAL_NUMBER_OF_VERTICES;
  start_triangle_list = old_grid_per_lambda[lambda_diffuse_int].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES=old_grid_per_lambda[lambda_diffuse_int].TOTAL_NUMBER_OF_TRIANGLES;

  tau_diffuse_back = _ZELENJAK_tau_over_grid (startpoint_diffuse,
		      endpoint_diffuse,lambda_diffuse_int,FALSE);

  if (tau_diffuse_back<20.0 || tau_diffuse_back<INPUT.TAU_LIMIT) {
    /* this is exposed to the stellar radiation - exit */
    start_vertex_list = v_diffuse;
    start_triangle_list = t_diffuse;
    TOTAL_NUMBER_OF_VERTICES=TOTAL_NUMBER_OF_VERTICES_diffuse;
    TOTAL_NUMBER_OF_TRIANGLES=TOTAL_NUMBER_OF_TRIANGLES_diffuse;
    return;
  }

  printf("(%13.11E,%13.11E) T=%lf ",V_current->x,V_current->z,V_current->T_dust[0]);

#ifdef blabla
  /* lambda that corresonds to the local temperature */
  lambda_diffuse = 2900.0/V_current->prev->T_dust[0];
  for (lambda_diffuse_int=0; lambda_diffuse_int<lambda_Max; 
	      lambda_diffuse_int++) 
       if (Lambda_Grid[lambda_diffuse_int]>=lambda_diffuse) break;
  start_vertex_list = old_grid_per_lambda[lambda_diffuse_int].start_vertex;
    TOTAL_NUMBER_OF_VERTICES=old_grid_per_lambda[lambda_diffuse_int].TOTAL_NUMBER_OF_VERTICES;
  start_triangle_list = old_grid_per_lambda[lambda_diffuse_int].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES=old_grid_per_lambda[lambda_diffuse_int].TOTAL_NUMBER_OF_TRIANGLES;

  /* find tau toward the current point and toward z=0 */
  endpoint_diffuse.x = V_current->prev->x;
  endpoint_diffuse.y = 0.0;
  endpoint_diffuse.z = 0.0;
  tau_diffuse_front = _ZELENJAK_tau_over_grid (startpoint_diffuse,
		      endpoint_diffuse,lambda_diffuse_int,FALSE);
  if (V_current->z==0.0) endpoint_diffuse.z =  V_current->prev->prev->z;
  else endpoint_diffuse.z = V_current->z;
  tau_diffuse_back = _ZELENJAK_tau_over_grid (startpoint_diffuse,
		      endpoint_diffuse,lambda_diffuse_int,FALSE);
  if (V_current->z==0.0) tau_diffuse_front += tau_diffuse_back;

  /* find the diffuse temperature */
  T_diffuse = tau_diffuse_front - tau_diffuse_back + 0.6666666;
  T_diffuse /= (tau_diffuse_front + 0.6666666);
  T_diffuse = pow (T_diffuse,0.25);
  if (V_current->z==0.0) T_diffuse = T_diffuse * V_current->prev->prev->T_dust[0];
    else T_diffuse = T_diffuse * V_current->prev->T_dust[0];
  printf("prevT=%lf lam=%d tauB=%lf tauF=%lf Tdiffuse=%lf\n",
         V_current->prev->T_dust[0],lambda_diffuse_int,tau_diffuse_back,
         tau_diffuse_front,T_diffuse);
#endif

  /************************************************/
  /************* FLUX CALCULATION *****************/

  /* unit vector toward the disk interior */
  endpoint_diffuse.x = V_current->prev->z / rho;
  endpoint_diffuse.y = 0.0;
  endpoint_diffuse.z = - V_current->prev->x / rho;

  endpoint_diffuse.x = 0.0;
  endpoint_diffuse.y = 0.0;
  endpoint_diffuse.z = - 1.0;
 

  Fdown = dvector(0,lambda_Max-1);
  Fup = dvector(0,lambda_Max-1);
  J = dvector(0,lambda_Max-1);
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
     {Fdown[lambda_k]=0.0;Fup[lambda_k]=0.0;J[lambda_k]=0.0;}

  /* find the most detailed grid (the largest number of vertices) */
  lambda_max_N = 0; 
  lambda_diffuse_int = old_grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > lambda_diffuse_int) {
      lambda_max_N = lambda_k;
      lambda_diffuse_int = old_grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    }

  start_vertex_list = old_grid_per_lambda[lambda_max_N].start_vertex;
    TOTAL_NUMBER_OF_VERTICES=old_grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  start_triangle_list = old_grid_per_lambda[lambda_max_N].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES=old_grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

  _GOROSTAS_create_integral_directions(V_current->prev);
  printf("rays=%d:",total_number_of_spherical_triangles);fflush(stdout);

  /* loop over angular integrals */
  for (angular_rays_counter=0;
       angular_rays_counter<total_number_of_spherical_triangles;
       angular_rays_counter++) {

    ray_direction = integral_directions[angular_rays_counter].direction;

    angular_weight = 
        integral_directions[angular_rays_counter].weight*2.0/(4.0*PI); /* factor 2 
                                          comes from the axial symmetry */

    /* include the flux part of the weight: */ 
    dummy = ray_direction.x * endpoint_diffuse.x + 
            ray_direction.y * endpoint_diffuse.y + 
            ray_direction.z * endpoint_diffuse.z;
    angular_flux_weight = - angular_weight * dummy;
           
    /* is this flux into the disk or out of the disk? */
    if (dummy < 0.0) flag_plus_minus = 1; /* outward flux */
    else if (dummy > 0.0) flag_plus_minus = -1; /* inward flux */
    else flag_plus_minus = 0; /* exactly on the edge */

    /* calculate the intensities */
    //printf("%.3d:",angular_rays_counter);fflush(stdout);printf("\b\b\b\b");
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {

      tau_diffuse_front = _VILOZMAJ_Intensity_Integral(lambda_k,startpoint_diffuse,
		      ray_direction,&Intensity_em,&Intensity_sca,TRUE);

      /*flux */
      if (flag_plus_minus == 1) {
	Fdown[lambda_k] +=  (angular_flux_weight*(Intensity_sca+Intensity_em));
      }
      else if (flag_plus_minus == -1) {
	Fup[lambda_k] +=  (angular_flux_weight*(Intensity_sca+Intensity_em));
      }
      else {
	Fdown[lambda_k] += (0.5*angular_flux_weight*(Intensity_sca+Intensity_em));
	Fup[lambda_k] += (0.5*angular_flux_weight*(Intensity_sca+Intensity_em));
      }

      /* angular averaged intensity */
      J[lambda_k] += (angular_weight*(Intensity_sca+Intensity_em));

    } /* end of the lambda loop */ 

  } /* end of the loop over angular integrals */

  /* destroy the angular rays */
  total_number_of_spherical_triangles=0;

  free_dvector(Fdown,0,lambda_Max-1);
  free_dvector(Fup,0,lambda_Max-1);
  free_dvector(J,0,lambda_Max-1);

  /* bolometric flux */
  Fdown_bol = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Fdown);
  Fup_bol = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,Fup);
  J_bol = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,J);
  printf("\n......Fdown=%13.11E Fup=%13.11E Fd/Fu=%lf J=%13.11E Fdown/J=%lf ", 
           Fdown_bol, Fup_bol, Fdown_bol/Fup_bol, J_bol, Fdown_bol/J_bol);

  /****** end of flux calculation **********/
  /************************************************/

  /* temperature correction */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
        V_current->T_dust[dust_i]>0.0)
    V_current->T_dust[dust_i] *= pow(-Fdown_bol/Fup_bol,0.25);

  printf("Tnew=%lf\n",V_current->T_dust[0]);


  start_vertex_list = v_diffuse;
  start_triangle_list = t_diffuse;
  TOTAL_NUMBER_OF_VERTICES=TOTAL_NUMBER_OF_VERTICES_diffuse;
  TOTAL_NUMBER_OF_TRIANGLES=TOTAL_NUMBER_OF_TRIANGLES_diffuse;
  return;
}
