/* LELUYA - PERUNIKA program block */
/* This part of the LELUYA code is generating the computational grid */

/* This function goes through the list of triangles and removes 
   the edge-triangles, (triangles which are partially over one dust component).
   If a triangle is not the edge-triangle, but it has more than three vertices,
   then it is split and replaced by new triangles. 

   Finally, when we have only triangles in the grid, replace the triangles 
   which have vertices at x<0.0 by the new ones which don't have this.
  */

#include "PERUNIKA.h"

/* function which makes the search for a dust edge point from the point1
   toward the point2 into the direction 'direction' */ 
struct _vertex_structure *_PERUNIKA_create_point_at_dust_edge 
                                 (struct _triangle_structure *parent_triangle, 
                                  struct _vertex_structure *point1, 
                                  struct _vertex_structure *point2, 
                                  int direction,int lambda_computational_grid,
                                  int status);

/* function which creates a vertex at x=0.0 between the given two vertices */
struct _vertex_structure *_PERUNIKA_create_point_at_x_zero
        (struct _vertex_structure *point1, 
                 struct _vertex_structure *point2, 
                 int lambda_computational_grid,int status);

/* function which moves a triangle back into the analysis */
void _PERUNIKA_move_triangle_back ( struct _triangle_structure *current_triangle,
							    struct _vertex_structure *v0);

/****************************************************************************/
/****************************************************************************/
int _PERUNIKA_edge_triangles_removal (int lambda_computational_grid,int status){

  /*   
  lambda_computational_grid = wavelength at which we need to create 
                    the computational grid
  status          = if status = 0 or 1, the temperature and energy density 
                    values will be taken from the initial approximation arrays
                    if status = 2, then they will be taken from the grid
                    built in the previous iteration
  */

  short int UP, DOWN; /* 1 (TRUE) or 0 (FALSE) = orientation of a triangle */
  int error; /* error flag */
  /* dust component counter */
  int dust_i;
  /* vertices at the dust edge: *new_vxy is a vertex on the side xy, where
     the vertex x is a real point and the vertex y is a fake point */
  struct _vertex_structure *new_v12, *new_v21, *new_v23, *new_v32, 
                           *new_v31, *new_v13;
  /* a triangle used in the loop over the triangle list */
  struct _triangle_structure *current_triangle;
  /* the next triangle in the loop over triangles */
  struct _triangle_structure *next_triangle;
  /* flags which show if the triangle's sides are going over the dust 
     components' edge, even though the vertices are not fake */
  int dust_edge1, dust_edge2, dust_edge3;
  /* a new triangle (if it is necessary) */
  struct _triangle_structure *new_triangle;
  /* a polygon built from the vetices on the edge triangle and 
     the edge vertices */
  struct _polygon_structure *polygon;
  /* a vertex in the polygon */
  struct _polygon_structure *polygon_vertex;
  /* a vertex in the loop over the vertices on the triangle's sides */
  struct _vertex_structure *current_vertex;
  /* distance of triangle vertices to the dust surface */
  double distance_v1_to_edge,distance_v2_to_edge,distance_v3_to_edge;
  /* the fraction of the minimal grid cell size that shows how close a point
     can be to the dust surface */
  double min_distance_to_surface = 0.2; 

  if (DEBUG.SCREEN_MESSAGE >= 1 )
    printf("process#%.3d:... removal of the edge-triangles \n",my_MPI_rank);

  /* first mark all triangles as UP or DOWN. This is needed for the -R flag,
     and it does not hurt to do it without that flag */
  current_triangle = start_triangle_list;
  do {
    if (current_triangle->v2->z > current_triangle->v1->z &&
        current_triangle->v2->z > current_triangle->v3->z &&
        current_triangle->v2->x < current_triangle->v1->x)
      current_triangle->ID_number = 1; /* UP */
    else
      current_triangle->ID_number = -1; /* DOWN */    
    current_triangle = current_triangle->prev;
  } while (current_triangle!=start_triangle_list);

  /* start the loop over triangles */
  current_triangle = start_triangle_list;
  do {

    /* the next triangle will be: */
    next_triangle = current_triangle->prev;

    /* is it UP or DOWN triangle? */
    if (current_triangle->ID_number>0) {UP=1; DOWN=0;}
    else if (current_triangle->ID_number<0) {UP=0; DOWN=1;}
	else {
       /* this is a triangle that we have already analyzed and it does not
	      need it again */
	   current_triangle  = next_triangle;
	   continue;
	}
    /* current_triangle->ID_number=0; */

    /* if the triangle's point is fake only because it has x<0.0, and not
       because its Eta, change its type into the real type
       (don't do that if the whole triangle has x<0.0) */
    if (current_triangle->v1->x >= 0.0 || current_triangle->v2->x >= 0.0 ||
        current_triangle->v3->x >= 0.0 ){
      if (current_triangle->v1->x<0.0 && current_triangle->v1->Eta>0.0)
    current_triangle->v1->type=1;
      if (current_triangle->v2->x<0.0 && current_triangle->v2->Eta>0.0)
    current_triangle->v2->type=1;
      if (current_triangle->v3->x<0.0 && current_triangle->v3->Eta>0.0)
    current_triangle->v3->type=1;
    }

    /* check first if we have a special case of the dust component edge when
       the vertices are not fake */
    dust_edge1=dust_edge2=dust_edge3=0;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (current_triangle->v1->type != 3 && current_triangle->v2->type != 3 &&
          current_triangle->v1->T_dust[dust_i]*
          current_triangle->v2->T_dust[dust_i]<0.0) dust_edge1=dust_i+1;
      if (current_triangle->v2->type != 3 && current_triangle->v3->type != 3 &&
          current_triangle->v2->T_dust[dust_i]*
          current_triangle->v3->T_dust[dust_i]<0.0) dust_edge2=dust_i+1;
      if (current_triangle->v3->type != 3 && current_triangle->v1->type != 3 &&
          current_triangle->v3->T_dust[dust_i]*
          current_triangle->v1->T_dust[dust_i]<0.0) dust_edge3=dust_i+1;
    }
    
    if (INPUT.TAU_SURFACE_SCALE > 1.0) 
      /* identify triangles with one or two level=668 vertices */
      if (!(dust_edge1 || dust_edge2 || dust_edge3)) 
        if (current_triangle->v1->level==668 || current_triangle->v2->level==668  ||
            current_triangle->v3->level==668) {
          if (current_triangle->v1->level==668 && current_triangle->v2->level==668) 
            {dust_edge2=TRUE; dust_edge3=TRUE;}
          else if (current_triangle->v2->level==668 && current_triangle->v3->level==668) 
            {dust_edge3=TRUE; dust_edge1=TRUE;}
          else if (current_triangle->v3->level==668 && current_triangle->v1->level==668) 
            {dust_edge1=TRUE; dust_edge2=TRUE;}
          else if (current_triangle->v1->level==668) {
            dust_edge1=TRUE;dust_edge3=TRUE;
            /* check for a more important dust edge */ 
	        if ((current_triangle->v2->type != 3 && current_triangle->v3->type == 3) ||
                (current_triangle->v2->type == 3 && current_triangle->v3->type != 3)) {
	          if (current_triangle->v2->type != 3)   dust_edge1=FALSE;
	          if (current_triangle->v3->type != 3)   dust_edge3=FALSE;
			}
		  }
          else if (current_triangle->v2->level==668) {
	        dust_edge2=TRUE;dust_edge1=TRUE;
	        /* check for a more important dust edge */ 
	        if ((current_triangle->v1->type != 3 && current_triangle->v3->type == 3) ||
                (current_triangle->v1->type == 3 && current_triangle->v3->type != 3)) {
	          if (current_triangle->v1->type != 3)   dust_edge1=FALSE;
	          if (current_triangle->v3->type != 3)   dust_edge2=FALSE;
			}
		  }
          else if (current_triangle->v3->level==668) {
	        dust_edge3=TRUE;dust_edge2=TRUE;
	        /* check for a more important dust edge */ 
	        if ((current_triangle->v2->type != 3 && current_triangle->v1->type == 3) ||
                (current_triangle->v2->type == 3 && current_triangle->v1->type != 3)) {
	          if (current_triangle->v2->type != 3)   dust_edge2=FALSE;
	          if (current_triangle->v1->type != 3)   dust_edge3=FALSE;
			}
		  }
		}

    /* is this triangle an edge-triangle? */
    if (current_triangle->v1->type == 3 ||
        current_triangle->v2->type == 3 ||
        current_triangle->v3->type == 3 || 
        dust_edge1 || dust_edge2 || dust_edge3) {

      new_v12=new_v21=new_v23=new_v32=new_v31=new_v13=NULL;

      /* find the sides which are crossing the dust edge and create a point
         at the position of that crossing */
         
      /* vertex 1 is real, vertex 2 is fake OR vertex 1 has the dust component
         dust_edge1-1 while the vertex 2 doesn't have it */
      if (current_triangle->v1->level==668 && 
          (current_triangle->v2->type==3 || dust_edge1) ) {
		  new_v12 = current_triangle->v1; /* this is level=668 vertex on the edge */
      }
      else 
      if ((current_triangle->v1->type!=3 && current_triangle->v2->type==3) ||
          (dust_edge1 && current_triangle->v1->T_dust[dust_edge1-1]>0.0 &&
           current_triangle->v1->level!=668 && current_triangle->v2->level!=668)){
        if (DOWN) new_v12=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v1,current_triangle->v2,2,
                 lambda_computational_grid,status);
        if ( UP ) new_v12=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v1,current_triangle->v2,3,
                 lambda_computational_grid,status);
      }
      /* vertex 2 is real, vertex 1 is fake OR vertex 2 has the dust component
         dust_edge1-1 while the vertex 1 doesn't have it */
      if (current_triangle->v2->level==668 && 
          (current_triangle->v1->type==3 || dust_edge1) ) {
		 new_v21 = current_triangle->v2; /* this is level=668 vertex on the edge */
      }
      else
      if ((current_triangle->v1->type==3 && current_triangle->v2->type!=3) ||
          (dust_edge1 && current_triangle->v2->T_dust[dust_edge1-1]>0.0 &&
           current_triangle->v2->level!=668 && current_triangle->v1->level!=668)){
        if (DOWN) new_v21=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v2,current_triangle->v1,5,
                 lambda_computational_grid,status);
        if ( UP ) new_v21=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v2,current_triangle->v1,6,
                 lambda_computational_grid,status);
      }
      /* vertex 2 is real, vertex 3 is fake OR vertex 2 has the dust component
         dust_edge2-1 while the vertex 3 doesn't have it */
      if (current_triangle->v2->level==668 && 
          (current_triangle->v3->type==3 || dust_edge2) ) {
         new_v23 = current_triangle->v2; /* this is level=668 vertex on the edge */
      }
      else 
      if ((current_triangle->v2->type!=3 && current_triangle->v3->type==3) ||
          (dust_edge2 && current_triangle->v2->T_dust[dust_edge2-1]>0.0 &&
           current_triangle->v2->level!=668 && current_triangle->v3->level!=668)){
        if (DOWN) new_v23=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v2,current_triangle->v3,4,
                 lambda_computational_grid,status);
        if ( UP ) new_v23=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v2,current_triangle->v3,5,
                 lambda_computational_grid,status);
      }
      /* vertex 3 is real, vertex 2 is fake  OR vertex 3 has the dust component
         dust_edge2-1 while the vertex 2 doesn't have it */
      if (current_triangle->v3->level==668 && 
          (current_triangle->v2->type==3 || dust_edge2) ) {
         new_v32 = current_triangle->v3; /* this is level=668 vertex on the edge */
      }
      else 
      if ((current_triangle->v2->type==3 && current_triangle->v3->type!=3) ||
          (dust_edge2 && current_triangle->v3->T_dust[dust_edge2-1]>0.0 &&
           current_triangle->v3->level!=668 && current_triangle->v2->level!=668)){
        if (DOWN) new_v32=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v3,current_triangle->v2,1,
                 lambda_computational_grid,status);
        if ( UP ) new_v32=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v3,current_triangle->v2,2,
                 lambda_computational_grid,status);
      }
      /* vertex 3 is real, vertex 1 is fake  OR vertex 3 has the dust component
         dust_edge3-1 while the vertex 1 doesn't have it */
      if (current_triangle->v3->level==668 && 
          (current_triangle->v1->type==3 || dust_edge3) ) {
         new_v31 = current_triangle->v3; /* this is level=668 vertex on the edge */
      }
      else 
      if ((current_triangle->v3->type!=3 && current_triangle->v1->type==3) ||
          (dust_edge3 && current_triangle->v3->T_dust[dust_edge3-1]>0.0 &&
           current_triangle->v3->level!=668 && current_triangle->v1->level!=668)){
        if (DOWN) new_v31=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v3,current_triangle->v1,6,
                 lambda_computational_grid,status);
        if ( UP ) new_v31=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v3,current_triangle->v1,1,
                 lambda_computational_grid,status);
      }
      /* vertex 1 is real, vertex 3 is fake OR vertex 1 has the dust component
         dust_edge3-1 while the vertex 3 doesn't have it */
      if (current_triangle->v1->level==668 && 
          (current_triangle->v3->type==3 || dust_edge3) ) {
         new_v13 = current_triangle->v1; /* this is level=668 vertex on the edge */
      }
      else 
      if ((current_triangle->v3->type==3 && current_triangle->v1->type!=3) ||
          (dust_edge3 && current_triangle->v1->T_dust[dust_edge3-1]>0.0 &&
           current_triangle->v1->level!=668 && current_triangle->v3->level!=668)){
        if (DOWN) new_v13=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v1,current_triangle->v3,3,
                 lambda_computational_grid,status);
        if ( UP ) new_v13=_PERUNIKA_create_point_at_dust_edge 
        (current_triangle,current_triangle->v1,current_triangle->v3,4,
                 lambda_computational_grid,status);
      }

      /* now we know where the edge points should be, but we also have to know
         how to connect them. I don't know a better way than treat each example
         separately. */

if (FALSE) 
if ((current_triangle->v1->x > 7.78 && current_triangle->v1->x < 8.15 &&
      current_triangle->v1->z > 0.92  && current_triangle->v1->z < 1.27) ||
     (current_triangle->v2->x > 7.78 && current_triangle->v2->x < 8.15 &&
      current_triangle->v2->z > 0.92  && current_triangle->v2->z < 1.27) ||
     (current_triangle->v3->x > 7.78 && current_triangle->v3->x < 8.15 &&
      current_triangle->v3->z > 0.92  && current_triangle->v3->z < 1.27) ) {


if (UP) printf("UP:"); else printf("DN:");
printf("T1=%d T2=%d T3=%d ",current_triangle->v1->type,current_triangle->v2->type,
                            current_triangle->v3->type);
printf("L1=%d L2=%d L3=%d ",current_triangle->v1->level,current_triangle->v2->level,
                            current_triangle->v3->level);
printf("edge1=%d edge2=%d edge3=%d\n",dust_edge1,dust_edge2,dust_edge3);
printf("^^^^^[%lf,%lf][%lf,%lf][%lf,%lf]\n",
current_triangle->v1->x,current_triangle->v1->z,
current_triangle->v2->x,current_triangle->v2->z,
current_triangle->v3->x,current_triangle->v3->z);
if (new_v12!=NULL) {printf("V12[%lf,%lf]",new_v12->x,new_v12->z);}
if(new_v12==current_triangle->v1) printf("=1 ");
if(new_v12==current_triangle->v2) printf("=2 ");
if (new_v21!=NULL) {printf("V21[%lf,%lf]",new_v21->x,new_v21->z);}
if(new_v21==current_triangle->v1) printf("=1 ");
if(new_v21==current_triangle->v2) printf("=2 ");
if (new_v13!=NULL) {printf("V13[%lf,%lf]",new_v13->x,new_v13->z);}
if(new_v13==current_triangle->v1) printf("=1 ");
if(new_v13==current_triangle->v3) printf("=3 ");
if (new_v31!=NULL) {printf("V31[%lf,%lf]",new_v31->x,new_v31->z);}
if(new_v31==current_triangle->v1) printf("=1 ");
if(new_v31==current_triangle->v3) printf("=3 ");
if (new_v23!=NULL) {printf("V23[%lf,%lf]",new_v23->x,new_v23->z);}
if(new_v23==current_triangle->v3) printf("=3 ");
if(new_v23==current_triangle->v2) printf("=2 ");
if (new_v32!=NULL) {printf("V32[%lf,%lf]",new_v32->x,new_v32->z);}
if(new_v32==current_triangle->v3) printf("=3 ");
if(new_v32==current_triangle->v2) printf("=2 ");
printf("\n");
}


      /************************************************************************/
      /************************************************************************/      
      /* vertices 2 and 3 are out of the dust
 
              V3--------------V2                   V2        
               \              /                   /  \
                \            /                   /    \ 
         new_v13 O**********O new_v12           /      O new_v12
                  **********                   /      ***
                   ********                   /      *****
                    ******                   /      *******
                     ****                   /      *********
                      V1                   V3-----O********V1
                                               new_v13
      */
      if (new_v13!=NULL && new_v12!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v1->level!=668 &&
            current_triangle->v1->level!=667 &&
            current_triangle->v1->x!=0.0 && current_triangle->v1->z!=0.0) {
          if (new_v12==current_triangle->v1 || new_v13==current_triangle->v1)
            distance_v1_to_edge = 0.0;
          else 
            distance_v1_to_edge = _MATH_LIB_distance_point_to_line
                                            (current_triangle->v1,new_v12,new_v13);
          if (distance_v1_to_edge < min_distance_to_surface * 
               _ZELENJAK_size_limit(current_triangle->v1->x,current_triangle->v1->z)) {
             /* current_triangle->v1 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             if (_MATH_LIB_distance(current_triangle->v1->x,current_triangle->v1->z,
                                    new_v12->x,new_v12->z) <
                 _MATH_LIB_distance(current_triangle->v1->x,current_triangle->v1->z,
                                    new_v13->x,new_v13->z) ){
               current_triangle->v1->x = new_v12->x; 
               current_triangle->v1->z = new_v12->z;
             } else {
               current_triangle->v1->x = new_v13->x; 
               current_triangle->v1->z = new_v13->z;
             }
             current_triangle->v1->type = 2;
             current_triangle->v1->level = 668;
             /* check if there were some triangles with v1 already analyzed */
			 if (distance_v1_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v1);
                next_triangle = current_triangle->prev;
			 }
             if (new_v12!=current_triangle->v1 && new_v12!=current_triangle->v2) {
               /* remove new_v12 */
               if (UP) {new_v12->v3->v6 = new_v12->v6;new_v12->v6->v3 = new_v12->v3;}
               if (DOWN) {new_v12->v2->v5 = new_v12->v5;new_v12->v5->v2 = new_v12->v2;}
               _PERUNIKA_destroy_vertex(new_v12);
               new_v12 = current_triangle->v1;
             } /* end of removing new_v12 */
             if (new_v13!=current_triangle->v1 && new_v13!=current_triangle->v3) {
               /* remove new_v13 */
               if (UP) {new_v13->v1->v4 = new_v13->v4;new_v13->v4->v1 = new_v13->v1;}
               if (DOWN) {new_v13->v3->v6 = new_v13->v6;new_v13->v6->v3 = new_v13->v3;}
               _PERUNIKA_destroy_vertex(new_v13);
               new_v13 = current_triangle->v1;
             } /* end of removing new_v13 */
          }
        } /* end of: if (INPUT.TAU_SURFACE_SCALE > 1.0) */

        if (current_triangle->v1->level==668) {
          /* this vertex has been moved to the surface */
          if (current_triangle->v3->type!=3 && current_triangle->v2->type!=3) {
            if (UP) 
              _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v1,
                             current_triangle->v3,current_triangle->v2,4,3,2,
                             current_triangle->v1,lambda_computational_grid, status);
            if (DOWN) 
              _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v1,
                             current_triangle->v3,current_triangle->v2,3,2,1,
                             current_triangle->v1,lambda_computational_grid, status);
          }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
          current_triangle = next_triangle;
          continue; /* go to the next triangle */
        }

        if (new_v13!=current_triangle->v1 && new_v12!=current_triangle->v1){

          /***********************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v12,new_v13,current_triangle->v1,
                               current_triangle->v1,5,6,0,current_triangle->v1,
                               lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v12,new_v13,current_triangle->v1,
                               current_triangle->v1,6,1,0,current_triangle->v1,
                               lambda_computational_grid, status);
          }
          /***********************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v1;
            if (DOWN) current_vertex = current_triangle->v1->v2;
            if ( UP ) current_vertex = current_triangle->v1->v3;
            while (current_vertex!= new_v12) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v12;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v13;
            if (DOWN) current_vertex = new_v13->v6;
            if ( UP ) current_vertex = new_v13->v1;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");

          }/* end of (-R -1.0) flag: grid without surface recognition */ 
        }

        if ( current_triangle->v1->type!=3 && current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3 ) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN) {
              if (new_v13!=current_triangle->v3 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v3,
                               current_triangle->v2,3,2,1,current_triangle->v1,
                               lambda_computational_grid, status);
	      else if (new_v13==current_triangle->v3 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v2,
                               current_triangle->v2,1,2,0,current_triangle->v2,
                               lambda_computational_grid, status);
	      else if (new_v13!=current_triangle->v3 && new_v12==current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v3,
                               current_triangle->v3,3,4,0,current_triangle->v3,
                               lambda_computational_grid, status);
	    }
	    if (UP) {
              if (new_v13!=current_triangle->v3 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v3,
                               current_triangle->v2,4,3,2,current_triangle->v1,
                               lambda_computational_grid, status);
	      else if (new_v13==current_triangle->v3 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v2,
                               current_triangle->v2,2,3,0,current_triangle->v2,
                               lambda_computational_grid, status);
	      else if (new_v13!=current_triangle->v3 && new_v12==current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v13,new_v12,current_triangle->v3,
                               current_triangle->v3,4,5,0,current_triangle->v3,
                               lambda_computational_grid, status);
	    }
	  }
          else if (new_v13!=current_triangle->v3 && new_v12!=current_triangle->v2){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v3;
            if (new_v13 != current_triangle->v3) {
              if (DOWN) current_vertex = current_triangle->v3->v6;
              if ( UP ) current_vertex = current_triangle->v3->v1;
              while (current_vertex!= new_v13) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v6;
                if ( UP ) current_vertex = current_vertex->v1;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v13;
            }
            if (new_v12 != current_triangle->v2) {
              if (new_v12!=new_v13) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex =  new_v12;
              }
              if (DOWN) current_vertex = new_v12->v2;
              if ( UP ) current_vertex = new_v12->v3;
              while (current_vertex != current_triangle->v2) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v2;
                if ( UP ) current_vertex = current_vertex->v3;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v2;
            if (DOWN) current_vertex = current_triangle->v2->v4;
            if ( UP ) current_vertex = current_triangle->v2->v5;
            while (current_vertex != current_triangle->v3) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }
        }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 

      /************************************************************************/
      /************************************************************************/      
      /* vertices 1 and 3 are out of the dust
                   new_v23            
              V3-----O********V2                   V2
               \      *********                   ****
                \      *******                   ******
                 \      *****                   ********
                  \      ***           new_v23 O********O new_v21
                   \      O new_v21           /          \
                    \    /                   /            \
                     \  /                   /              \
                      V1                   V3--------------V1
      */
      else if (new_v21!=NULL && new_v23!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0  && current_triangle->v2->level!=668
            && current_triangle->v2->level!=667 &&
            current_triangle->v2->x!=0.0 && current_triangle->v2->z!=0.0) {
          if (new_v23==current_triangle->v2 || new_v21==current_triangle->v2)
            distance_v2_to_edge = 0.0;
          else 
            distance_v2_to_edge = _MATH_LIB_distance_point_to_line
                                        (current_triangle->v2,new_v21,new_v23);
          if (distance_v2_to_edge < min_distance_to_surface * 
            _ZELENJAK_size_limit(current_triangle->v2->x,current_triangle->v2->z)) {
             /* current_triangle->v2 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             if (_MATH_LIB_distance(current_triangle->v2->x,current_triangle->v2->z,
                                    new_v21->x,new_v21->z) <
                 _MATH_LIB_distance(current_triangle->v2->x,current_triangle->v2->z,
                                    new_v23->x,new_v23->z) ){
               current_triangle->v2->x = new_v21->x; 
               current_triangle->v2->z = new_v21->z;
             } else {
               current_triangle->v2->x = new_v23->x; 
               current_triangle->v2->z = new_v23->z;
             }
             current_triangle->v2->type = 2;
             current_triangle->v2->level = 668;
             /* check if there were some triangles with v2 already analyzed */
			 if (distance_v2_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v2);
                next_triangle = current_triangle->prev;
			 }
             if (new_v21!=current_triangle->v2 && new_v21!=current_triangle->v1) {
               /* remove new_v21 */
               if (UP) {new_v21->v3->v6 = new_v21->v6;new_v21->v6->v3 = new_v21->v3;}
               if (DOWN) {new_v21->v2->v5 = new_v21->v5;new_v21->v5->v2 = new_v21->v2;}
               _PERUNIKA_destroy_vertex(new_v21);
               new_v21 = current_triangle->v2;
             } /* end of removing new_v21 */
             if (new_v23!=current_triangle->v2 && new_v23!=current_triangle->v3) {
               /* remove new_v23 */
               if (DOWN) {new_v23->v1->v4 = new_v23->v4;new_v23->v4->v1 = new_v23->v1;}
               if (UP) {new_v23->v2->v5 = new_v23->v5;new_v23->v5->v2 = new_v23->v2;}
               _PERUNIKA_destroy_vertex(new_v23);
               new_v23 = current_triangle->v2;
             } /* end of removing new_v23 */
          }
        } /* end of: if (INPUT.TAU_SURFACE_SCALE > 1.0) */

        if (current_triangle->v2->level==668) {
          /* this vertex has been moved to the surface */
          if (current_triangle->v3->type!=3 && current_triangle->v1->type!=3) {
            if (UP)
              _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v2,
                             current_triangle->v1,current_triangle->v3,6,5,4,
                             current_triangle->v2,lambda_computational_grid, status);
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v2,
                             current_triangle->v1,current_triangle->v3,5,4,3,
                             current_triangle->v2,lambda_computational_grid, status);
          }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
          current_triangle = next_triangle;
          continue; /* go to the next triangle */
        }

        if (new_v21!=current_triangle->v2 && new_v23!=current_triangle->v2){
          /********************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v23,new_v21,current_triangle->v2,
                               current_triangle->v2,1,2,0,current_triangle->v2,
                               lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v23,new_v21,current_triangle->v2,
                               current_triangle->v2,2,3,0,current_triangle->v2,
                               lambda_computational_grid, status);
          }
          /*********************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v2;
            if (DOWN) current_vertex = current_triangle->v2->v4;
            if ( UP ) current_vertex = current_triangle->v2->v5;
            while (current_vertex!= new_v23) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v23;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v21;
            if (DOWN) current_vertex = new_v21->v2;
            if ( UP ) current_vertex = new_v21->v3;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");

          }/* end of (-R -1.0) flag: grid without surface recognition */ 

        }
        if ( current_triangle->v1->type!=3 &&
             current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN) {
              if (new_v21!=current_triangle->v1 && new_v23!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v1,
                               current_triangle->v3,5,4,3,current_triangle->v2,
                               lambda_computational_grid, status);
	      else if (new_v21==current_triangle->v1 && new_v23!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v3,
                               current_triangle->v3,3,4,0,current_triangle->v3,
                               lambda_computational_grid, status);
	      else if (new_v21!=current_triangle->v1 && new_v23==current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v1,
                               current_triangle->v1,5,6,0,current_triangle->v1,
                               lambda_computational_grid, status);
	    }
	    if (UP) {
              if (new_v21!=current_triangle->v1 && new_v23!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v1,
                               current_triangle->v3,6,5,4,current_triangle->v2,
                               lambda_computational_grid, status);
	      else if (new_v21==current_triangle->v1 && new_v23!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v3,
                               current_triangle->v3,4,5,0,current_triangle->v3,
                               lambda_computational_grid, status);
	      else if (new_v21!=current_triangle->v1 && new_v23==current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v21,new_v23,current_triangle->v1,
                               current_triangle->v1,6,1,0,current_triangle->v1,
                               lambda_computational_grid, status);
	    }
	  }
          else if (new_v21!=current_triangle->v1 && new_v23!=current_triangle->v3){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v1;
            if (new_v21 != current_triangle->v1) {
              if (DOWN) current_vertex = current_triangle->v1->v2;
              if ( UP ) current_vertex = current_triangle->v1->v3;
              while (current_vertex!= new_v21) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v2;
                if ( UP ) current_vertex = current_vertex->v3;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v21;
            }
            if (new_v23 != current_triangle->v3) {
              if (new_v23!=new_v21) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex =  new_v23;
              }
              if (DOWN) current_vertex = new_v23->v4;
              if ( UP ) current_vertex = new_v23->v5;
              while (current_vertex != current_triangle->v3) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v4;
                if ( UP ) current_vertex = current_vertex->v5;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v3;
            if (DOWN) current_vertex = current_triangle->v3->v6;
            if ( UP ) current_vertex = current_triangle->v3->v1;
            while (current_vertex != current_triangle->v1) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }
	}
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 

      /************************************************************************/
      /************************************************************************/      
      /* vertices 1 and 2 are out of the dust
                   new_v32            
              V3******O-------V2                 V2 
               *******        /                 /  \
                *****        /                 /    \
                 ***        /                 /      \ 
          new_v31 O        /         new_v32 O        \
                   \      /                 ***        \
                    \    /                 *****        \
                     \  /                 *******        \
                      V1                 V3******O--------V1
                                              new_v31
      */
      else if (new_v31!=NULL && new_v32!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v3->level!=668
            && current_triangle->v3->level!=667 &&
            current_triangle->v3->x!=0.0 && current_triangle->v3->z!=0.0) {
          if (new_v32==current_triangle->v3 || new_v31==current_triangle->v3)
            distance_v3_to_edge = 0.0;
          else 
            distance_v3_to_edge = _MATH_LIB_distance_point_to_line
               (current_triangle->v3,new_v31,new_v32);
          if (distance_v3_to_edge < min_distance_to_surface * 
            _ZELENJAK_size_limit(current_triangle->v3->x,current_triangle->v3->z)) {
             /* current_triangle->v3 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             if (_MATH_LIB_distance(current_triangle->v3->x,current_triangle->v3->z,
                                    new_v31->x,new_v31->z) <
                 _MATH_LIB_distance(current_triangle->v3->x,current_triangle->v3->z,
                                    new_v32->x,new_v32->z) ){
               current_triangle->v3->x = new_v31->x; 
               current_triangle->v3->z = new_v31->z;
             } else {
               current_triangle->v3->x = new_v32->x; 
               current_triangle->v3->z = new_v32->z;
             }
             current_triangle->v3->type = 2;
             current_triangle->v3->level = 668;
              /* check if there were some triangles with v3 already analyzed */
			 if (distance_v3_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v3);
                next_triangle = current_triangle->prev;
			 }
             if (new_v31!=current_triangle->v3 && new_v31!=current_triangle->v1) {
               /* remove new_v31 */
               if (UP) {new_v31->v1->v4 = new_v31->v4;new_v31->v4->v1 = new_v31->v1;}
               if (DOWN) {new_v31->v3->v6 = new_v31->v6;new_v31->v6->v3 = new_v31->v3;}
               _PERUNIKA_destroy_vertex(new_v31);
               new_v31 = current_triangle->v3;
             } /* end of removing new_v31 */
             if (new_v32!=current_triangle->v3 && new_v32!=current_triangle->v2) {
               /* remove new_v32 */
               if (DOWN) {new_v32->v1->v4 = new_v32->v4;new_v32->v4->v1 = new_v32->v1;}
               if (UP) {new_v32->v2->v5 = new_v32->v5;new_v32->v5->v2 = new_v32->v2;}
               _PERUNIKA_destroy_vertex(new_v32);
               new_v32 = current_triangle->v3;
             } /* end of removing new_v32 */
          }
        } /* end of: if (INPUT.TAU_SURFACE_SCALE > 1.0) */

        if (current_triangle->v3->level==668) {
          /* this vertex has been moved to the surface */
          if (current_triangle->v1->type!=3 && current_triangle->v2->type!=3) {
            if (UP)
              _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v3,
                             current_triangle->v2,current_triangle->v1,2,1,6,
                             current_triangle->v3,lambda_computational_grid, status);
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v3,
                             current_triangle->v2,current_triangle->v1,1,6,5,
                             current_triangle->v3,lambda_computational_grid, status);
          }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
          current_triangle = next_triangle;
          continue; /* go to the next triangle */
        }

        if (new_v31!=current_triangle->v3 && new_v32!=current_triangle->v3){
          /***************************************************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v31,new_v32,current_triangle->v3,
                               current_triangle->v3,3,4,0,current_triangle->v3,
                          lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v31,new_v32,current_triangle->v3,
                               current_triangle->v3,4,5,0,current_triangle->v3,
                          lambda_computational_grid, status);
          }
          /***************************************************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */

            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v3;
            if (DOWN) current_vertex = current_triangle->v3->v6;
            if ( UP ) current_vertex = current_triangle->v3->v1;
            while (current_vertex!= new_v31) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v31;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v32;
            if (DOWN) current_vertex = new_v32->v4;
            if ( UP ) current_vertex = new_v32->v5;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");

          }/* end of (-R -1.0) flag: grid without surface recognition */ 
        }

        if ( current_triangle->v1->type!=3 &&
             current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3 ) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN) {
              if (new_v32!=current_triangle->v2 && new_v31!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v2,
                               current_triangle->v1,1,6,5,current_triangle->v3,
                               lambda_computational_grid, status);
	      else if (new_v32==current_triangle->v2 && new_v31!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v1,
                               current_triangle->v1,1,2,0,current_triangle->v1,
                               lambda_computational_grid, status);
	      else if (new_v32!=current_triangle->v2 && new_v31==current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v2,
                               current_triangle->v2,5,6,0,current_triangle->v2,
                               lambda_computational_grid, status);
	    }
	    if (UP) {
              if (new_v32!=current_triangle->v2 && new_v31!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v2,
                               current_triangle->v1,2,1,6,current_triangle->v3,
                               lambda_computational_grid, status);
	      else if (new_v32==current_triangle->v2 && new_v31!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v1,
                               current_triangle->v1,6,1,0,current_triangle->v1,
                               lambda_computational_grid, status);
	      else if (new_v32!=current_triangle->v2 && new_v31==current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v32,new_v31,current_triangle->v2,
                               current_triangle->v2,2,3,0,current_triangle->v2,
                               lambda_computational_grid, status);
	    }
	  }
          else if (new_v32!=current_triangle->v2 && new_v31!=current_triangle->v1){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v2;
            if (new_v32 != current_triangle->v2) {
              if (DOWN) current_vertex = current_triangle->v2->v4;
              if ( UP ) current_vertex = current_triangle->v2->v5;
              while (current_vertex!= new_v32) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v4;
                if ( UP ) current_vertex = current_vertex->v5;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v32;
            }
            if (new_v31 != current_triangle->v1) {
              if (new_v31!=new_v32) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex =  new_v31;
              }
              if (DOWN) current_vertex = new_v31->v6;
              if ( UP ) current_vertex = new_v31->v1;
              while (current_vertex != current_triangle->v1) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v6;
                if ( UP ) current_vertex = current_vertex->v1;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v1;
            if (DOWN) current_vertex = current_triangle->v1->v2;
            if ( UP ) current_vertex = current_triangle->v1->v3;
            while (current_vertex != current_triangle->v2) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }
	}
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 

      /************************************************************************/
      /************************************************************************/      
      /* vertex 1 is out of the dust
                          
              V3**************V2                   V2
               ****************                   ****
                **************                   ******
                 ************                   ********
          new_v31 O********O new_v21           *********O new_v21
                   \      /                   ********** \
                    \    /                   **********   \
                     \  /                   **********     \
                      V1                   V3*******O-------V1
                                                 new_v31
      */
      else if (new_v31!=NULL && new_v21!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v2->level !=668
            && current_triangle->v2->level !=667 &&
            current_triangle->v2->x!=0.0 && current_triangle->v2->z!=0.0) {
          if (new_v21==current_triangle->v2) distance_v2_to_edge = 0.0;
          else distance_v2_to_edge = 
              _MATH_LIB_distance_point_to_line(current_triangle->v2,new_v21,new_v31) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v2->x,current_triangle->v2->z));
          if (distance_v2_to_edge < 1.0) {
             /* current_triangle->v2 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v2->x = new_v21->x; 
             current_triangle->v2->z = new_v21->z;
             current_triangle->v2->type = 2;
             current_triangle->v2->level = 668;
             /* check if there were some triangles with v2 already analyzed */
			 if (distance_v2_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v2);
                next_triangle = current_triangle->prev;
			 }
             if (new_v21!=current_triangle->v2) {
               /* remove new_v21 */
               if (UP) {new_v21->v3->v6 = new_v21->v6;new_v21->v6->v3 = new_v21->v3;}
               if (DOWN) {new_v21->v2->v5 = new_v21->v5;new_v21->v5->v2 = new_v21->v2;}
               _PERUNIKA_destroy_vertex(new_v21);
               new_v21 = current_triangle->v2;
             }
           }
        } /* end of checking the distance of current_triangle->v2 to the surface */
        
        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v3->level !=668
            && current_triangle->v3->level !=667 &&
            current_triangle->v3->x!=0.0 && current_triangle->v3->z!=0.0) {
          if (new_v31==current_triangle->v3) distance_v3_to_edge = 0.0;
          else distance_v3_to_edge =
              _MATH_LIB_distance_point_to_line(current_triangle->v3,new_v21,new_v31) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v3->x,current_triangle->v3->z));
          if (distance_v3_to_edge < 1.0) {
             /* current_triangle->v3 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v3->x = new_v31->x; 
             current_triangle->v3->z = new_v31->z;
             current_triangle->v3->type = 2;
             current_triangle->v3->level = 668;
             /* check if there were some triangles with v3 already analyzed */
			 if (distance_v3_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v3);
                next_triangle = current_triangle->prev;
			 }
             if (new_v31!=current_triangle->v3) { 
               /* remove new_v31 */
               if (UP) {new_v31->v4->v1 = new_v31->v1;new_v31->v1->v4 = new_v31->v4;}
               if (DOWN) {new_v31->v3->v6 = new_v31->v6;new_v31->v6->v3 = new_v31->v3;}
               _PERUNIKA_destroy_vertex(new_v31);
               new_v31 = current_triangle->v3;
             }
           }
        } /* end of checking the distance of current_triangle->v3 to the surface */

        if (current_triangle->v2->level==668 || current_triangle->v3->level==668) {
          /* at least one of these vertices has been moved to the surface */
          /* if one vertex has level=668, but the other does not, then split this
              into two triangles: */
          if (current_triangle->v2->level==668 && current_triangle->v3->level!=668) {
	         if (current_triangle->v1->type!=3){
                new_triangle= _PERUNIKA_create_triangle(&error);
                new_triangle->v1=current_triangle->v1;
                new_triangle->v2=current_triangle->v2;
                new_triangle->v3=new_v31;
			 }
	         else new_triangle = NULL; 
             current_triangle->v1 = new_v31;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,4,5,0,
                             current_triangle->v3,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v2,new_triangle->v3,
                             new_triangle->v1,new_triangle->v1,6,1,0,
                             new_triangle->v1,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,3,4,0,
                             current_triangle->v3,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v2,new_triangle->v3,
                             new_triangle->v1,new_triangle->v1,5,6,0,
                             new_triangle->v1,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
          }
          else if (current_triangle->v3->level==668 && current_triangle->v2->level!=668) {
	         if (current_triangle->v1->type!=3){
               new_triangle= _PERUNIKA_create_triangle(&error);
               new_triangle->v1=current_triangle->v1;
               new_triangle->v2=new_v21;
               new_triangle->v3=current_triangle->v3;
			 }
	         else new_triangle = NULL; 
             current_triangle->v1 = new_v21;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,2,3,0,
                             current_triangle->v2,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v2,new_triangle->v3,
                             new_triangle->v1,new_triangle->v1,6,1,0,
                             new_triangle->v1,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,1,2,0,
                             current_triangle->v2,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v2,new_triangle->v3,
                             new_triangle->v1,new_triangle->v1,5,6,0,
                             new_triangle->v1,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
          }
          /* both vertices are on the surface: */
          else {
             if (UP) if (current_triangle->v1->type!=3)
                 _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,6,1,0,
                             current_triangle->v1,lambda_computational_grid, status);
             if (DOWN) if (current_triangle->v1->type!=3)
                 _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,5,6,0,
                             current_triangle->v1,lambda_computational_grid, status);
          }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
          /* go to the next triangle */
          current_triangle = next_triangle;
          continue; 
        }

        if (new_v21!=current_triangle->v2 || new_v31!=current_triangle->v3){

          /***************************************************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v31,new_v21,current_triangle->v3,
                               current_triangle->v2,3,2,1,current_triangle->v1,
                          lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v31,new_v21,current_triangle->v3,
                               current_triangle->v2,4,3,2,current_triangle->v1,
                          lambda_computational_grid, status);
          }
          /***************************************************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v3;
            if (new_v31!=current_triangle->v3) {
              if (DOWN) current_vertex = current_triangle->v3->v6;
              if ( UP ) current_vertex = current_triangle->v3->v1;
              while (current_vertex!= new_v31) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v6;
                if ( UP ) current_vertex = current_vertex->v1;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v31;
            }
            if (new_v21!=current_triangle->v2) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v21;
              if (DOWN) current_vertex = new_v21->v2;
              if ( UP ) current_vertex = new_v21->v3;
              while (current_vertex!= current_triangle->v2) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v2;
                if ( UP ) current_vertex = current_vertex->v3;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v2;
            if (DOWN) current_vertex = current_triangle->v2->v4;
            if ( UP ) current_vertex = current_triangle->v2->v5;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }/* end of (-R -1.0) flag: grid without surface recognition */ 

        }

        if ( current_triangle->v1->type!=3 &&
             current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3 ) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN && new_v31!=current_triangle->v1 && new_v21!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v21,new_v31,current_triangle->v1,
                               current_triangle->v1,5,6,0,current_triangle->v1,
                               lambda_computational_grid, status);
	    if (UP && new_v31!=current_triangle->v1 && new_v21!=current_triangle->v1)
_PERUNIKA_surface_grid_refinement (new_v21,new_v31,current_triangle->v1,
                               current_triangle->v1,6,1,0,current_triangle->v1,
                               lambda_computational_grid, status);
	  }
          else if (new_v31!=current_triangle->v1 && new_v21!=current_triangle->v1){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v1;
            if (DOWN) current_vertex = current_triangle->v1->v2;
            if ( UP ) current_vertex = current_triangle->v1->v3;
            while (current_vertex!= new_v21) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v21;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v31;
            if (DOWN) current_vertex = new_v31->v6;
            if ( UP ) current_vertex = new_v31->v1;
            while (current_vertex != current_triangle->v1) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }
	}
        /* we have built new triangles, destroy the old one */ 
        error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 

      /************************************************************************/
      /************************************************************************/      
      /* vertex 2 is out of the dust
                    new_v32
              V3*******O------V2                   V2
               **********     /                   /  \
                **********   /                   /    \
                 ********** /                   /      \
                  *********O new_v12   new_v32 O********O new_v12
                   ********                   ************
                    ******                   **************
                     ****                   ****************
                      V1                   V3**************V1
      */
      else if (new_v32!=NULL && new_v12!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v1->level !=668
            && current_triangle->v1->level !=667 &&
            current_triangle->v1->x!=0.0 && current_triangle->v1->z!=0.0) {
          if (new_v12==current_triangle->v1) distance_v1_to_edge = 0.0;
          else distance_v1_to_edge = 
              _MATH_LIB_distance_point_to_line(current_triangle->v1,new_v12,new_v32) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v1->x,current_triangle->v1->z));
          if (distance_v1_to_edge < 1.0) {
             /* current_triangle->v1 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v1->x = new_v12->x; 
             current_triangle->v1->z = new_v12->z;
             current_triangle->v1->type = 2;
             current_triangle->v1->level = 668;
             /* check if there were some triangles with v1 already analyzed */
			 if (distance_v1_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v1);
                next_triangle = current_triangle->prev;
			 }
             if (new_v12!=current_triangle->v1) {
               /* remove new_v12 */
               if (UP) {new_v12->v3->v6 = new_v12->v6;new_v12->v6->v3 = new_v12->v3;}
               if (DOWN) {new_v12->v2->v5 = new_v12->v5;new_v12->v5->v2 = new_v12->v2;}
               _PERUNIKA_destroy_vertex(new_v12);
               new_v12 = current_triangle->v1;
             }
           }
        }

        if (INPUT.TAU_SURFACE_SCALE > 1.0 && current_triangle->v3->level !=668
            && current_triangle->v3->level !=667 &&
            current_triangle->v3->x!=0.0 && current_triangle->v3->z!=0.0) {
          if (new_v32==current_triangle->v3) distance_v3_to_edge = 0.0;
          else distance_v3_to_edge =
              _MATH_LIB_distance_point_to_line(current_triangle->v3,new_v12,new_v32) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v3->x,current_triangle->v3->z));
          if (distance_v3_to_edge < 1.0) {
             /* current_triangle->v3 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v3->x = new_v32->x; 
             current_triangle->v3->z = new_v32->z;
             current_triangle->v3->type = 2;
             current_triangle->v3->level = 668;
             /* check if there were some triangles with v3 already analyzed */
			 if (distance_v3_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v3);
                next_triangle = current_triangle->prev;
			 }
             if (new_v32!=current_triangle->v3) {
               /* remove new_v32 */
               if (UP) {new_v32->v2->v5 = new_v32->v5;new_v32->v5->v2 = new_v32->v2;}
               if (DOWN) {new_v32->v1->v4 = new_v32->v4;new_v32->v4->v1 = new_v32->v1;}
               _PERUNIKA_destroy_vertex(new_v32);
               new_v32 = current_triangle->v3;
             }
           }
         }  
        
         if (current_triangle->v1->level==668 || current_triangle->v3->level==668) {
           /* at least one of these vertices has been moved to the surface */
          
           /* if one vertex has level=668, but the other does not, then split this
              into two triangles */
           if (current_triangle->v1->level==668 && current_triangle->v3->level!=668) {
	     if (current_triangle->v2->type!=3){
               new_triangle = _PERUNIKA_create_triangle(&error);
               new_triangle->v1=current_triangle->v1;
               new_triangle->v2=current_triangle->v2;
               new_triangle->v3=new_v32;
	     }
	     else new_triangle = NULL; 
             current_triangle->v2 = new_v32;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,4,5,0,
                             current_triangle->v3,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v3,new_triangle->v1,
                             new_triangle->v2,new_triangle->v2,2,3,0,
                             new_triangle->v2,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,3,4,0,
                             current_triangle->v3,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v3,new_triangle->v1,
                             new_triangle->v2,new_triangle->v2,1,2,0,
                             new_triangle->v2,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
           }
           else if (current_triangle->v3->level==668 && current_triangle->v1->level!=668) {
	     if (current_triangle->v2->type!=3){
               new_triangle = _PERUNIKA_create_triangle(&error);
               new_triangle->v1=new_v12;
               new_triangle->v2=current_triangle->v2;
               new_triangle->v3=current_triangle->v3;
	     }
	     else new_triangle = NULL; 
             current_triangle->v2 = new_v12;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,6,1,0,
                             current_triangle->v1,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v3,new_triangle->v1,
                             new_triangle->v2,new_triangle->v2,2,3,0,
                             new_triangle->v2,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,5,6,0,
                             current_triangle->v1,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v3,new_triangle->v1,
                             new_triangle->v2,new_triangle->v2,1,2,0,
                             new_triangle->v2,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
           }
          /* both vertices are on the surface: */
          else {
             if (UP) if (current_triangle->v2->type!=3)
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,2,3,0,
                             current_triangle->v2,lambda_computational_grid, status);
             if (DOWN) if (current_triangle->v2->type!=3)
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,1,2,0,
                             current_triangle->v2,lambda_computational_grid, status);
          }
          /* we have built new triangles, destroy the old one */ 
          error = _PERUNIKA_destroy_triangle (current_triangle);
          /* go to the next triangle */
          current_triangle = next_triangle;
          continue;           
        } 
        
        if (new_v12!=current_triangle->v1 || new_v32!=current_triangle->v3){
          /***************************************************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v12,new_v32,current_triangle->v1,
                   current_triangle->v3,5,4,3,current_triangle->v2,
                   lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v12,new_v32,current_triangle->v1,
                   current_triangle->v3,6,5,4,current_triangle->v2,
                   lambda_computational_grid, status);
          }
          /***************************************************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */ 
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v1;
            if (new_v12!=current_triangle->v1) {
              if (DOWN) current_vertex = current_triangle->v1->v2;
              if ( UP ) current_vertex = current_triangle->v1->v3;
              while (current_vertex!= new_v12) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v2;
                if ( UP ) current_vertex = current_vertex->v3;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v12;
            }
            if (new_v32!=current_triangle->v3) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v32;
              if (DOWN) current_vertex = new_v32->v4;
              if ( UP ) current_vertex = new_v32->v5;
              while (current_vertex!= current_triangle->v3) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v4;
                if ( UP ) current_vertex = current_vertex->v5;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v3;
            if (DOWN) current_vertex = current_triangle->v3->v6;
            if ( UP ) current_vertex = current_triangle->v3->v1;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");

          }/* end of (-R -1.0) flag: grid without surface recognition */ 
        }
        if ( current_triangle->v1->type!=3 &&
             current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3 ) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN && new_v32!=current_triangle->v2 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v32,new_v12,current_triangle->v2,
                               current_triangle->v2,1,2,0,current_triangle->v2,
                               lambda_computational_grid, status);
	    if (UP && new_v32!=current_triangle->v2 && new_v12!=current_triangle->v2)
_PERUNIKA_surface_grid_refinement (new_v32,new_v12,current_triangle->v2,
                               current_triangle->v2,2,3,0,current_triangle->v2,
                               lambda_computational_grid, status);
	  }
          else if (new_v32!=current_triangle->v2 && new_v12!=current_triangle->v2){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v2;
            if (DOWN) current_vertex = current_triangle->v2->v4;
            if ( UP ) current_vertex = current_triangle->v2->v5;
            while (current_vertex!= new_v32) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v32;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v12;
            if (DOWN) current_vertex = new_v12->v2;
            if ( UP ) current_vertex = new_v12->v3;
            while (current_vertex != current_triangle->v2) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          } 
	}
        /* we have built new triangles, destroy the old one */ 
        error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 

      /************************************************************************/
      /************************************************************************/      
      /* vertex 3 is out of the dust
                    new_v23
              V3------O*******V2               V2
               \     **********               ****
                \   **********               ******
                 \ **********               ********
          new_v13 O*********       new_v23 O*********
                   ********               / **********
                    ******               /   **********
                     ****               /     **********
                      V1               V3------O*******V1
                                            new_v13
      */
      else if (new_v13!=NULL && new_v23!=NULL) {

        /* if we look for a very fine surface grid, check if some points are
           too close to the dust surface */
        if (INPUT.TAU_SURFACE_SCALE > 1.0  && current_triangle->v1->level !=668
            && current_triangle->v1->level !=667 &&
            current_triangle->v1->x!=0.0 && current_triangle->v1->z!=0.0) {
          if (new_v13==current_triangle->v1) distance_v1_to_edge = 0.0;
          else distance_v1_to_edge = 
              _MATH_LIB_distance_point_to_line(current_triangle->v1,new_v13,new_v23) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v1->x,current_triangle->v1->z));
          if (distance_v1_to_edge < 1.0) {
             /* current_triangle->v1 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v1->x = new_v13->x; 
             current_triangle->v1->z = new_v13->z;
             current_triangle->v1->type = 2;
             current_triangle->v1->level = 668;
             /* check if there were some triangles with v1 already analyzed */
			 if (distance_v1_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v1);
                next_triangle = current_triangle->prev;
			 }
             if (new_v13!=current_triangle->v1) {
               /* remove new_v13 */
               if (UP) {new_v13->v1->v4 = new_v13->v4;new_v13->v4->v1 = new_v13->v1;}
               if (DOWN) {new_v13->v3->v6 = new_v13->v6;new_v13->v6->v3 = new_v13->v3;}
               _PERUNIKA_destroy_vertex(new_v13);
               new_v13 = current_triangle->v1;
             }
           }
        }
        
        if (INPUT.TAU_SURFACE_SCALE > 1.0  && current_triangle->v2->level !=668
            && current_triangle->v2->level !=667 &&
            current_triangle->v2->x!=0.0 && current_triangle->v2->z!=0.0) {
          if (new_v23==current_triangle->v2) distance_v2_to_edge = 0.0;
          else distance_v2_to_edge =
              _MATH_LIB_distance_point_to_line(current_triangle->v2,new_v13,new_v23) /
                  (min_distance_to_surface * 
                  _ZELENJAK_size_limit(current_triangle->v2->x,current_triangle->v2->z));
          if (distance_v2_to_edge < 1.0) {
             /* current_triangle->v2 is too close to the dust surface, therefore we
                will move it exactly to the surface */
             current_triangle->v2->x = new_v23->x; 
             current_triangle->v2->z = new_v23->z;
             current_triangle->v2->type = 2;
             current_triangle->v2->level = 668;
             /* check if there were some triangles with v2 already analyzed */
			 if (distance_v2_to_edge>0.0){
                _PERUNIKA_move_triangle_back (current_triangle,current_triangle->v2);
                next_triangle = current_triangle->prev;
			 }
             if (new_v23!=current_triangle->v2) {
               /* remove new_v23 */
               if (UP) {new_v23->v2->v5 = new_v23->v5;new_v23->v5->v2 = new_v23->v2;}
               if (DOWN) {new_v23->v1->v4 = new_v23->v4;new_v23->v4->v1 = new_v23->v1;}
               _PERUNIKA_destroy_vertex(new_v23);
               new_v23 = current_triangle->v2;
             }
           }
         }
         
         if (current_triangle->v1->level==668 || current_triangle->v2->level==668) {
           /* at least one of these vertices has been moved to the surface */

           /* if one vertex has level=668, but the other does not, then split this
              into two triangles */
          if (current_triangle->v1->level==668 && current_triangle->v2->level!=668) {
	     if (current_triangle->v3->type!=3){
               new_triangle = _PERUNIKA_create_triangle(&error);
               new_triangle->v1=current_triangle->v1;
               new_triangle->v3=current_triangle->v3;
               new_triangle->v2=new_v23;
	     }
	     else new_triangle = NULL; 
             current_triangle->v3 = new_v23;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,2,3,0,
                             current_triangle->v2,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v1,new_triangle->v2,
                             new_triangle->v3,new_triangle->v3,4,5,0,
                             new_triangle->v3,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v3,current_triangle->v1,
                             current_triangle->v2,current_triangle->v2,1,2,0,
                             current_triangle->v2,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v1,new_triangle->v2,
                             new_triangle->v3,new_triangle->v3,3,4,0,
                             new_triangle->v3,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
           }
           else if (current_triangle->v2->level==668 && current_triangle->v1->level!=668) {
     	     if (current_triangle->v3->type!=3){
               new_triangle = _PERUNIKA_create_triangle(&error);
               new_triangle->v1=new_v13;
               new_triangle->v2=current_triangle->v2;
               new_triangle->v3=current_triangle->v3;
			 }
	         else new_triangle = NULL; 
             current_triangle->v3 = new_v13;
             if (UP) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,6,1,0,
                             current_triangle->v1,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v1,new_triangle->v2,
                             new_triangle->v3,new_triangle->v3,4,5,0,
                             new_triangle->v3,lambda_computational_grid, status);
             }
             if (DOWN) {
               _PERUNIKA_surface_grid_refinement (current_triangle->v2,current_triangle->v3,
                             current_triangle->v1,current_triangle->v1,5,6,0,
                             current_triangle->v1,lambda_computational_grid, status);
               if (new_triangle!=NULL)
                 _PERUNIKA_surface_grid_refinement (new_triangle->v1,new_triangle->v2,
                             new_triangle->v3,new_triangle->v3,3,4,0,
                             new_triangle->v3,lambda_computational_grid, status);
             }
			 /* we have built new triangles, destroy the old one */
             if (new_triangle!=NULL) error = _PERUNIKA_destroy_triangle (new_triangle);
           }
           /* both vertices are on the surface: */
           else {
             if (UP)   if (current_triangle->v3->type!=3)
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,4,5,0,
                             current_triangle->v3,lambda_computational_grid, status);
             if (DOWN) if (current_triangle->v3->type!=3)
               _PERUNIKA_surface_grid_refinement (current_triangle->v1,current_triangle->v2,
                             current_triangle->v3,current_triangle->v3,3,4,0,
                             current_triangle->v3,lambda_computational_grid, status);
           }
           /* we have built new triangles, destroy the old one */ 
           error = _PERUNIKA_destroy_triangle (current_triangle);
           /* go to the next triangle */
           current_triangle = next_triangle;
           continue;           
        } 
        
        if (new_v13!=current_triangle->v1 || new_v23!=current_triangle->v2){

          /************************************/
          /* if we look for a very fine surface grid (input flag -R): */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
            if (DOWN)
              _PERUNIKA_surface_grid_refinement (new_v23,new_v13,current_triangle->v2,
                   current_triangle->v1,1,6,5,current_triangle->v3,
                   lambda_computational_grid, status);
            if (UP)
              _PERUNIKA_surface_grid_refinement (new_v23,new_v13,current_triangle->v2,
                   current_triangle->v1,2,1,6,current_triangle->v3,
                   lambda_computational_grid, status);
          }
          /*************************************/
          else {

            /* usual grid refinement...  (flag -R -1.0) */ 
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v2;
            if (new_v23!=current_triangle->v2) {
              if (DOWN) current_vertex = current_triangle->v2->v4;
              if ( UP ) current_vertex = current_triangle->v2->v5;
              while (current_vertex!= new_v23) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v4;
                if ( UP ) current_vertex = current_vertex->v5;
              }
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v23;
            }
            if (new_v13!=current_triangle->v1) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex =  new_v13;
              if (DOWN) current_vertex = new_v13->v6;
              if ( UP ) current_vertex = new_v13->v1;
              while (current_vertex!= current_triangle->v1) {
                polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
                polygon_vertex->vertex = current_vertex;
                if (DOWN) current_vertex = current_vertex->v6;
                if ( UP ) current_vertex = current_vertex->v1;
              }
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex = current_triangle->v1;
            if (DOWN) current_vertex = current_triangle->v1->v2;
            if ( UP ) current_vertex = current_triangle->v1->v3;
            while (current_vertex!= polygon->vertex) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v2;
              if ( UP ) current_vertex = current_vertex->v3;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");

          }/* end of (-R -1.0) flag: grid without surface recognition */ 
        }
        if ( current_triangle->v1->type!=3 &&
             current_triangle->v2->type!=3 &&
             current_triangle->v3->type!=3 ) {
          /* a situation when the whole triangle is in the dust */
          if (INPUT.TAU_SURFACE_SCALE > 1.0) {
	    if (DOWN && new_v23!=current_triangle->v3 && new_v13!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v13,new_v23,current_triangle->v3,
                               current_triangle->v3,3,4,0,current_triangle->v3,
                               lambda_computational_grid, status);
	    if (UP && new_v23!=current_triangle->v3 && new_v13!=current_triangle->v3)
_PERUNIKA_surface_grid_refinement (new_v13,new_v23,current_triangle->v3,
                               current_triangle->v3,4,5,0,current_triangle->v3,
                               lambda_computational_grid, status);
	  }
          else if (new_v23!=current_triangle->v3 && new_v13!=current_triangle->v3){
            polygon = NULL;
            polygon =_PERUNIKA_create_polygon_point (polygon, &error);
            polygon->vertex = current_triangle->v3;
            if (DOWN) current_vertex = current_triangle->v3->v6;
            if ( UP ) current_vertex = current_triangle->v3->v1;
            while (current_vertex!= new_v13) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v6;
              if ( UP ) current_vertex = current_vertex->v1;
            }
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v13;
            polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
            polygon_vertex->vertex =  new_v23;
            if (DOWN) current_vertex = new_v23->v4;
            if ( UP ) current_vertex = new_v23->v5;
            while (current_vertex != current_triangle->v3) {
              polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
              polygon_vertex->vertex = current_vertex;
              if (DOWN) current_vertex = current_vertex->v4;
              if ( UP ) current_vertex = current_vertex->v5;
            }
            error = _PERUNIKA_split_polygon (polygon);
            if (error) printf("POLYGON ERROR!!!!!\n");
          }
	}
        /* we have built new triangles, destroy the old one */ 
        error = _PERUNIKA_destroy_triangle (current_triangle);
      }                 
      /************************************************************************/
      /************************************************************************/      

    } /* end of the edge-triangle */
    else {
      /* this is an ordinary triangle. Check if it has additional vertices 
     on its sides. */
      if ( ( DOWN && ( current_triangle->v1->v2!=current_triangle->v2 ||
                       current_triangle->v2->v4!=current_triangle->v3 ||
                       current_triangle->v3->v6!=current_triangle->v1 )) ||
            ( UP  && ( current_triangle->v1->v3!=current_triangle->v2 ||
                       current_triangle->v2->v5!=current_triangle->v3 ||
                       current_triangle->v3->v1!=current_triangle->v1 ))) {
    /* OK, this triangle has to be transformed into a polygon and split
           into smaller triangles */
    polygon = NULL;
        polygon =_PERUNIKA_create_polygon_point (polygon, &error);
        polygon->vertex = current_triangle->v1;
    if (DOWN) current_vertex = current_triangle->v1->v2;
        if ( UP ) current_vertex = current_triangle->v1->v3;
        while (current_vertex!= current_triangle->v2) {
          polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
      polygon_vertex->vertex = current_vertex;
          if (DOWN) current_vertex = current_vertex->v2;
          if ( UP ) current_vertex = current_vertex->v3;
    }
        polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
    polygon_vertex->vertex = current_triangle->v2;
    if (DOWN) current_vertex = current_triangle->v2->v4;
        if ( UP ) current_vertex = current_triangle->v2->v5;
        while (current_vertex!= current_triangle->v3) {
          polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
      polygon_vertex->vertex = current_vertex;
          if (DOWN) current_vertex = current_vertex->v4;
          if ( UP ) current_vertex = current_vertex->v5;
    }
        polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
    polygon_vertex->vertex = current_triangle->v3;
    if (DOWN) current_vertex = current_triangle->v3->v6;
        if ( UP ) current_vertex = current_triangle->v3->v1;
        while (current_vertex!= current_triangle->v1) {
          polygon_vertex = _PERUNIKA_create_polygon_point(polygon,&error);
      polygon_vertex->vertex = current_vertex;
          if (DOWN) current_vertex = current_vertex->v6;
          if ( UP ) current_vertex = current_vertex->v1;
    }
        error = _PERUNIKA_split_polygon (polygon);
        if (error) printf("POLYGON ERROR!!!!!\n");
        error = _PERUNIKA_destroy_triangle (current_triangle);
      }
    }

    current_triangle = next_triangle;
  } while (current_triangle != start_triangle_list);


  /* I had to skip my racquetball game because of a severe thunderstorm and
     a small tornado ...
     Anyway, let's continue with this code:
     now we have only triangles in the grid, and we have to check if some
     of them have vertices with x<0.0 */

  /* start the loop over triangles */
  current_triangle = start_triangle_list;
  do {
    /* the next triangle will be: */
    next_triangle = current_triangle->prev;

    /* mark the vertices with x<0.0 with the fake type */
    if (current_triangle->v1->x<0.0) current_triangle->v1->type=3;
    if (current_triangle->v2->x<0.0) current_triangle->v2->type=3;
    if (current_triangle->v3->x<0.0) current_triangle->v3->type=3;

    /* if all three vertices have x<0.0 then destroy the triangle */
    if (current_triangle->v1->type == 3 &&
        current_triangle->v2->type == 3 &&
        current_triangle->v3->type == 3)
      error=_PERUNIKA_destroy_triangle (current_triangle);
    /* if at least one vertex has x<0.0, cut the triangle */
    else if (current_triangle->v1->type == 3 ||
        current_triangle->v2->type == 3 ||
        current_triangle->v3->type == 3){

      /* find the triangle's sides which are passing over the x=0.0 line */
      new_v12=new_v21=new_v23=new_v32=new_v31=new_v13=NULL;
      /* vertex 1 has x>=0.0, vertex 2 has x<0.0 */
      if (current_triangle->v1->type!=3 && current_triangle->v2->type==3)
        new_v12=_PERUNIKA_create_point_at_x_zero 
        (current_triangle->v1,current_triangle->v2,
                 lambda_computational_grid,status);
      /* vertex 2 has x>=0.0, vertex 1 has x<0.0 */
      if (current_triangle->v1->type==3 && current_triangle->v2->type!=3)
        new_v21=_PERUNIKA_create_point_at_x_zero 
        (current_triangle->v2,current_triangle->v1,
                 lambda_computational_grid,status);
      /* vertex 2 has x>=0.0, vertex 3 has x<0.0 */
      if (current_triangle->v2->type!=3 && current_triangle->v3->type==3)
        new_v23=_PERUNIKA_create_point_at_x_zero 
        (current_triangle->v2,current_triangle->v3,
                 lambda_computational_grid,status);
      /* vertex 3 has x>=0.0, vertex 2 has x<0.0 */
      if (current_triangle->v2->type==3 && current_triangle->v3->type!=3)
        new_v32=_PERUNIKA_create_point_at_x_zero
        (current_triangle->v3,current_triangle->v2,
                 lambda_computational_grid,status);
      /* vertex 3 has x>=0.0, vertex 1 has x<0.0 */
      if (current_triangle->v3->type!=3 && current_triangle->v1->type==3)
        new_v31=_PERUNIKA_create_point_at_x_zero 
        (current_triangle->v3,current_triangle->v1,
                 lambda_computational_grid,status);
      /* vertex 1 has x>=0.0, vertex 3 has x<0.0 */
      if (current_triangle->v3->type==3 && current_triangle->v1->type!=3)
        new_v13=_PERUNIKA_create_point_at_x_zero
        (current_triangle->v1,current_triangle->v3,
                 lambda_computational_grid,status);

      /* we know where the x=0.0 points are, split the triangle now */

      /* vertices 2 and 3 are out of the dust
 
              V3-------V2                   V2        
                \      /                   /  \ 
         new_v13 O****O new_v12           /    O new_v12
                  ****                   /    ***
                   V1                   V3---O***V1
                                           new_v13
      */
      if (new_v13!=NULL && new_v12!=NULL) {
        if (new_v13!=current_triangle->v1 && new_v12!=current_triangle->v1){
          current_triangle->v2=new_v12;
          current_triangle->v3=new_v13;
	}
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }
      /* vertices 1 and 3 are out of the dust
                new_v23            
              V3---O***V2                  V2  
               \   ***                   ****
                \   O new_v21   new_v23 O****O new_v21
                 \ /                   /      \
                  V1                  V3------V1
      */
      else if (new_v21!=NULL && new_v23!=NULL) {
        if (new_v23!=current_triangle->v2 && new_v21!=current_triangle->v2){
          current_triangle->v1=new_v21;
          current_triangle->v3=new_v23;
	}
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }
      /* vertices 1 and 2 are out of the dust
                new_v32            
              V3**O----V2                V2 
               ***    /                 /  \
        new_v31 O    /         new_v32 O    \
                 \  /                 ***    \
                  V1                 V3**O----V1
                                       new_v31
      */
      else if (new_v31!=NULL && new_v32!=NULL) {
        if (new_v31!=current_triangle->v3 && new_v32!=current_triangle->v3){
          current_triangle->v1=new_v31;
          current_triangle->v2=new_v32;
	}
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }
      /* vertex 1 is out of the dust
                          
                  V3******V2                   V2
                   ********                   ****
            new_v31 O****O new_v21           *****O new_v21
                     \  /                   *****  \
                      V1                   V3**O----V1
                                             new_v31
      */
      else if (new_v31!=NULL && new_v21!=NULL) {
        if (new_v31!=current_triangle->v1 && new_v21!=current_triangle->v2
            && new_v31!=new_v21){
          new_triangle= _PERUNIKA_create_triangle(&error);
          new_triangle->v1=new_v31;
          new_triangle->v2=new_v21;
          new_triangle->v3=current_triangle->v2;
	}
        if (new_v31!=current_triangle->v3)
          current_triangle->v1=new_v31;
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }
      /* vertex 2 is out of the dust
                new_v32
              V3**O---V2                   V2
               *****  /                   /  \
                *****O new_v12   new_v32 O****O new_v12
                 ****                   ********
                  V1                   V3******V1
      */
      else if (new_v32!=NULL && new_v12!=NULL) {
        if (new_v12!=current_triangle->v2 && new_v32!=current_triangle->v3
            && new_v12!=new_v32){
          new_triangle= _PERUNIKA_create_triangle(&error);
          new_triangle->v1=new_v12;
          new_triangle->v2=new_v32;
          new_triangle->v3=current_triangle->v3;
	}
        if (new_v12!=current_triangle->v1)
          current_triangle->v2=new_v12;
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }
      /* vertex 3 is out of the dust
                new_v23
              V3--O***V2               V2
               \ ******               ****
        new_v13 O*****       new_v23 O*****
                 ****               / ******
                  V1               V3--O****V1
                                     new_v13
      */
      else if (new_v13!=NULL && new_v23!=NULL) {
        if (new_v23!=current_triangle->v3 && new_v13!=current_triangle->v1
            && new_v23!=new_v13){
          new_triangle= _PERUNIKA_create_triangle(&error);
          new_triangle->v1=new_v23;
          new_triangle->v2=new_v13;
          new_triangle->v3=current_triangle->v1;
        }
        if (new_v23!=current_triangle->v2)
          current_triangle->v3=new_v23;
        else error=_PERUNIKA_destroy_triangle (current_triangle);
      }

    } /* end of the triangle with at least one vertex in the x<0.0 region */

    current_triangle = next_triangle;
    } while (current_triangle != start_triangle_list);


  /* numerate the vertices: useful for visualization where the vertex ID
     numbers are required */
  current_vertex=start_vertex_list;
  dust_i=0;
  do {
    current_vertex->identification_number=dust_i;
    dust_i++;
    current_vertex = current_vertex->next;
  } while (current_vertex!=start_vertex_list);

  return 0;
} /* end of "_PERUNIKA_edge_triangles_removal" */



/****************************************************************************/
/****************************************************************************/

/* function which makes the search for a dust edge point from the point1
   toward the point2 into the direction 'direction' */
struct _vertex_structure *_PERUNIKA_create_point_at_dust_edge 
                                 (struct _triangle_structure *parent_triangle, 
                                  struct _vertex_structure *point1, 
                                  struct _vertex_structure *point2, 
                                  int direction,int lambda_computational_grid,
                                  int status){

  /* The points 1 and 2 belong to the triangle 'parent_triangle'.
     We search for the dust edge into the direction 'direction', toward 
     the point B:

              direction
        o---------<---------o
     point2               point1

     On that line we find already existing vertices pointA and pointB which
     are one before and one after the still unknown edge point
     (usually they are the same as point1 and point2)

        o---<--o-----<-------?-------<-----o-------<-----o
     point2  pointB      edge_point      pointA        point1


     The dust edge can already have a point. In this case we don't need to 
     do anything, just return this point. If the edge point doesn't exist, 
     we should find it:
                  
               empty space             dust component 1 
        o--------------------------0***********************o
     pointB                    edge_point                pointA
    (type=3)                    (type=2)                (type=1)


              dust component 1       dust components 1 & 2
        o**************************0***********************o
     pointB                    edge_point                pointA
    (type=1)                    (type=2)                (type=1)


    
    There is one special situation:
    if the distance of the point B from the origin is more then Rout, then
       we know that the edge_point should have a distance Rout or less,
  */
 
  /* error flag */
  int error;
  /* temperature error from interpolation */
  double temp_error;
  /* dust component counter */
  int dust_i;
  /* the edge_point which we are looking for */
  struct _vertex_structure *edge_point;
  /*the nearby points to the edge_point (they are surrounding the edge point)*/
  struct _vertex_structure *pointA, *pointB;
  /* a flag which is used whenever we compare the dust existence at two 
     different points */
  int edge_flag;
  /* a unit vector which tells me where to go during the search */
  double direction_x, direction_z;
  /* a step by which we search for the edge point */
  double search_step;
  /* the dust temperature at the point considered one step earlier during
     the search for the dust edge */
  double *T_dust_before;
  /* a distance of the edge_point from the origin */
  double edge_distance;
  /* distance between the points pointA and pointB */
  double distance_AB;
  /* a small distance below which we can not resolve a difference between 
     the point A or B and the edge_point */
  double small_distance; 
  /* the point immediately after the edge, used for the analysis about 
     the type of the edge point: due to the dust density or due to 
     the temperature */
  double adjacent_to_edge_x, adjacent_to_edge_z, adjacent_to_edge_temperature;


  if (direction<1 || direction>6) return NULL;

  /* check first if the edge_point already exists */
  edge_point = point1;
  while (edge_point != point2) {
    /* special case: point2 has x<0 */
    if (edge_point->type == 2) break;
    switch (direction) {
    case 1 : edge_point=edge_point->v1;break;
    case 2 : edge_point=edge_point->v2;break;
    case 3 : edge_point=edge_point->v3;break;
    case 4 : edge_point=edge_point->v4;break;
    case 5 : edge_point=edge_point->v5;break;
    case 6 : edge_point=edge_point->v6;break;
    default : break;
    }
  }

  if (edge_point == point2 && edge_point->type != 2) {
    /* OK, the edge_point doesn't exist, create a new one */

    /* find the point from which we will start the search for the edge point */
    pointA = point1; edge_flag=0;
    do {
      switch (direction) {
      case 1 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v1->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      case 2 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v2->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      case 3 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v3->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      case 4 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v4->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      case 5 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v5->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      case 6 : for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
             if (pointA->v6->T_dust[dust_i]*
             pointA->T_dust[dust_i] < 0.0) edge_flag = 1;
               break;
      default : break;
      }
      if (!edge_flag){
        switch (direction) {
        case 1 : pointA=pointA->v1;break;
        case 2 : pointA=pointA->v2;break;
        case 3 : pointA=pointA->v3;break;
        case 4 : pointA=pointA->v4;break;
        case 5 : pointA=pointA->v5;break;
        case 6 : pointA=pointA->v6;break;
        default : break;
        }
      }
    }while(pointA!= point2 && (!edge_flag));
    
    /* create the direction vector and set up the pointB*/
    switch (direction) {
    case 1 : direction_x= 1.0; direction_z= 0.0      ; pointB=pointA->v1;break;
    case 2 : direction_x= 0.5; direction_z= SQRT3/2.0; pointB=pointA->v2;break;
    case 3 : direction_x=-0.5; direction_z= SQRT3/2.0; pointB=pointA->v3;break;
    case 4 : direction_x=-1.0; direction_z= 0.0      ; pointB=pointA->v4;break;
    case 5 : direction_x=-0.5; direction_z=-SQRT3/2.0; pointB=pointA->v5;break;
    case 6 : direction_x= 0.5; direction_z=-SQRT3/2.0; pointB=pointA->v6;break;
    default : break;
    }
    if (pointA==point2) {
      printf("\nWARNING!! Erroneous search for an edge point!\n");
      pointB=point2;
    }

    /* search for the dust edge as long as the step in this search is
       more than 10% of the spatial resolution at that distance */
    edge_point = _PERUNIKA_create_vertex (&error);
    edge_point->x = pointA->x;
    edge_point->z = pointA->z;
    edge_point->level = -666;

    T_dust_before = dvector(0,Max_Dust_Type-1);
  temp_error=-1.0; /* flag for not calculating T_error in _PERUNIKA_temperature */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      T_dust_before[dust_i]=_PERUNIKA_temperature(edge_point->x,
                                     edge_point->z,dust_i,status,&temp_error);

    search_step = _ZELENJAK_size_limit(pointA->x,pointA->z);
    distance_AB = _MATH_LIB_distance(pointA->x,pointA->z,pointB->x,pointB->z);
    if (distance_AB < 50.0*search_step) search_step = distance_AB/50.0;
    small_distance = 0.1*search_step; /* define the small distance */ 
small_distance = 1E-7;
    while (search_step > small_distance) {
      edge_point->x += (direction_x*search_step);
      edge_point->z += (direction_z*search_step);
      edge_flag = 0;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
  temp_error=-1.0; /* flag for not calculating T_error in _PERUNIKA_temperature */
        edge_point->T_dust[dust_i]=_PERUNIKA_temperature(edge_point->x,
                                    edge_point->z,dust_i,status,&temp_error);
        if (edge_point->T_dust[dust_i]*T_dust_before[dust_i] < 0.0)
      edge_flag = 1;
      }

      if (edge_flag) {
    /* we have passed over the edge in this step */ 
        edge_point->level = 666;
    /* step back */
        edge_point->x -= (direction_x*search_step);
        edge_point->z -= (direction_z*search_step);
        search_step /= 2.0;
      }
      else {
        if (edge_point->level < 0) {
          search_step = _ZELENJAK_size_limit(edge_point->x,edge_point->z);
          if (distance_AB < 50.0*search_step) search_step = distance_AB/50.0;
          small_distance = 0.1*search_step;
small_distance = 1E-7;
        }
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
          T_dust_before[dust_i] = edge_point->T_dust[dust_i]; 
      }
    } /* end of the search */

    free_dvector(T_dust_before,0,Max_Dust_Type-1);
    adjacent_to_edge_x = edge_point->x + (direction_x*2.0*search_step);
    adjacent_to_edge_z = edge_point->z + (direction_z*2.0*search_step);

    edge_distance = _MATH_LIB_distance(0.0,0.0,edge_point->x,edge_point->z);
    /* check if the distance from the origin is more than Rout */
    if (edge_distance > Rout) {
      edge_point->x = edge_point->x * Rout / edge_distance;
      edge_point->z = edge_point->z * Rout / edge_distance;
    }

    /* find the temperature and Eta at the edge_point */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      temp_error=1.0; /* flag for calculating T_error in _PERUNIKA_temperature */
      edge_point->T_dust[dust_i]=_PERUNIKA_temperature(edge_point->x,
                                    edge_point->z,dust_i,status,&temp_error);
      if (status>1) edge_point->T_error[dust_i]=temp_error;
    }
    edge_point->Eta=Eta(edge_point->x,edge_point->z,
                        lambda_computational_grid,edge_point->T_dust,&error);
    /* now we have to find out if this edge point is due to the sublimation
       temperature or due to the given dust density profile */
    edge_point->type = 2;
    edge_point->level = 666;

  temp_error=-1.0; /* flag for not calculating T_error in _PERUNIKA_temperature */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      adjacent_to_edge_temperature = _PERUNIKA_temperature
    (adjacent_to_edge_x,adjacent_to_edge_z,dust_i,status,&temp_error);
      if (adjacent_to_edge_temperature*edge_point->T_dust[dust_i] < 0.0) 
    if (Dust_Density(adjacent_to_edge_x,adjacent_to_edge_z,dust_i,NULL,&error)
        > 0.0) 
      if (edge_point->T_dust[dust_i]>0.0 && 
              fabs(edge_point->T_dust[dust_i]-Dust_Sublim_Temp[dust_i])/
          Dust_Sublim_Temp[dust_i] < 0.2) /*20%error*/
          edge_point->level = 1000+dust_i;
    }

    small_distance = _ZELENJAK_size_limit(pointA->x,pointA->z);
    if (distance_AB < 50.0*small_distance) small_distance=0.1*distance_AB/50.0;
    else small_distance *= 0.1;
    /* check if the pointA is the answer */ 
    if (_MATH_LIB_distance(pointA->x,pointA->z,edge_point->x,edge_point->z) <=
        small_distance) {
      pointA->x = edge_point->x;
      pointA->z = edge_point->z;
      /* update the temperature and Eta at the pointA */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        temp_error=1.0; /* flag for calculating T_error in _PERUNIKA_temperature */
        pointA->T_dust[dust_i]=_PERUNIKA_temperature(pointA->x,pointA->z,
                                                   dust_i,status,&temp_error);
    if (status>1) pointA->T_error[dust_i]=temp_error;
      }
      pointA->Eta=Eta(pointA->x,pointA->z,lambda_computational_grid,
                        pointA->T_dust,&error);
      pointA->level = edge_point->level;
      pointA->type = 2;
      error = _PERUNIKA_destroy_vertex(edge_point);
      return pointA;
    }

    small_distance = _ZELENJAK_size_limit(pointB->x,pointB->z);
    if (distance_AB < 50.0*small_distance) small_distance=0.1*distance_AB/50.0;
    else small_distance *= 0.1;
    /* check if the pointB is the answer */ 
    if ( distance_AB - small_distance <
         _MATH_LIB_distance(pointA->x,pointA->z,edge_point->x,edge_point->z)) {
      pointB->x = edge_point->x;
      pointB->z = edge_point->z;
      pointB->level = edge_point->level;
      pointB->type = 2;
      error = _PERUNIKA_destroy_vertex(edge_point);
      /* check if the distance of the pont B from the origin is more 
         than Rout */
      edge_distance = _MATH_LIB_distance(0.0,0.0,pointB->x,pointB->z);
      if (edge_distance > Rout) {
        pointB->x = pointB->x * Rout / edge_distance;
        pointB->z = pointB->z * Rout / edge_distance;
      }
      /* update the temperature and Eta at the pointB */
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
        temp_error=1.0; /* flag for calculating T_error in _PERUNIKA_temperature */
        pointB->T_dust[dust_i]=_PERUNIKA_temperature(pointB->x,pointB->z,
                                                 dust_i,status,&temp_error);
    if (status>1) pointB->T_error[dust_i]=temp_error;
      }
      pointB->Eta=Eta(pointB->x,pointB->z,lambda_computational_grid,
                      pointB->T_dust,&error);
      return pointB;
    }

    /* set the pointers to adjacent points at the edge_point */
    switch (direction) {
    case 1 : edge_point->v1=pointA->v1;
             edge_point->v4=pointA;
             pointA->v1=edge_point;
             edge_point->v1->v4=edge_point;break;
    case 2 : edge_point->v2=pointA->v2;
             edge_point->v5=pointA;
             pointA->v2=edge_point;
             edge_point->v2->v5=edge_point;break;
    case 3 : edge_point->v3=pointA->v3;
             edge_point->v6=pointA;
             pointA->v3=edge_point;
             edge_point->v3->v6=edge_point;break;
    case 4 : edge_point->v4=pointA->v4;
             edge_point->v1=pointA;
             pointA->v4=edge_point;
             edge_point->v4->v1=edge_point;break;
    case 5 : edge_point->v5=pointA->v5;
             edge_point->v2=pointA;
             pointA->v5=edge_point;
             edge_point->v5->v2=edge_point;break;
    case 6 : edge_point->v6=pointA->v6;
             edge_point->v3=pointA;
             pointA->v6=edge_point;
             edge_point->v6->v3=edge_point;break;
    default : break;
    }

  } /* end of the case where the edge_point doesn't exist */

  /* return the edge_point , either a new one or an old one */
  return edge_point;

}

/****************************************************************************/
/****************************************************************************/

struct _vertex_structure *_PERUNIKA_create_point_at_x_zero
        (struct _vertex_structure *point1, 
                 struct _vertex_structure *point2, 
                 int lambda_computational_grid,int status){

  /* the line which connects the points 1 and 2 intersects the x=0.0 line:

           point2  |
              \    |
               \   |
                \  |
                 \ |
        X < 0.0   \|     X > 0.0
                   \
                   |\
                   | point1

  We have to find the intersection! */ 

  int dust_i,error;
  double temp_error;
  struct _vertex_structure *edge_point;
  struct _vertex_structure *check_point;
  double distance12, check_distance;
  double direction_x, direction_z;

  if (point1->x < 0.0  ||   point2->x >= 0.0) return NULL;
  if (point1->x == 0.0) return point1;

  distance12= _MATH_LIB_distance(point1->x,point1->z,point2->x,point2->z);
  if (distance12==0.0) return NULL;

  direction_x = (point2->x - point1->x) / distance12;
  direction_z = (point2->z - point1->z) / distance12;

  edge_point = _PERUNIKA_create_vertex (&error);
  edge_point->x = 0.0;
  edge_point->z = point2->z - (point2->x*direction_z/direction_x);

  /* check if this vertex already exists */
  check_point = start_vertex_list;
  do {
    if (check_point->x == 0.0 && check_point!=edge_point) 
      if (fabs(check_point->z - edge_point->z) < 
           0.0001*_ZELENJAK_size_limit(edge_point->x,edge_point->z)){
    error = _PERUNIKA_destroy_vertex(edge_point);
    return check_point;
      }    
    check_point = check_point->next;
  } while (check_point != start_vertex_list);

  /* calculate all other properties at the new vertex */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    temp_error=1.0; /* flag for calculating T_error in _PERUNIKA_temperature */
    edge_point->T_dust[dust_i]=_PERUNIKA_temperature(edge_point->x,
                                  edge_point->z,dust_i,status,&temp_error);
    if (status>1) edge_point->T_error[dust_i]=temp_error;
  }
  edge_point->Eta=Eta(edge_point->x,edge_point->z,
                      lambda_computational_grid,edge_point->T_dust,&error);
  if (edge_point->Eta <=0.0) {
    error = _PERUNIKA_destroy_vertex(edge_point);
    return point1;
  }

  edge_point->level = 666;
  edge_point->type = 1;

  return edge_point;

}

/****************************************************************************/
/****************************************************************************/

/* function which moves a triangle back into the analysis */
void _PERUNIKA_move_triangle_back ( struct _triangle_structure *current_triangle,
							   struct _vertex_structure *v0) {
   /* v0 is the vertex that turned into level=668 and now we have to 
      check if there is an alrady analyzed triangle with this vertex */

   struct _triangle_structure *t, *t_next;

   t = start_triangle_list;
   while (current_triangle != t) {
	   t_next = t->prev;
	   if (t->v1 == v0 || t->v2 == v0 || t->v3 == v0){
		   if (t==start_triangle_list) start_triangle_list=t->prev;
		   t->next->prev = t->prev;
		   t->prev->next = t->next;
		   current_triangle->prev->next = t;
		   t->prev = current_triangle->prev;
		   current_triangle->prev = t;
		   t->next = current_triangle;
	   }
	   t = t_next;
   }
}
