/* LELUYA - PERUNIKA program block */
/* This block of the LELUYA code is generating the computational grid */

/* This function is interpolating the initial energy density from the
   approximated energy density solution (a point source, without diffuse dust 
   radiation) */

#include "PERUNIKA.h"

double _PERUNIKA_initial_energy_density(double x,double z,int lambda_k){
  /* x,z = the point's coordinates
     lambda_k = wavelength's index

  The interpolation is performed over the initial energy density array:
       double  initial_energy_array[radial_j][angular_n][lambda_k]
  where 
         radial_j  = 0,...,_INITIAL_GRID_ARRAY_r-1
         angular_n = 0,...,_INITIAL_GRID_ARRAY_theta-1
  */
 
  int radial_j;           /* counter over the radial grid steps */
  int angular_n;          /* counter over the angular grid steps */
  double r, theta;        /* r=sqrt(x*x+z*z), theta=arccos(z/r) */
  double r_j,r_j1;        /* radial distances of the grid points 
                             radial_j and radial_j+1  */
  double theta_n,theta_n1;/* angular position of the grid points 
                             angular_n and angular_n+1 */
  double U_jn,U_j1n,U_jn1,U_j1n1; /* energy densities at the points: 
                             (radial_j,angular_n),(radial_j+1,angular_n),
                             (radial_j,angular_n+1),(radial_j+1,angular_n+1) */
  double Ua, Ub;          /* energy densities at intermediate points:
                             Ua: between 
                              (radial_j,angular_n),(radial_j+1,angular_n)
                             Ub: between 
                              (radial_j,angular_n+1),(radial_j+1,angular_n+1)*/
  double U;               /* energy density at (x,z) */
  double r_m;             /* r_m=sqrt(r_j * r_j1)   */
  double step_j, step_n;  /* radial (logarithmic) and angular (linear) 
                             grid steps */

  /* first check if (x,z) is out of the dust region */
  r = sqrt(x*x+z*z);
  if (r>Rout) return 0.0;
  /* check if the point is inside of the r=1 cavity */
  if (r<1.0) return 0.0;

  /* check if x or z are out of the computational domain */
  if ( x<0 ) x *= -1.0;
  if ( z<0 ) z *= -1.0;

  /* check if the angle is out of the initial solution range */
  theta = acos(z/r);
  if (theta < _INITIAL_GRID_ARRAY_theta_min ) return 0.0;
  if (theta > _INITIAL_GRID_ARRAY_theta_max ) 
    theta = _INITIAL_GRID_ARRAY_theta_max;

  /* find the position of the point in the energy density array */
  radial_j = (int)((_INITIAL_GRID_ARRAY_r-1)*log(r)/log(Rout));

  if (radial_j==_INITIAL_GRID_ARRAY_r-1) radial_j=_INITIAL_GRID_ARRAY_r - 2 ;
  angular_n = (int)(
           (_INITIAL_GRID_ARRAY_theta-1)*(theta-_INITIAL_GRID_ARRAY_theta_min)
           /(_INITIAL_GRID_ARRAY_theta_max - _INITIAL_GRID_ARRAY_theta_min));
  if (angular_n == _INITIAL_GRID_ARRAY_theta-1) 
    angular_n = _INITIAL_GRID_ARRAY_theta - 2;

  /* find the grid steps */
  step_j = exp(log(Rout) / (_INITIAL_GRID_ARRAY_r-1));
  step_n = (_INITIAL_GRID_ARRAY_theta_max-_INITIAL_GRID_ARRAY_theta_min)
           / (_INITIAL_GRID_ARRAY_theta-1);

  /* find the radii */
  r_j = exp(radial_j*log(step_j));
  r_j1 = step_j*r_j;

  /* find the angles */
  theta_n = angular_n*step_n+_INITIAL_GRID_ARRAY_theta_min;
  theta_n1 = theta_n+step_n;

  /* energy densities */
  U_jn  = (double) initial_energy_array[radial_j][angular_n][lambda_k];
  U_j1n = (double) initial_energy_array[radial_j+1][angular_n][lambda_k];
  U_jn1 = (double) initial_energy_array[radial_j][angular_n+1][lambda_k];
  U_j1n1= (double) initial_energy_array[radial_j+1][angular_n+1][lambda_k];

  /* find the energy density Ua 
     between U_jn(radial_j,angular_n),U_j1n(radial_j+1,angular_n) */  
  if (U_jn > 0.0 && U_j1n > 0.0)
    Ua = exp(log(U_j1n/U_jn)*log(r/r_j)/log(r_j1/r_j) + log(U_jn));
  else if (U_jn <= 0.0 && U_j1n > 0.0) {
    r_m = r_j + r_j1 - sqrt(r_j*r_j1);
    Ua = U_j1n*(r-r_m)/(r_j1-r_m);
    if (Ua <= 0.0) Ua = -1.0;
    else Ua = U_j1n;
  }
  else if (U_jn > 0.0 && U_j1n <= 0.0) {
    r_m = sqrt(r_j*r_j1);
    Ua = U_jn*(r-r_m)/(r_j-r_m);
    if (Ua <= 0.0) Ua = -1.0;
    else Ua = U_jn;
  }
  else Ua = -1.0;

  /* find the energy density Ub 
     between U_jn1(radial_j,angular_n+1),U_j1n1(radial_j+1,angular_n+1) */
  if (U_jn1 > 0.0 && U_j1n1 > 0.0)
    Ub = exp(log(U_j1n1/U_jn1)*log(r/r_j)/log(r_j1/r_j) + log(U_jn1));
  else if (U_jn1 <= 0.0 && U_j1n1 > 0.0) {
    r_m = r_j + r_j1 - sqrt(r_j*r_j1);
    Ub = U_j1n1*(r-r_m)/(r_j1-r_m);
    if (Ub <= 0.0) Ub = -1.0;
    else Ub = U_j1n1;
  }
  else if (U_jn1 > 0.0 && U_j1n1 <= 0.0) {
    r_m = sqrt(r_j*r_j1);
    Ub = U_jn1*(r-r_m)/(r_j-r_m);
    if (Ub <= 0.0) Ub = -1.0;
    else Ub = U_jn1;
  }
  else Ub = -1.0;

  /* find the energy density at (x,z) */
  if (Ua <= 0.0 && Ub <= 0.0 ) U = 0.0;
  else if (Ua > 0.0 && Ub > 0.0) 
    U = (Ub-Ua)*(theta-theta_n)/step_n + Ua;
  else if (Ua <= 0.0 && Ub > 0.0) {
    /* U = Ub*(theta-theta_n-0.5*step_n)/(0.5*step_n); */
    if (theta > theta_n+0.5*step_n ) U = Ub;
    else U = 0.0;
  }
  else if (Ua > 0.0 && Ub <= 0.0) {
    /* U = -Ua*(theta-theta_n-0.5*step_n)/(0.5*step_n); */
    if (theta < theta_n+0.5*step_n ) U = Ua;
    else U = 0.0;
  }
  else U= 0.0;

  return U;

} /* end of the function _PERUNIKA_initial_temperature */
