/* LELUYA - PERUNIKA program block */
/* This block of the LELUYA code is generating the computational grid */

/* This function is interpolating the initial dust temperature from the
   approximated temperature */

#include "PERUNIKA.h"

double _PERUNIKA_initial_temperature (double x, double z, int dust_i){
  /* x,z = the point's coordinates
     dust_i  = dust component

  The interpolation is performed over the initial temperature arrays:
       double  initial_temperature_array[dust_i][radial_j][angular_n]
  where 
         radial_j  = 0,...,_INITIAL_GRID_ARRAY_r-1
         angular_n = 0,...,_INITIAL_GRID_ARRAY_theta-1
  */
 
  int radial_j;           /* counter over the radial grid steps */
  int angular_n;          /* counter over the angular grid steps */
  double r, theta;        /* r=sqrt(x*x+z*z), theta=arccos(z/r) */
  double r_j,r_j1;        /* radial distances of the grid points 
                             radial_j and radial_j+1  */
  double theta_n,theta_n1;/* angular position of the grid points 
                             angular_n and angular_n+1 */
  double T_jn,T_j1n,T_jn1,T_j1n1; /* temperatures at the points: 
                             (radial_j,angular_n),(radial_j+1,angular_n),
                             (radial_j,angular_n+1),(radial_j+1,angular_n+1) */
  double Ta, Tb;          /* temperatures at intermediate points:
                             Ta: between 
                              (radial_j,angular_n),(radial_j+1,angular_n)
                             Tb: between 
                              (radial_j,angular_n+1),(radial_j+1,angular_n+1)*/
  double T;               /* temperature at (x,z) */
  double r_m;             /* r_m=sqrt(r_j * r_j1)   */
  double step_n;          /* angular (linear) grid steps */
  double density;         /* dust density at the point (x,z) */
  int    err;             /* error flag */

  /* first check if (x,z) is out of the dust region */
  r = sqrt(x*x+z*z);
  if (r>Rout+0.000001) return -1.0;
  /* check if the point is inside of the r=1 cavity */
  if (r<1.0) {return -1.0;}

  /* check if x or z are out of the computational domain */
  if ( x<0 ) {x *= -1.0;}
  if ( z<0 ) {z *= -1.0;}

  /* check if the angle is out of the initial solution range */
  theta = acos(z/r);
  if (theta < _INITIAL_GRID_ARRAY_theta_min ) {return -1.0;}
  if (theta > _INITIAL_GRID_ARRAY_theta_max ) 
    theta = _INITIAL_GRID_ARRAY_theta_max;

  /* special treatment of flared disks */
  if (INPUT.DUST_DENSITY_TYPE[dust_i]==8) {
    z = 0.0;
    r = x;
    theta = _INITIAL_GRID_ARRAY_theta_max;    
  }

  /* find the position of the point in the temperature array */
  radial_j = 0;
  while (r>initial_radial_array[radial_j+1] && radial_j<_INITIAL_GRID_ARRAY_r-1) radial_j++;
  if (radial_j == _INITIAL_GRID_ARRAY_r-1) radial_j--;

  angular_n = (int)(
           (_INITIAL_GRID_ARRAY_theta-1)*(theta-_INITIAL_GRID_ARRAY_theta_min)
           /(_INITIAL_GRID_ARRAY_theta_max - _INITIAL_GRID_ARRAY_theta_min));
  if (angular_n == _INITIAL_GRID_ARRAY_theta-1) 
    angular_n = _INITIAL_GRID_ARRAY_theta - 2; 

  /* find the grid steps */
  step_n = (_INITIAL_GRID_ARRAY_theta_max-_INITIAL_GRID_ARRAY_theta_min)
           /(_INITIAL_GRID_ARRAY_theta-1);

  /* find the radii */
  r_j = initial_radial_array[radial_j];
  r_j1 = initial_radial_array[radial_j+1];

  /* find the angles */
  theta_n = angular_n*step_n+_INITIAL_GRID_ARRAY_theta_min;
  theta_n1 = theta_n+step_n;

  /* temperatures */
  T_jn  = (double) initial_temperature_array[dust_i][radial_j][angular_n];
  T_j1n = (double) initial_temperature_array[dust_i][radial_j+1][angular_n];
  T_jn1 = (double) initial_temperature_array[dust_i][radial_j][angular_n+1];
  T_j1n1= (double) initial_temperature_array[dust_i][radial_j+1][angular_n+1];

  /* find the temperature Ta 
     between T_jn(radial_j,angular_n),T_j1n(radial_j+1,angular_n) */  
  if (T_jn > 0.0 && T_j1n > 0.0)
    Ta = exp(log(T_j1n/T_jn)*log(r/r_j)/log(r_j1/r_j) + log(T_jn));
  else if (T_jn <= 0.0 && T_j1n > 0.0) {
    /*    r_m = r_j + r_j1 - sqrt(r_j*r_j1);
    Ta = T_j1n*(r-r_m)/(r_j1-r_m);
    if (Ta <= 0.0) Ta = -1.0;
    else Ta = T_j1n; */

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0 && T_j1n <= Dust_Sublim_Temp[dust_i]) Ta = T_j1n;
    else Ta = -1.0; */
    Ta = -1.0;
  }
  else if (T_jn > 0.0 && T_j1n <= 0.0) {
    /* r_m = sqrt(r_j*r_j1);
    Ta = T_jn*(r-r_m)/(r_j-r_m);
    if (Ta <= 0.0) Ta = -1.0;
    else Ta = T_jn;*/

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0 && T_jn <= Dust_Sublim_Temp[dust_i]) Ta = T_jn;
    else Ta = -1.0; */
    Ta = -1.0;
  }
  else Ta = -1.0;

  /* find the temperature Tb 
     between T_jn1(radial_j,angular_n+1),T_j1n1(radial_j+1,angular_n+1) */
  if (T_jn1 > 0.0 && T_j1n1 > 0.0)
    Tb = exp(log(T_j1n1/T_jn1)*log(r/r_j)/log(r_j1/r_j) + log(T_jn1));
  else if (T_jn1 <= 0.0 && T_j1n1 > 0.0) {
    /* r_m = r_j + r_j1 - sqrt(r_j*r_j1);
    Tb = T_j1n1*(r-r_m)/(r_j1-r_m);
    if (Tb <= 0.0) Tb = -1.0; 
    else Tb = T_j1n1;*/

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0 && T_j1n1 <= Dust_Sublim_Temp[dust_i]) Tb = T_j1n1;
    else Tb = -1.0; */
    Tb = -1.0;
  }
  else if (T_jn1 > 0.0 && T_j1n1 <= 0.0) {
    /*r_m = sqrt(r_j*r_j1);
    Tb = T_jn1*(r-r_m)/(r_j-r_m);
    if (Tb <= 0.0) Tb = -1.0;
    else Tb = T_jn1;*/

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0 && T_jn1 <= Dust_Sublim_Temp[dust_i]) Tb = T_jn1;
    else Tb = -1.0; */
    Tb = -1.0;
  }
  else Tb = -1.0;

  /* find the temperature at (x,z) */
  if (Ta <= 0.0 && Tb <= 0.0 ) T = -1.0;
  else if (Ta > 0.0 && Tb > 0.0) 
    T = (Tb-Ta)*(theta-theta_n)/step_n + Ta;
  else if (Ta <= 0.0 && Tb > 0.0) {
    /* T = Tb*(theta-theta_n-0.5*step_n)/(0.5*step_n); */
    /*if (theta > theta_n+0.5*step_n ) T = Tb;
    else T = -1.0;*/

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0) T = Tb;
    else T = -1.0; */
    T = -1.0;
  }
  else if (Ta > 0.0 && Tb <= 0.0) {
    /* T = -Ta*(theta-theta_n-0.5*step_n)/(0.5*step_n); */
    /*if (theta < theta_n+0.5*step_n ) T = Ta;
    else T = -1.0;*/

    /* density = Dust_Density(x,z,dust_i,NULL,&err);
    if (density > 0.0) T = Ta;
    else T = -1.0; */
    T = -1.0;
  }
  else T=-1.0;

  if (T<=0.0) T= -1.0;

  return T;

} /* end of the function _PERUNIKA_initial_temperature */
