/* LELUYA - PERUNIKA program block */
/* This part of the LELUYA code is generating the computational grid */

/* This function is the main function of the PERUNIKA block. It is called whenever 
   a new grid has to be calculated.  The grid can be built from the initial 
   approximations (before the iterations start),or from a previous iteration.*/

#include "PERUNIKA.h"

int _PERUNIKA_main (int status) {
  /* 
  status = 0 : calculate the grid from the initial approximation
                  (a point source, neglected diffuse radiation )
  status = 1 : read the initial approximation from a file 
  status = 2 : calculate the grid from a previous iteration
  */

  int err;              /* error flags */
  int dust_i;                         /* counter over the dust components */
  int lambda_k;                       /* counter over the lambda grid steps */
  int lambda_kk;                      /* counter over the lambda grid steps */
  short int split_flag;               /* used for calculation of the optical
                                         depth */
  struct _vertex_structure *new_vertex1,
                           *new_vertex2, 
                           *new_vertex3; /* vertices of a new triangle */
  struct _triangle_structure *new_triangle; /* new triangle (zero-level) */
  struct _quadrangle_structure *quadrangle,*next_quadrangle; /* used in 
                                          destruction of the quadrangle list */
  char imagename[256];

  int *Tau_Lambda_Sort; /* this is a list of indexes of wavelengths sorted by
                           the optical depth */
  int tau_lambda_k;     /* counter over the sorted lambda grid */

  /* variables for calculation of the finite source size flux */
  double Tau_center; 
  struct _3DVector StartPoint,EndPoint;
  double source_correction;

  /* variables for interpolation of Utot_new from old to the new grids */
  struct _vertex_structure *loop_vertex;
  double x,z;
  struct _triangle_structure *interpolation_triangle;
  double c1,c2,c3;
  int ID1,ID2,ID3;
  void _PERUNIKA_main_Interpolation_Coefficient 
       (double x1, double y1, double x2, double y2, double x3, double y3, 
        double x, double y, double *c1, double *c2, double *c3);

  /* I/O variables */
  FILE *DUMP_FILE;       /* file with dumped grids */
  char dumpfilename[256];

#ifdef _2DD_MPI_
  int tag = 0;
  MPI_Status Receive_Status, Status;
  int vertex_ID;
  int job_ID;
#endif

  start_old_vertex_list = NULL;
  start_old_triangle_list = NULL;

  /*-------------------------------------------------------------------------*/
  if (status <= 1) {
    /* if status = 1 then we need the inital temperature */
    if (status==1 || DEBUG.CONTINUE==2 || DEBUG.CONTINUE==200) {
      initial_energy_array=d3tensor(0,_INITIAL_GRID_ARRAY_r-1,
                                    0,_INITIAL_GRID_ARRAY_theta-1,
                                    0,lambda_Max-1);
      initial_temperature_array=d3tensor(0,Max_Dust_Type-1,
                                         0,_INITIAL_GRID_ARRAY_r-1,
                                         0,_INITIAL_GRID_ARRAY_theta-1);
      _PERUNIKA_initialize_initial_arrays(status);
    }
    /* create the list of X-Z grids (one X-Z grid per wavelength) */
    grid_per_lambda = grid_per_lambda_vector(0,lambda_Max-1);
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      grid_per_lambda[lambda_k].lambda = Lambda_Grid[lambda_k];
      grid_per_lambda[lambda_k].tau=_PERUNIKA_optical_depth_distance(
	   sin(Theta_Total_Value), cos(Theta_Total_Value),
           Rout*sin(Theta_Total_Value),Rout*cos(Theta_Total_Value), 
           lambda_k,status,&split_flag);
      grid_per_lambda[lambda_k].Ustar = NULL; /* stellar energy density */
      grid_per_lambda[lambda_k].Usca  = NULL; /* dust scattering energy */
      grid_per_lambda[lambda_k].Uem   = NULL; /* dust emission energy */
      grid_per_lambda[lambda_k].Utot_new = NULL;/* updated total energy */
      grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES = 0;
      grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 0;
      grid_per_lambda[lambda_k].start_vertex=NULL;
      grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES=0;
      grid_per_lambda[lambda_k].start_triangle=NULL;
    }
  }

  /* update the dust density if required */
  /* ... this works only if a new grid is calculated: */
  if (status>1)
    /* ... this works only if the flag for update is active: */
    if (DEBUG.DENSITY_UPDATE == 1) 
      /* ... this works only if the grid is azimuthal: */
      if (DEBUG.AZIMUTHAL_GRID>=0)  {
        /* find if any of the dust types is a flared disk */
        err = FALSE;
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
            if (INPUT.DUST_DENSITY_TYPE[dust_i]==8) /* Flared Disk */
              {err=TRUE; break;}
        /* ... this works only if the dust geometry is flared disk: */
        if (err) {
          if (DEBUG.INPUT_DENSITY_UPDATE < 0) {
            /* create ln(density) space */
             _GOROSTAS_set_logarithmic_density_at_pixels(0);
	  }
          /* update the flare disk's density */
          _GOROSTAS_update_density ();
	}
  }

  /* create the list of old grids */ 
  if (status>1) old_grid_per_lambda = grid_per_lambda_vector(0,lambda_Max-1);


  /* set TOTAL_NUMBER_OF_VERTICES and TOTAL_NUMBER_OF_TRIANGLES to zero at
     each wavelength and use this as an indication of non-updated grid */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    /* destroy the old energy density arrays at each wavelength */
    if (status>1) {
    TOTAL_NUMBER_OF_VERTICES = 
       grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
    /* we are going to preserve the energy densities */ 
    old_grid_per_lambda[lambda_k].Ustar=grid_per_lambda[lambda_k].Ustar;
    old_grid_per_lambda[lambda_k].Usca=grid_per_lambda[lambda_k].Usca;
    old_grid_per_lambda[lambda_k].Uem=grid_per_lambda[lambda_k].Uem;
    old_grid_per_lambda[lambda_k].Utot_new=grid_per_lambda[lambda_k].Utot_new;
    grid_per_lambda[lambda_k].Ustar = NULL;
    grid_per_lambda[lambda_k].Usca  = NULL;
    grid_per_lambda[lambda_k].Uem   = NULL; 
    grid_per_lambda[lambda_k].Utot_new = NULL; 
    /* memorize the old grids */ 
    old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES = 
      grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
    old_grid_per_lambda[lambda_k].start_vertex = 
      grid_per_lambda[lambda_k].start_vertex;
    old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES = 
      grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;
    old_grid_per_lambda[lambda_k].start_triangle = 
      grid_per_lambda[lambda_k].start_triangle;
    }
    /* mark the lists as non-updated */ 
    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES = 0;
    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES=0;
  }

  /* create the list of sorted wavelength indexes */
  Tau_Lambda_Sort = ivector(0,lambda_Max-1);
  for (tau_lambda_k=0; tau_lambda_k<lambda_Max; tau_lambda_k++)
    Tau_Lambda_Sort[tau_lambda_k]=tau_lambda_k;
  for (tau_lambda_k=0; tau_lambda_k<lambda_Max-1; tau_lambda_k++)
    for (lambda_k=tau_lambda_k+1; lambda_k<lambda_Max; lambda_k++){
      if (grid_per_lambda[Tau_Lambda_Sort[tau_lambda_k]].tau <
	  grid_per_lambda[Tau_Lambda_Sort[lambda_k]].tau){
	lambda_kk = Tau_Lambda_Sort[tau_lambda_k];
	Tau_Lambda_Sort[tau_lambda_k] = Tau_Lambda_Sort[lambda_k];
	Tau_Lambda_Sort[lambda_k] = lambda_kk;
      }
    }	

  if (status>1) printf("NEW GRIDS BUILD FROM THE OLD GRID FOR #lambda=%d\n",
                       Tau_Lambda_Sort[0]);
	  
  /* make the loop over wavelengths now, and update the grids (if status>1) or
     create the initial grids (if status<=1) */
  for (tau_lambda_k=0; tau_lambda_k<lambda_Max; tau_lambda_k++) {

    lambda_k = Tau_Lambda_Sort[tau_lambda_k];

    /* create a new grid if this wavelength needs it */ 
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES==0){

      if (DEBUG.SCREEN_MESSAGE >= 1 ) {
	printf("\nprocess#%.3d:Grid generation at lambda[%.3d]=%lf TAU=%lf\n",
               my_MPI_rank,
	       lambda_k,Lambda_Grid[lambda_k],grid_per_lambda[lambda_k].tau);
      }

    if (status>1) {
      /* initialize some variables */
/*
      start_old_vertex_list = old_grid_per_lambda[lambda_k].start_vertex;
      start_old_triangle_list = old_grid_per_lambda[lambda_k].start_triangle;
*/
      start_old_vertex_list = old_grid_per_lambda[Tau_Lambda_Sort[0]].start_vertex;
      start_old_triangle_list = old_grid_per_lambda[Tau_Lambda_Sort[0]].start_triangle;
      start_quadrangle_list = NULL;
      /* check if the sublimation surfaces need to be corrected */
      err = _PERUNIKA_sublimation_extension();
      /*sprintf(imagename,"2D_dusty.Sublimation_shells_with_quadrangles.lambda=%.3d",lambda_k);
       err = _VODENJAK_plot_dust_edge_vrml(imagename,lambda_k);
      */
    }

    /*----------------------------------------------------------------------*/
    /* initialize the global grid variables */
    start_vertex_list = NULL;
    start_triangle_list = NULL;
    TOTAL_NUMBER_OF_VERTICES = 0;
    TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 0;
    TOTAL_NUMBER_OF_TRIANGLES= 0;
    new_vertex1=new_vertex2=new_vertex3=NULL;
    new_triangle=NULL;

    if ( DEBUG.AZIMUTHAL_GRID < 0) {
      /* recursive, self-adaptive triangular grid */ 
      /**************************************************************************/

      /* create the zero-level triangle */
      if (DEBUG.SCREEN_MESSAGE >= 3 ) {
        printf("process#%.3d: PERUNIKA: Creates the zero-level triangle\n",
	       my_MPI_rank);
      }
      /* point 1 */
      new_vertex1 = _PERUNIKA_create_vertex (&err);
      new_vertex1->level=0;
      new_vertex1->type =3;  /* fake point */
      new_vertex1->x = 2*Rout/SQRT3;
      new_vertex1->z = 0.0;

      /* point 2 */
      new_vertex2 = _PERUNIKA_create_vertex (&err);
      new_vertex2->level=0;
      new_vertex2->type =3;  /* fake point */
      new_vertex2->x = Rout/(2*SQRT3);
      new_vertex2->z = 3*Rout/2;

      /* point 3 */
      new_vertex3 = _PERUNIKA_create_vertex (&err);
      new_vertex3->level=0;
      new_vertex3->type =3;  /* fake point */
      new_vertex3->x = -Rout/SQRT3;
      new_vertex3->z = 0.0;

      /* create the zero-level triangle */
      new_triangle = _PERUNIKA_create_triangle (&err);
      new_triangle->v1 = new_vertex1;
      new_triangle->v2 = new_vertex2;
      new_triangle->v3 = new_vertex3;
      new_triangle->level = 0;

      /* connect the new vertices */
      new_vertex1->v3 = new_vertex2;  /*        v2         */
      new_vertex1->v4 = new_vertex3;  /*       /  \        */
      new_vertex2->v6 = new_vertex1;  /*      /    \       */
      new_vertex2->v5 = new_vertex3;  /*     /      \      */
      new_vertex3->v2 = new_vertex2;  /*    /        \     */
      new_vertex3->v1 = new_vertex1;  /*   v3---------v1   */

      /* start with the generation of the grid by analyzing 
         the zero-level triangle */
      if (DEBUG.SCREEN_MESSAGE >= 1 ) {
        printf("\nprocess#%.3d:... making the hexagonal grid ...\n",my_MPI_rank);
      }
      err = _PERUNIKA_analyze_triangle (new_triangle,lambda_k,status);
      if (err==1) {
        printf("process#%.3d:WARNING: the grid has not been finished: too many\n",
               my_MPI_rank);
        printf("process#%.3d:grid points (more than 10000 points), lambda=%lf\n",
               my_MPI_rank,Lambda_Grid[lambda_k]);
       return 1;
      }
      else if (err>1) return 1;
      /* remove the edge triangles and shortcut the others */
      err = _PERUNIKA_edge_triangles_removal (lambda_k,status);
      /**************************************************************************/
    }
    else {
      /* azimuthal triangular grid */
      /**************************************************************************/
      if (DEBUG.SCREEN_MESSAGE >= 1 ) {
        printf("\nprocess#%.3d:... making the azimuthal grid ...\n",my_MPI_rank);
      }
      err = _PERUNIKA_create_azimuthal_grid (lambda_k,status);
      if (err==1) {
        printf("process#%.3d:WARNING: azimuthal grid generation failed\n",my_MPI_rank);
        return 1;
      }
      /**************************************************************************/
    }

    /* remove the fake vertices and sort the triangle lists at each vertex */
    err = _PERUNIKA_clean_the_vertex_list();
    if (err>0) {
  printf("process#%.3d:ERROR in _PERUNIKA_clean_the_vertex_list() at lambda=%lf\n",
	     my_MPI_rank,Lambda_Grid[lambda_k]);
      return 1;
    }

    TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 0;
    loop_vertex=start_vertex_list;
    do {
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
              loop_vertex->T_dust[dust_i]>0.0)
	    {TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES++;break;}
	loop_vertex = loop_vertex->next;
    } while (loop_vertex!=start_vertex_list); 
    if (DEBUG.SCREEN_MESSAGE >= 1 ) {
      printf("process#%.3d:... total number of the grid vertices = %d\n",
	     my_MPI_rank,TOTAL_NUMBER_OF_VERTICES);
      printf("process#%.3d:... with %d of them not empty.\n",
	     my_MPI_rank,TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES);
      printf("process#%.3d:... total number of the grid triangles = %d\n",
	     my_MPI_rank,TOTAL_NUMBER_OF_TRIANGLES);
    }

  
    /*   
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    sprintf(imagename,"2D_dusty.Eta.grid.proc=%.3d.lambda=%.3d.dust%.3d",
	      my_MPI_rank,lambda_k,dust_i);
      err = _VODENJAK_plot_vrml(imagename,0,dust_i,lambda_k);
    sprintf(imagename,"2D_dusty.Eta.surface.proc=%.3d.lambda=%.3d.dust%.3d",
	      my_MPI_rank,lambda_k,dust_i);
      err = _VODENJAK_plot_vrml(imagename,1,dust_i,lambda_k);
    }
    */ 

    /* if we work within the iteration loop, then we are updating the existing
       grids, which have to be destroyed here */
    if (status>1) { /* status=2 */
      /* clean the quadrangle list */
      quadrangle = start_quadrangle_list;
      while (quadrangle != NULL) {
	next_quadrangle = quadrangle->next;
        if (quadrangle->new_next_edge_vertex!=quadrangle->new_edge_vertex) err = TRUE;
        else err = FALSE;
	/* destroy quadrangle->new_next_edge_vertex */
	if (quadrangle->new_next_edge_vertex->identification_number<0) {
     free_dvector(quadrangle->new_next_edge_vertex->T_dust,0,Max_Dust_Type-1);
          quadrangle->new_next_edge_vertex->prev->next =
            quadrangle->new_next_edge_vertex->next;
          quadrangle->new_next_edge_vertex->next->prev = 
	    quadrangle->new_next_edge_vertex->prev;
          free (quadrangle->new_next_edge_vertex);
	}
	/* destroy quadrangle->new_edge_vertex */
        if (err)
	if (quadrangle->new_edge_vertex->identification_number<0) {
     free_dvector(quadrangle->new_edge_vertex->T_dust,0,Max_Dust_Type-1);
          quadrangle->new_edge_vertex->prev->next =
            quadrangle->new_edge_vertex->next;
          quadrangle->new_edge_vertex->next->prev = 
	    quadrangle->new_edge_vertex->prev;
          free (quadrangle->new_edge_vertex);
	}
       /* destroy quadrangle */
	free(quadrangle);
	quadrangle = next_quadrangle;
      }
      start_quadrangle_list = NULL;

      /* before we continue, we should find at which wavelengths we have used
         the old grids and then put NULL grid pointers at these wavelengths */
      for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++)
	if (grid_per_lambda[lambda_kk].start_vertex==
	    old_grid_per_lambda[lambda_k].start_vertex)
	  {
	    grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES=0;
	    grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES=0;
	    grid_per_lambda[lambda_kk].start_vertex=NULL;
	    grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_TRIANGLES=0;
	    grid_per_lambda[lambda_kk].start_triangle=NULL;
	  }
    } /* end status>1 */

    /* store the grid at wavelengths which show similar total optical depth */
grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES=TOTAL_NUMBER_OF_VERTICES;
grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES=
                                           TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
grid_per_lambda[lambda_k].start_vertex=start_vertex_list;
grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES=TOTAL_NUMBER_OF_TRIANGLES;
grid_per_lambda[lambda_k].start_triangle=start_triangle_list;
/* create the energy density arrays and set them to zero*/
grid_per_lambda[lambda_k].Ustar=dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_k].Usca =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_k].Uem  =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_k].Utot_new =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
for (err=0; err<TOTAL_NUMBER_OF_VERTICES; err++) {
  grid_per_lambda[lambda_k].Ustar[err]=0.0;
  grid_per_lambda[lambda_k].Usca[err]=0.0;
  grid_per_lambda[lambda_k].Uem[err]=0.0;
  grid_per_lambda[lambda_k].Utot_new[err]=0.0;
}
    for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) 
      if (lambda_k!=lambda_kk) {
	if (grid_per_lambda[lambda_k].tau <= Tau_Min_Grid) {
	  if (grid_per_lambda[lambda_kk].tau <= Tau_Min_Grid) {
	    /*if (grid_per_lambda[lambda_kk].start_vertex!=NULL)*/
            if (my_MPI_rank==0)
	    if (DEBUG.SCREEN_MESSAGE >= 1 ) 
	      printf("process#%.3d: copy grid to lambda[%.3d]=%lf TAU=%lf\n",
		     my_MPI_rank,lambda_kk,Lambda_Grid[lambda_kk],
		     grid_per_lambda[lambda_kk].tau);
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES=TOTAL_NUMBER_OF_VERTICES;
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES=
                                           TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
grid_per_lambda[lambda_kk].start_vertex=start_vertex_list;
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_TRIANGLES=TOTAL_NUMBER_OF_TRIANGLES;
grid_per_lambda[lambda_kk].start_triangle=start_triangle_list;
/* create the energy density arrays and set them to zero*/
grid_per_lambda[lambda_kk].Ustar=dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Usca =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Uem  =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Utot_new =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
for (err=0; err<TOTAL_NUMBER_OF_VERTICES; err++) {
  grid_per_lambda[lambda_kk].Ustar[err]=0.0;
  grid_per_lambda[lambda_kk].Usca[err]=0.0;
  grid_per_lambda[lambda_kk].Uem[err]=0.0;
  grid_per_lambda[lambda_kk].Utot_new[err]=0.0;
}
	  }
	}
	else if (grid_per_lambda[lambda_kk].tau<=grid_per_lambda[lambda_k].tau 
		 && grid_per_lambda[lambda_kk].tau >
		    grid_per_lambda[lambda_k].tau*Grid_Tolerance_Tau_Factor
		 && grid_per_lambda[lambda_kk].tau > Tau_Min_Grid) {
	  /*if (grid_per_lambda[lambda_kk].start_vertex!=NULL)*/
           if (my_MPI_rank==0)
            if (DEBUG.SCREEN_MESSAGE >= 1 ) 
	      printf("process#%.3d: copy grid to lambda[%.3d]=%lf TAU=%lf\n",
		     my_MPI_rank,lambda_kk,Lambda_Grid[lambda_kk],
		     grid_per_lambda[lambda_kk].tau);
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES=TOTAL_NUMBER_OF_VERTICES;
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES=
                                           TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
grid_per_lambda[lambda_kk].start_vertex=start_vertex_list;
grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_TRIANGLES=TOTAL_NUMBER_OF_TRIANGLES;
grid_per_lambda[lambda_kk].start_triangle=start_triangle_list;
/* create the energy density arrays and set them to zero*/
grid_per_lambda[lambda_kk].Ustar=dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Usca =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Uem  =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
grid_per_lambda[lambda_kk].Utot_new =dvector(0,TOTAL_NUMBER_OF_VERTICES-1);
for (err=0; err<TOTAL_NUMBER_OF_VERTICES; err++) {
  grid_per_lambda[lambda_kk].Ustar[err]=0.0;
  grid_per_lambda[lambda_kk].Usca[err]=0.0;
  grid_per_lambda[lambda_kk].Uem[err]=0.0;
  grid_per_lambda[lambda_kk].Utot_new[err]=0.0;
}
	}

      }/* end of the loop over lambda_kk */ 

    }/* end of: if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES==0) */

  } /* end of the loop over tau_lambda_sort */

  /* destroy the list of indexes */
  free_ivector(Tau_Lambda_Sort,0,lambda_Max-1);

  /* if we work with ln(density) then calculate these logarithms */
  if (status>1) 
    if (DEBUG.DENSITY_UPDATE>=0) 
      _GOROSTAS_set_logarithmic_density_at_pixels(1);

  /* copy Utot_new from the old grids to the new ones */
  if (status>1) {
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      loop_vertex = grid_per_lambda[lambda_k].start_vertex;
      do {
        x = loop_vertex->x; 
        z = loop_vertex->z; 
        /* find triangle in the old grid which contains this loop_vertex */
        start_triangle_list = old_grid_per_lambda[lambda_k].start_triangle; 
        err = _ZELENJAK_locate_point_in_grid(x,z,&interpolation_triangle);
	if (interpolation_triangle!=NULL) {
          /* find the interpolation coefficients */
          _PERUNIKA_main_Interpolation_Coefficient 
            (interpolation_triangle->v1->x, interpolation_triangle->v1->z,
             interpolation_triangle->v2->x, interpolation_triangle->v2->z,
             interpolation_triangle->v3->x, interpolation_triangle->v3->z,
             x, z, &c1, &c2, &c3);
	  /* interpolate */
          ID1 = interpolation_triangle->v1->identification_number;
	  ID2 = interpolation_triangle->v2->identification_number;
	  ID3 = interpolation_triangle->v3->identification_number;

        grid_per_lambda[lambda_k].Ustar[loop_vertex->identification_number]=
            c1*old_grid_per_lambda[lambda_k].Ustar[ID1]+
            c2*old_grid_per_lambda[lambda_k].Ustar[ID2]+
            c3*old_grid_per_lambda[lambda_k].Ustar[ID3];
        grid_per_lambda[lambda_k].Usca[loop_vertex->identification_number]=
            c1*old_grid_per_lambda[lambda_k].Usca[ID1]+
            c2*old_grid_per_lambda[lambda_k].Usca[ID2]+
            c3*old_grid_per_lambda[lambda_k].Usca[ID3];
        grid_per_lambda[lambda_k].Uem[loop_vertex->identification_number]=
            c1*old_grid_per_lambda[lambda_k].Uem[ID1]+
            c2*old_grid_per_lambda[lambda_k].Uem[ID2]+
            c3*old_grid_per_lambda[lambda_k].Uem[ID3];
        grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=
            c1*old_grid_per_lambda[lambda_k].Utot_new[ID1]+
            c2*old_grid_per_lambda[lambda_k].Utot_new[ID2]+
            c3*old_grid_per_lambda[lambda_k].Utot_new[ID3];
	}
	else {
          /* the old and new grid are incompatible: use a nearby point */
          start_vertex_list = old_grid_per_lambda[lambda_k].start_vertex;
          err = _ZELENJAK_locate_nearby_point_in_grid(x,z,&new_vertex1);
     grid_per_lambda[lambda_k].Ustar[loop_vertex->identification_number]=
       old_grid_per_lambda[lambda_k].Ustar[new_vertex1->identification_number];
     grid_per_lambda[lambda_k].Usca[loop_vertex->identification_number]=
       old_grid_per_lambda[lambda_k].Usca[new_vertex1->identification_number];
     grid_per_lambda[lambda_k].Uem[loop_vertex->identification_number]=
       old_grid_per_lambda[lambda_k].Uem[new_vertex1->identification_number];
     grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=
       old_grid_per_lambda[lambda_k].Utot_new[new_vertex1->identification_number];
	}
	/* next vertex in the new grid */ 
	loop_vertex = loop_vertex->next;
      } while (loop_vertex != grid_per_lambda[lambda_k].start_vertex);
      TOTAL_NUMBER_OF_VERTICES=
                        old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
      free_dvector(old_grid_per_lambda[lambda_k].Ustar,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(old_grid_per_lambda[lambda_k].Usca,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(old_grid_per_lambda[lambda_k].Uem,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
      free_dvector(old_grid_per_lambda[lambda_k].Utot_new,
		   0,TOTAL_NUMBER_OF_VERTICES-1);
    } /* end of the loop over lambda_k */ 
  }

  /* destroy the old grids */
  if (status>1) {
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      start_old_vertex_list = old_grid_per_lambda[lambda_k].start_vertex;
      start_old_triangle_list = old_grid_per_lambda[lambda_k].start_triangle;
      for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) 
	if (old_grid_per_lambda[lambda_kk].start_vertex==start_old_vertex_list)
	  {
	    old_grid_per_lambda[lambda_kk].start_vertex=NULL;
	    old_grid_per_lambda[lambda_kk].start_triangle=NULL;
	  }
     _PERUNIKA_destroy_vertex_triangle_list(start_old_vertex_list,start_old_triangle_list);
    }
    free_grid_per_lambda_vector(old_grid_per_lambda,0,lambda_Max-1);
  }


  /* if we create the grid for the first time then: 
      if (DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200) then use the energy densities
     based only on the flux form the finite stellar source, or
      if (DEBUG.CONTINUE==100 || DEBUG.CONTINUE==200)then use the Planck function
       for the given temperature at each vertex. */
    /* NOTE: if (status==1 && DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200) then 
             the energy density should reflect the given input temperatures. Thus, 
             at least one iteration with all grid points involved should be performed. */
  if (status<=1) {
    /* this energy density is imoprtant only if we want the initial temperature
       based on this energy */
    err = _ZELENJAK_Find_Source_Radius();
    if (DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200)
      printf("USE FINITE SOURCE SIZE FLUX AT ALL GRID POINTS\n");

#ifdef _2DD_MPI_
  MPI_Barrier(MPI_communicator);
#endif

    if (my_MPI_rank==0) {

    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
printf("WORKING ON LAMBDA#=%d\n",lambda_k);fflush(stdout);
      start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
      start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
      loop_vertex = start_vertex_list;
      do {

#ifdef _2DD_MPI_
        MPI_Recv(&lambda_kk,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,
		 &Receive_Status);
	job_ID = Receive_Status.MPI_SOURCE;
        if (lambda_kk>=0) {
          MPI_Recv(&vertex_ID,1,MPI_INT,job_ID,tag,MPI_communicator,&Status);
          MPI_Recv(&source_correction,1,MPI_DOUBLE,job_ID,tag,MPI_communicator,&Status);
          grid_per_lambda[lambda_kk].Ustar[vertex_ID]=source_correction;
          grid_per_lambda[lambda_kk].Utot_new[vertex_ID]=source_correction;
	}
	MPI_Send(&lambda_k,1,MPI_INT,job_ID,tag,MPI_communicator);
	MPI_Send(&(loop_vertex->identification_number),1,MPI_INT,
                 job_ID,tag,MPI_communicator);
	loop_vertex = loop_vertex->next;
      } while (loop_vertex!=start_vertex_list);    
    } /* end of lambda_k loop */ 
    /* receive the remaing data */ 
    for (ID1=1; ID1<total_MPI_process; ID1++) {
        MPI_Recv(&lambda_kk,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,
		 &Receive_Status);
	job_ID = Receive_Status.MPI_SOURCE;
        MPI_Recv(&vertex_ID,1,MPI_INT,job_ID,tag,MPI_communicator,&Status);
        MPI_Recv(&source_correction,1,MPI_DOUBLE,job_ID,tag,MPI_communicator,&Status);
        grid_per_lambda[lambda_kk].Ustar[vertex_ID]=source_correction;
        grid_per_lambda[lambda_kk].Utot_new[vertex_ID]=source_correction;
        lambda_kk=-1; vertex_ID=-1;
	MPI_Send(&lambda_kk,1,MPI_INT,job_ID,tag,MPI_communicator);
	MPI_Send(&vertex_ID,1,MPI_INT,job_ID,tag,MPI_communicator);
    }

    } /* end of (my_MPI_rank==0) */
    else { 

      /* send the inital flag to the zero process */
      lambda_kk=-1;
      MPI_Send(&lambda_kk,1,MPI_INT,0,tag,MPI_communicator);
      /* get the indices */ 
      MPI_Recv(&lambda_k,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
      MPI_Recv(&vertex_ID,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
      while (lambda_k>=0) {
	/* set the variables according to the indices */ 
        start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
        start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
        loop_vertex = start_vertex_list;
	while (loop_vertex->identification_number!=vertex_ID) 
                                     	  loop_vertex = loop_vertex->next;
#endif

	/* check if this is empty space */
	err = FALSE; 
	for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	  if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
              loop_vertex->T_dust[dust_i]>0.0)
	    {err=TRUE; c1=loop_vertex->T_dust[dust_i]; break;}

	if (err && DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200) {
          /*printf("vertex#=%d \n",loop_vertex->identification_number);fflush(stdout);*/
	  StartPoint.x = loop_vertex->x;
	  StartPoint.y = 0.0;
	  StartPoint.z = loop_vertex->z;
	  EndPoint.y = sqrt(StartPoint.x*StartPoint.x+StartPoint.z*StartPoint.z);
	  EndPoint.x = Source_Radius*StartPoint.x/EndPoint.y;
	  EndPoint.z = Source_Radius*StartPoint.z/EndPoint.y;
	  EndPoint.y = 0.0; 
	  Tau_center = _ZELENJAK_tau_over_grid(StartPoint,EndPoint,lambda_k,FALSE);
	  source_correction=Finite_Source_Flux_Correction(StartPoint.x,StartPoint.z,
	          Tau_center,lambda_k,start_vertex_list,start_triangle_list); 
	  source_correction*=Total_Source_Flux(lambda_k,StartPoint.x,StartPoint.z,&err);
#ifdef _2DD_MPI_
          MPI_Send(&lambda_k,1,MPI_INT,0,tag,MPI_communicator);
	  MPI_Send(&(loop_vertex->identification_number),1,MPI_INT,
                   0,tag,MPI_communicator);
          MPI_Send(&source_correction,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
          grid_per_lambda[lambda_k].Ustar[loop_vertex->identification_number]=
                                                                source_correction;
	  /* 
          grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=
	  _PERUNIKA_initial_energy_density(loop_vertex->x,loop_vertex->z,lambda_k); */
          grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=
                                                                source_correction;
#endif
	}
	else if (err && (DEBUG.CONTINUE==100 || DEBUG.CONTINUE==200)) {
	  c2 = c1/Dust_Sublim_Temp[dust_i];
	  c2 = c2*c2; c2 = c2*c2; 
	  /* make sure that this energy density will give correct Psi later on */
	  source_correction =
	    _MATH_LIB_Planck(c1,Lambda_Grid[lambda_k])*(loop_vertex->x*loop_vertex->x +
                             loop_vertex->z*loop_vertex->z)*Psi*c2;
#ifdef _2DD_MPI_
          MPI_Send(&lambda_k,1,MPI_INT,0,tag,MPI_communicator);
	  MPI_Send(&(loop_vertex->identification_number),1,MPI_INT,
                   0,tag,MPI_communicator);
          MPI_Send(&source_correction,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
          grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=
                                                                source_correction;
#endif
	}
	else {
#ifdef _2DD_MPI_
	  source_correction = 0.0;
          MPI_Send(&lambda_k,1,MPI_INT,0,tag,MPI_communicator);
	  MPI_Send(&(loop_vertex->identification_number),1,MPI_INT,
                   0,tag,MPI_communicator);
          MPI_Send(&source_correction,1,MPI_DOUBLE,0,tag,MPI_communicator);
#else
	  grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number]=0.0;
#endif
	}

#ifdef _2DD_MPI_
        /* get the new indices */ 
        MPI_Recv(&lambda_k,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
        MPI_Recv(&vertex_ID,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
      } /* end of (lambda_k>=0) */
#else
	loop_vertex = loop_vertex->next;
      } while (loop_vertex!=start_vertex_list);    
    }
#endif

    } /* end of (my_MPI_rank==0) or (my_MPI_rank!=0) */ 
#ifdef _2DD_MPI_
  MPI_Barrier(MPI_communicator);
#endif


    if (DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200 && my_MPI_rank==0) {
      /* if this is status=0 (grid generation for the first time), calculate
         new Psi, the initial dust temperature, and update the source radius */
      if (status==0) {
        err = _GOROSTAS_Find_Psi ();
        if (err>0) return 1;
        err = _VILOZMAJ_Find_New_Temperature();
        if (err>0) return 1;
        err = _ZELENJAK_Find_Source_Radius();
        if (err>0) return 1;
      }
      /* find tempertures below _INITIAL_GRID_ARRAY_min_temperature */
      printf("MINIMUM TEMPERATURE REQUIRED = %lf\n",_INITIAL_GRID_ARRAY_min_temperature);
      printf("DUST TEMPERATURE CAN NOT GO ABOVE 1.01*SUBLIMATION TEMPERATURE\n");
      printf("(correction in progress...)\n");
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
        err=0;
        TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES =
                    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
        loop_vertex = grid_per_lambda[lambda_k].start_vertex;
        do {
	  split_flag = FALSE;
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
                loop_vertex->T_dust[dust_i]>0.0) {
	      if (loop_vertex->T_dust[dust_i]<_INITIAL_GRID_ARRAY_min_temperature){
	        loop_vertex->T_dust[dust_i]=_INITIAL_GRID_ARRAY_min_temperature;
                err++;
              }
	      if (loop_vertex->T_dust[dust_i]>1.01*Dust_Sublim_Temp[dust_i]) {
		/* remove dust which is too hot */
                loop_vertex->T_dust[dust_i] = -1.0;
                loop_vertex->T_error[dust_i]=1E30;
		split_flag = TRUE;
	      }
	    }
	  if (split_flag) {
	    /* recognize if this is a new empty space point */ 
	    split_flag = TRUE;
            for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
		    loop_vertex->T_dust[dust_i]>0.0) {split_flag=FALSE;break;}
	    if (split_flag) TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES--;
	  }
	  loop_vertex = loop_vertex->next;
        } while (loop_vertex!=grid_per_lambda[lambda_k].start_vertex);
	if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES!=
            TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES) {
	  printf("lambda_k=%d : new number of not empty grid points = %d\n",lambda_k,
		 TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES);
          grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES =
	     TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
	}
        if (err>0) printf("lambda_k=%d #of T<%lf points=%d\n",lambda_k,
			  _INITIAL_GRID_ARRAY_min_temperature,err);
      }
      printf("(correction completed)\n");
    } /* end of: if (DEBUG.CONTINUE!=100 && DEBUG.CONTINUE!=200) */

  }

  /* if status=0 or 1 then free the memory allocated by the initial arrays */
  if (status <= 1) {
      free_d3tensor(initial_energy_array,0,_INITIAL_GRID_ARRAY_r-1,
                                       0,_INITIAL_GRID_ARRAY_theta-1,
                                       0,lambda_Max-1);
      free_d3tensor(initial_temperature_array,0,Max_Dust_Type-1,
                                       0,_INITIAL_GRID_ARRAY_r-1,
                                       0,_INITIAL_GRID_ARRAY_theta-1);
      free_dvector(initial_radial_array,0,_INITIAL_GRID_ARRAY_r-1);
  }

  /* check for deeply embedded vertices */ 
  if (INPUT.DEEPLY_EMBEDDED) _PERUNIKA_mark_optically_deep_vertices();

  /* if we create the grid for the first time then make a VRML model */
  if ((status <= 1 || DEBUG.CONTINUE == 204) && my_MPI_rank==0) {
    /* find the wavelength with the largest number of vertices */
    lambda_kk = 0;
    for (lambda_k=1; lambda_k<lambda_Max; lambda_k++){
      if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES >
	  grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES)
	lambda_kk = lambda_k;
    }

    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
 sprintf(dumpfilename,"2DD_interactive_grid_max_at_lambda=%.3d_START_%.2d",
	      lambda_kk,dust_i);
      err = _VODENJAK_plot_interactive_vrml(dumpfilename,dust_i,lambda_kk);
    }

    /* find the wavelength with the smallest number of vertices */
    lambda_kk = 0;
    for (lambda_k=1; lambda_k<lambda_Max; lambda_k++){
      if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES <
	  grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES)
	lambda_kk = lambda_k;
    }
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
 sprintf(dumpfilename,"2DD_interactive_grid_min_at_lambda=%.3d_START_%.2d",
	      lambda_kk,dust_i);
      err = _VODENJAK_plot_interactive_vrml(dumpfilename,dust_i,lambda_kk);
      err = _VODENJAK_plot_PovRay(dumpfilename,dust_i,lambda_kk);

    }
  }

  /* if we have to STOP here, dump the data and the grids */
  if (DEBUG.STOP==202) {
    if (my_MPI_rank==0) {
      _PESJANEK_Dump_Data_STOP2 ();
#ifdef _2DD_MPI_
      sprintf(dumpfilename,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_MPIOUT_PATH);
#else
      sprintf(dumpfilename,"%s/LELUYA.dump.STOP202.2dd",_LELUYA_OUTPUT_PATH);
#endif
      /* dump the grids */
      DUMP_FILE = fopen (dumpfilename,"w");
      if (DUMP_FILE==NULL) {
        printf("process#%.3d:---- WARNING!! Can not open the dump file:\n",
	       my_MPI_rank);
        printf("process#%.3d: %s\n",my_MPI_rank,dumpfilename);
      }
      else {
        _PESJANEK_Dump_Data_STOP202(DUMP_FILE);
      }
      fclose(DUMP_FILE);
    }
    printf("process#%.3d:WARNING!! The program is stopped due to the user's command line flag:\n", my_MPI_rank);
    printf("-s 202\n     location of the stop: after the grid generation.\n");
    return 1;
  }

  return 0;
} /**** end of the _PERUNIKA_main() function ***/


/* subroutine for calculation of the 3D liner interpolation coefficients */
void _PERUNIKA_main_Interpolation_Coefficient 
       (double x1, double y1, double x2, double y2, double x3, double y3, 
        double x, double y, double *c1, double *c2, double *c3){

  double denominator;
  denominator = (x2-x1)*(y3-y1) - (x3-x1)*(y2-y1);
  if (denominator == 0.0) {
    *c1=0.0; *c2=0.0; *c3=0.0; return;
  }
  *c1 = 1 + ( (x-x1)*(y2-y3)+(y-y1)*(x3-x2) )/denominator;
  *c2 = ( (x-x1)*(y3-y1)-(y-y1)*(x3-x1) )/denominator;
  *c3 = ( (y-y1)*(x2-x1)-(x-x1)*(y2-y1) )/denominator;
  return;
}
