/* LELUYA - PERUNIKA program block */

/* This function is activated by the flag -W. It looks for optically
   deeply embedded vertices and mark them (with Terror=0.0) so that
   they are not used for radiative transfer later on. */

/* sketch of the algorithm:
   - loop over grids
   | - find the smallest tau among the wavelengths sharing this grid
   |   and use this wavelength in the next steps
   | - loop over grid vertices
   | | - find the angular grid 
   | | - loop over the angular grid
   | | | - check for tau along the ray
   | | | - if tau<tau_limit then break
   | | - endloop
   | | - destroy the angular grid
   | | - if (tau > tau_limit) 
   | | | - mark this vertex as deeply embedded
   | | | - go over all wavelngths wit the same grid and mark this same
   | | |   vertex as deeply embedded
   | | - endif
   | - endloop
   | - mark all wavelengths with this same grid as "checked"
   | - go to the next "unchecked" grid
   - endloop
   - remove "checked" flag from all grids
*/

#include "PERUNIKA.h"

/************************************************************************/
/* this subroutine calculates the optical depth up to the dust surface */ 
double _PERUNIKA_mark_optically_deep_tau_to_surface (struct _3DVector startpoint, 
			       struct _3DVector direction, int lambda_k);
/************************************************************************/

int  _PERUNIKA_mark_optically_deep_vertices(void){

  int err;/* error flag */
  int dust_i, lambda_k, lambda_kk; /* various counters */
  int lambda_tau_min; /* wavelength with the smallest optical depth*/
  struct _vertex_structure   *loop_vertex; /* a grid point */
  double tau_ray; /*optical depth along a ray from loop_vertex */
  struct _3DVector direction; /* direction of tau calculation (a ray) */
  struct _3DVector startpoint; /* vector rpresentation of loop_vertex */ 
  int N_ray_total = 24; /* total number of rays around loop_vertex */
  int N_ray; /* counter over rays */
  double angle; /* angle for calculating the ray direction */
  int Is_this_empty_space; /* true/false: is a vertex in empty space or not? */
  int counter_deep_vertex; /* counter of optically deep vertices */

  printf("###############\n");
  printf("Searching for optically deep vertices\n");

  /* loop over all grids */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
   if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > 0) {

   printf("... CHECKING GRID FOR LAMBDA=%lf\n",Lambda_Grid[lambda_k]);

   /* set the counter of "deep vertices" to zero */
   counter_deep_vertex = 0;

   /* find the minimum tau among the same grids */
   lambda_tau_min = lambda_k;
   for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) 
     if (grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES > 0 &&
         grid_per_lambda[lambda_kk].start_vertex ==
                             grid_per_lambda[lambda_k].start_vertex)
       if (grid_per_lambda[lambda_kk].tau < grid_per_lambda[lambda_tau_min].tau)
	 lambda_tau_min = lambda_kk;
 
   /* set the grid variables */ 
   start_triangle_list = grid_per_lambda[lambda_tau_min].start_triangle; 
   start_vertex_list = grid_per_lambda[lambda_tau_min].start_vertex;

   /* loop over vertices */
   loop_vertex = start_vertex_list;
   do {

     /* check of the empty space vartices */
     Is_this_empty_space = TRUE;
     for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
       if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
	   loop_vertex->T_dust[dust_i]>0.0)
	 Is_this_empty_space = FALSE;

     if (!Is_this_empty_space) {
       /* this is not empty space */
       /* printf("Vertex ID=%d [%lf,%lf]\n",loop_vertex->identification_number,
	  loop_vertex->x,loop_vertex->z); */

       /* look around the vertex and find the smallest optical depth
          (actually, just find one depth smaller than INPUT.TAU_LIMIT) */
       direction.y = 0.0;
       startpoint.y = 0.0;
       for (N_ray=0; N_ray<N_ray_total; N_ray++) {
         angle = 2*PI*N_ray/N_ray_total;
         direction.x = sin(angle);
         direction.z = cos(angle);
         startpoint.x = loop_vertex->x;
         startpoint.z = loop_vertex->z;
         tau_ray = _PERUNIKA_mark_optically_deep_tau_to_surface (startpoint,
					     direction, lambda_tau_min);
         if (tau_ray < INPUT.TAU_LIMIT) break;
       }

       /* mark this vertex as "deeply embeded" if the smallest tau
	  around it is bigger than the given limit */
       if (tau_ray >= INPUT.TAU_LIMIT) {
	 /*
	 printf("Deep vertex #%d at lambda_k=%d\n",loop_vertex->identification_number,
		lambda_tau_min);
	 */
         counter_deep_vertex++;
         for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
	   loop_vertex->T_error[dust_i] = 0.0;
       }
     }

     /* next grid vertex */
     loop_vertex = loop_vertex->next;
   } while (loop_vertex != start_vertex_list);

   /* mark all wavelengths with this same grid as "checked"*/
   for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++)
     if (grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES > 0 &&
         grid_per_lambda[lambda_kk].start_vertex ==
         	 grid_per_lambda[lambda_k].start_vertex)
       grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES *= -1;

   printf(".... total number of 'deep vertices'=%d\n",counter_deep_vertex);

  } /* end of loop over all grids */

  /* remove the flags */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1;

  return 0;
}

/************************************************************************/
/************************************************************************/
/* this subroutine calculates the optical depth up to the dust surface */ 
/* it is identical to _ZELENJAK_tau_over_grid, except for small details:
    - it does not need endpoint becaue direction is given
    - it stops when all dust temperatures are negative 
      or tau>INPUT.TAU_LIMIT
    - it does not need current_point_xz because it works in the x-z plane
  
   The version of _ZELENJAK_tau_over_grid is as of Jul 10, 2003, and if 
   some improvements are introduced in _ZELENJAK_tau_over_grid, this 
   subroutine has to be updated. However, it does not require such a high
   precision as  _ZELENJAK_tau_over_grid. 
*/

double _PERUNIKA_mark_optically_deep_tau_to_surface (struct _3DVector startpoint, 
				  struct _3DVector direction, int lambda_k){
  /* calculate the optical depth from startpoint to the surface in 
     direction "direction" */
  struct _3DVector current_point; /* a point in 3D space during integration */
  struct _3DVector old_current_point; /* current_point from previous step */
  struct _3DVector next_point; /* point after current_point */
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _triangle_structure *old_current_triangle; /* from previous step */
  struct _triangle_structure *next_triangle; /* location of next_point */
  double distance_total; /* distance between startpoint and endpoint */
  double current_distance; /* distance between startpoint and current_point */
  double distance_step; /* integration spatial step */
  double distance_step_refinement; /* refinement of the integral step */
  int step_function; /* flag which shows if we are crossing a step in Eta */ 
  double tau; /* optical depth distance between startpoint and current_point */
  double tau_step; /* integration tau step */
  double Eta0, Eta1; /* Eta values at current_point and one point back */
  double *temperature; /* dust temperature at old current_point */
  double *new_temperature; /* dust temperature at current_point */
  int dust_i; /* dust component counter */
  int err; /* error flag */
  double A,B,C,distance_old,Eta_old; /* quadratic optical depth interpolation*/
  int quadratic; /* flag which shows if the quadratic interpolation is used */

  double dummy; 

  temperature = dvector(0,Max_Dust_Type-1);  
  new_temperature = dvector(0,Max_Dust_Type-1);  

  /* just make sure this is a unit vector in the x-z plane */ 
  dummy = _MATH_LIB_distance(0.0,0.0,direction.x,direction.z);
  direction.x /= dummy;
  direction.y = 0.0;
  direction.z /= dummy;
  
  /* initial values */
  distance_total=2*Rout;

  current_point = startpoint;

  err = _ZELENJAK_locate_point_in_grid(current_point.x,current_point.z,
				      &current_triangle);
  tau = 0.0;
  current_distance = 0.0;

  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
  if (current_triangle!=NULL)  
   for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
      if (current_triangle->v1->T_dust[dust_i]>0.0 && 
          current_triangle->v2->T_dust[dust_i]>0.0 && 
          current_triangle->v3->T_dust[dust_i]>0.0){
        temperature[dust_i]= _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
                           current_triangle->v1->T_dust[dust_i],
			   current_triangle->v2->x,current_triangle->v2->z,
                           current_triangle->v2->T_dust[dust_i],
			   current_triangle->v3->x,current_triangle->v3->z,
                           current_triangle->v3->T_dust[dust_i],
			   current_point.x,current_point.z);
        if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
      }
      else temperature[dust_i]=-1.0;
    }

  Eta0 = Eta(current_point.x,current_point.z,lambda_k,temperature,&err);

  /* quadratic interpolation */
  distance_old = 0.0;
  Eta_old = 0.0;
  C = 0.0;
  quadratic=0;

  /* now we can start with the integral */
  while (current_distance<distance_total){
    old_current_point=current_point;
    old_current_triangle=current_triangle; 
    distance_step_refinement = 1.0; 
    /* loop in case of crossing a dust edge (step function) */ 
    do {
    distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,direction,
                          distance_step_refinement,&next_point,&next_triangle);

    current_distance += distance_step;
    current_point = next_point;
    current_triangle = next_triangle;

    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                               new_temperature[dust_i]=-1.0;
    quadratic = 0;
    if (current_triangle!=NULL){
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
	  quadratic = 1;
          if (current_triangle->v1->T_dust[dust_i]>0.0 && 
              current_triangle->v2->T_dust[dust_i]>0.0 && 
              current_triangle->v3->T_dust[dust_i]>0.0){
            new_temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point.x,current_point.z);
	    if (new_temperature[dust_i]==0.0) new_temperature[dust_i]=-1.0;
	  }
          else new_temperature[dust_i]=-1.0;
	}
    }

    /* check the interpolation flag */
    step_function=FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (temperature[dust_i]*new_temperature[dust_i]<0.0) {
	quadratic = 0;
	step_function=TRUE;
      }

    Eta1=Eta(current_point.x,current_point.z,lambda_k,new_temperature,&err);

    /* optical depth interpolation*/
    if (quadratic == 0 || C==0.0) {
      /* linear interpolation */ 
      tau_step = (Eta0+Eta1)*distance_step*0.5;
      if (quadratic == 1 && C==0.0) C=1.0; /* this is a special case when we
   use linear interpolation only because it was also used in previous step */
      else C=0.0;
    }
    else {
      /* quadratic interpolation */
      if (Eta0==0.0 && Eta1==0.0) tau_step = 0.0;
      else {
        B = (Eta0 - Eta_old)/(current_distance - distance_step - distance_old);
        C = (Eta1 - Eta_old)/(current_distance - distance_old) - B;
        C /= distance_step;
        B = B - C*(current_distance - distance_step + distance_old);
        A = Eta_old - distance_old*(B + C*distance_old);
        tau_step = distance_step*(A + B*(current_distance - 0.5*distance_step)+
	           C*(current_distance*current_distance - 
                      current_distance*distance_step + 
                      distance_step*distance_step/3.0));
        /* failed quadrature */
        if (tau_step<0.0) tau_step = (Eta0+Eta1)*distance_step*0.5;
      }
    }

    /* this "if" detects a step function with a significant influence 
       on the optical depth */
    if (step_function) {
      if (tau_step>0.01) {
	current_distance -= distance_step;
	current_point=old_current_point;
	current_triangle=old_current_triangle;
	distance_step_refinement *= 2.0; 
      }
      else step_function=FALSE;
    }

    } while (step_function);

    tau += tau_step;

    /* check if this is the end of tau integration (dust surface reached)*/
    step_function=TRUE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (new_temperature[dust_i]>0.0) {step_function=FALSE; break;}
    if (step_function) break;

    /* stop if we are already optically deeper than the given limit */
    if (tau>=INPUT.TAU_LIMIT) break;

    /* prepare Eta for the next step */
    Eta_old = Eta0;
    Eta0 = Eta1;
    distance_old = current_distance - distance_step;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      temperature[dust_i]=new_temperature[dust_i];
  }

  free_dvector(temperature,0,Max_Dust_Type-1);
  free_dvector(new_temperature,0,Max_Dust_Type-1);

  return tau;
}
