/* LELUYA - PERUNIKA program block */
/* This part of the LELUYA code is generating the computational grid */

/* This function updates the sublimation surface and creates the grid
   quadrangles. These quadrangles are extension at the sublimation surface
   points which have the temperature below sublimation 

   There are three possible types of "quadrangles", due to the temperatures 
   higher and lower than the sublimation temperature Tsub:
   (the quadrangle's side is zero if the temperature is higher than Tsub)

      TYPE 1                   TYPE 2                     TYPE 3
   (quadrangle)            (quadrangle with one    (quadrangle with one
                            side equal to zero)     side equal to zero)
 
vertex \*************   vertex \*************       vertex \*************
 T<Tsub O************    T>Tsub O************        T<Tsub O************
       / \***dust****           |\***dust****              / \***dust****
      /   \**********           | \**********             /   \**********
     /q    \*********           |  \*********            /     \*********
    /  u    \********           | q \********           /       \********
   O    a    \*******           | u  \*******          /  quad-  \*******
    \    d    \******           | a   \******         /  rangle   \******
     \    r    \*****           | d    \*****        /             \*****
      \    a    \****           | r     \****       O---------------O*vertex
       \    n    \***           | a      O*vertex                    \** T>Tsub
        \    g    O*vertex      | n     / \** T<Tsub                  \**  
         \    l  / \* T<Tsub    | g    /   \* 
          \    e/               | l   /
           \   /                | e  /
            \ /                 |   /
             O                  |  /
                                | /
                                |/
                                O 
                                    
   The general description of a quadrangle:

                                            ****************************
   new_next_edge_vertex ----<------ next_edge_vertex********************
          |                                |****************************
          |                                |******** region where ******
          V                                ^***** the dust component ***
          |                                |*********** exists *********
          |                                |****************************
    new_edge_vertex ----->----------- edge_vertex***********************
                                   (current_vertex)*********************
                                            ****************************

   The new vertices are not created if the temperature at the edge vertices is
   higher than the sublimation temperature, (see the quadrangles TYPE 2&3):
  if (edge_vertex->T[dust_i]     >Tsub) new_edge_vertex = edge_vertex
  if (next_edge_vertex->T[dust_i]>Tsub) new_next_edge_vertex = next_edge_vertex

   The quadrangles have their pointer list: start_quadrangle_list
   This list is destroyed when the grid generation is finished (see the 
   function _PERUNIKA_main ).
*/

#include "PERUNIKA.h" 

/* function which calculates the new vertex from a given sublimation 
   edge vertex */
struct _vertex_structure *_PERUNIKA_sublimation_extension_vertex (int dust_i,
                                      struct _vertex_structure *prev_vertex,
                                      struct _vertex_structure *vertex,
				      struct _vertex_structure *next_vertex);

/****************************************************************************/
int _PERUNIKA_sublimation_extension (void) {

  int dust_i; /* a dust component */
  int create_quadrangle; /* a flag */
  struct _quadrangle_structure *quadrangle;
  struct _vertex_structure *current_vertex, *next_vertex;
  struct _vertex_structure *next_edge_vertex,*prev_edge_vertex;
  struct _vertex_structure *next_next_edge_vertex;
  struct _vertex_triangles *vertex_triangle; /* used in a search for the 
                                                adjacent triangles */
  struct _vertex_structure *vertex_in, *vertex_out;/* see the sketch below */
  struct _triangle_structure *triangle_in, *triangle_out;
  double a1,b1,c1,a2,b2,c2; 

  if (DEBUG.SCREEN_MESSAGE >= 1 ) 
    printf("process#%.3d:correction of sublimation surfaces ...\n",
           my_MPI_rank);

  /* go over each dust component separately */
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++){
    /* search over the old vertex list for the sublimation surface (vertices)*/
    current_vertex = start_old_vertex_list;
    do {
      next_vertex = current_vertex->prev;

/*printf("%.3d:vertex=%.4d (level=%d)\n",my_MPI_rank,current_vertex->identification_number,
       current_vertex->level);*/

      /* check first if we deal with a sublimation edge vertex and if 
         the temperature at that point is below the sublimation temperature */
      if (current_vertex->type==2 && current_vertex->level >=1000 &&
          current_vertex->T_dust[dust_i] > 0.0) {

	/* this point has to have an adjacent edge vertex and triangles 
           like this sketch :

              next_edge_vertex
                    O---------------O vertex_in
                    |\ triangle_in /        with vertex_in->T[dust_i] > 0
                    | \           /
                    |  \         /
                    |   \       /
            triangle_out \     /
                    |     \   /
                    |      \ /
                    O-------O current_vertex
              vertex_out
          with vertex_out->T[dust_i] <= 0

	*/
	vertex_in=vertex_out=NULL;
	triangle_in=triangle_out=NULL;
	next_edge_vertex = NULL;
	vertex_triangle = current_vertex->triangles;
	while (vertex_triangle!=NULL && next_edge_vertex==NULL) {
	  triangle_in = vertex_triangle->triangle;
	  vertex_in = _PERUNIKA_adjacent_vertex_in_triangle
	    (current_vertex,triangle_in);
	  next_edge_vertex=_PERUNIKA_adjacent_antivertex_in_triangle
	    (current_vertex,triangle_in);
	  if (next_edge_vertex->type!=2 || next_edge_vertex->level<1000 ||
	      vertex_in->T_dust[dust_i]<=0.0 ||
              next_edge_vertex->T_dust[dust_i]<=0.0) next_edge_vertex = NULL; 
	  vertex_triangle = vertex_triangle->next;
	}

	if (next_edge_vertex!=NULL)
	  if (next_edge_vertex->T_dust[dust_i]>0.0)
            triangle_out=_PERUNIKA_adjacent_triangle(current_vertex,triangle_in);
	if (triangle_out!=NULL)
	  vertex_out = _PERUNIKA_adjacent_antivertex_in_triangle
	    (current_vertex,triangle_out);

	/* now we know everything about the relevant vertices and triangles */
	/* check if need to create a quadrangle */
	create_quadrangle = 0;
	if (next_edge_vertex!=NULL){
	  if (vertex_out==NULL && vertex_in->T_dust[dust_i]>0.0)
	    create_quadrangle = 1;
          else if (vertex_out->T_dust[dust_i]<=0.0 && 
              vertex_in->T_dust[dust_i]>0.0 && (current_vertex->T_dust[dust_i]<
              (1.0-_SUBLIMATION_TOLERANCE_)*Dust_Sublim_Temp[dust_i] ||
              next_edge_vertex->T_dust[dust_i] < (1.0-_SUBLIMATION_TOLERANCE_)*
              Dust_Sublim_Temp[dust_i])) 
	  create_quadrangle = 1; 
	}
	/* create a new quadrangle */
	if (create_quadrangle) {

	  quadrangle = (struct _quadrangle_structure *) malloc 
	    (sizeof(struct _quadrangle_structure));
	  if (start_quadrangle_list==NULL) quadrangle->next=NULL;
	  else quadrangle->next=start_quadrangle_list;
	  start_quadrangle_list = quadrangle;
	  quadrangle->edge_vertex = current_vertex;
	  quadrangle->next_edge_vertex = next_edge_vertex;
          quadrangle->dust_component=dust_i;

	  /* check if we need an extension at next_edge_vertex */
	  if (next_edge_vertex->T_dust[dust_i] >
	      (1.0-_SUBLIMATION_TOLERANCE_)*Dust_Sublim_Temp[dust_i])
	    quadrangle->new_next_edge_vertex = next_edge_vertex;
	  else{
	    /* we need to find an edge vertex next to the next_edge_vertex:

                    \**************************
           next_next_edge_vertex **************
                      \************ dust ******
                       \***********************
                 next_edge_vertex *************
                      /  \*********************
                     /    \********************
                    /  current_vertex *********
                   /        \******************
                  /
       quadrangle->new_next_edge_vertex

	    */
	    next_next_edge_vertex = NULL;
	    vertex_triangle = next_edge_vertex->triangles;
	    while (vertex_triangle!=NULL && next_next_edge_vertex==NULL) {
	      next_next_edge_vertex=_PERUNIKA_adjacent_antivertex_in_triangle
		(next_edge_vertex,vertex_triangle->triangle);
	      if (next_edge_vertex->x==0.0 && 
		  (next_next_edge_vertex->x!=0.0 || 
                  next_next_edge_vertex==current_vertex ||
		   next_next_edge_vertex->T_dust[dust_i]<=0.0))
		next_next_edge_vertex = NULL;
              else if (next_edge_vertex->x!=0.0 &&
                 (next_next_edge_vertex->type!=2 || 
                  next_next_edge_vertex==current_vertex ||
                  next_next_edge_vertex->T_dust[dust_i]<=0.0)) 
		next_next_edge_vertex = NULL;
	      vertex_triangle = vertex_triangle->next;
	    }
	    if (next_next_edge_vertex!=NULL)
	      quadrangle->new_next_edge_vertex =
                _PERUNIKA_sublimation_extension_vertex
		(dust_i,current_vertex,next_edge_vertex,next_next_edge_vertex);
	    else quadrangle->new_next_edge_vertex = next_edge_vertex;
	  } /* end of extension at next_edge_vertex */

	  /* check if we need an extension at current_vertex */
	  if (current_vertex->T_dust[dust_i] >
	      (1.0-_SUBLIMATION_TOLERANCE_)*Dust_Sublim_Temp[dust_i])
	    quadrangle->new_edge_vertex = current_vertex;
	  else{
	    /* we need to find an edge vertex prior the current_vertex:

                    \**************************
              next_edge_vertex ****************
                      \************ dust ******
                       \***********************
                   current_vertex *************
                      /  \*********************
                     /    \********************
                    /  prev_edge_vertex *******
                   /        \******************
                  /
       quadrangle->new_edge_vertex

	    */
	    prev_edge_vertex = NULL;
	    vertex_triangle = current_vertex->triangles;
	    while (vertex_triangle!=NULL && prev_edge_vertex==NULL) {
	      prev_edge_vertex=_PERUNIKA_adjacent_vertex_in_triangle
		(current_vertex,vertex_triangle->triangle);
	      if (current_vertex->z==0.0 && 
		  (prev_edge_vertex->z!=0.0 || 
                   prev_edge_vertex==current_vertex ||
		   prev_edge_vertex->T_dust[dust_i]<=0.0))
		prev_edge_vertex = NULL;
              else if (current_vertex->z!=0.0 &&
                       (prev_edge_vertex->type!=2 || 
                        prev_edge_vertex==next_edge_vertex))
		/*  || prev_edge_vertex->T_dust[dust_i]<=0.0)) */ 
                prev_edge_vertex = NULL;
	      vertex_triangle = vertex_triangle->next;
	    }
	    if (prev_edge_vertex!=NULL)
	      quadrangle->new_edge_vertex = _PERUNIKA_sublimation_extension_vertex
		(dust_i,prev_edge_vertex,current_vertex,next_edge_vertex);
	    else quadrangle->new_edge_vertex = current_vertex;
	  } /* end of extension at current_vertex */


	  if (quadrangle->new_next_edge_vertex == next_edge_vertex &&
              quadrangle->new_edge_vertex == current_vertex) {
	    /* this quadrangle has a zero surface, destroy it ! */ 
	    start_quadrangle_list = quadrangle->next;
	    free(quadrangle);
            quadrangle=NULL;
	  }
	  else if (quadrangle->new_next_edge_vertex != next_edge_vertex &&
              quadrangle->new_edge_vertex != current_vertex)
	    /* check if the quadrangle is concave */
	    /* there are two possible conacave quadrangles :

        ***********************           *************************
        ***********************           *************************
        ----------O************           ------------O************
                  |\***dust****                       |\***dust**** 
                  | \**********                       | \**********     
                  |  \*********                       |  \*********   
                  |   \********                       |   \********   
                  |    \*******      quadrangle->     |    \******* 
                  |     \******  new_next_edge_vertex O     \******   
                  |      \*****                      /       \*****
           quadrangle->   \****                     /         \**** 
         new_edge_vertex   \***                    /           \***     
                  |   O-----O**                   O-------------O**        
                  |  /       \*              quadrangle->        \*         
                  | /                      new_edge_vertex        
                  |/                                        
                  O                                       
            quadrangle->                                
        new_next_edge_vertex


            detect the concave angles :   */
	    if (_MATH_LIB_nu_value(quadrangle->new_edge_vertex->x,
				   quadrangle->new_edge_vertex->z, 
				   current_vertex->x, current_vertex->z, 
				   quadrangle->new_next_edge_vertex->x,
				   quadrangle->new_next_edge_vertex->z) 
                < 0.0  ||
		_MATH_LIB_nu_value(quadrangle->new_next_edge_vertex->x,
				   quadrangle->new_next_edge_vertex->z, 
				   quadrangle->new_edge_vertex->x,
				   quadrangle->new_edge_vertex->z, 
				   next_edge_vertex->x, next_edge_vertex->z) 
                < 0.0  ){
	      /* find the intersection between the lines 
 line: a1*x + b1*z + c1 = 0 (current_vertex, quadrangle->new_edge_vertex) 
                  AND
 line: a2*x + b2*z + c2 = 0 (quadrangle->new_next_edge_vertex,next_edge_vertex)
	      */
	      if (current_vertex->x == quadrangle->new_edge_vertex->x)
		{a1=1.0; b1=0.0; c1= - current_vertex->x;}
	      else if (current_vertex->z == quadrangle->new_edge_vertex->z)
		{a1=0.0; b1=1.0; c1= - current_vertex->z;}
	      else {
		a1=1.0; 
		b1= - (current_vertex->x - quadrangle->new_edge_vertex->x) /
		  (current_vertex->z - quadrangle->new_edge_vertex->z);
		c1 = -a1*current_vertex->x - b1*current_vertex->z;
	      }
	      if (quadrangle->new_next_edge_vertex->x == next_edge_vertex->x)
		{a2=1.0; b2=0.0; c2= - next_edge_vertex->x;}
	      else if(quadrangle->new_next_edge_vertex->z==next_edge_vertex->z)
		{a2=0.0; b2=1.0; c2= - next_edge_vertex->z;}
	      else {
		a2=1.0; 
		b2= -(next_edge_vertex->x-quadrangle->new_next_edge_vertex->x)/
		  (next_edge_vertex->z - quadrangle->new_next_edge_vertex->z);
		c2 = -a2*next_edge_vertex->x - b2*next_edge_vertex->z;
	      }
	      /* find the intersection point */
	      quadrangle->new_next_edge_vertex->x=(b1*c2-b2*c1)/(a1*b2-a2*b1);
	      quadrangle->new_next_edge_vertex->z=(c1*a2-c2*a1)/(a1*b2-a2*b1);
	      /* destroy the data at quadrangle->new_edge_vertex and make it
                 the same as quadrangle->new_next_edge_vertex */
              quadrangle->new_edge_vertex->prev->next =
                                             quadrangle->new_edge_vertex->next;
	      quadrangle->new_edge_vertex->next->prev = 
                                             quadrangle->new_edge_vertex->prev;
              free_dvector(quadrangle->new_edge_vertex->T_dust,
                           0,Max_Dust_Type-1);
              free(quadrangle->new_edge_vertex);
	      quadrangle->new_edge_vertex = quadrangle->new_next_edge_vertex;
	    } /* and of the concave quadrangle correction */

	  if (quadrangle!=NULL) {
	    /* create the bounding box */
	    quadrangle->bound_box_down_left_x = 
              DMIN(current_vertex->x,quadrangle->next_edge_vertex->x);
            quadrangle->bound_box_down_left_z = 
              DMIN(current_vertex->z,quadrangle->next_edge_vertex->z);
            quadrangle->bound_box_up_right_x  = 
              DMAX(current_vertex->x,quadrangle->next_edge_vertex->x);
            quadrangle->bound_box_up_right_z  = 
              DMAX(current_vertex->z,quadrangle->next_edge_vertex->z);

	    quadrangle->bound_box_down_left_x = 
              DMIN(quadrangle->bound_box_down_left_x,
		   quadrangle->new_next_edge_vertex->x);
            quadrangle->bound_box_down_left_z = 
              DMIN(quadrangle->bound_box_down_left_z,
		   quadrangle->new_next_edge_vertex->z);
            quadrangle->bound_box_up_right_x  = 
              DMAX(quadrangle->bound_box_up_right_x ,
		   quadrangle->new_next_edge_vertex->x);
            quadrangle->bound_box_up_right_z  = 
              DMAX(quadrangle->bound_box_up_right_z ,
		   quadrangle->new_next_edge_vertex->z);

	    quadrangle->bound_box_down_left_x = 
              DMIN(quadrangle->bound_box_down_left_x,
		   quadrangle->new_edge_vertex->x);
            quadrangle->bound_box_down_left_z = 
              DMIN(quadrangle->bound_box_down_left_z,
		   quadrangle->new_edge_vertex->z);
            quadrangle->bound_box_up_right_x  = 
              DMAX(quadrangle->bound_box_up_right_x ,
		   quadrangle->new_edge_vertex->x);
            quadrangle->bound_box_up_right_z  = 
              DMAX(quadrangle->bound_box_up_right_z ,
		   quadrangle->new_edge_vertex->z);
	  }

	} /* end of a new quadrangle creation */

      } /* end of the work with a sublimation edge vertex */ 
      current_vertex = next_vertex;
    } while (current_vertex != start_old_vertex_list);
  } /* end of the dust component dust_i */

  /* now, when the whole extension (quadrangle) structure has been created,
     calculate the energy densities at the extension points */

  /* missing..... */


  if (DEBUG.SCREEN_MESSAGE >= 1 )
    printf("process#%.3d: ... end of the correction of sublimation surfaces\n",
           my_MPI_rank);

  return 0;

} /* end of _PERUNIKA_sublimation_extension (void) */


/****************************************************************************/
/****************************************************************************/

struct _vertex_structure *_PERUNIKA_sublimation_extension_vertex (int dust_i,
                                      struct _vertex_structure *prev_vertex,
                                      struct _vertex_structure *vertex,
                                      struct _vertex_structure *next_vertex){
  /* function which calculates the new vertex from a given sublimation 
   edge vertex.
   
   Geometry:                     Z
                                 |
      observer in -Y direction   |
                                 *----X

            \***************************************
        next_vertex ********************************
	      \*********** dust ******************** 
	       \************************************
	        \***********************************
              vertex--------<-------- prev_vertex --
                /
               /
              /
             /
        new_vertex
  */
  struct _vector_structure {double x,z;};

  struct _vertex_structure *new_vertex;
  /* unit vector from vertex to new_vertex */
  struct _vector_structure direction;
  /* unit vector from vertex to prev_vertex */
  struct _vector_structure direction_prev;
  /* unit vector from vertex to next_vertex */
  struct _vector_structure direction_next;
  /* distance between 'vertex' and 'new_vertex' */
  double new_vertex_distance;

  struct _vector_structure search_new_point;
  double dummy,search_step,small_distance;
  int dust_counter;
  int error ;

  direction.x = 0.0;
  direction.z = 0.0; 

  if (vertex->x==0.0 || next_vertex->x==0.0 || prev_vertex->x==0.0) {
    if (prev_vertex->x==0.0 && next_vertex->x==0.0) return vertex;
    else {
      if (prev_vertex->x!=0.0 && next_vertex->x==0.0)
	{direction.x=0.0; direction.z=-1.0;}
      else if (prev_vertex->x==0.0 && next_vertex->x!=0.0)
	{direction.x=0.0; direction.z=1.0;}
      else return vertex;
    }
  }

  if (vertex->z==0.0 || next_vertex->z==0.0 || prev_vertex->z==0.0) {
    if (next_vertex->z==0.0 && prev_vertex->z==0.0) return vertex;
    else {
      if (prev_vertex->z==0.0 && next_vertex->z!=0.0)
	{direction.x=-1.0; direction.z=0.0;}
      else if (prev_vertex->z!=0.0 && next_vertex->z==0.0)
	{direction.x=1.0; direction.z=0.0;}
      else return vertex;
    }		      
  }

  if (direction.x==0.0 && direction.z==0.0) {
#ifdef NOT_WORKING

    /* THERE IS A PROBLEM WITH THIS SYBLIMATION EXTENSION. WHENEVER THE SURFACE
       IS BUMPY, THIS DIRECTION WILL OFTEN CRASH LELUYA. THUS I REPLACED IT 
       WITH SIMPLE RADIAL EXTENSION */

    /* find a vector from vertex to prev_vertex */
    direction_prev.x = prev_vertex->x - vertex->x;
    direction_prev.z = prev_vertex->z - vertex->z;
    dummy = sqrt ( direction_prev.x*direction_prev.x +
                   direction_prev.z*direction_prev.z);
    direction_prev.x /= dummy;
    direction_prev.z /= dummy;

    /* find a vector from vertex to next_vertex */
    direction_next.x = next_vertex->x - vertex->x;
    direction_next.z = next_vertex->z - vertex->z;
    dummy = sqrt ( direction_next.x*direction_next.x +
                   direction_next.z*direction_next.z);
    direction_next.x /= dummy;
    direction_next.z /= dummy;

    /* find a vector from vertex to new_vertex */
    /* vector(dummy) = vector(direction_next) X vector(direction_prev) */ 
    dummy=direction_prev.x*direction_next.z-direction_prev.z*direction_next.x;
    if (dummy==0.0) {
      /* vector(direction) = vector(direction_next) X unit_vector(Y) */ 
      direction.x = -direction_next.z;
      direction.z =  direction_next.x;
    }
    else {
      /* if (dummy>0) 
           vector(direction) = -vector(direction_next) - vector(direction_prev)
       if (dummy<0) 
           vector(direction) =  vector(direction_next) + vector(direction_prev)
      */
      direction.x = -direction_next.x - direction_prev.x;
      direction.z = -direction_next.z - direction_prev.z;
      if (dummy < 0.0) {direction.x *=-1.0;direction.z *=-1.0;} 
      dummy = sqrt ( direction.x*direction.x + direction.z*direction.z );
      direction.x /= dummy;
      direction.z /= dummy;
    }
#endif
    /* radial extension */ 
    direction.x =  - vertex->x;
    direction.z =  - vertex->z;
    dummy = sqrt ( direction.x*direction.x + direction.z*direction.z );
    direction.x /= dummy;
    direction.z /= dummy;
  }
  


  /* create a new_vertex */
  new_vertex = (struct _vertex_structure *) malloc 
               (sizeof(struct _vertex_structure));
  start_old_vertex_list->prev->next = new_vertex;
  new_vertex->prev = start_old_vertex_list->prev;
  start_old_vertex_list->prev = new_vertex;
  new_vertex->next = start_old_vertex_list;

  new_vertex->identification_number = -1;
  new_vertex->level = 1000+dust_i;
  new_vertex->type = 2;
  new_vertex->v1=new_vertex->v2=new_vertex->v3=new_vertex->v4=new_vertex->v5
                =new_vertex->v6=NULL;
  new_vertex->T_dust = dvector(0,Max_Dust_Type-1);
  for (dust_counter=0; dust_counter<Max_Dust_Type; dust_counter++) 
      new_vertex->T_dust[dust_counter] = -1.0;
  new_vertex->Eta = 0.0;
  new_vertex->triangles = NULL;


  /* interpolated distance between vertex and new_vertex */
  /* Temperature = const / sqrt( radius ) */
  dummy = vertex->T_dust[dust_i]/Dust_Sublim_Temp[dust_i];
  dummy = dummy*dummy*dummy*dummy;;
  new_vertex_distance = _MATH_LIB_distance(vertex->x,vertex->z,0.0,0.0) *
                  (1.0-dummy);
  new_vertex->x = vertex->x + (direction.x*new_vertex_distance);
  new_vertex->z = vertex->z + (direction.z*new_vertex_distance);
  new_vertex->T_dust[dust_i] = Dust_Sublim_Temp[dust_i]; 

  /* check if the line vertex-new_vertex is passing over a dust edge or 
     x=0 or z=0 lines */ 
  search_step = _ZELENJAK_size_limit(vertex->x,vertex->z);
  if (new_vertex_distance < 50.0*search_step) 
    search_step = new_vertex_distance/50.0;
  search_new_point.x = vertex->x;
  search_new_point.z = vertex->z;

  small_distance = 0.1*search_step; /* define the small distance */ 
  while (search_step > small_distance && _MATH_LIB_distance(vertex->x,
	 vertex->z,search_new_point.x,search_new_point.z) < 
         new_vertex_distance) {
    search_new_point.x += (direction.x*search_step);
    search_new_point.z += (direction.z*search_step);
    if (Dust_Density(search_new_point.x,search_new_point.z,dust_i,NULL,&error)
        <=0.0 || search_new_point.x<0.0 || search_new_point.z<0.0) {
      /* we have passed over the dust edge or X=0 or Z=0 lines in this step */ 
      /* step back */
      if (search_new_point.x<0.0 || search_new_point.z<0.0) {
        if (search_new_point.x<0.0) {
	  search_new_point.x = 0.0;
          search_new_point.z = vertex->z - vertex->x*direction.z/direction.x;
        }
        if (search_new_point.z<0.0) {
	  search_new_point.z = 0.0;
	  search_new_point.x = vertex->x - vertex->z*direction.x/direction.z;
        }
	if (Dust_Density(search_new_point.x,search_new_point.z,dust_i,NULL,&error)
            >0.0) break;
      }
      search_new_point.x -= (direction.x*search_step);
      search_new_point.z -= (direction.z*search_step);
      search_step /= 2.0;
    }
    else {
      search_step = _ZELENJAK_size_limit(search_new_point.x,search_new_point.z);
      if (new_vertex_distance < 50.0*search_step) 
	search_step = new_vertex_distance/50.0;
      small_distance = 0.1*search_step;
    }
  } /* end of checking for a dust edge */

  /* check if the new_vertex is the same as vertex */  
  small_distance = _ZELENJAK_size_limit(vertex->x,vertex->z);
  if (new_vertex_distance < 50.0*small_distance) 
    small_distance = 0.1*new_vertex_distance/50.0;
  else small_distance *= 0.1;
  dummy = _MATH_LIB_distance(search_new_point.x,search_new_point.z,
                             vertex->x,vertex->z);
  if (dummy < small_distance) {
    new_vertex->prev->next = new_vertex->next;
    new_vertex->next->prev = new_vertex->prev;
    free_dvector(new_vertex->T_dust,0,Max_Dust_Type-1);
    free(new_vertex);
    return vertex;
  }

  /* check if we need to correct the data in new_vertex */
  if (dummy < new_vertex_distance) {
    new_vertex->x = search_new_point.x;
    new_vertex->z = search_new_point.z;
  }

  return new_vertex;

}

