/* LELUYA - PERUNIKA program block */

/* This function refines the grid close to the dust density surface. */

/* It works on a surface quadrangle, where "vertex_in1" and "vertex_in2" 
   are inside the dust cloud, while "vertex_edge1" and "vertex_edge2" are 
   two vertices on the surface of this cloud:
   
                      vertex_in1
       vertex_edge1 O-----O-----O vertex_tip
                     \     \   /
                      \     \ /
                       \     O vertex_in2
                        \   /
                         \ / 
                          O vetex_edge2

  Additional vertices are created between vertex_edge1 and vertex_in1 (we call
  this "direction_1") and between vertex_edge2 and vertex_in2 (this is
  called "direction_2"). 

  Once we have these new vertices, they have to be connected into a triangular 
  grid. For that purpose, we need the intersection between direction_1 and 
  direction_2, called vertex_tip. We also need the direction from vertex_in1
  to vertex_in2, called direction_3. If direction_3 is zero then 
  vertex_in1 = vertex_in2 = vertex_tip, this is just a triangle. 
*/

#include "PERUNIKA.h"
struct _vertex_structure *_PERUNIKA_surface_grid_refinement_direction
                                        (struct _vertex_structure *v, int d);
int _PERUNIKA_surface_grid_refinement_remove_vertex
                                        (struct _vertex_structure *v, int d);




int _PERUNIKA_surface_grid_refinement (struct _vertex_structure   *vertex_edge1,
                                   struct _vertex_structure   *vertex_edge2,
                                   struct _vertex_structure   *vertex_in1,
                                   struct _vertex_structure   *vertex_in2,
                                   int  direction_1, int direction_2,
                                   int  direction_3,
                                   struct _vertex_structure   *vertex_tip,
                                   int lambda_computational_grid, int status){

  /* We will search for new vertices while integrating over the optical
     depth. A new point is set whenever the optical depth, measured from the
     dust surface, is between tau_prev*INPUT.TAU_SURFACE_SCALE*(1-epsilon_tau) and
     tau_prev*INPUT.TAU_SURFACE_SCALE*(1+epsilon_tau), where tau_prev is 
     the optical depth at the previously encountered vertex, 
     INPUT.TAU_SURFACE_SCALE is given with the flag -S (usually between
     2.0 and 3.0), and epsilon_tau is the allowed error. */
  double tau, tau_prev, epsilon_tau;
  double delta_tau; /* step in integration */
  double tau_new;   /* = tau_prev*INPUT.TAU_SURFACE_SCALE */
  double tau_zero;  /* the first tau_new (because it can not be zero) */
  double L,x,z,delta_L; /* spatial dimensions during integration */ 
  struct _3DVector integral_dir; /* direction of the optical depth integration */
  struct _3DVector surface_dir; /* direction of the dust surface */
  double eta_0, eta_1; /* Eta values during the integration */ 
  double *temperature,temp_error; /* needed for integration */

  int dust_i;  /* dust type */
  int i; /* just a counter, baby  */
  int direction; /* direction under consideration */
  struct _vertex_structure *vertex_start, *vertex_end; /* bounding vertices
                                                      along one direction */
  double L_tot; /* = || vertex_end - vertex_start || */
  
  struct _vertex_structure *v_0, *v_1; /* vertices during integration */
  double L_10; /* = || v_1 - v_0 || */
  struct _vertex_structure *vn, *vp, *vn1, *vp1; /* loop vertices used in 
                            making the triangular grid */
  struct _vertex_structure *vn_end, *vp_end; /* end vertices for tirangulation */
  int direction_n, direction_p; /* directions for grid generation */ 
  struct _triangle_structure *new_triangle; 
  double Rn, Rp; /* cosine of angle between a line and the dust surface */
  double k_base,k_np1,k_pn1; 
  double cos_angle_n, cos_angle_p;

  int level; /* this is used whenever we need to check the level of a vertex */
  /* a triangle used in the loop over the triangle list */
  struct _triangle_structure *current_triangle;

  int forcing_point_flag;
  int error;

/*  printf("EDGE:\n");
  printf("[%lf,%lf][%lf,%lf][%lf,%lf][%lf,%lf]\n",vertex_edge1->x,vertex_edge1->z,
   vertex_edge2->x,vertex_edge2->z,vertex_in1->x,vertex_in1->z,
         vertex_in2->x, vertex_in2->z);*/


  /* set some "global" parameters */
  tau_zero = 0.1;
  epsilon_tau = 0.1;
  integral_dir.x = integral_dir.y = integral_dir.z = 0.0;
  surface_dir.x = surface_dir.y = surface_dir.z = 0.0;

  /* prepare for the inegration */
  temperature = dvector(0,Max_Dust_Type-1);  
  temp_error=-1.0; /* flag for not calculating T_error in _PERUNIKA_temperature */

  /* find the surface direction vector */
  if (vertex_edge1!=vertex_edge2) {
    surface_dir.x = vertex_edge2->x - vertex_edge1->x;
    surface_dir.z = vertex_edge2->z - vertex_edge1->z;
  } else {
    /* in this case, we have to find an adjacent grid point */
    current_triangle = start_triangle_list;
    do {
      if (current_triangle->v1==vertex_edge1 || current_triangle->v2==vertex_edge1 || 
          current_triangle->v3==vertex_edge1) {
        v_1 = NULL;
        if (current_triangle->v1!=vertex_edge1 && current_triangle->v1->type==2)
               v_1=current_triangle->v1;
        else  if (current_triangle->v2!=vertex_edge1 && current_triangle->v2->type==2)
               v_1=current_triangle->v2;
        else  if (current_triangle->v3!=vertex_edge1 && current_triangle->v3->type==2)
               v_1=current_triangle->v3;
        if (v_1!=NULL ){
           surface_dir.x = v_1->x - vertex_edge1->x;
           surface_dir.z = v_1->z - vertex_edge1->z;
           break;
        }
      }
      current_triangle = current_triangle->prev;
    } while (current_triangle != start_triangle_list);
  }
  k_base = sqrt(surface_dir.x*surface_dir.x+surface_dir.z*surface_dir.z);
  if (k_base!=0.0) {
    surface_dir.x /= k_base;
    surface_dir.z /= k_base;
  }
  else {surface_dir.x=surface_dir.z=0.0;}

  /* create a loop with i = direction_1 or i = direction_2 */
  for (i=1; i<=2; i++) {

    /* set the intergation parameters */
    if (i==1) {
      direction = direction_1; 
      vertex_start = vertex_edge1;
      vertex_end = vertex_in1;
    }
    else {
      direction = direction_2; 
      vertex_start = vertex_edge2;
      vertex_end = vertex_in2;
      surface_dir.x *= -1;
      surface_dir.z *= -1;
    }

    /* check if we work with a weird case when the input quadrangle is actually 
       a triangle */
    if (vertex_start == vertex_end) continue; 

    /* check if any vertex between vertex_start and vertex_end has level=667 */
    if (direction>0) {
      v_0 = _PERUNIKA_surface_grid_refinement_direction(vertex_start,direction);
      level =666;
      while (v_0 != vertex_end) {
        if (v_0->level == 667) {level=667; break;} 
        v_0 = _PERUNIKA_surface_grid_refinement_direction(v_0,direction);
      }
      /* go to the next direction if this one already has level=667 vertices */
      if (level==667) continue;
    }
    
    /* start with the optical depth integral */
    /* First, prepare some variables needed in the integral */ 
    v_0 = vertex_start;
    v_1 = _PERUNIKA_surface_grid_refinement_direction(v_0,direction);
    if (v_1==NULL) v_1=vertex_end;
    L_10 = sqrt((v_0->x-v_1->x)*(v_0->x-v_1->x) + (v_0->z-v_1->z)*(v_0->z-v_1->z));
    if (v_1==vertex_end) L_tot = L_10;
    else L_tot = sqrt((v_0->x-vertex_end->x)*(v_0->x-vertex_end->x) + 
                      (v_0->z-vertex_end->z)*(v_0->z-vertex_end->z));
    L =0.0; tau=0.0; tau_prev=0.0; tau_new=tau_zero;
    x = v_0->x;  z = v_0->z;
    integral_dir.x = (vertex_end->x - x)/L_tot; /* direction of integration */ 
    integral_dir.z = (vertex_end->z - z)/L_tot;

    /* find the cosine of the angle between the integral direction and the normal
       to the dust surface */
    k_base = surface_dir.x*integral_dir.x+surface_dir.z*integral_dir.z;
    k_base = sqrt (1.0 - k_base*k_base);
    if (k_base==0.0) k_base = 1.0;

    /* find the initial Eta value */ 
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        temperature[dust_i]=_PERUNIKA_temperature(x,z,dust_i,status,&temp_error);
      else temperature[dust_i]=-1.0;
    }
    eta_0 = Eta(x,z,lambda_computational_grid,temperature,&error);
    delta_L = L_10 / 40.0; /* initial guess */
    forcing_point_flag = TRUE; /* force one point close to the surface */

    /* now start with the loop over tau (integral) */
    do {

      /* find new delta_L and delta_tau */ 
      delta_L *= 2.0;
      do {
        x += (delta_L*integral_dir.x); 
        z += (delta_L*integral_dir.z);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
            temperature[dust_i]=_PERUNIKA_temperature(x,z,dust_i,status,&temp_error);
          else temperature[dust_i]=-1.0;
        }
        eta_1 = Eta(x,z,lambda_computational_grid,temperature,&error);
        delta_tau = k_base * (eta_0+eta_1)*delta_L/2.0;
        x -= (delta_L*integral_dir.x);
        z -= (delta_L*integral_dir.z);
        if (tau+delta_tau > tau_new) delta_L*=0.5; /* refine the step */ 
      } while (tau+delta_tau > tau_new);

      /* first check if we crossed over the next vertex on the line */ 
      if (L+delta_L > L_10 - 1E-15) {

        /* set the step variables to be exactly up to the vertex v_1 */
        delta_L = L_10-L;
        L=L_10;  x += (delta_L*integral_dir.x); z += (delta_L*integral_dir.z);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
            temperature[dust_i]=_PERUNIKA_temperature(x,z,dust_i,status,&temp_error);
          else temperature[dust_i]=-1.0;
        }
        eta_1 = Eta(x,z,lambda_computational_grid,temperature,&error);
        delta_tau = k_base * (eta_0+eta_1)*delta_L/2.0;
        tau += delta_tau; 
        /* let's check if we created a new vertex too close to this exiting one */
        if (v_0->level == 667)
          if (tau < (1+epsilon_tau)*tau_prev) {
            /* remove v_0 vertex */
            _PERUNIKA_surface_grid_refinement_remove_vertex(v_0,direction);
          }
        /* check if the integral is over */
        if (L >= L_tot) break;
        /* to avoid problems with grid generation later on in other subroutines, 
           do not allow new grid points beyond any non-667 vertex */
        if (v_1->level!=667) {
			/* set all these grid points to level=667 */
            while (v_1 != vertex_end) {
			  if (v_1!=NULL) v_1->level = 667;
			  v_1 = _PERUNIKA_surface_grid_refinement_direction(v_1,direction);
			}
			break;
		}
        tau_prev = tau; tau_new = INPUT.TAU_SURFACE_SCALE*tau;
        v_0 = v_1; 
        v_1 = _PERUNIKA_surface_grid_refinement_direction(v_0,direction);
        if (v_1==NULL) v_1 = vertex_end;
        L_10 = sqrt((vertex_start->x-v_1->x)*(vertex_start->x-v_1->x) + 
                    (vertex_start->z-v_1->z)*(vertex_start->z-v_1->z));
	delta_L = L_10 / 40.0; /* initial guess */
      }
      else {
        /* so, we have not crossed over any vertex */
        L += delta_L; tau += delta_tau;  
        x += (delta_L*integral_dir.x); z += (delta_L*integral_dir.z);
        /* check if we need to create a vertex (factor 7 is arbitrary) */
	if (L>L_10/7.0 && forcing_point_flag) tau_new=tau;
        if (tau > tau_new*(1.0-epsilon_tau) && 
            tau < tau_new*(1.0+epsilon_tau) && L != L_10) {

          /* create a new vertex and give it level = 667 */
          vn = _PERUNIKA_create_vertex (&error);
          vn->x = x; vn->z = z;
          switch (direction) {
            case 1: vn->v1=v_1; vn->v4=v_0; v_0->v1=vn; v_1->v4=vn; break;
            case 2: vn->v2=v_1; vn->v5=v_0; v_0->v2=vn; v_1->v5=vn; break;
            case 3: vn->v3=v_1; vn->v6=v_0; v_0->v3=vn; v_1->v6=vn; break;
            case 4: vn->v4=v_1; vn->v1=v_0; v_0->v4=vn; v_1->v1=vn; break;
            case 5: vn->v5=v_1; vn->v2=v_0; v_0->v5=vn; v_1->v2=vn; break;
            case 6: vn->v6=v_1; vn->v3=v_0; v_0->v6=vn; v_1->v3=vn; break;
            default: break;
          }
          vn->level = 667;
          vn->type = 1;
          v_0 = vn;
          /* find the temperature and Eta for this vertex */
          for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
            if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
              temp_error = 1.0;
              temperature[dust_i]=_PERUNIKA_temperature(x,z,dust_i,status,&temp_error);
              vn->T_error[dust_i] = temp_error;
            }
            else temperature[dust_i]=-1.0;
            vn->T_dust[dust_i] = temperature[dust_i];
          } 
          temp_error = -1.0;
          vn->Eta = Eta(x,z,lambda_computational_grid,temperature,&error);
          /* update the loop variables */ 
          tau_prev = tau; 
          /* 0.8 is arbitrary chosen */ 
          if (forcing_point_flag && tau<tau_zero*0.8) tau_new = tau_zero;
          else tau_new = INPUT.TAU_SURFACE_SCALE*tau;
          forcing_point_flag=FALSE;
        }
      }

      eta_0 = eta_1; /* this is for the next step */ 
    } while ( L < L_tot); /* end of the integral */

    /********************************************************************************/
    /****** forcing new vertices ****************************************************
    if (vertex_end==_PERUNIKA_surface_grid_refinement_direction(vertex_start,direction) &&
        vertex_end->level != 667) {

      error = FALSE;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 && vertex_end->T_dust[dust_i]>0.0) 
          {error=TRUE;break;}

      if (error) {
        v_0 = vertex_start; v_1 = vertex_end;
        vn = _PERUNIKA_create_vertex (&error);
        vn->x = v_0->x*0.5 + v_1->x*0.5; 
        vn->z = v_0->z*0.5 + v_1->z*0.5;
        switch (direction) {
            case 1: vn->v1=v_1; vn->v4=v_0; v_0->v1=vn; v_1->v4=vn; break;
            case 2: vn->v2=v_1; vn->v5=v_0; v_0->v2=vn; v_1->v5=vn; break;
            case 3: vn->v3=v_1; vn->v6=v_0; v_0->v3=vn; v_1->v6=vn; break;
            case 4: vn->v4=v_1; vn->v1=v_0; v_0->v4=vn; v_1->v1=vn; break;
            case 5: vn->v5=v_1; vn->v2=v_0; v_0->v5=vn; v_1->v2=vn; break;
            case 6: vn->v6=v_1; vn->v3=v_0; v_0->v6=vn; v_1->v3=vn; break;
            default: break;
        }
        vn->level = 667;
        vn->type = 1;
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
            temp_error = 1.0;
            temperature[dust_i]=_PERUNIKA_temperature(x,z,dust_i,status,&temp_error);
            vn->T_error[dust_i] = temp_error;
          }
          else temperature[dust_i]=-1.0;
          vn->T_dust[dust_i] = temperature[dust_i];
        } 
        temp_error = -1.0;
        vn->Eta = Eta(x,z,lambda_computational_grid,temperature,&error);
      }

    }
    ********************************************************************************/

  } /* end of the "direction" loop */


  /* set the vertices to be unmovable later on */
  if (vertex_edge1->level != 668) vertex_edge1->level = 667;
  if (vertex_edge2->level != 668) vertex_edge2->level = 667;
  if (vertex_in1->level != 668) vertex_in1->level=667;
  if (vertex_in2->level != 668) vertex_in2->level=667;

  /* we need to create a triangular grid now */

  /* prepare the grid variables*/
  v_0 = vertex_tip;
  if (direction_3 == 0) {
    /* this means: vertex_in1=vertex_in2=vertex_tip */
    vn_end = vertex_edge1; 
    vp_end = vertex_edge2;
    if (direction_1 < 0) direction_n=-1; 
    else direction_n = (direction_1 + 2)%6 + 1;
    if (direction_2 < 0) direction_p=-1; 
    else direction_p = (direction_2 + 2)%6 + 1;    
    vn = _PERUNIKA_surface_grid_refinement_direction(v_0,direction_n);
    if (vn==NULL) vn = vn_end;
    vp = _PERUNIKA_surface_grid_refinement_direction(v_0,direction_p);
    if (vp==NULL) vp = vp_end;
  }
  else {
    vn_end = vertex_in2;
    vp_end = vertex_in1;
    direction_n = direction_2;
    direction_p = direction_1;
    vn = vertex_edge2;
    vp = vertex_edge1; 
  }

  vn1 = vn; vp1 =vp;

  if (vn != vn_end) {
    vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
    if (vn1==NULL) vn1 = vn_end;
  }
  if (vp != vp_end) {
    vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
    if (vp1==NULL) vp1 = vp_end;
  }

  /* create a triangle (v_0,vn,vp) if vertex_in1=vertex_in2=vertex_tip */
  if (direction_3 == 0)
    if (v_0!=vn && v_0!=vp && vn!=vp) {
      new_triangle = _PERUNIKA_create_triangle (&error);
      new_triangle->v1 = v_0;
      new_triangle->v2 = vn;
      new_triangle->v3 = vp;
  }
  /* create a triangle (v_0, vn1,vp1) if vertex_edge1==vertex_edge2=vertex_tip */
  if (vertex_edge1==vertex_edge2) {
    new_triangle = _PERUNIKA_create_triangle (&error);
    new_triangle->v1 = v_0;
    new_triangle->v2 = vn1;
    new_triangle->v3 = vp1;
    if (vn1 == vn_end && vp1 == vp_end) {
      if (vertex_in2 == 
               _PERUNIKA_surface_grid_refinement_direction(vertex_in1,direction_3)) {
        /* this was a single triangle */
        free_dvector(temperature,0,Max_Dust_Type-1);  
        return 0;
      }
      else {
        /* there are vertices between vertex_in1 and vertex_in2 */
        /* Without them this would be just a single triangle. */ 
        error=_PERUNIKA_destroy_triangle(new_triangle);
	new_triangle = NULL; 
      }
    }
    vn = vn1; vp = vp1;
    if (vn1 != vn_end) {
      vn1 = _PERUNIKA_surface_grid_refinement_direction(vn1,direction_n);
      if (vn1==NULL) vn1 = vn_end;
    }
    if (vp1 != vp_end) {
      vp1 = _PERUNIKA_surface_grid_refinement_direction(vp1,direction_p);
      if (vp1==NULL) vp1 = vp_end;
    }
  }

  /* find the slope of the dust edge */
  k_base = 0.0;
  if (vertex_edge1!=vertex_edge2) {
    if (vertex_edge1->x == vertex_edge2->x) k_base = 1.0E6;
    else k_base = (vertex_edge2->z-vertex_edge1->z)/(vertex_edge2->x-vertex_edge1->x);
  } else {
    /* in this case, we have to find an adjacent grid point */
    current_triangle = start_triangle_list;
    do {
      if (current_triangle->v1==vertex_edge1 || current_triangle->v2==vertex_edge1 || 
          current_triangle->v3==vertex_edge1) {
        v_1 = NULL;
        if (current_triangle->v1!=vertex_edge1 && current_triangle->v1->type==2)
               v_1=current_triangle->v1;
        else  if (current_triangle->v2!=vertex_edge1 && current_triangle->v2->type==2) 
               v_1=current_triangle->v2;
        else  if (current_triangle->v3!=vertex_edge1 && current_triangle->v3->type==2) 
               v_1=current_triangle->v3;
        if (v_1!=NULL ){
           if (vertex_edge1->x == v_1->x) k_base = 1.0E6;
           else k_base = (v_1->z-vertex_edge1->z)/(v_1->x-vertex_edge1->x);
           break;
        }
      }
      current_triangle = current_triangle->prev;
    } while (current_triangle != start_triangle_list);
  }

  /* start creating the grid */
  while ((direction_3==0 && (vn != vn_end || vp != vp_end))  ||
         (direction_3!=0 && vn!=vn_end && vp!=vp_end) ) {
    /* check if the n-side reached the end */
    if (vn == vn_end) {
      v_0 = vp; vp = vp1;
      if (vp != vp_end) {
        vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
        if (vp1==NULL) vp1 = vp_end;
      }
    }
    /* check if the p-side reached the end */
    else if (vp == vp_end) {
      v_0 = vn; vn = vn1;
      if (vn != vn_end) {
        vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
        if (vn1==NULL) vn1 = vn_end;
      }
    }
    /* decide about what triangle to make */
    else {

      /* find the slope of the line vn to vp1 */ 
      if (vn->x == vp1->x) k_np1 = 1.0E6;
      else k_np1 = (vn->z-vp1->z)/(vn->x-vp1->x);
      /* find the slope of the line vp to vn1 */ 
      if (vp->x == vn1->x) k_pn1 = 1.0E6;
      else k_pn1 = (vp->z-vn1->z)/(vp->x-vn1->x);
      /* find the cosine of the nagle between these lines and the baseline */
      Rn = fabs((1.0+k_base*k_np1)/sqrt((1.0+k_base*k_base)*(1.0+k_np1*k_np1)));
      Rp = fabs((1.0+k_base*k_pn1)/sqrt((1.0+k_base*k_base)*(1.0+k_pn1*k_pn1)));

      /* find the angles at vn and vp */
      cos_angle_n = _MATH_LIB_nu_value (vn->x,vn->z,vn1->x,vn1->z,vp->x,vp->z) - 1;
      cos_angle_p = _MATH_LIB_nu_value (vp->x,vp->z,vn->x,vn->z,vp1->x,vp1->z) - 1;

      if (cos_angle_n*cos_angle_p<-100.0) { /* THIS IS TURNED OFF (SEE BELOW ALSO) */
        /* choose the line */ 
        if (cos_angle_n>0) {
          v_0 = vn; vn = vn1; 
          if (vn != vn_end) {
            vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
            if (vn1==NULL) vn1 = vn_end;
          }
        }
        else {
          v_0 = vp; vp = vp1;
          if (vp != vp_end) {
            vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
            if (vp1==NULL) vp1 = vp_end;
          }
        }
      }/* end of choosing based on cos_angle_n and cos_angle_p */
      else {
        /* choose the line */ 
        if (Rn<Rp) {
          v_0 = vn; vn = vn1; 
          if (vn != vn_end) {
            vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
            if (vn1==NULL) vn1 = vn_end;
          }
        }
        else {
          v_0 = vp; vp = vp1;
          if (vp != vp_end) {
            vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
            if (vp1==NULL) vp1 = vp_end;
          }
        }
      } /* end of choosing based on Rn and Rp */
    }

    /* create a triangle */ 
    if (v_0!=vn && v_0!=vp && vn!=vp) {
      new_triangle = _PERUNIKA_create_triangle (&error);
      new_triangle->v1 = v_0;
      new_triangle->v2 = vn;
      new_triangle->v3 = vp;

/*    printf("Triangle[%lf,%lf][%lf,%lf][%lf,%lf]\n", 
           new_triangle->v1->x,  new_triangle->v1->z,
           new_triangle->v2->x,  new_triangle->v2->z,
           new_triangle->v3->x,  new_triangle->v3->z);*/
    }

  }

  /* check if we need to consider vertices along direction_3 */
  if (direction_3 != 0) {
    /* there are only two options: vn==vn_end OR vp==vp_end */
    if (vn==vn_end) {
      v_0 = vertex_in1;
      if (direction_1<0) direction_n=-1;
      else direction_n = (direction_1+2)%6+1;
      direction_p = direction_3;
      vn_end = vp;
      /* check for a very special case when vertex_edge1==vertex_edge2=vertex_tip
         and without vertices between vertex_tip and vertex_in1, and without vertices
         between vertex_tip and vertex_in2 */
      if (vp==vp_end) vn_end = vertex_tip;
      vp_end = vertex_in2;
    }
    else {
      v_0 = vertex_in2;
      if (direction_3<0) direction_n=-1;
      else direction_n = (direction_3+2)%6+1;
      if (direction_2<0) direction_p=-1;
      else direction_p = (direction_2+2)%6+1;
      vn_end = vertex_in1;
      vp_end = vn;
    }
    vn =  _PERUNIKA_surface_grid_refinement_direction(v_0,direction_n);
    if (vn==NULL) vn = vn_end;
    vp =  _PERUNIKA_surface_grid_refinement_direction(v_0,direction_p);
    if (vp==NULL) vp = vp_end;

    /* now create triangle (v0,vn,vp) */
    if (v_0!=vn && v_0!=vp && vn!=vp) {
      new_triangle = _PERUNIKA_create_triangle (&error);
      new_triangle->v1 = v_0;
      new_triangle->v2 = vn;
      new_triangle->v3 = vp;
    }
    /* if there is more vertices for the grid, connect them: */
    vp1 = vp; vn1 = vn;
    if (vn != vn_end) vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
    if (vp != vp_end) vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
     
    while (vn!=vn_end || vp!=vp_end) {

      /* check if the n-side reached the end */
      if (vn == vn_end) {
        v_0 = vp; vp = vp1;
        if (vp != vp_end) {
          vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
          if (vp1==NULL) vp1 = vp_end;
        }
      }
      /* check if the p-side reached the end */
      else if (vp == vp_end) {
        v_0 = vn; vn = vn1;
        if (vn != vn_end) {
          vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
          if (vn1==NULL) vn1 = vn_end;
        }
      }
      /* decide about what triangle to make */
      else {

        /* find the slope of the line vn to vp1 */
        if (vn->x == vp1->x) k_np1 = 1.0E6;
        else k_np1 = (vn->z-vp1->z)/(vn->x-vp1->x);
        /* find the slope of the line vp to vn1 */
        if (vp->x == vn1->x) k_pn1 = 1.0E6;
        else k_pn1 = (vp->z-vn1->z)/(vp->x-vn1->x);
        /* find the cosine of the nagle between these lines and the baseline */
        Rn = fabs((1.0+k_base*k_np1)/sqrt((1.0+k_base*k_base)*(1.0+k_np1*k_np1)));
        Rp = fabs((1.0+k_base*k_pn1)/sqrt((1.0+k_base*k_base)*(1.0+k_pn1*k_pn1)));

        /* find the angles at vn and vp */
        cos_angle_n = _MATH_LIB_nu_value (vn->x,vn->z,vn1->x,vn1->z,vp->x,vp->z) - 1;
        cos_angle_p = _MATH_LIB_nu_value (vp->x,vp->z,vn->x,vn->z,vp1->x,vp1->z) - 1;

        if (cos_angle_n*cos_angle_p<-100.0) {/* THIS IS TURNED OFF (SEE ABOVE ALSO) */
          /* choose the line */ 
          if (cos_angle_n>0) {
            v_0 = vn; vn = vn1; 
            if (vn != vn_end) {
              vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
              if (vn1==NULL) vn1 = vn_end;
            }
          }
          else {
            v_0 = vp; vp = vp1;
            if (vp != vp_end) {
              vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
              if (vp1==NULL) vp1 = vp_end;
            }
          }
        }/* end of choosing based on cos_angle_n and cos_angle_p */
        else {
          /* choose the line */
          if (Rn<Rp) {
            v_0 = vn; vn = vn1;
            if (vn != vn_end) {
              vn1 = _PERUNIKA_surface_grid_refinement_direction(vn,direction_n);
              if (vn1==NULL) vn1 = vn_end;
            }
          }
          else {
            v_0 = vp; vp = vp1;
            if (vp != vp_end) {
              vp1 = _PERUNIKA_surface_grid_refinement_direction(vp,direction_p);
              if (vp1 == NULL) vp1=vp_end;
            }
          }
        } /* end of choosing based on Rn and Rp */
      }
      /* create a triangle */
      if (v_0!=vn && v_0!=vp && vn!=vp) {
        new_triangle = _PERUNIKA_create_triangle (&error);
        new_triangle->v1 = v_0;
        new_triangle->v2 = vn;
        new_triangle->v3 = vp;
      }
    } /* end of connecting vertices along direction_3 */ 

  } /* end of if (direction_3!=0) */ 


  free_dvector(temperature,0,Max_Dust_Type-1);  
  return 0;

}

/******************************************************************************/
struct _vertex_structure *_PERUNIKA_surface_grid_refinement_direction
                                          (struct _vertex_structure *v, int d){

  switch (d) {
    case 1: return (v->v1);
    case 2: return (v->v2);
    case 3: return (v->v3);
    case 4: return (v->v4);
    case 5: return (v->v5);
    case 6: return (v->v6);
    default: return (NULL);
  }
  
  return (NULL);
}

/******************************************************************************/
int _PERUNIKA_surface_grid_refinement_remove_vertex
                                         (struct _vertex_structure *v, int d){

  int error;

  switch (d) {
    case 1: v->v4->v1=v->v1;
            v->v1->v4=v->v4;
            break;
    case 2: v->v5->v2=v->v2;
            v->v2->v5=v->v5;
            break;
    case 3: v->v6->v3=v->v3;
            v->v3->v6=v->v6;
            break;
    case 4: v->v1->v4=v->v4;
            v->v4->v1=v->v1;
            break;
    case 5: v->v2->v5=v->v5;
            v->v5->v2=v->v2;
            break;
    case 6: v->v3->v6=v->v6;
            v->v6->v3=v->v3;
            break;
    default: break;
  }

  error = _PERUNIKA_destroy_vertex(v);

  return error;
}
