/* LELUYA - VILOZMAJ program block */
/* This function calculates new dust temperatures at each grid vertex from
   a given energy density at that point. The calculation is performed on a
   grid with the most of vertices*/

/* The algorithm works in the following way:

   - create the energy density array (from 0 to lambda_Max-1)
   - find a grid with the largest number of vetices (call it grid_N_max)
   - loop over vertices in grid_N_max
   | - copy the energy densities at this vertex to the energy density array
   |   at all wavelengths which share this grid
   | - mark these wavelengths as "used"
   | - loop over lambda
   | | - if (lambda has not been "used")
   | | | - find triangle in grid[lambda] which contains this vertex point
   | | | - find the interpolation coefficients
   | | | - interpolate the energy densities at this point to the energy 
   | | |   density array at all wavelengths which share this grid[lambda]
   | | |   and mark these wavelengths as "used"
   | | end
   | end
   | - remove the "used" marker from all wavelengths
   | - copy the temperatures at the vertex into "old_temperatures"
   | - find new temperatures from the energy density array
   | - compare the old and new temperatures and calculate the error 
   end
   - mark the wavelengths which share grid_N_max as "used"
   - loop over lambda
   | - if (lambda has not been "used")
   | | - loop over vertices in grid[lambda]
   | | | - find triangle in grid_N_max which contains this vertex
   | | | - interpolate temperatures at the vertex from the triangle
   | | | - compare the old and new temperatures and calculate the error 
   | | end
   | | - mark all wavelengths which share this grid[lambda] as "used"
   | end
   end
   - remove the "used" marker from all wavelengths
   - destroy the energy density array (from 0 to lambda_Max-1)

*/


#include "VILOZMAJ.h"
void _VILOZMAJ_Find_New_Temperature_Interpolation_Coefficient 
       (double x1, double y1, double x2, double y2, double x3, double y3, 
        double x, double y, double *c1, double *c2, double *c3);


/***************************************************************************/
int _VILOZMAJ_Find_New_Temperature( void ){

  int err, err_location;/* error flags */
  int dust_i, lambda_k, lambda_kk; /* various counters */
  int lambda_N_max; /* wavelength with the most grid points */
  int N_max;     /* number of vertices at lambda_N_max */
  struct _vertex_structure   *loop_vertex; /* a grid point */
  double x, z;        /* coordinates of loop_vertex */
  struct _triangle_structure *interpolation_triangle;
  struct _vertex_structure   *interpolation_vertex;
  double *U;          /* total energy density array */
  double c1,c2,c3;    /* interpolation coefficients */
  double *old_temperature;
  FILE *DUMP_FILE; /* dumping the spectra at the grid vertices */
  char filename[255];
  double Ustar,Usca,Uem,Utotal; /* energy densities used for dumping the
                                   spectra into a file */

  /* flag in search for temperatures at verticies at non-compatible grids */
  int flag_T0;
  struct _vertex_structure   *nearby_vertex; /* vertex to copy temperature from */

  /* create the energy density array */
  U = dvector(0, lambda_Max-1);
  old_temperature = dvector(0, Max_Dust_Type-1);
  
  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
printf("\nprocess=%.3d: _VILOZMAJ_Find_New_Temperature subroutine:\n",my_MPI_rank);
printf("calculation of temperatures from the updated energy densities\n\n");
  }

  /* find a grid with the largest number of points (let's call it grid_N_max)*/
  lambda_N_max = 0;
  N_max = grid_per_lambda[lambda_N_max].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) {
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > N_max) {
      N_max = grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
      lambda_N_max = lambda_k;
    }
  }

  if (DEBUG.SCREEN_MESSAGE >= 2 ) {
    printf("A wavelength with the most number of grid points: %E\n",
           Lambda_Grid[lambda_N_max]);
  }

  /* loop over vertices in grid_N_max */
  loop_vertex = grid_per_lambda[lambda_N_max].start_vertex;
  do {

    /* is this vertex just empty space? */
    flag_T0 = FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
	if (loop_vertex->T_dust[dust_i]>0.0) {flag_T0 = TRUE;break;}

    if (flag_T0){

      flag_T0 = TRUE;
      /* check if this vretex belongs to the current temperature bin */ 
      if (TOTAL_NUMBER_OF_GRID_BINS>1 && Bin_Type==1) {
        flag_T0=FALSE;
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
	      loop_vertex->T_dust[dust_i] >
                              Temperature_Bins[GLOBAL_BIN_COUNTER+1] &&
              loop_vertex->T_dust[dust_i] <=
                              Temperature_Bins[GLOBAL_BIN_COUNTER])
          {flag_T0=TRUE; break;}
      }
      /* check if this vretex belongs to the current radial bin */ 
      if (TOTAL_NUMBER_OF_GRID_BINS>1 && Bin_Type==2) {
        flag_T0=FALSE;
        x = sqrt(loop_vertex->x*loop_vertex->x+loop_vertex->z*loop_vertex->z);
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
	      loop_vertex->T_dust[dust_i] > 0.0 &&
              x  > Radial_Bins[GLOBAL_BIN_COUNTER] &&
              x <= Radial_Bins[GLOBAL_BIN_COUNTER+1])
          {flag_T0=TRUE; break;}
      }
    } /* end of if (not empty space ) */

    /* check if this point was skipped because of its small error */
    if (flag_T0) {
      err = FALSE;
      if (INPUT.SKIP_POINT!=0)
        if (INPUT.SKIP_POINT==2 ||
	    (INPUT.SKIP_POINT==1 && GLOBAL_ITERATION_COUNTER>0)) {
	  err = TRUE;
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	    if (fabs(loop_vertex->T_error[dust_i]) > T_error_tolerance &&
	        INPUT.DUST_DENSITY_TYPE[dust_i]!=99999 &&
	        loop_vertex->T_dust[dust_i]>0.0)
	      err = FALSE;
        }
      if (err) flag_T0=FALSE;
    }

    if (flag_T0) {
    /* this vertex belong to the current temperature bin and thus require 
         temperature update */

    if (DEBUG.SCREEN_MESSAGE >= 2 ) 
      printf("FIND TEMPERATURE AT VERTEX #%d\n",loop_vertex->identification_number);

    /* copy the temperatures at this vertex into "old_temperatures" */
    for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
      old_temperature[dust_i] = loop_vertex->T_dust[dust_i];

    /* specify the point and set the energy density array to zero */
    x = loop_vertex->x; 
    z = loop_vertex->z; 
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) U[lambda_k]=0.0;

    /* copy the energy densities at this vertex to the energy density array
       at all wavelengths which share this grid */
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
      if (grid_per_lambda[lambda_k].start_vertex ==
	  grid_per_lambda[lambda_N_max].start_vertex) {
	U[lambda_k] = 
	grid_per_lambda[lambda_k].Utot_new[loop_vertex->identification_number];
	/* mark this wavelength as "used" */
	grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1.0; 
      }
    }

    /* interpolation for the wavelengths which do not have grid_N_max */
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
     if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > 0){ 

       /* find triangle in grid[lambda] which contains this vertex point */
       start_triangle_list = grid_per_lambda[lambda_k].start_triangle; 
       start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
       err = _ZELENJAK_locate_point_in_grid(x,z,&interpolation_triangle);
/* 
       if (interpolation_triangle==NULL)
         err = _ZELENJAK_locate_nearby_point_in_grid(x,z,&interpolation_triangle);
       if (interpolation_triangle==NULL){
          printf("process#%.3d: TERRIBLE ERROR: 
_VILOZMAJ_Find_New_Temperature can not locate a triangle for:",my_MPI_rank);
          printf("process#%.3d: vertex=%.4d, lambda=%E\n",
my_MPI_rank,loop_vertex->identification_number,Lambda_Grid[lambda_k]);
       }
*/
       if (interpolation_triangle!=NULL) {

       /* find the interpolation coefficients */
       _VILOZMAJ_Find_New_Temperature_Interpolation_Coefficient 
         (interpolation_triangle->v1->x, interpolation_triangle->v1->z,
	  interpolation_triangle->v2->x, interpolation_triangle->v2->z,
	  interpolation_triangle->v3->x, interpolation_triangle->v3->z,
          x, z, &c1, &c2, &c3);

       /* interpolate the energy densities at (x,z) at all wavelengths which 
          share this grid[lambda_k] and mark these wavelengths as "used" */
       for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) {
	 if (grid_per_lambda[lambda_k].start_vertex ==
	     grid_per_lambda[lambda_kk].start_vertex) {
	   U[lambda_kk] = c1*grid_per_lambda[lambda_kk].Utot_new[
                            interpolation_triangle->v1->identification_number]
	                + c2*grid_per_lambda[lambda_kk].Utot_new[
                            interpolation_triangle->v2->identification_number]
	                + c3*grid_per_lambda[lambda_kk].Utot_new[
			    interpolation_triangle->v3->identification_number];
	   /* mark this wavelength as "used" */
	   grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES *= -1.0;
	 }
       } /* end of  wavelengths which share this grid[lambda_k] */ 

       } /* end of interpolation_triangle!=NULL */
       else {
         /* set the temperature to zero and take care of it later */
         for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
           if (loop_vertex->T_dust[dust_i]>0.0 && 
                INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) 
             loop_vertex->T_dust[dust_i]=0.0;
         break;
       }

    } /* end of interpolation */

    /* remove the "used" marker from all wavelengths */  
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
     if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES<0) 
       grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1.0;

    /* find new temperatures from the energy density array */
    for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
      if (loop_vertex->T_dust[dust_i]>0.0 && 
         INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {

       loop_vertex->T_dust[dust_i] = _GOROSTAS_Find_Temperature (x,z,
               0.5*loop_vertex->T_dust[dust_i],2.0*loop_vertex->T_dust[dust_i],
               dust_i,U,&err);

       if (err>0) {
         err_location=10;
         _PESJANEK__Error_Message (err_location,err);
         printf("Temperatures calculated on the grid at wavelength=%E\n",
	        Lambda_Grid[lambda_N_max]);
         printf("(iteration step %d, dust component #%d)\n",
                GLOBAL_ITERATION_COUNTER,dust_i);
         printf("vertex ID=%d (x=%E z=%E)\n",loop_vertex->identification_number,
	        x,z);
         return 1;
       }

       /* compare the old and new temperatures and calculate the error */
       if ((!INPUT.DEEPLY_EMBEDDED) || loop_vertex->T_error[dust_i]!=0.0) {
         loop_vertex->T_error[dust_i] = (loop_vertex->T_dust[dust_i] -
			   old_temperature[dust_i]) / old_temperature[dust_i];
         if (INPUT.TEMPERATURE_ACCELERATE_min_err>=0.0) 
           if (fabs(loop_vertex->T_error[dust_i])>=
               INPUT.TEMPERATURE_ACCELERATE_min_err) {
             loop_vertex->T_dust[dust_i] = old_temperature[dust_i] * ( 1 +
             INPUT.TEMPERATURE_ACCELERATE_scale * loop_vertex->T_error[dust_i]);
             loop_vertex->T_error[dust_i] *= INPUT.TEMPERATURE_ACCELERATE_scale;
	   }
       }
       if (DEBUG.SCREEN_MESSAGE > 1 )
	 printf(".... [dust=%d] old temperature=%lf, new=%lf\n",dust_i,
		old_temperature[dust_i],loop_vertex->T_dust[dust_i]);
    }

    }/* end of the vertex within the current temperature bin - if it is not
      then the temperature is not updated */ 

    /* end of the loop over the grid vertices */
    loop_vertex = loop_vertex->next;
  } while (loop_vertex != grid_per_lambda[lambda_N_max].start_vertex);

  /* look for points which are not shared among the grids */
  flag_T0 = 1; /* the flag which shows if there is zero (error) temperatures 
                  in the grid */
  while (flag_T0) {
    flag_T0 = 0;
    loop_vertex = grid_per_lambda[lambda_N_max].start_vertex;
    do {
      for (dust_i=0; dust_i < Max_Dust_Type; dust_i++) 
        if (loop_vertex->T_dust[dust_i]==0.0 && 
            INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
	  if (DEBUG.SCREEN_MESSAGE > 1) printf("Look for temperature from adjacent triangles\n");
          nearby_vertex = _ZELENJAK_Find_Adjacent_Temperature (loop_vertex,dust_i);
          if (nearby_vertex==NULL) {
            flag_T0 = 1;
	  }
          else {
            loop_vertex->T_dust[dust_i]=nearby_vertex->T_dust[dust_i];
            loop_vertex->T_error[dust_i]=nearby_vertex->T_error[dust_i];
          }
        }
      loop_vertex = loop_vertex->next;
    } while (loop_vertex != grid_per_lambda[lambda_N_max].start_vertex);
  }


  /* Save the energy densities at each vertex into a file */
  if (DEBUG.SCREEN_MESSAGE > 2 ) {
    printf("   ... dumping the spectra at grid vertices\n");
    sprintf(filename,
	    "%s/2D_dusty.Spectrum_at_vertices.iter=.process=%.2d.dat",
	    _LELUYA_OUTPUT_PATH,my_MPI_rank);
    DUMP_FILE=fopen(filename,"w");
    loop_vertex = grid_per_lambda[lambda_N_max].start_vertex;
    do {
      x = loop_vertex->x;
      z = loop_vertex->z;
      fprintf(DUMP_FILE,"vertex# %.4d x= %E z= %E Nlambda= %.3d\n",
	      loop_vertex->identification_number,x,z,lambda_Max);
      fprintf(DUMP_FILE,"lambda  Ustar(old) Usca(old) Uem(old) ");
      fprintf(DUMP_FILE,"Utotal(old) Utotal(new) ");
      fprintf(DUMP_FILE,"interpolation_triangle: ID, vertices\n");
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
	fprintf(DUMP_FILE,"%E ",Lambda_Grid[lambda_k]);
	/* find triangle in grid[lambda] which contains this vertex point */
	start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
	err = _ZELENJAK_locate_point_in_grid(x,z,&interpolation_triangle);
        if (interpolation_triangle==NULL){
          start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
          err = _ZELENJAK_locate_nearby_point_in_grid(x,z,&interpolation_vertex);
          Ustar = grid_per_lambda[lambda_k].Ustar[interpolation_vertex->identification_number];
          Usca = grid_per_lambda[lambda_k].Usca[interpolation_vertex->identification_number];
          Uem = grid_per_lambda[lambda_k].Uem[interpolation_vertex->identification_number];
          Utotal = grid_per_lambda[lambda_k].Utot_new[interpolation_vertex->identification_number];
          fprintf(DUMP_FILE,"%E %E %E %E %E ",Ustar,Usca,Uem,(Ustar+Usca+Uem),Utotal);
          fprintf(DUMP_FILE,"%.4d -1 -1 -1\n",interpolation_vertex->identification_number);
/*
printf("process#%.3d: TERRIBLE ERROR: _VILOZMAJ_Find_New_Temperature can not locate a triangle for:",my_MPI_rank);
printf("process#%.3d: vertex=%.4d, lambda=%E\n",my_MPI_rank,loop_vertex->identification_number,Lambda_Grid[lambda_k]);
*/
         }
        else {

	/* find the interpolation coefficients */
	_VILOZMAJ_Find_New_Temperature_Interpolation_Coefficient 
	  (interpolation_triangle->v1->x, interpolation_triangle->v1->z,
	   interpolation_triangle->v2->x, interpolation_triangle->v2->z,
	   interpolation_triangle->v3->x, interpolation_triangle->v3->z,
	   x, z, &c1, &c2, &c3);

	/* interpolate the energy densities at this vertex */
	Ustar =           c1*grid_per_lambda[lambda_k].Ustar[
                            interpolation_triangle->v1->identification_number]
	                + c2*grid_per_lambda[lambda_k].Ustar[
                            interpolation_triangle->v2->identification_number]
	                + c3*grid_per_lambda[lambda_k].Ustar[
			    interpolation_triangle->v3->identification_number];
	fprintf(DUMP_FILE,"%E ",Ustar);
	Usca  =           c1*grid_per_lambda[lambda_k].Usca[
                            interpolation_triangle->v1->identification_number]
	                + c2*grid_per_lambda[lambda_k].Usca[
                            interpolation_triangle->v2->identification_number]
	                + c3*grid_per_lambda[lambda_k].Usca[
			    interpolation_triangle->v3->identification_number];
	fprintf(DUMP_FILE,"%E ",Usca);
	Uem   =           c1*grid_per_lambda[lambda_k].Uem[
                            interpolation_triangle->v1->identification_number]
	                + c2*grid_per_lambda[lambda_k].Uem[
                            interpolation_triangle->v2->identification_number]
	                + c3*grid_per_lambda[lambda_k].Uem[
			    interpolation_triangle->v3->identification_number];
	fprintf(DUMP_FILE,"%E ",Uem);
	fprintf(DUMP_FILE,"%E ",(Ustar+Usca+Uem));
	Utotal=           c1*grid_per_lambda[lambda_k].Utot_new[
                            interpolation_triangle->v1->identification_number]
	                + c2*grid_per_lambda[lambda_k].Utot_new[
                            interpolation_triangle->v2->identification_number]
	                + c3*grid_per_lambda[lambda_k].Utot_new[
			    interpolation_triangle->v3->identification_number];
	fprintf(DUMP_FILE,"%E ",Utotal);
	fprintf(DUMP_FILE,"%.4d %.4d %.4d %.4d\n",
                interpolation_triangle->ID_number,
		interpolation_triangle->v1->identification_number,
		interpolation_triangle->v2->identification_number,
		interpolation_triangle->v3->identification_number);
        }
      }
      fprintf(DUMP_FILE,"\n\n");
      loop_vertex = loop_vertex->next;
    } while (loop_vertex != grid_per_lambda[lambda_N_max].start_vertex);
    fclose(DUMP_FILE);
  }


  /* OK, we have the new temperatures, but only on grid_N_max */
  /* Copy these temperatures to the other grids */

  /* mark the wavelengths which share grid_N_max as "used" */
   for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
     if (grid_per_lambda[lambda_k].start_vertex ==
	  grid_per_lambda[lambda_N_max].start_vertex)
       grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1.0;


  /* Copy the grid_N_max temperatures to the other grids */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > 0){ 

      /* loop over vertices in grid[lambda_k] */
      loop_vertex = grid_per_lambda[lambda_k].start_vertex;
      do {

        /* specify the point */
        x = loop_vertex->x; 
        z = loop_vertex->z; 

        /* find triangle in grid_N_max which contains this loop_vertex */
        start_triangle_list = grid_per_lambda[lambda_N_max].start_triangle; 
        err = _ZELENJAK_locate_point_in_grid(x,z,&interpolation_triangle);

/*
        if (interpolation_triangle==NULL) {
          start_vertex_list = grid_per_lambda[lambda_N_max].start_vertex;
          err = _ZELENJAK_locate_nearby_point_in_grid(x,z,&interpolation_triangle);
	}
*/
        if (interpolation_triangle!=NULL) {

          /* find the interpolation coefficients */
          _VILOZMAJ_Find_New_Temperature_Interpolation_Coefficient 
            (interpolation_triangle->v1->x, interpolation_triangle->v1->z,
	     interpolation_triangle->v2->x, interpolation_triangle->v2->z,
	     interpolation_triangle->v3->x, interpolation_triangle->v3->z,
             x, z, &c1, &c2, &c3);

  	  /* interpolate temperatures at the vertex from the triangle */
          for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
	    if (loop_vertex->T_dust[dust_i]>0.0 && 
                INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
	      loop_vertex->T_dust[dust_i] = 
                                c1*interpolation_triangle->v1->T_dust[dust_i]+
                                c2*interpolation_triangle->v2->T_dust[dust_i]+
                                c3*interpolation_triangle->v3->T_dust[dust_i];
              loop_vertex->T_error[dust_i] = 
                                c1*interpolation_triangle->v1->T_error[dust_i]+
                                c2*interpolation_triangle->v2->T_error[dust_i]+
                                c3*interpolation_triangle->v3->T_error[dust_i];
	    }
        }
        else for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
          if (loop_vertex->T_dust[dust_i]>0.0 && 
              INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
            loop_vertex->T_dust[dust_i]=0.0;
	  }
	loop_vertex = loop_vertex->next;
      } while (loop_vertex != grid_per_lambda[lambda_k].start_vertex);

      /* look for points which are not shared among the grids */
      flag_T0 = 1; /* the flag which shows if there is zero (error) temperatures
                  in the grid */
      while (flag_T0) {
        flag_T0 = 0;
        loop_vertex = grid_per_lambda[lambda_k].start_vertex;
        do {
          for (dust_i=0; dust_i < Max_Dust_Type; dust_i++)
            if (loop_vertex->T_dust[dust_i]==0.0 && 
              INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
            nearby_vertex = _ZELENJAK_Find_Adjacent_Temperature (loop_vertex,dust_i);
            if (nearby_vertex==NULL) {
              flag_T0 = 1;
              loop_vertex->T_dust[dust_i]=0.1;
              loop_vertex->T_error[dust_i]=1.0;
	    }
            else {
              loop_vertex->T_dust[dust_i]=nearby_vertex->T_dust[dust_i];
              loop_vertex->T_error[dust_i]=nearby_vertex->T_error[dust_i];
              /* copy the energy density */
              for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++)
               if (grid_per_lambda[lambda_k].start_vertex ==
                                           grid_per_lambda[lambda_kk].start_vertex)
                 grid_per_lambda[lambda_kk].Utot_new[loop_vertex->identification_number]=
                   grid_per_lambda[lambda_kk].Utot_new[nearby_vertex->identification_number];
            }
          }
          loop_vertex = loop_vertex->next;
        } while (loop_vertex != grid_per_lambda[lambda_k].start_vertex);
      }


      /* mark all wavelengths which share this grid[lambda_k] as "used" */
      for (lambda_kk=0; lambda_kk<lambda_Max; lambda_kk++) 
	if (grid_per_lambda[lambda_k].start_vertex ==
	    grid_per_lambda[lambda_kk].start_vertex)
	  grid_per_lambda[lambda_kk].TOTAL_NUMBER_OF_VERTICES *= -1.0;

    }
  } /* end of the loop over lambda_k */


  /* remove the "used" marker from all wavelengths */  
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1.0;

  free_dvector(old_temperature, 0, Max_Dust_Type-1);
  free_dvector(U,0, lambda_Max-1);

  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("--- end of _VILOZMAJ_Find_New_Temperature subroutine\n");
  }

  return 0;
} /* end of _VILOZMAJ_Find_New_Temperature */

/*************************************************************************/
/* subroutine for calculation of the 3D liner interpolation coefficients */
void _VILOZMAJ_Find_New_Temperature_Interpolation_Coefficient 
       (double x1, double y1, double x2, double y2, double x3, double y3, 
        double x, double y, double *c1, double *c2, double *c3){

  double denominator;
  denominator = (x2-x1)*(y3-y1) - (x3-x1)*(y2-y1);
  if (denominator == 0.0) {
    *c1=0.0; *c2=0.0; *c3=0.0; return;
  }
  *c1 = 1 + ( (x-x1)*(y2-y3)+(y-y1)*(x3-x2) )/denominator;
  *c2 = ( (x-x1)*(y3-y1)-(y-y1)*(x3-x1) )/denominator;
  *c3 = ( (y-y1)*(x2-x1)-(x-x1)*(y2-y1) )/denominator;
  return;
}

/*************************************************************************/
