/* LELUYA - VILOZMAJ program block */
/* This function calculates the intensity on a line given by the start point 
   and the edge of Rout sphere (computational domain) */ 

/* the algorithm is the same as calculation of optical depth over the grid
   (_ZELENJAK_tau_over_grid), expect that now we add intensity integrals into 
   the loops */

/* It is using integration over e^-Tau, not over Tau */

#include "VILOZMAJ.h"


double _VILOZMAJ_Intensity_Integral(int lambda_k, struct _3DVector start_point, 
			     struct _3DVector integral_direction,
			     double *Intensity_em, double *Intensity_sca,
                             int use_old_grid){

  /* use_old_grid = TRUE  --> use old_grid_per_lambda[] variables 
     use_old_grid = FALSE --> use grid_per_lambda[] variables 
  */

  double Emission, Scattering; /* intensities */
  struct _3DVector end_point; /* end point of the integral */
  struct _3DVector direction_center; /* direction from a point to the center */
  struct _3DVector current_point; /* a point in 3D space during integration */
  struct _3DVector old_current_point; /* ... current_point from previous step */
  struct _3DVector next_point; /* point after current_point */
  struct _3DVector current_point_xz; /* cylindrical coordinates */
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _triangle_structure *old_current_triangle; /* ... from previous step */
  struct _triangle_structure *next_triangle; /* location of next_point */
  double start_point_sqr;/* |start_point|^2  */ 
  double distance_total; /* distance between start_point and end_point */
  double current_distance; /* distance between start_point and current_point */
  double distance_step; /* spatial integration step */
  double tau; /*optical depth distance between start_point and current_point */
  double tau_step; /* integration tau step */
  double distance_step_refinement; /* refinement of the integral step */
  int tau_edge_flag; /* flag which shows if tau_step is crossing a dust edge */
  double Eta0, Eta1; /* Eta values at current_point and one point back */
  double *albedo_old,*albedo_new;/* albedo at current_point and point back */
  double *temperature; /* dust temperature at current_point */
  double *old_temperature; /* dust temperature at previous current_point */
  int dust_i; /* dust component counter */
  int err; /* error flag */

  double sqr_distance; /* sqr(distance from a point to the center) */ 
  double intensity_inter; /* an intensity interpolated from a triangle */
  double Iem_new, Iem_old; /* intermediate values of the emission intensity 
			      integral */
  double Isca_new,Isca_old; /* intermediate values of the scattering intensity 
			      integral */
  /* e_{-Tau} integral elements: 
    We use N_tau_substep steps over the e_-Tau function between two points
    on the radiative transfer integral ray. The source function is represented 
    by linear interpolation between these two points. */
  double Sum_point_old, Sum_point_new; /* factors that multiply the source 
                                    function at two integral points */
  int N_tau_substep = 4; /* number of optical depth substeps */
  int substep_n;  /* substep counter */
  double substep; /* substep spatial length */ 
  struct _3DVector subpoint1, subpoint2; /* two substep points */ 
  double *exp_tau_substep; /* list of exp(-tau) elements in the sum */
  int interpolate; /* flag which shows if ETA needs to be interpolated or not */
  double SubEta0,SubEta1; /* ETA value at two adjacent subpoints */

  double dummy; 

  Emission = 0.0;
  Scattering = 0.0;

  temperature = dvector(0,Max_Dust_Type-1); 
  old_temperature = dvector(0,Max_Dust_Type-1);
  albedo_old = dvector(0,Max_Dust_Type-1); 
  albedo_new = dvector(0,Max_Dust_Type-1); 

  if (use_old_grid) {
    start_vertex_list = old_grid_per_lambda[lambda_k].start_vertex;
    TOTAL_NUMBER_OF_VERTICES=old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
    start_triangle_list = old_grid_per_lambda[lambda_k].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES=old_grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;
  }
  else {
    start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
    TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
    start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;
  }

  current_point.x = current_point.y = current_point.z = 0.0;

  /* list of optical depth substeps */
  exp_tau_substep = dvector (0,N_tau_substep);

  /* normalize integral_direction just in case it is not a unit vector */ 
  dummy =   _MATH_LIB_3Ddistance(integral_direction,current_point);
  integral_direction.x /= dummy;
  integral_direction.y /= dummy;
  integral_direction.z /= dummy;

  /* find the end point */ 
  start_point_sqr = start_point.x*start_point.x +  start_point.y*start_point.y + 
                    start_point.z*start_point.z;
  distance_total = sqrt(start_point_sqr);
  direction_center.x = -start_point.x/distance_total;
  direction_center.y = -start_point.y/distance_total;
  direction_center.z = -start_point.z/distance_total;


  dummy = integral_direction.x*direction_center.x + 
          integral_direction.y*direction_center.y +
          integral_direction.z*direction_center.z;

  if (dummy >= 0.0 && Source_Radius*Source_Radius > 
                     distance_total* distance_total * (1.0 - dummy*dummy) ) {
    /******* the ray goes through the source *****/
    distance_total=distance_total*dummy - sqrt(Source_Radius*Source_Radius -
                              distance_total*distance_total*(1.0-dummy*dummy));
  }
  else {
    /******* the ray does not go through the source ********/
    distance_total=distance_total*dummy+
               sqrt(Rout*Rout-distance_total*distance_total*(1.0-dummy*dummy));
  }

  if (distance_total == 0.0) return 0.0;

  end_point.x=start_point.x+distance_total*integral_direction.x;
  end_point.y=start_point.y+distance_total*integral_direction.y;
  end_point.z=start_point.z+distance_total*integral_direction.z;

  /* prepare the variables for integration */ 
  current_point = start_point;

  current_point_xz.x = _MATH_LIB_distance(0.0,0.0,start_point.x,start_point.y);
  current_point_xz.y = 0.0;
  if (start_point.z >= 0.0) current_point_xz.z = start_point.z;
  else current_point_xz.z = - start_point.z;

  err = _ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
				      &current_triangle);

  tau = 0.0;
  current_distance = 0.0;

  if (current_triangle==NULL) 
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
  else for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    if (current_triangle->v1->T_dust[dust_i]>0.0 && 
        current_triangle->v2->T_dust[dust_i]>0.0 && 
        current_triangle->v3->T_dust[dust_i]>0.0){
      temperature[dust_i]= _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
                           current_triangle->v1->T_dust[dust_i],
			   current_triangle->v2->x,current_triangle->v2->z,
                           current_triangle->v2->T_dust[dust_i],
			   current_triangle->v3->x,current_triangle->v3->z,
                           current_triangle->v3->T_dust[dust_i],
			   current_point_xz.x,current_point_xz.z);
      if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
    }
    else temperature[dust_i]=-1.0;
  }
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
    old_temperature[dust_i]=temperature[dust_i];

  Eta0 = Eta(current_point_xz.x,current_point_xz.z,lambda_k,temperature,&err);

  /* find the initial integral step values */
  Iem_old = 0.0;
  Isca_old = 0.0;
  sqr_distance = start_point_sqr;

  if (current_triangle==NULL) 
    intensity_inter = 0.0;
  else {
         if (use_old_grid) {
    intensity_inter  = _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v1->identification_number],
			   current_triangle->v2->x,current_triangle->v2->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v2->identification_number],
			   current_triangle->v3->x,current_triangle->v3->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v3->identification_number],
			   current_point_xz.x,current_point_xz.z);
         }
         else {
    intensity_inter  = _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v1->identification_number],
			   current_triangle->v2->x,current_triangle->v2->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v2->identification_number],
			   current_triangle->v3->x,current_triangle->v3->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v3->identification_number],
			   current_point_xz.x,current_point_xz.z);
          }
  }

  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    if (temperature[dust_i]>0.0) {

      albedo_old[dust_i]=albedo(current_point_xz.x,current_point_xz.z,
                                  lambda_k,dust_i,&err);
      dummy = temperature[dust_i]/Dust_Sublim_Temp[0];
      dummy *= dummy;dummy *= dummy;
      dummy *= _MATH_LIB_Planck(temperature[dust_i],Lambda_Grid[lambda_k]);
      Iem_old = Iem_old + dummy*upsilon (current_point_xz.x,current_point_xz.z, 
		lambda_k,dust_i,temperature, &err) * (1.0-albedo_old[dust_i]);
      Isca_old = Isca_old + upsilon (current_point_xz.x,current_point_xz.z,
		lambda_k,dust_i,temperature, &err) * albedo_old[dust_i] *
                intensity_inter;
      /* in Isca_old we skiped the term: start_point_sqr/sqr_distance=1.0 */
    }
    else albedo_old[dust_i]=0.0; 
  }
  /* now we can start with the integral */
  distance_step_refinement=1.0;
  while (current_distance<distance_total){

    old_current_point = current_point;
    old_current_triangle = current_triangle;
    /* memorize the old point */
    subpoint1 = current_point;
    do { /* this do-loop will check the tau step over the dust edge */ 

    distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,
	 integral_direction,distance_step_refinement,&next_point,&next_triangle);

    /* check if the step is too big */
    if (distance_step_refinement < 1.0 && 
        current_triangle!=next_triangle) {
       distance_step_refinement=1.0;
       /* search again for the distance step */
       distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,
         integral_direction,distance_step_refinement,&next_point,&next_triangle);
    }


    /* check if we reached the end_point */
    if (current_distance + distance_step > distance_total) {
      distance_step = distance_total - current_distance;
      current_distance = distance_total;
      current_point = end_point;
      current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
					    current_point.y);
      current_point_xz.y = 0.0;
      if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
      else current_point_xz.z = - current_point.z;
      err=_ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
					&current_triangle);
    }
    else { /* ...or we have not reached the end_point */ 
      current_distance += distance_step;
      current_point = next_point;
      current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
					    current_point.y);
      current_point_xz.y = 0.0;
      if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
      else current_point_xz.z = - current_point.z;
      current_triangle = next_triangle;
    }

    if (current_triangle==NULL){
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
    }
    else for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (current_triangle->v1->T_dust[dust_i]>0.0 && 
          current_triangle->v2->T_dust[dust_i]>0.0 && 
          current_triangle->v3->T_dust[dust_i]>0.0){
        temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point_xz.x,current_point_xz.z);
        if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
      }
      else temperature[dust_i]=-1.0;
    }

    if (current_triangle!=NULL)
      Eta1=Eta(current_point_xz.x,current_point_xz.z,lambda_k,temperature,&err);
    else Eta1 = 0.0; 

    /* check if we are crossing the dust edge */ 
    tau_edge_flag=FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
      if (temperature[dust_i]*old_temperature[dust_i]<0.0)
	tau_edge_flag=TRUE;

    if(tau_edge_flag) {
      if ((Eta0+Eta1)*distance_step*0.5>0.01 && current_distance!=distance_total) {
	current_distance -= distance_step;
	current_point=old_current_point;
	current_triangle=old_current_triangle;
	distance_step_refinement *= 2.0; 
      }
      else tau_edge_flag=FALSE;
    }

    } while(tau_edge_flag);

    /* find new albedo */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (temperature[dust_i]>0.0) 
	albedo_new[dust_i] = albedo(current_point_xz.x,current_point_xz.z,
                                    lambda_k,dust_i,&err);
      else
	albedo_new[dust_i] = 0.0;


 if (current_triangle!=NULL){

    /* We are using N_tau_substep substeps to calculate tau */
    if (N_tau_substep<=1) {
      /* linear optical depth interpolation*/
      tau_step = (Eta0+Eta1)*distance_step*0.5;
      Sum_point_old = 0.5*exp(-tau)*(1.0-exp(-tau_step));
      Sum_point_new = Sum_point_old;
    }
    else {
      /* First, we need to know if each substep will have its own eta, because
         if not, we will interpolate it. */
      interpolate = FALSE;
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
        if ((albedo_new[dust_i]>0.0 && albedo_old[dust_i]==0.0) ||
            (albedo_new[dust_i]==0.0 && albedo_old[dust_i]>0.0))
	  {interpolate = TRUE; break;}
      /* size of a substep */
      substep = distance_step / N_tau_substep;
      exp_tau_substep[0] = exp(-tau);
      SubEta0 = Eta0;
      /* loop over the substeps */
      tau_step=0.0;
      for (substep_n=1; substep_n<=N_tau_substep; substep_n++) {
	if (interpolate) {
	  dummy = (Eta1-Eta0)/distance_step;
	  SubEta1 = dummy*substep_n*substep + Eta0;
	}
	else {
	  dummy = substep_n*substep;
	  subpoint2.x = subpoint1.x + dummy*integral_direction.x;
	  subpoint2.y = subpoint1.y + dummy*integral_direction.y;
	  subpoint2.z = subpoint1.z + dummy*integral_direction.z;
	  subpoint2.x=_MATH_LIB_distance(0.0,0.0,subpoint2.x,subpoint2.y);
	  if (subpoint2.z < 0.0) subpoint2.z = - subpoint2.z;
	  SubEta1=Eta(subpoint2.x,subpoint2.z,lambda_k,temperature,&err);
	}
	tau_step += ((SubEta0+SubEta1)*0.5*substep);
	exp_tau_substep[substep_n] = exp_tau_substep[0]*exp(-tau_step);
	SubEta0 = SubEta1;
      }
      /* Now, we have exp(optical depth) at each subpoint.
	 Calculate Sum_point_new and Sum_point_old */
      Sum_point_old = Sum_point_new = 0.0;
      for (substep_n=0; substep_n<N_tau_substep; substep_n++) {
	SubEta0 = exp_tau_substep[substep_n] + exp_tau_substep[substep_n+1];
	SubEta1 = exp_tau_substep[substep_n] - exp_tau_substep[substep_n+1];
	Sum_point_old = Sum_point_old + 
                  (SubEta0 - 2.0*exp_tau_substep[N_tau_substep])*SubEta1;
	Sum_point_new = Sum_point_new + 
		  (2.0*exp_tau_substep[0]-SubEta0)*SubEta1;
      }
      dummy = 2.0*(exp_tau_substep[0] - exp_tau_substep[N_tau_substep]);
      if (dummy!=0.0) {
	Sum_point_old /= dummy;
	Sum_point_new /= dummy;
      }
    } /* end of substep tau integration */


    /* calculate the intensity integrals */
    Iem_new = 0.0;
    Isca_new = 0.0;
    sqr_distance = current_point.x*current_point.x +
                   current_point.y*current_point.y +
                   current_point.z*current_point.z;

         if (use_old_grid) {
    intensity_inter  = _MATH_LIB_plane_interpolation 
                        (current_triangle->v1->x,current_triangle->v1->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v1->identification_number],
                          current_triangle->v2->x,current_triangle->v2->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v2->identification_number],
			  current_triangle->v3->x,current_triangle->v3->z,
old_grid_per_lambda[lambda_k].Utot_new[current_triangle->v3->identification_number],
			  current_point_xz.x,current_point_xz.z);
	 }
	 else {
    intensity_inter  = _MATH_LIB_plane_interpolation 
                        (current_triangle->v1->x,current_triangle->v1->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v1->identification_number],
                          current_triangle->v2->x,current_triangle->v2->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v2->identification_number],
			  current_triangle->v3->x,current_triangle->v3->z,
grid_per_lambda[lambda_k].Utot_new[current_triangle->v3->identification_number],
			  current_point_xz.x,current_point_xz.z);
	 }

    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (temperature[dust_i]>0.0) {
	dummy = temperature[dust_i]/Dust_Sublim_Temp[0];
        dummy *= dummy;dummy *= dummy;
        dummy *= _MATH_LIB_Planck(temperature[dust_i],Lambda_Grid[lambda_k]);
        Iem_new =Iem_new + dummy*upsilon (current_point_xz.x,current_point_xz.z, 
		lambda_k,dust_i,temperature, &err) * (1.0-albedo_new[dust_i]);
        Isca_new = Isca_new + upsilon (current_point_xz.x,current_point_xz.z,
		lambda_k,dust_i,temperature, &err) * albedo_new[dust_i] *
	        (start_point_sqr/sqr_distance)*intensity_inter;
    }

    Emission += (Iem_new*Sum_point_new+Iem_old*Sum_point_old);
    Scattering += (Isca_new*Sum_point_new+Isca_old*Sum_point_old);
    tau += tau_step;

/*if (lambda_k==16) printf("distance=%lf (X=%lf z=%lf) tau=%lf tau_step=%lf
D_step=%lf Eta0=%E Eta1=%E Emm=%E
Sca=%E\n",current_distance,current_point_xz.x,current_point_xz.z,tau,tau_step,distance_step,Eta0,Eta1,Emission,Scattering);*/

 }
 else {
   Eta1 = 0.0;
   Isca_new = Iem_new = 0.0;
 }

    /* prepare Eta for the next step */
    if (old_current_triangle == current_triangle && 
        current_triangle!=NULL) {
       distance_step_refinement /= 2.0;
    }
    else distance_step_refinement=1.0;
    Eta0 = Eta1;
    Isca_old = Isca_new;
    Iem_old = Iem_new;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++){
      albedo_old[dust_i] = albedo_new[dust_i];
      old_temperature[dust_i]=temperature[dust_i];
    }

    /* check for the optical depth limit */
    if (INPUT.TAU_LIMIT>0.0) 
      if (tau<INPUT.TAU_LIMIT) current_distance=distance_total+1.0;
  }

  Emission = Emission * (Psi*start_point_sqr);

  *Intensity_em=Emission;
  *Intensity_sca = Scattering;

  free_dvector(exp_tau_substep,0,N_tau_substep);
  free_dvector(temperature,0,Max_Dust_Type-1);
  free_dvector(old_temperature,0,Max_Dust_Type-1);
  free_dvector(albedo_old,0,Max_Dust_Type-1);
  free_dvector(albedo_new,0,Max_Dust_Type-1);

  /* debugger option *
  tau_step = _VODENJAK_tau_over_grid (start_point, end_point, lambda_k);
  *****************/

  return tau;
}
