/* LELUYA - VILOZMAJ program block */
/* This function calculates luminosity at various radii and checks if
   it is conserved. If it is not then the radiative transfer solution is 
   not correct. */

/* Detailed algorithm is given in the notes. Here we describe only the major
   components of it:

  - calculate the source luminosity
  - create the radial grid 
  - loop over the radii
  | - loop over the theta angle
  | | - calculate stellar flux 
  | - end
  - end
  - if process=0 then send a job to each process
  - else receive a job
  - if process=0
  | loop while at least one process is active 
  | | - wait for data from a process
  | | - send a new job to this process
  | end
  else
  | loop while there is a job to do
  | | - loop over the angular integral rays
  | | | - find the intensity along the rays and add them into the diffuse flux
  | | - end
  | | - calculate diffuse flux
  | end
  end
  - print/save the results (or make a decision about the calculation error)
  - destroy the luminosity array (pointer lists)
*/

#include "VILOZMAJ.h"

int _VILOZMAJ_Luminosity_shells (void) {

  int _LUMINOSITY_RADIAL_GRID_ = 20; /* total number of shells between (and
                                       including) 1 and Rout
                           NOTE: users should have access to this number */
  double *luminosity_radii; /* radial grid from 0 to _LUMINOSITY_RADIAL_GRID_
                             where luminosity_radii[0]=Source_Radius, 
                             luminosity_radii[1] = 1.1, and
                             luminosity_radii[_LUMINOSITY_RADIAL_GRID_]=Rout */
  struct _luminosity_cell_structure *luminosity_cells; 
                            /* array of small rings on the shells at which 
                             the flux is calculated. These rings are then 
                             integrated to get the luminosity for each shell.*/
  struct _luminosity_cell_structure *luminosity_patch; /* one of the cells 
			     in an luminosity pointer list */ 

  double radial_step;    /* log-step used in the radial grid */
  int radius_i;       /* counter over radii */

  double theta_loop, theta_loop_minus, theta_loop_plus; /* variables in 
                        the theta loop over the surface of shells. The flux 
                        is calculated at 'theta_loop' angle and it covers 
                        the range [theta_loop-0.5*theta_loop_minus, 
                        theta_loop+0.5*theta_loop_plus] */
  struct _3DVector rho, rho_unit; /* position of a point on a shell (and the
				     unit vector of this position) */
  struct _vertex_structure *rho_vertex; /* rho vector transformed into 
                         a vertex because of _GOROSTAS_create_integral_directions
                         function */
  struct _3DVector ray_direction; /* direction of the intensity integral*/

  int angular_rays_counter; /* counter of the angluar rays at a point */

  double angular_flux_weight; /* weight in the angular integral */
  int flag_plus_minus; /* flag which shows inward and outward flux at a point*/
  double Intensity_sca, Intensity_em; /*intensities calculated along an angular
				        integral ray */
  struct _3DVector step_direction; /* vector over the surface of a shell */
  double step_distance; /* distance over the surface of a shell */
  int lambda_max_N; /* wavelength with the largest number of vertices */
  struct _triangle_structure *rho_triangle; /* triangle at the point rho 
				       (in the lambda_max_N grid) */
  struct _triangle_structure *rho_triangle_next; /* next traingle for rho */

  double tau_lambda;  /* optical depth at some wavelength */
  int lambda_k;       /* counter over wavelengths */
  int err;            /* error flag */

  int job_flag; /* this flag will keep track of necessity for the loop over the jobs */
  int *process_status; /* flags which show the job status of the processes */
  int p;        /* counter over the processes */
  int theta_counter; /* counter over the theta angle */ 

  double dummy;

  /* output files */
  FILE *flux_file, *lum_file;
  char filename[256];
  double luminosity_value; /* luminosity at some shell */
  double flux_value;       /* bolometric flux at some shell and angle */

#ifdef _2DD_MPI_
  int tag = 0;
  int receive_flag;
  MPI_Status Receive_Status, Status;
  int MPI_radius_i, MPI_theta_counter;
#endif

  if (my_MPI_rank==0)
    printf("#%.3d: ....Luminosity calculation\n",my_MPI_rank);

  /* create the list of radii and luminosity arrays for shells */ 
  luminosity_radii = dvector(0,_LUMINOSITY_RADIAL_GRID_);
  luminosity_cells = luminosity_cell_vector(0,_LUMINOSITY_RADIAL_GRID_);
  process_status = ivector(0,total_MPI_process-1);

  /* create rho_vertex */
  rho_vertex = (struct _vertex_structure *) malloc 
                                            (sizeof(struct _vertex_structure));
  rho_vertex->prev = rho_vertex->next = NULL;
  rho_vertex->identification_number=0;
  rho_vertex->v1 = rho_vertex->v2 = rho_vertex->v3 = rho_vertex->v4 =
                   rho_vertex->v5 = rho_vertex->v6 = NULL;
  rho_vertex->level = 0;
  rho_vertex->type  = 0;
  rho_vertex->x = rho_vertex->z = 0.0;
  rho_vertex->T_dust = NULL;
  rho_vertex->T_error = NULL;
  rho_vertex->Eta = 0.0;
  rho_vertex->triangles = NULL;


  /* create the source luminosity at Source_Radius */
  luminosity_radii[0] = Source_Radius;
  luminosity_cells[0].next = luminosity_cells[0].prev = &(luminosity_cells[0]);
  luminosity_cells[0].Fstar      = dvector(0,lambda_Max-1);
  luminosity_cells[0].Fsca_plus  = dvector(0,lambda_Max-1);
  luminosity_cells[0].Fsca_minus = dvector(0,lambda_Max-1);
  luminosity_cells[0].Fem_plus   = dvector(0,lambda_Max-1);
  luminosity_cells[0].Fem_minus  = dvector(0,lambda_Max-1);

  /* find the most detailed grid (the largest number of vertices) */
  lambda_max_N = 0; 
  err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > err) {
      lambda_max_N = lambda_k;
      err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    }

  start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
  TOTAL_NUMBER_OF_VERTICES =
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
	   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
  start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
  TOTAL_NUMBER_OF_TRIANGLES = 
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

  /* create 3D cloud of points */
  err = _GOROSTAS_create_3D_cloud ();

  /* NOTE: for now we work only with the point source approximation.
     Change this source luminosity in case of a finite source. */
  luminosity_cells[0].ID = 0;
  luminosity_cells[0].theta = PI/4.0; 
  luminosity_cells[0].theta_minus = 0.0; 
  luminosity_cells[0].theta_plus = 0.5*PI;
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    luminosity_cells[0].Fstar[lambda_k] = 
                                  Find_Source_Flux(lambda_k,0.0,1.0,0.0,&err);
    luminosity_cells[0].Fsca_plus[lambda_k]  = 0.0;
    luminosity_cells[0].Fsca_minus[lambda_k] = 0.0;
    luminosity_cells[0].Fem_plus[lambda_k]   = 0.0;
    luminosity_cells[0].Fem_minus[lambda_k]  = 0.0;
  }

  /* create the radial grid */
  luminosity_radii[1] = 1.1;
  luminosity_radii[_LUMINOSITY_RADIAL_GRID_] = Rout;
  radial_step = exp(log(luminosity_radii[_LUMINOSITY_RADIAL_GRID_]/luminosity_radii[1])
               / (_LUMINOSITY_RADIAL_GRID_-1));
  for (radius_i=2; radius_i<=_LUMINOSITY_RADIAL_GRID_-1;radius_i++) 
    luminosity_radii[radius_i] = radial_step*luminosity_radii[radius_i-1];

  /* start with calculations of the luminosity */

  /* radial grid loop */
  for (radius_i=_LUMINOSITY_RADIAL_GRID_; radius_i>=1; radius_i--) {
 
    /* the initial theta-loop varaibles: */
    theta_loop = 0.0;
    theta_loop_minus = 0.0; theta_loop_plus = 0.0;

    /* loop over theta angle */
    do {
      theta_loop += theta_loop_plus;
      if (theta_loop > 0.5*PI) {theta_loop = 0.5*PI; theta_loop_plus = 0.0;}

      /* point on the shell */
      rho_unit.x = sin(theta_loop);
      rho_unit.y = 0.0;
      rho_unit.z = cos(theta_loop);
      rho.x = rho_unit.x * luminosity_radii[radius_i];
      rho.y = rho_unit.y * luminosity_radii[radius_i];
      rho.z = rho_unit.z * luminosity_radii[radius_i];

      if (my_MPI_rank==0)  printf("R=%E : theta=%7.4lf x=%E z=%E\n",
                   luminosity_radii[radius_i],theta_loop*180.0/PI,rho.x,rho.z);

      /* create a new luminosity cell and incorporate it into 
         the luminosity_cells[radius_i] pointer list */
      if (theta_loop == 0.0) {
        luminosity_cells[radius_i].next = luminosity_cells[radius_i].prev = 
                                                 &(luminosity_cells[radius_i]);
        luminosity_cells[radius_i].ID = 0;
	luminosity_cells[radius_i].Fstar      = dvector(0,lambda_Max-1);
	luminosity_cells[radius_i].Fsca_plus  = dvector(0,lambda_Max-1);
	luminosity_cells[radius_i].Fsca_minus = dvector(0,lambda_Max-1);
	luminosity_cells[radius_i].Fem_plus   = dvector(0,lambda_Max-1);
	luminosity_cells[radius_i].Fem_minus  = dvector(0,lambda_Max-1);
	luminosity_patch = &(luminosity_cells[radius_i]);
      }
      else {
        luminosity_patch = (struct _luminosity_cell_structure *) malloc
                                  (sizeof(struct _luminosity_cell_structure));
        luminosity_cells[radius_i].prev->next = luminosity_patch;
        luminosity_patch->prev = luminosity_cells[radius_i].prev;
        luminosity_patch->next = &(luminosity_cells[radius_i]);
        luminosity_cells[radius_i].prev = luminosity_patch;
        luminosity_patch->ID = luminosity_patch->prev->ID + 1;
	luminosity_patch->Fstar      = dvector(0,lambda_Max-1);
	luminosity_patch->Fsca_plus  = dvector(0,lambda_Max-1);
	luminosity_patch->Fsca_minus = dvector(0,lambda_Max-1);
	luminosity_patch->Fem_plus   = dvector(0,lambda_Max-1);
	luminosity_patch->Fem_minus  = dvector(0,lambda_Max-1);
      }

#ifndef _2DD_MPI_
  /* This flux calculation is active ONLY if we work with just one process */

      /* find the stellar flux */ 
      ray_direction.x = Source_Radius*rho_unit.x;
      ray_direction.y = Source_Radius*rho_unit.y;
      ray_direction.z = Source_Radius*rho_unit.z;

      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
	start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
 TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
        start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
 TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;
        tau_lambda = _ZELENJAK_tau_over_grid (rho,ray_direction,lambda_k,FALSE);
        luminosity_patch->Fstar[lambda_k] =  
  Finite_Source_Flux_Correction(rho.x,rho.z,tau_lambda,lambda_k,start_vertex_list,
                                start_triangle_list)*
               Total_Source_Flux(lambda_k,rho.x,rho.z,&err);
	luminosity_patch->Fsca_plus[lambda_k]=0.0;
	luminosity_patch->Fsca_minus[lambda_k]=0.0;
	luminosity_patch->Fem_plus[lambda_k]=0.0;
	luminosity_patch->Fem_minus[lambda_k]=0.0;
      }

      /* find the angular rays at rho for integration and flux calculation */
      start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
      TOTAL_NUMBER_OF_VERTICES =
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
      TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
	   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
      start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
      TOTAL_NUMBER_OF_TRIANGLES = 
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

      
      rho_vertex->x = rho.x;
      rho_vertex->z = rho.z;
      _GOROSTAS_create_integral_directions(rho_vertex);

      /* loop over angular integrals */
      for (angular_rays_counter=0;
           angular_rays_counter<total_number_of_spherical_triangles;
           angular_rays_counter++) {

	ray_direction = integral_directions[angular_rays_counter].direction;

        angular_flux_weight = integral_directions[angular_rays_counter].weight*
                              2.0/(4.0*PI); /* factor 2 
                                               comes from the axial symmetry */

	/* include the flux part of the weight: */ 
	dummy = ray_direction.x * rho_unit.x + 
                ray_direction.y * rho_unit.y + 
  	        ray_direction.z * rho_unit.z;
	angular_flux_weight = - angular_flux_weight * dummy;

        /* if (dummy<=0.0) printf("Impact b= %E ",
            (luminosity_radii[radius_i]*sqrt(1.0-dummy*dummy))); */
           
	/* is this flux into the shall or out of the shell? */
	if (dummy < 0.0) flag_plus_minus = 1; /* outward flux */
        else flag_plus_minus = -1;            /* inward flux */

	/* calculate the intensities */
	for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
	  tau_lambda = _VILOZMAJ_Intensity_Integral(lambda_k,rho,
		      ray_direction,&Intensity_em,&Intensity_sca,FALSE);
	  if (flag_plus_minus == 1) {
	    luminosity_patch->Fsca_plus[lambda_k]  += 
	                                   (angular_flux_weight*Intensity_sca);
	    luminosity_patch->Fem_plus[lambda_k]   += 
	                                   (angular_flux_weight*Intensity_em);
	  }
	  else {
	    luminosity_patch->Fsca_minus[lambda_k] += 
	                                   (angular_flux_weight*Intensity_sca);
	    luminosity_patch->Fem_minus[lambda_k]  +=
	                                   (angular_flux_weight*Intensity_em);
	  }

	} /* end of the lambda loop */ 

      } /* end of the loop over angular integrals */

      /* 
      flux_value = 0.0;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fstar);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fem_plus);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fem_minus);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fsca_plus);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fsca_minus);
      flux_value += dummy;
      printf("Flux=%E\n",flux_value);
      */ 
      

      /* destroy the angular rays */
      total_number_of_spherical_triangles=0;

#endif

      if (theta_loop < 0.5*PI) {

        /* find the triangle which contains the point rho */ 
        start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
        TOTAL_NUMBER_OF_VERTICES =
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
        start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
        TOTAL_NUMBER_OF_TRIANGLES = 
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

        err = _ZELENJAK_locate_point_in_grid(rho.x,rho.z,&rho_triangle); 

        /* find step to the next point */
        ray_direction.x = ray_direction.z = 0.0;
        ray_direction.y = 1.0;
        step_direction = _MATH_LIB_3D_vector_product (ray_direction,rho_unit);

        step_distance = _ZELENJAK_3Dstep (rho,rho_triangle,step_direction,1.0,
                                         &ray_direction,&rho_triangle_next);
        step_distance *= 2.0; /* let's use a larger step */ 

        step_distance = 1.0 - step_distance*step_distance /
                   (2.0*luminosity_radii[radius_i]*luminosity_radii[radius_i]);

        theta_loop_plus = acos(step_distance);

	/* be sure that the step is less than 10deg */
        if (theta_loop_plus > 0.174533) theta_loop_plus = 0.174533;
        /* be sure that the step is less than 1deg */
        /*if (theta_loop_plus > 0.0174533) theta_loop_plus = 0.0174533;*/

        /* set the theta angles in luminosity_patch */
        luminosity_patch->theta = theta_loop;
        luminosity_patch->theta_minus = theta_loop - 0.5*theta_loop_minus;
        if (theta_loop+theta_loop_plus > 0.5*PI) {
	  luminosity_patch->theta_plus  = theta_loop+0.5*(0.5*PI-theta_loop);
	  theta_loop_minus = 0.5*PI-theta_loop; 
	}
        else {luminosity_patch->theta_plus  = theta_loop + 0.5*theta_loop_plus;
	      theta_loop_minus = theta_loop_plus;}
      }
      else {
        luminosity_patch->theta = theta_loop;
        luminosity_patch->theta_minus = theta_loop - 0.5*theta_loop_minus;
	luminosity_patch->theta_plus  = 0.5*PI;
      } 

    } while(theta_loop < 0.5*PI);

  } /* end of the radial grid loop */


#ifdef _2DD_MPI_
  /* parallel version of the flux calculation at each cell */
  radius_i=1;
  theta_counter = 0;
  process_status[0]= -1;

  if (my_MPI_rank==0) {
    /* send jobs to other porcesses */
    for (p=1; p<total_MPI_process; p++) {
      if (radius_i<=_LUMINOSITY_RADIAL_GRID_) {
	process_status[p]=1;
        MPI_Send(&radius_i,1,MPI_INT,p,tag,MPI_communicator);
        MPI_Send(&theta_counter,1,MPI_INT,p,tag,MPI_communicator);
        theta_counter++;
        if (theta_counter>luminosity_cells[radius_i].prev->ID) 
             { radius_i++; theta_counter=0; }
      }
      else {
        job_flag=-1;
        MPI_Send(&job_flag,1,MPI_INT,p,tag,MPI_communicator);
        process_status[p]=-1;
      }
    }
  }
  else {
    /* receive jobs from the process 0 */
    MPI_Recv(&job_flag,1,MPI_INT,0,tag,MPI_communicator,&Status);
    if (job_flag>=0) {
      radius_i = job_flag;
      job_flag = 1;
      MPI_Recv(&theta_counter,1,MPI_INT,0,tag,MPI_communicator,&Status);
    }
  }

  if (my_MPI_rank==0) {
    /* keep sending the jobs until they run out */

    job_flag=-1; p=1; /* find out if we have more jobs to send */ 
    while (job_flag==-1 && p<total_MPI_process) {
      if (process_status[p]==1) job_flag=1;
      p++;
    }

    while (job_flag==1) {
      /* receive a signal from any process */
      MPI_Recv(&MPI_radius_i,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,
	       &Receive_Status);
      MPI_Recv(&MPI_theta_counter,1,MPI_INT,Receive_Status.MPI_SOURCE,
               tag,MPI_communicator,&Status);

      /* printf("000: radius_i=%d, theta=%d from %.3d\n",MPI_radius_i,MPI_theta_counter,
	 Receive_Status.MPI_SOURCE); */

      /* find the luminosity cell of (MPI_radius_i,MPI_theta_counter) */
      luminosity_patch = &(luminosity_cells[MPI_radius_i]);
      while (luminosity_patch->ID != MPI_theta_counter) 
	luminosity_patch = luminosity_patch->next;
      /* receive the fluxes */ 
      MPI_Recv(luminosity_patch->Fstar,lambda_Max,MPI_DOUBLE,
               Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
      MPI_Recv(luminosity_patch->Fsca_plus,lambda_Max,MPI_DOUBLE,
               Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
      MPI_Recv(luminosity_patch->Fsca_minus,lambda_Max,MPI_DOUBLE,
               Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
      MPI_Recv(luminosity_patch->Fem_plus,lambda_Max,MPI_DOUBLE,
               Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
      MPI_Recv(luminosity_patch->Fem_minus,lambda_Max,MPI_DOUBLE,
               Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);

      /* send the next job to this process */
      if (radius_i<=_LUMINOSITY_RADIAL_GRID_) {
	process_status[Receive_Status.MPI_SOURCE]=1;
        MPI_Send(&radius_i,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
        MPI_Send(&theta_counter,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
        theta_counter++;
        if (theta_counter>luminosity_cells[radius_i].prev->ID) 
             { radius_i++; theta_counter=0; }
      }
      else {
        job_flag=-1;
        MPI_Send(&job_flag,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,MPI_communicator);
        process_status[Receive_Status.MPI_SOURCE]=-1;
      }
      /* find out if we still need to wait for some job to finish */
      job_flag=-1; p=1; 
      while (job_flag==-1 && p<total_MPI_process) {
	if (process_status[p]==1) job_flag=1;
	p++;
      }
    } /* end of job_flag==1 */ 
  }
  else {
    /* calculate the fluxes (when my_MPI_rank!=0) */
    while (job_flag==1) {

      /* printf("#%.2d: radius_i=%d, theta=%d\n",my_MPI_rank,radius_i,theta_counter); */

      /* find the flux for (radius_i, theta_counter) */
      luminosity_patch = &(luminosity_cells[radius_i]);
      while (luminosity_patch->ID != theta_counter) 
	luminosity_patch = luminosity_patch->next;

      /***********************************/ 
      /****** flux calculation ***********/
      rho_unit.x = sin(luminosity_patch->theta);
      rho_unit.y = 0.0;
      rho_unit.z = cos(luminosity_patch->theta);
      rho.x = rho_unit.x * luminosity_radii[radius_i];
      rho.y = rho_unit.y * luminosity_radii[radius_i];
      rho.z = rho_unit.z * luminosity_radii[radius_i];

      /* find the stellar flux */ 
      ray_direction.x = Source_Radius*rho_unit.x;
      ray_direction.y = Source_Radius*rho_unit.y;
      ray_direction.z = Source_Radius*rho_unit.z;

      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
	start_vertex_list = grid_per_lambda[lambda_k].start_vertex;
 TOTAL_NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
        start_triangle_list = grid_per_lambda[lambda_k].start_triangle;
 TOTAL_NUMBER_OF_TRIANGLES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_TRIANGLES;
        tau_lambda = _ZELENJAK_tau_over_grid (rho,ray_direction,lambda_k,FALSE);
        luminosity_patch->Fstar[lambda_k] =  
  Finite_Source_Flux_Correction(rho.x,rho.z,tau_lambda,lambda_k,start_vertex_list,
                                start_triangle_list)*
               Total_Source_Flux(lambda_k,rho.x,rho.z,&err);
	luminosity_patch->Fsca_plus[lambda_k]=0.0;
	luminosity_patch->Fsca_minus[lambda_k]=0.0;
	luminosity_patch->Fem_plus[lambda_k]=0.0;
	luminosity_patch->Fem_minus[lambda_k]=0.0;
      }

      /* find the angular rays at rho for integration and flux calculation */
        start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
      TOTAL_NUMBER_OF_VERTICES =
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
      TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES = 
	   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_NOT_EMPTY_VERTICES;
      start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
      TOTAL_NUMBER_OF_TRIANGLES = 
                     grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;

      rho_vertex->x = rho.x;
      rho_vertex->z = rho.z;
      _GOROSTAS_create_integral_directions(rho_vertex);

      /* loop over angular integrals */
      for (angular_rays_counter=0;
           angular_rays_counter<total_number_of_spherical_triangles;
           angular_rays_counter++) {

	ray_direction = integral_directions[angular_rays_counter].direction;

        angular_flux_weight = 
 integral_directions[angular_rays_counter].weight*2.0/(4.0*PI); /* factor 2 
                                          comes from the axial symmetry */

        /* a special case when the whole ray exists in the X-Z plane: */ 
        /* if (ray_direction.y==0.0) angular_flux_weight /= 2.0;*/

	/* include the flux part of the weight: */ 
	dummy = ray_direction.x * rho_unit.x + 
                ray_direction.y * rho_unit.y + 
  	        ray_direction.z * rho_unit.z;
	angular_flux_weight = - angular_flux_weight * dummy;
           
	/* is this flux into the shell or out of the shell? */
	if (dummy < 0.0) flag_plus_minus = 1; /* outward flux */
        else flag_plus_minus = -1;            /* inward flux */

	/* calculate the intensities */
	for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
	  tau_lambda = _VILOZMAJ_Intensity_Integral(lambda_k,rho,
		      ray_direction,&Intensity_em,&Intensity_sca,FALSE);
	  if (flag_plus_minus == 1) {
	    luminosity_patch->Fsca_plus[lambda_k]  += 
	                                   (angular_flux_weight*Intensity_sca);
	    luminosity_patch->Fem_plus[lambda_k]   += 
	                                   (angular_flux_weight*Intensity_em);
	  }
	  else {
	    luminosity_patch->Fsca_minus[lambda_k] += 
	                                   (angular_flux_weight*Intensity_sca);
	    luminosity_patch->Fem_minus[lambda_k]  +=
	                                   (angular_flux_weight*Intensity_em);
	  }
	} /* end of the lambda loop */ 

      } /* end of the loop over angular integrals */

      /* destroy the angular rays */
      total_number_of_spherical_triangles=0;

      /****** end of flux calculation **********/
      /*****************************************/

      /* send the data to the process 0 */ 
      MPI_Send(&radius_i,1,MPI_INT,0,tag,MPI_communicator);
      MPI_Send(&theta_counter,1,MPI_INT,0,tag,MPI_communicator);

      MPI_Send(luminosity_patch->Fstar,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(luminosity_patch->Fsca_plus,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(luminosity_patch->Fsca_minus,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(luminosity_patch->Fem_plus,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(luminosity_patch->Fem_minus,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);

      /* receive the next job */
      MPI_Recv(&job_flag,1,MPI_INT,0,tag,MPI_communicator,&Status);
      if (job_flag>=0) {
	radius_i = job_flag;
	job_flag = 1;
	MPI_Recv(&theta_counter,1,MPI_INT,0,tag,MPI_communicator,&Status);
      }

    } /* end of while (job_flag) */ 
  }

  /* wait for other processes to finish */
  MPI_Barrier(MPI_communicator);
#endif
  /* destry the 3D cloud of points (used for calculating integral directions) */
  err=_GOROSTAS_destroy_3D_cloud ();

  if (my_MPI_rank==0) {

  /* print the fluxes and luminosities */
#ifdef _2DD_MPI_
  sprintf(filename,"%s/Luminosity_shells.dat",_LELUYA_MPIOUT_PATH);
#else
  sprintf(filename,"%s/Luminosity_shells.dat",_LELUYA_OUTPUT_PATH);
#endif

  lum_file = fopen(filename,"w");
  fprintf(lum_file,"  radius  luminosity\n");
  /* loop over radii */ 
  for (radius_i=0; radius_i<=_LUMINOSITY_RADIAL_GRID_;radius_i++) {


#ifdef _2DD_MPI_
    sprintf(filename,"%s/Luminosity_flux_r=%6.4E.dat",_LELUYA_MPIOUT_PATH,
	    luminosity_radii[radius_i]);
#else
    sprintf(filename,"%s/Luminosity_flux_r=%6.4E.dat",_LELUYA_OUTPUT_PATH,
	    luminosity_radii[radius_i]);
#endif

    flux_file = fopen(filename,"w");
    fprintf(flux_file,"theta        theta_minus  theta_plus   flux_star    flux_em_out  flux_em_in    flux_sca_out flux_sca_in   flux_total\n");

    luminosity_value = 0.0;
    /* loop over the surface of a shell */ 
    luminosity_patch = &(luminosity_cells[radius_i]);
    do {
      fprintf(flux_file,"%E %E %E ",luminosity_patch->theta,
              luminosity_patch->theta_minus,luminosity_patch->theta_plus);
      flux_value = 0.0;
      /* stellar part */
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fstar);
      fprintf(flux_file,"%E ",dummy);
      flux_value += dummy;
      /* emission part */
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fem_plus);
      fprintf(flux_file,"%E ",dummy);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fem_minus);
      fprintf(flux_file,"%E ",dummy);
      flux_value += dummy;
      /* scattering part */
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fsca_plus);
      fprintf(flux_file,"%E ",dummy);
      flux_value += dummy;
      dummy = _MATH_LIB_simpson(lambda_Max,0,lambda_Max-1,Lambda_Grid,
				luminosity_patch->Fsca_minus);
      fprintf(flux_file,"%E ",dummy);
      flux_value += dummy;
      /* total flux */
      fprintf(flux_file,"%E\n",flux_value);

      /* luminosity */
      luminosity_value = luminosity_value+( cos(luminosity_patch->theta_minus)
                         - cos(luminosity_patch->theta_plus) )*flux_value;

      luminosity_patch = luminosity_patch->next;
    } while(luminosity_patch!= &(luminosity_cells[radius_i]));

    fclose(flux_file);


    fprintf(lum_file,"%E  %E\n",luminosity_radii[radius_i],luminosity_value);
  }
  fclose(lum_file);

  } /* end of:   if (my_MPI_rank==0) */ 

  /* free the memory allocated by luminosity_radii and luminosity_cells*/
  for (radius_i=0; radius_i<=_LUMINOSITY_RADIAL_GRID_;radius_i++) {
    luminosity_patch = luminosity_cells[radius_i].next;
    while (luminosity_patch!= &(luminosity_cells[radius_i]) ){
      luminosity_cells[radius_i].next = luminosity_patch->next;
      free_dvector(luminosity_patch->Fstar,0,lambda_Max-1);
      free_dvector(luminosity_patch->Fsca_plus,0,lambda_Max-1);
      free_dvector(luminosity_patch->Fsca_minus,0,lambda_Max-1);
      free_dvector(luminosity_patch->Fem_plus,0,lambda_Max-1);
      free_dvector(luminosity_patch->Fem_minus,0,lambda_Max-1);
      free(luminosity_patch);
      luminosity_patch = luminosity_cells[radius_i].next;
    }
    free_dvector(luminosity_cells[radius_i].Fstar,0,lambda_Max-1);
    free_dvector(luminosity_cells[radius_i].Fsca_plus,0,lambda_Max-1);
    free_dvector(luminosity_cells[radius_i].Fsca_minus,0,lambda_Max-1);
    free_dvector(luminosity_cells[radius_i].Fem_plus,0,lambda_Max-1);
    free_dvector(luminosity_cells[radius_i].Fem_minus,0,lambda_Max-1);
  }

  free_dvector(luminosity_radii,0,_LUMINOSITY_RADIAL_GRID_);
  free_luminosity_cell_vector(luminosity_cells,0,_LUMINOSITY_RADIAL_GRID_);
  free_ivector(process_status,0,total_MPI_process-1);
  /* free rho_vertex */
  free(rho_vertex);

  return 1;
}
