/* LELUYA - VILOZMAJ program block */
/* This function creates the image grid and intensities for a given viewing
   angle. Integral of the intensity over the grid gives the total observed
   flux. */

/* Detailed algorithm is given in the notes. Here we describe only the major
   components of it:

   - find Source_Radius, find the wavelength with the most detailed grid
   - create the source image: image grid from 0 to Source_Radius
   - create the initial grid for radius > Source_Radius
   - loop while a refinement is needed
   | - find how many vertices is visible through each image cell
   | - if the number of visible points is more than a given limit, split
   |   the cell into smaller cells, and kill the cell
   - end
   - calculate intensities at each image cell, correct the cells with one side
     in the other hemisphere (symmetry requirement)

*/

#include "VILOZMAJ.h"

/****************************************************************************/
/* subroutine: split a cell into 9 sub-cells ********************************/
struct _image_cell_structure *_VILOZMAJ_Output_Image_Flux_Split_Cell (
                              struct _image_cell_structure *Image_cell,
                              struct _image_cell_structure *Image,
                              int *TOTAL_NUMBER_OF_IMAGE_CELLS, 
                              int Refinement_matrix[][3]);

/****************************************************************************/

int _VILOZMAJ_Output_Image_Flux(double theta_view,int *TOTAL_NUMBER_OF_IMAGE_CELLS,
			    struct _image_cell_structure **Image_){
  /* Constants which determine the image grid properties */
  /* NOTE: These constants are currently local variables. In the future, it
     would be better if users can change them if necessary, thus they should
     be global variables. */
  int N_Source_r_image = 15; /* number of radial source image grid points */
  int N_Source_phi_image=19; /* number of angular source image grid points */
  double Ri1=0.1;            /* Radius (Ri1*Rout) at which the cell density 
                                is changing */
  int N_r_image   = 270;     /* 150 = initial number of radial image grid points 
                                up to the radius Ri1 */
  int N_phi_image = 180 ;     /* 45 = initial number of angular image grid points 
                                up to the radius Ri1 */
  int N_r_image_1=50;           /* 40 =initial number of radial image grid points
                                between Ri1*Rout and Rout */
  int N_phi_image_1 = 60;    /* 180 = initial number of angular image grid points 
                                between Ri1*Rout and Rout */
  int limit_POINTS_IN_CELL = 60; /* 60 = each image cell should have no more than
                         this number of 3D grid vertices visible through it */
  double grid_refinement_factor = 1; /* if refined, cells will use:
                            (int)(grid_refinement_factor*limit_POINTS_IN_CELL) */
  /* increased precision along given radial directions : */
  double refine_angle1 = 177.99*RAD1; /* increased number of cells between these...*/
  double refine_angle2 = 82.01*RAD1; /* ... two angles. */
  double refine_angle3 = PI - refine_angle2; /* don't change this */
  double refine_angle4 = PI - refine_angle1; /* don't change this */
  int refine_angle_N = 10; /* number of sub-angles within one angular step */
  int max_refine_angle_N;
  /*********/
  int N_phi_vertex_grid = 36; /* 36 = global phi grid in the refinement loop */ 
  int lambda_max_N; /* wavelength with the largest optical depth */
  struct _3DVector observer_direction; /* direction of the observer */

  struct _image_cell_structure *Image; /* the image */
  struct _image_cell_structure *Image_cell; /* a cell in the image */
  struct _image_cell_structure *Image_cell1; /* a cell in the image */
  int lambda_k;                             /* counter over wavelengths */
  double r_polar1, r_polar2, phi_polar1,phi_polar2; /*polar image coordinates*/
  double r_polar_Int, phi_polar_Int;  /* point of intensity in an image cell */
  int r_polar_counter, phi_polar_counter;   /* counters in the polar loops */
  int N_phi_image_limit;               /* limit for phi_polar_counter */
  double r_polar_step;                 /* log step of the initial image grid */

  int refine_grid; /* flag in the refinement loop which shows if the refinement
		      is necessary or not */
  struct _vertex_structure   *loop_vertex; /* vertex in the refinement loop */
  int phi_counter_3D;  /* phi angle counter in the refinement loop */
  double phi_angle_3D; /*  = phi_counter_3D*PI/N_phi_vertex_grid */ 
  int z_symmetry;      /* flag which shows + or - hemisphere */
  struct _3DVector point_3D; /* a point in the refinement loop */

  struct _3DVector start_point; /* start point for an intensity integral */
  struct _3DVector integral_direction; /* direction of an intensity integral */

  /* variables in case of Output_Lambda_Image>0.0 (image at a specific wavelength
     is required) */
  int Out_lambda_k; /* a lambda grid number which represents Output_Lambda_Image */
  int Out_lambda_k1,Out_lambda_k2; /* lambda grid numbers which represents wavelenths
                                      for interpolation of Output_Lambda_Image */
  double Int_interpolate_em,Int_interpolate_sca,Int_interpolate_star; /* interpolated 
                                                                  intensities/flux */ 

  int err;            /* error flag */
  double dummy,cos_phi1,cos_phi2,sin_phi1,sin_phi2; 
  struct _3DVector dummy_point;
  /* shows which subcells are created */
  int Refinement_matrix[3][3] = {{0,0,0},{0,0,0},{0,0,0}};

  /* total flux */
  double *total_flux_em,*total_flux_sca,*total_flux_star ; 

  /* output flux file */
  FILE *flux_file;
  char filename[256];
  int counter,counter1;

#ifdef _2DD_MPI_
  int tag = 0;
  MPI_Status Receive_Status, Status;
  int *Process_Status;
  int ID;
  double *Iem11, *Iem12, *Iem13, *Iem21, *Iem22, *Iem23, *Iem31, *Iem32, *Iem33;
  double *Isca11, *Isca12, *Isca13, *Isca21, 
         *Isca22, *Isca23, *Isca31, *Isca32, *Isca33;
  double *Tau11, *Tau12, *Tau13, *Tau21, *Tau22, *Tau23, *Tau31, *Tau32, *Tau33;
  double *Average_I_total;
#define _INTENSITY_VARIATION_OVER_THE_CELL_SURFACE_ 30
  double Ilow = 1.0/_INTENSITY_VARIATION_OVER_THE_CELL_SURFACE_;
  double Ihigh= _INTENSITY_VARIATION_OVER_THE_CELL_SURFACE_;
  double received_r1, received_r2, received_phi1,  received_phi2;
  double r_subcell_step, phi_subcell_step;
#endif

  Out_lambda_k1=Out_lambda_k2=Out_lambda_k=-1;
  if (Output_Lambda_Image>0.0) {
    /* find the lambda grid number for Output_Lambda_Image */
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
      if (Lambda_Grid[lambda_k]==Output_Lambda_Image) {
        Out_lambda_k = lambda_k;
        break;
      }
    if (Out_lambda_k <0) {
      /* find the lambda grid numbers for interpolation */
      Out_lambda_k2 = 0; 
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
	if (Lambda_Grid[lambda_k]<Output_Lambda_Image) {
	  Out_lambda_k1++;
	  Out_lambda_k2++;
	}
      if (Out_lambda_k1<0 || Out_lambda_k2>=lambda_Max) {
printf("ERROR: The required wavelength (%E) is not part of the wavelength-grid\n",
        Output_Lambda_Image);
        return 1;
      }
    }
  }


  if (my_MPI_rank==0) {
    total_flux_em = dvector(0,lambda_Max-1);
    total_flux_sca = dvector(0,lambda_Max-1);
    total_flux_star = dvector(0,lambda_Max-1);
  }
#ifdef _2DD_MPI_
    Process_Status = ivector(0,total_MPI_process-1);
    Iem11 = dvector(0,lambda_Max-1);
    Iem12 = dvector(0,lambda_Max-1);
    Iem13 = dvector(0,lambda_Max-1);
    Iem21 = dvector(0,lambda_Max-1);
    Iem22 = dvector(0,lambda_Max-1);
    Iem23 = dvector(0,lambda_Max-1);
    Iem31 = dvector(0,lambda_Max-1);
    Iem32 = dvector(0,lambda_Max-1);
    Iem33 = dvector(0,lambda_Max-1);
    Isca11 = dvector(0,lambda_Max-1);
    Isca12 = dvector(0,lambda_Max-1);
    Isca13 = dvector(0,lambda_Max-1);
    Isca21 = dvector(0,lambda_Max-1);
    Isca22 = dvector(0,lambda_Max-1);
    Isca23 = dvector(0,lambda_Max-1);
    Isca31 = dvector(0,lambda_Max-1);
    Isca32 = dvector(0,lambda_Max-1);
    Isca33 = dvector(0,lambda_Max-1);
    Tau11 = dvector(0,lambda_Max-1);
    Tau12 = dvector(0,lambda_Max-1);
    Tau13 = dvector(0,lambda_Max-1);
    Tau21 = dvector(0,lambda_Max-1);
    Tau22 = dvector(0,lambda_Max-1);
    Tau23 = dvector(0,lambda_Max-1);
    Tau31 = dvector(0,lambda_Max-1);
    Tau32 = dvector(0,lambda_Max-1);
    Tau33 = dvector(0,lambda_Max-1);
    Average_I_total = dvector(0,lambda_Max-1);
#endif

  dummy = theta_view*180.0/PI;
  if (DEBUG.SCREEN_MESSAGE >= 1 ) {
    printf("Find the image with inclination = %-5.2lf\n",dummy);
  }
  fflush(stdout);

  /* find Source_Radius */
  err = _ZELENJAK_Find_Source_Radius();

  /* direction of the observer */
  observer_direction.x = sin(theta_view);
  observer_direction.y = 0.0;
  observer_direction.z = cos(theta_view);

  /* find the wavelength with the largest number of vertices
     (this wavelength will be used for calculation of the image grid) */

  lambda_max_N = 0;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++){
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES >
        grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES)
      lambda_max_N = lambda_k;
  }

  /*****************************************************************/
  /* create the first image cell at r=0 */
  *TOTAL_NUMBER_OF_IMAGE_CELLS = 0;

  Image = (struct _image_cell_structure *) malloc
                                        (sizeof(struct _image_cell_structure));
  Image->next = Image->prev = Image;
  Image->r1 = 0.0; 
  Image->r2 = Source_Radius/N_Source_r_image;
  Image->phi1 = 0; Image->phi2 = PI;
  Image->Intensity_em   = dvector (0,lambda_Max-1);
  Image->Intensity_sca  = dvector (0,lambda_Max-1);
  Image->Intensity_star = dvector (0,lambda_Max-1);
  Image->tau = dvector (0,lambda_Max-1);
  Image->number_of_visible_points = 0;
  Image->old_number_of_visible_points = 0;

#ifdef _2DD_MPI_
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
    if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
	Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k) {
      Image->tau[lambda_k] = 0.0;
      Image->Intensity_em[lambda_k] = 0.0;
      Image->Intensity_sca[lambda_k] = 0.0;
      Image->Intensity_star[lambda_k] = 0.0;
    }
#else
  start_point.x = Rout*observer_direction.x;
  start_point.y = Rout*observer_direction.y;
  start_point.z = Rout*observer_direction.z;
  /* direction of integration */
  integral_direction.x = - observer_direction.x;
  integral_direction.y = - observer_direction.y;
  integral_direction.z = - observer_direction.z;
  /* intensities */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
    if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
	Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k) {
      Image->tau[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                           integral_direction,&(Image->Intensity_em[lambda_k]),
                           &(Image->Intensity_sca[lambda_k]),FALSE);
      dummy = 1.0;
      Image->Intensity_star[lambda_k] = exp(-Image->tau[lambda_k])*
                 Find_Source_Flux (lambda_k,observer_direction.z,dummy,
				   Image->r1,&err);
      dummy = Rout/Source_Radius;
      dummy *= dummy; dummy *= 4.0;
      Image->Intensity_star[lambda_k] *= dummy;
    }
#endif

  (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;


  /*****************************************************************/
  /* create image of the source */

  for (r_polar_counter=1; r_polar_counter<N_Source_r_image;r_polar_counter++){
    for (phi_polar_counter=0; phi_polar_counter<=N_Source_phi_image;
	 phi_polar_counter++){

      Image_cell = (struct _image_cell_structure *) malloc
                                        (sizeof(struct _image_cell_structure));
      Image->prev->next = Image_cell;
      Image_cell->prev = Image->prev;
      Image_cell->next = Image;
      Image->prev = Image_cell;

      Image_cell->r1 = r_polar_counter*Source_Radius/N_Source_r_image;
      Image_cell->r2 = (r_polar_counter+1)*Source_Radius/N_Source_r_image;
      Image_cell->phi1 = (2*phi_polar_counter-1)*PI/(2.0*N_Source_phi_image); 
      Image_cell->phi2 = (2*phi_polar_counter+1)*PI/(2.0*N_Source_phi_image);
      Image_cell->Intensity_em   = dvector (0,lambda_Max-1);
      Image_cell->Intensity_sca  = dvector (0,lambda_Max-1);
      Image_cell->Intensity_star = dvector (0,lambda_Max-1);
      Image_cell->tau = dvector (0,lambda_Max-1);
      Image_cell->number_of_visible_points = 0;
      Image_cell->old_number_of_visible_points = 0;

#ifdef _2DD_MPI_
      /* if under MPI then calculate intensities later */ 
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
	    Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k) {
          Image_cell->tau[lambda_k] = 0.0;
          Image_cell->Intensity_em[lambda_k] = 0.0;
          Image_cell->Intensity_sca[lambda_k] = 0.0;
          Image_cell->Intensity_star[lambda_k] = 0.0;
        }
#else
      /* a point on the source surface */
      r_polar_Int = sqrt(Image_cell->r1 * Image_cell->r2);
      phi_polar_Int = 0.5*(Image_cell->phi1+Image_cell->phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
                                 Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k){
          Image_cell->tau[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,
                           start_point,integral_direction,
                           &(Image_cell->Intensity_em[lambda_k]),
                           &(Image_cell->Intensity_sca[lambda_k]),FALSE);
	  dummy = 1.0;
          Image_cell->Intensity_star[lambda_k] = exp(-Image_cell->tau[lambda_k])*
                 Find_Source_Flux (lambda_k,observer_direction.z,dummy,
				   r_polar_Int,&err);
	  dummy = Rout/Source_Radius;
	  dummy *= dummy; dummy *= 4.0;
          Image_cell->Intensity_star[lambda_k] *= dummy;
        }
#endif

      (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
    }
  }


#ifdef _2DD_MPI_
  /*****************************************************************/
  /* find the intensities of the source */
  if (my_MPI_rank==0) {
    for (ID=1; ID<total_MPI_process; ID++) Process_Status[ID]=FALSE;
    counter = 0;
    /* loop until all cells are sent to other processes */
    do {
      /* find available process */
      ID = 0;
      for (err=1; err<total_MPI_process; err++) 
        if (Process_Status[err]==FALSE) {ID=err; break;}
      if (ID>0) {
	/* send info about the cell to process ID */
        MPI_Send(&counter,1,MPI_INT,ID,tag,MPI_communicator);
        /* go to another cell */
	Process_Status[ID]=TRUE;
        counter++;
      }
      else {
        /* all processes are busy: receive data from one of them */
        MPI_Recv(&err,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,&Receive_Status);
        ID = Receive_Status.MPI_SOURCE;
        Image_cell = Image; counter1 = 0;
        while(err!=counter1) {Image_cell = Image_cell->next; counter1++;}
	/* receive intensities */
   MPI_Recv((Image_cell->tau),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_em),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_sca),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_star),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
        /* make this process available */ 
        Process_Status[ID]=FALSE;
      }
    } while (counter < (*TOTAL_NUMBER_OF_IMAGE_CELLS));

    /* terminate processes and receive what is still to be received */
    for (ID=1; ID<total_MPI_process; ID++){
      if (Process_Status[ID]) {
	/* receive data */
        MPI_Recv(&err,1,MPI_INT,ID,tag,MPI_communicator,&Receive_Status);
        Image_cell = Image; counter1 = 0;
        while(err!=counter1) {Image_cell = Image_cell->next; counter1++;}
	/* receive intensities */
   MPI_Recv((Image_cell->tau),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_em),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_sca),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
   MPI_Recv((Image_cell->Intensity_star),lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
        /* make this process available */ 
        Process_Status[ID]=FALSE;
      }
      /* terminate this process */
      err=-1;
      MPI_Send(&err,1,MPI_INT,ID,tag,MPI_communicator);
    }
  } 
  else {
    /* nonzero process */

    /* receive job from zero process */
    MPI_Recv(&counter,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
    while (counter>=0) {
      Image_cell = Image; counter1 = 0;
      while(counter!=counter1) {Image_cell = Image_cell->next; counter1++;}
      /****** calculate ******/
      /* a point on the source surface */
      r_polar_Int = sqrt(Image_cell->r1 * Image_cell->r2);
      phi_polar_Int = 0.5*(Image_cell->phi1+Image_cell->phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
                                 Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k){
          Image_cell->tau[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,
                           start_point,integral_direction,
                           &(Image_cell->Intensity_em[lambda_k]),
                           &(Image_cell->Intensity_sca[lambda_k]),FALSE);
	  dummy = 1.0;
          Image_cell->Intensity_star[lambda_k] = exp(-Image_cell->tau[lambda_k])*
                 Find_Source_Flux (lambda_k,observer_direction.z,dummy,
				   r_polar_Int,&err);
	  dummy = Rout/Source_Radius;
	  dummy *= dummy; dummy *= 4.0;
          Image_cell->Intensity_star[lambda_k] *= dummy;
        }
      /*******************/
      /* send it to zero process */
      MPI_Send(&counter,1,MPI_INT,0,tag,MPI_communicator);
      MPI_Send((Image_cell->tau),lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send((Image_cell->Intensity_em),lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send((Image_cell->Intensity_sca),lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send((Image_cell->Intensity_star),lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);

      /* receive job from zero process */
      MPI_Recv(&counter,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
    };    

  }
#endif

  printf("......source image: done\n");fflush(stdout);

  /*****************************************************************/
  /* create initial image grid for polar radius > source radius */
  r_polar_step = exp(log(Ri1*Rout/Source_Radius)/N_r_image);
  r_polar1 = Source_Radius; 
  r_polar2 = r_polar_step*Source_Radius;
  r_polar_counter = 1;
  while (r_polar_counter <= N_r_image + N_r_image_1) {

    if (r_polar_counter <= N_r_image) N_phi_image_limit=N_phi_image;
    else N_phi_image_limit=N_phi_image_1;

    for ( phi_polar_counter=0;  phi_polar_counter<=N_phi_image_limit; 
         phi_polar_counter++) {
         
      phi_polar1 = ((double)(phi_polar_counter) - 0.5)*PI/N_phi_image_limit;
      phi_polar2 = ((double)(phi_polar_counter) + 0.5)*PI/N_phi_image_limit;

      /* check for predefined grid refinement */
      if ((phi_polar1>=refine_angle1 && phi_polar2<=refine_angle2) ||
          (phi_polar1>=refine_angle3 && phi_polar2<=refine_angle4) )
        max_refine_angle_N = refine_angle_N;
      else max_refine_angle_N = 1;

      for (counter=0; counter<max_refine_angle_N; counter++) {

	if (max_refine_angle_N>1) {
          phi_polar1 = ((double)(phi_polar_counter) - 0.5 + 
             (double)(counter)/max_refine_angle_N)*PI/N_phi_image_limit;
          phi_polar2 = ((double)(phi_polar_counter) - 0.5 + 
             ((double)(counter)+1.0)/max_refine_angle_N)*PI/N_phi_image_limit;
	}

        /* create an image cell (r_polar1,r_polar2,phi_polar1,phi_polar2) */
        Image_cell = (struct _image_cell_structure *) malloc
                                        (sizeof(struct _image_cell_structure));
        Image->prev->next = Image_cell;
        Image_cell->prev = Image->prev;
        Image_cell->next = Image;
        Image->prev = Image_cell;

        Image_cell->r1 = r_polar1;
        Image_cell->r2 = r_polar2;
        Image_cell->phi1 = phi_polar1; 
        Image_cell->phi2 = phi_polar2;
        Image_cell->Intensity_em   = NULL;
        Image_cell->Intensity_sca  = NULL;
        Image_cell->Intensity_star = NULL;
        Image_cell->tau = NULL;
        /* (limit_POINTS_IN_CELL + 1) will activate the grid refinement */ 
        Image_cell->number_of_visible_points = 0;
        Image_cell->old_number_of_visible_points = limit_POINTS_IN_CELL + 1;

        (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
      } /* end of sub_phi-loop */

    }/* end of phi-loop */

    if (r_polar_counter == N_r_image) {
      /* change the radial step */ 
      r_polar_step = acos(Ri1)/N_r_image_1;
      r_polar1 = Ri1*Rout;
      r_polar2 = cos(acos(Ri1)-r_polar_step)*Rout;
    }
    else {
      if (r_polar_counter < N_r_image) {
	r_polar1 *= r_polar_step;
	r_polar2 *= r_polar_step;
      }
      else {
	r_polar1 = r_polar2;
	r_polar2 = cos(acos(Ri1)-(r_polar_counter+1-N_r_image)*r_polar_step)
                   *Rout;
      }
    }
    r_polar_counter++;
 
  } /* end of r-loop */ 

  printf("......initial image: done\n");

  /* start with the refinement loop */

  refine_grid = TRUE;

  while (refine_grid) {
    /* find out how many 3D vertices is visible through each image cell */
    /* set the vertex and triangle grid */
    start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
    TOTAL_NUMBER_OF_VERTICES = 
                   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;
    TOTAL_NUMBER_OF_TRIANGLES = 
                   grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;
    loop_vertex = start_vertex_list;
    do {
      /* loop over the 3D phi angle */
      for (phi_counter_3D=0; phi_counter_3D<=N_phi_vertex_grid; 
	   phi_counter_3D++){
	phi_angle_3D = phi_counter_3D*PI/N_phi_vertex_grid;

	/* loop over + or - z-axis */
	z_symmetry = 1;
	do {

	  /* determine the 3D point out of phi_angle_3D */
	  dummy_point.x = loop_vertex->x * cos(phi_angle_3D);
	  dummy_point.y = loop_vertex->x * sin(phi_angle_3D);
	  dummy_point.z = z_symmetry*loop_vertex->z;
	  /* transformation into the polar coordinate system */
	  point_3D.x = observer_direction.z*dummy_point.x -
                       observer_direction.x*dummy_point.z;
	  point_3D.y = dummy_point.y;
          point_3D.z = observer_direction.x*dummy_point.x +
                       observer_direction.z*dummy_point.z;
	  /* loop over the image cells in order to find a cell which contains
	     point_3D */
	  Image_cell = Image;
	  do {

            if (Image_cell->old_number_of_visible_points<0) 
	      err = (int)(grid_refinement_factor*limit_POINTS_IN_CELL);
            else err = limit_POINTS_IN_CELL;

	    if (abs(Image_cell->old_number_of_visible_points)>err){
	      dummy = sqrt(point_3D.x*point_3D.x + point_3D.y*point_3D.y);
	      if (Image_cell->r1 <= dummy && dummy <= Image_cell->r2){
		cos_phi1 = dummy*cos(Image_cell->phi1);
                cos_phi2 = dummy*cos(Image_cell->phi2);
	        if ( point_3D.y >= fabs(point_3D.x) && 
                     cos_phi2 <= point_3D.x && point_3D.x <= cos_phi1 )
		  {(Image_cell->number_of_visible_points)++; break;}
		sin_phi1 = dummy*sin(Image_cell->phi1);
                sin_phi2 = dummy*sin(Image_cell->phi2);
	        if ( point_3D.y < fabs(point_3D.x) )
                  if ( (point_3D.x>0.0 && sin_phi1<=point_3D.y 
			&& point_3D.y<=sin_phi2 )
		                       || 
		       (point_3D.x<0.0 && sin_phi2<=point_3D.y
			&& point_3D.y<=sin_phi1 ))
		    {(Image_cell->number_of_visible_points)++; break;}
	      }
	    }
	    Image_cell = Image_cell->next;
	  } while(Image_cell!=Image);

	  z_symmetry *= -1; 
	  if (loop_vertex->z == 0.0)  z_symmetry=1;/* a point on x-axis */
	} while(z_symmetry == -1);

	/* check if we need only one 3D point (a point on the z-axis) */
	if (loop_vertex->x == 0.0) phi_counter_3D=N_phi_vertex_grid+1;
      } /* end of the loop over 3D phi angle */

      loop_vertex = loop_vertex->next;
    } while(loop_vertex!= start_vertex_list);

    /* proceed with refinement, if necessary */
    refine_grid = FALSE;
    Image_cell = Image;
    do {
      if (Image_cell->old_number_of_visible_points<0) 
	/* cells born out of grid refinement have this negative */
        Image_cell->old_number_of_visible_points = 
                                       - Image_cell->number_of_visible_points;
      else 
        Image_cell->old_number_of_visible_points = 
                                         Image_cell->number_of_visible_points;
      /* if this is a cell born out of grid refinement than use higher
         resolution */
      if (Image_cell->old_number_of_visible_points<0) 
	err = (int)(grid_refinement_factor*limit_POINTS_IN_CELL);
      else err = limit_POINTS_IN_CELL;

      /*//////////////////////////////////////////////////
      //if (Image_cell->r1>Source_Radius && Image_cell->r1<200)
      //if ((Image_cell->phi1<(80*PI/180) && Image_cell->phi2>(80*PI/180) &&
      //     Image_cell->phi2-Image_cell->phi1 > 0.2*PI/180) ||
      //    (Image_cell->phi1<(100*PI/180) && Image_cell->phi2>(100*PI/180) &&
      //     Image_cell->phi2-Image_cell->phi1 > 0.2*PI/180)){
      // err = -2;
      // printf("cell[phi1=%lf, phi2=%lf, r1=%lf, r2=%lf]\n",Image_cell->phi1*180/PI,
      //        Image_cell->phi2*180/PI,Image_cell->r1,Image_cell->r2);
      //}
      ////////////////////////////////////////////////////*/
      if (Image_cell->number_of_visible_points > err){
	/* we need to create 9 sub-cells 

 d*d*d*r1=r2 -------------------------
             |  C31  |  C32  | C33   |
             |       |       |       |
      d*d*r1 -------------------------
             |  C21  |  C22  | C23   |
             |       |       |       |
        d*r1 -------------------------
             |  C11  |  C12  | C13   |
             |       |       |       |
         r1  -------------------------
            phi1  D+phi1  2D+phi1  3D+phi1=phi2
	*/
	refine_grid = TRUE;
        Image_cell = _VILOZMAJ_Output_Image_Flux_Split_Cell (Image_cell,Image,
                                   TOTAL_NUMBER_OF_IMAGE_CELLS,Refinement_matrix);
      } /* end of the Image_cell refinement */
      else {
	Image_cell->number_of_visible_points = 0;
	Image_cell = Image_cell->next;
      }

    } while(Image_cell!=Image);

  } /* end of the refinement loop */

  printf("......image refinement: done\n");
  printf("......calculate intensities (# of cells = %d)\n",
         (*TOTAL_NUMBER_OF_IMAGE_CELLS));
  fflush(stdout);

  /* finally, we have the image grid. Calculate the intensities. */
#ifdef _2DD_MPI_
  /***********************************************************/
  /******* multiple processes ********************************/
  MPI_Barrier(MPI_communicator);

  if (my_MPI_rank==0) {
    /**** zero process *****/
    /* set the cell identification system in place */
    Image->number_of_visible_points = (*TOTAL_NUMBER_OF_IMAGE_CELLS);
    Image_cell = Image->next;
    Image_cell->number_of_visible_points = 1;
    Image_cell = Image_cell->next;
    while (Image_cell!=Image) {
      Image_cell->number_of_visible_points = 
                    Image_cell->prev->number_of_visible_points + 1;
      Image_cell=Image_cell->next;
    }

    /* set the cell pointer to the first cell scheduled for analysis */
    for (ID=1; ID<total_MPI_process; ID++) Process_Status[ID]=FALSE;
    Image_cell = Image;
    while (Image_cell->Intensity_star != NULL) 
      {Image_cell=Image_cell->next;}
  
    /* loop as long as there is something to calculate */
    do {

      /* find if there is an idle process */
      for (ID=1; ID<total_MPI_process; ID++)
        if (!Process_Status[ID]) {break;}
      /* if ID is an idle process, give a job to it */
      if (ID<total_MPI_process) {
        /* check if all cells are already used */
        if (Image_cell!=Image) {
	  /* send this cell to the ID process */
          MPI_Send(&(Image_cell->number_of_visible_points),
                   1,MPI_INT,ID,tag,MPI_communicator);
          MPI_Send(&(Image_cell->r1),1,MPI_DOUBLE,ID,tag,MPI_communicator);
          MPI_Send(&(Image_cell->r2),1,MPI_DOUBLE,ID,tag,MPI_communicator);
          MPI_Send(&(Image_cell->phi1),1,MPI_DOUBLE,ID,tag,MPI_communicator);
          MPI_Send(&(Image_cell->phi2),1,MPI_DOUBLE,ID,tag,MPI_communicator);
	  if (Image_cell->Intensity_em==NULL) err = TRUE;
          else err = FALSE;
          MPI_Send(&err,1,MPI_INT,ID,tag,MPI_communicator);

          /* go to the next cell */
          Image_cell=Image_cell->next;
          /* mark this process as busy */
          Process_Status[ID]=TRUE;
	  continue;
        }
        else {
	  /* all cells used, check if we have to exit */
	  for (ID=1; ID<total_MPI_process; ID++) 
            if (Process_Status[ID]) break;
	  if (ID==total_MPI_process) break; /* EXIT !!! */
        }
      } /* end of: if (ID<total_MPI_process) */

      /* all processes are busy: receive data from one of them */
      MPI_Recv(&counter1,1,MPI_INT,MPI_ANY_SOURCE,tag,
               MPI_communicator,&Receive_Status);
      ID = Receive_Status.MPI_SOURCE;
      /* mark this process as free */
      Process_Status[ID]=FALSE;
      /* find the cell under the ordinal number counter1*/
      Image_cell1 = Image;
      do {
        Image_cell1=Image_cell1->next;
      } while (Image_cell1!=Image && counter1!=Image_cell1->number_of_visible_points); 
      /* we need to receive 9 intensities:
         d*d*d*r1=r2 -------------------------
                      |   31  |   32  |  33   |
                      |       |       |       |
               d*d*r1 -------------------------
                      |   21  |   22  |  23   |
                      |       |       |       |
                 d*r1 -------------------------
                      |   11  |   12  |  13   |
                      |       |       |       |
                  r1  -------------------------
                     phi1  D+phi1  2D+phi1  3D+phi1=phi2
      */
      MPI_Recv(Iem11,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem12,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem13,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem21,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      if (Image_cell1->Intensity_em==NULL)
         MPI_Recv(Iem22,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem23,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem31,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem32,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Iem33,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);

      MPI_Recv(Isca11,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca12,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca13,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca21,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      if (Image_cell1->Intensity_em==NULL)
         MPI_Recv(Isca22,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca23,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca31,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca32,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Isca33,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);

      /* find the average intensity */
      for (err=0;err<lambda_Max;err++) {
        Average_I_total[err]  =  Iem11[err]+Iem12[err]+Iem13[err]+
                                 Iem21[err]+           Iem23[err]+
                                 Iem31[err]+Iem32[err]+Iem33[err] ;
        Average_I_total[err]  =  Isca11[err]+Isca12[err]+Isca13[err]+
                                 Isca21[err]+            Isca23[err]+
                                 Isca31[err]+Isca32[err]+Isca33[err] ;
        if (Image_cell1->Intensity_em==NULL) {
          Average_I_total[err] += Iem22[err];
          Average_I_total[err] += Isca22[err];
	}
        else {
          Average_I_total[err] += Image_cell1->Intensity_em[err];
          Average_I_total[err] += Image_cell1->Intensity_sca[err];
          Iem22[err] = Image_cell1->Intensity_em[err];
          Isca22[err]= Image_cell1->Intensity_sca[err];
          Tau22[err]= Image_cell1->tau[err];
	}
      }
      for (err=0;err<lambda_Max;err++) Average_I_total[err] /= 9.0;

      MPI_Recv(Tau11,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau12,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau13,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau21,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      if (Image_cell1->Intensity_em==NULL)
         MPI_Recv(Tau22,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau23,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau31,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau32,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);
      MPI_Recv(Tau33,lambda_Max,MPI_DOUBLE,ID,tag,MPI_communicator,&Status);

      /* decide if this cell needs splitting */
      refine_grid = FALSE;
      r_polar1 = Image_cell1->r1*Image_cell1->r2;
      cos_phi1 = cos(0.5*(Image_cell1->phi1+Image_cell1->phi2));
      dummy_point.z = sqrt(r_polar1)*cos_phi1*observer_direction.x;
      dummy_point.x = sqrt(r_polar1 - dummy_point.z*dummy_point.z);
      if (Image_cell1->r2-Image_cell1->r1 > 
                                   _ZELENJAK_size_limit(dummy_point.x,dummy_point.z))
        for (err=0;err<lambda_Max;err++)
        if (Output_Lambda_Image<0.0 || Out_lambda_k==err || 
            Out_lambda_k1==err || Out_lambda_k2==err)
        if (Average_I_total[err]>0.0) {
          dummy = (Iem11[err] + Isca11[err])/Average_I_total[err];
          if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	  else {
            dummy = (Iem12[err] + Isca12[err])/Average_I_total[err];
            if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	    else {
              dummy = (Iem13[err] + Isca13[err])/Average_I_total[err];
              if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	      else {
                dummy = (Iem21[err] + Isca21[err])/Average_I_total[err];
                if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	        else {
                  dummy = (Iem22[err] + Isca22[err])/Average_I_total[err];
                  if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
        	  else {
                    dummy = (Iem23[err] + Isca23[err])/Average_I_total[err];
                    if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	            else {
                      dummy = (Iem31[err] + Isca31[err])/Average_I_total[err];
                      if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	              else {
                        dummy = (Iem32[err] + Isca32[err])/Average_I_total[err];
                        if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
	                else {
                          dummy = (Iem33[err] + Isca33[err])/Average_I_total[err];
                          if (dummy<Ilow || dummy>Ihigh) {refine_grid = TRUE; break; }
                        }
	              }
	            }
	          }
	        }
	      }
	    }
	  }
	}

      if (refine_grid) {
	/* So, dude, you want to split the cell... OK, split that damn cell!! */

	/* check first if we will need to update Image_cell pointer */
	refine_grid = FALSE;
        if (Image_cell == Image) {
          Image_cell=Image_cell->prev; 
          refine_grid = TRUE;
        }   
        if (Image_cell == Image_cell1) {
          Image_cell=Image_cell->prev;
          refine_grid=TRUE;
        }
        
        /* create new cells */
        Image_cell1 = _VILOZMAJ_Output_Image_Flux_Split_Cell (Image_cell1,Image,
                                    TOTAL_NUMBER_OF_IMAGE_CELLS,Refinement_matrix);
        Image_cell1 = Image;

        if (Refinement_matrix[2][2]) {
          /* sub-cell C33 (d*d*r1,r2, 2D+phi1, phi2) */
          Image_cell1 = Image_cell1->prev;
	  (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
	  for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem33[err];
	    Image_cell1->Intensity_sca[err]=Isca33[err];
	    Image_cell1->tau[err]=Tau33[err];
	  }
	}
        if (Refinement_matrix[1][2]) {
          /* sub-cell C23 (d*r1,d*d*r1, 2D+phi1,phi2) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
	  for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem23[err];
	    Image_cell1->Intensity_sca[err]=Isca23[err];
	    Image_cell1->tau[err]=Tau23[err];
	  }
	}
        if (Refinement_matrix[0][2]) {
          /* sub-cell C13 (r1,d*r1, 2D+phi1, phi2) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
          for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem13[err];
	    Image_cell1->Intensity_sca[err]=Isca13[err];
	    Image_cell1->tau[err]=Tau13[err];
	  }
	}
        if (Refinement_matrix[2][1]) {
          /* sub-cell C32 (d*d*r1,r2, D+phi1, 2D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
          for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem32[err];
	    Image_cell1->Intensity_sca[err]=Isca32[err];
	    Image_cell1->tau[err]=Tau32[err];
	  }
	}
        if (Refinement_matrix[1][1]) {
          /* sub-cell C22 (d*r1,d*d*r1, D+phi1, 2D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
	  for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem22[err];
	    Image_cell1->Intensity_sca[err]=Isca22[err];
	    Image_cell1->tau[err]=Tau22[err];
	  }
	}
        if (Refinement_matrix[0][1]) {
          /* sub-cell C12 (r1,d*r1, D+phi1, 2D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
	  for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem12[err];
	    Image_cell1->Intensity_sca[err]=Isca12[err];
	    Image_cell1->tau[err]=Tau12[err];
	  }
	}
        if (Refinement_matrix[2][0]) {
          /* sub-cell C31 (d*d*r1,r2, phi1, D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
          for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem31[err];
	    Image_cell1->Intensity_sca[err]=Isca31[err];
	    Image_cell1->tau[err]=Tau31[err];
	  }
	}
        if (Refinement_matrix[1][0]) {
          /* sub-cell C21 (d*r1,d*d*r1, phi1, D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
	  for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem21[err];
	    Image_cell1->Intensity_sca[err]=Isca21[err];
	    Image_cell1->tau[err]=Tau21[err];
	  }
	}
        if (Refinement_matrix[0][0]) {
          /* sub-cell C11 (r1,d*r1, phi1, D+phi1) */
          Image_cell1 = Image_cell1->prev;
         (Image->number_of_visible_points)++;
          Image_cell1->number_of_visible_points = Image->number_of_visible_points;
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
          Image_cell1->tau = dvector (0,lambda_Max-1);
          for (err=0;err<lambda_Max;err++) {
            Image_cell1->Intensity_em[err]=Iem11[err];
	    Image_cell1->Intensity_sca[err]=Isca11[err];
	    Image_cell1->tau[err]=Tau11[err];
	  }
	}
        
        /* go the the next cell if necessary */  
        if (refine_grid) {Image_cell=Image_cell->next;}
      }
      else {
	/* a-ha!! You don't want to split the cell! Good! */
	if (Image_cell1->Intensity_em==NULL) 
          Image_cell1->Intensity_em = dvector (0,lambda_Max-1);
	if (Image_cell1->Intensity_sca==NULL) 
          Image_cell1->Intensity_sca= dvector (0,lambda_Max-1);
	if (Image_cell1->tau==NULL) 
          Image_cell1->tau = dvector (0,lambda_Max-1);
        for (err=0;err<lambda_Max;err++) {
          Image_cell1->Intensity_em[err]= (Iem11[err]+Iem12[err]+Iem13[err]+
                                           Iem21[err]+Iem22[err]+Iem23[err]+
                                           Iem31[err]+Iem32[err]+Iem33[err])/9.0;
          Image_cell1->Intensity_sca[err]=(Isca11[err]+Isca12[err]+Isca13[err]+
                                           Isca21[err]+Isca22[err]+Isca23[err]+
                                           Isca31[err]+Isca32[err]+Isca33[err])/9.0;
          Image_cell1->tau[err] = Tau22[err];
	}
      }

    } while (Image_cell!=Image || ID<total_MPI_process);

    /* send -1 (exit) flag to all processes */
    counter = -1;
    for (ID=1; ID<total_MPI_process; ID++)
      MPI_Send(&counter,1,MPI_INT,ID,tag,MPI_communicator);

    /* scale the polar radius with Rout and check the phi angle */
    Image_cell = Image;
    do {
      Image_cell->r1 /= Rout; Image_cell->r2 /= Rout;
      if (Image_cell->phi1 < 0.0) Image_cell->phi1 = 0.0;
      if (Image_cell->phi2 > PI)  Image_cell->phi2 = PI;
      Image_cell = Image_cell->next;
    } while(Image_cell!=Image);

  }
  else {
    /***  non-zero processes ***/

    /* get the first flag from the zero process */
    MPI_Recv(&counter,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
    printf("%.3d: received cell #=%d\n",my_MPI_rank,counter);fflush(stdout);
    while (counter>0) {

      /* receive r1, r2, phi1, phi2 of the parent cell */
      MPI_Recv(&received_r1,1,MPI_DOUBLE,0,tag,MPI_communicator,&Receive_Status);
      MPI_Recv(&received_r2,1,MPI_DOUBLE,0,tag,MPI_communicator,&Receive_Status);
      MPI_Recv(&received_phi1,1,MPI_DOUBLE,0,tag,MPI_communicator,&Receive_Status);
      MPI_Recv(&received_phi2,1,MPI_DOUBLE,0,tag,MPI_communicator,&Receive_Status);
      
      /* do we need subcell[2][2] ? */
      MPI_Recv(&refine_grid,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);

      /* find the intensities */
      r_subcell_step = exp(log(received_r2/received_r1) / 3.0);
      phi_subcell_step = (received_phi2 - received_phi1)/3.0;
     
      /********* subcell [1][1] *********/
      r_polar_Int = sqrt(received_r1 * (r_subcell_step * received_r1));
      phi_polar_Int = 0.5*(received_phi1+(phi_subcell_step+received_phi1));
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
        Tau11[lambda_k] = Iem11[lambda_k] = Isca11[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau11[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem11[lambda_k]),&(Isca11[lambda_k]),FALSE);
      }
      /********* subcell [2][1] *********/
      r_polar_Int = sqrt((r_subcell_step *received_r1) * 
                         (r_subcell_step * r_subcell_step * received_r1));
      phi_polar_Int = 0.5*(received_phi1+(phi_subcell_step+received_phi1));
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau21[lambda_k] = Iem21[lambda_k] = Isca21[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau21[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem21[lambda_k]),&(Isca21[lambda_k]),FALSE);
      }
     /********* subcell [3][1] *********/
      r_polar_Int = sqrt((r_subcell_step * r_subcell_step *received_r1) * 
                         (received_r2));
      phi_polar_Int = 0.5*(received_phi1+(phi_subcell_step+received_phi1));
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau31[lambda_k] = Iem31[lambda_k] = Isca31[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau31[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem31[lambda_k]),&(Isca31[lambda_k]),FALSE);
      }

     /********* subcell [1][2] *********/
      r_polar_Int = sqrt(received_r1 * (r_subcell_step * received_r1));
      phi_polar_Int = 0.5*((phi_subcell_step+received_phi1) +
                           (2*phi_subcell_step+received_phi1));
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau12[lambda_k] = Iem12[lambda_k] = Isca12[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau12[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem12[lambda_k]),&(Isca12[lambda_k]),FALSE);
      }

      if (refine_grid) {
        /********* subcell [2][2] *********/
        r_polar_Int = sqrt((r_subcell_step *received_r1) * 
                         (r_subcell_step * r_subcell_step * received_r1));
        phi_polar_Int = 0.5*((phi_subcell_step+received_phi1) +
                           (2*phi_subcell_step+received_phi1));
        dummy_point.x = r_polar_Int*cos(phi_polar_Int);
        dummy_point.y = r_polar_Int*sin(phi_polar_Int);
        dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
        /* coordinate system transformation */ 
        start_point.x = observer_direction.z*dummy_point.x +
                        observer_direction.x*dummy_point.z;
        start_point.y = dummy_point.y;
        start_point.z = -observer_direction.x*dummy_point.x +
                        observer_direction.z*dummy_point.z;
        /* direction of integration */
        integral_direction.x = - observer_direction.x;
        integral_direction.y = - observer_direction.y;
        integral_direction.z = - observer_direction.z;
        /* intensities */
        for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau22[lambda_k] = Iem22[lambda_k] = Isca22[lambda_k] = 0.0;
          if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
              Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	    Tau22[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem22[lambda_k]),&(Isca22[lambda_k]),FALSE);
	}
     }
     
      /********* subcell [3][2] *********/
      r_polar_Int = sqrt((r_subcell_step * r_subcell_step *received_r1) * 
                         (received_r2));
      phi_polar_Int = 0.5*((phi_subcell_step+received_phi1) +
                           (2*phi_subcell_step+received_phi1));
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau32[lambda_k] = Iem32[lambda_k] = Isca32[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau32[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem32[lambda_k]),&(Isca32[lambda_k]),FALSE);
      }

      /********* subcell [1][3] *********/
      r_polar_Int = sqrt(received_r1 * (r_subcell_step * received_r1));
      phi_polar_Int = 0.5*((2*phi_subcell_step+received_phi1) +
                           received_phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau13[lambda_k] = Iem13[lambda_k] = Isca13[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau13[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem13[lambda_k]),&(Isca13[lambda_k]),FALSE);
      }

      /********* subcell [2][3] *********/
      r_polar_Int = sqrt((r_subcell_step *received_r1) * 
                         (r_subcell_step * r_subcell_step * received_r1));
      phi_polar_Int = 0.5*((2*phi_subcell_step+received_phi1) +
                           received_phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau23[lambda_k] = Iem23[lambda_k] = Isca23[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau23[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem23[lambda_k]),&(Isca23[lambda_k]),FALSE);
      }

     /********* subcell [3][3] *********/
      r_polar_Int = sqrt((r_subcell_step * r_subcell_step *received_r1) * 
                         (received_r2));
      phi_polar_Int = 0.5*((2*phi_subcell_step+received_phi1) +
                           received_phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++){
        Tau33[lambda_k] = Iem33[lambda_k] = Isca33[lambda_k] = 0.0;
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
            Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k)
	  Tau33[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,start_point,
                     integral_direction,&(Iem33[lambda_k]),&(Isca33[lambda_k]),FALSE);
      }

      /* send the results back to the zero process */
      MPI_Send(&counter,1,MPI_INT,0,tag,MPI_communicator);      

      MPI_Send(Iem11,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem12,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem13,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem21,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      if (refine_grid) 
        MPI_Send(Iem22,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem23,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem31,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem32,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Iem33,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      
      MPI_Send(Isca11,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca12,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca13,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca21,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      if (refine_grid) 
        MPI_Send(Isca22,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca23,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca31,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca32,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Isca33,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);

      MPI_Send(Tau11,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau12,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau13,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau21,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      if (refine_grid) 
        MPI_Send(Tau22,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau23,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau31,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau32,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);
      MPI_Send(Tau33,lambda_Max,MPI_DOUBLE,0,tag,MPI_communicator);

      /****************************************
      ########################################
      ****************************************/

      /* get the exit flag from the zero process */
      MPI_Recv(&counter,1,MPI_INT,0,tag,MPI_communicator,&Receive_Status);
      printf("%.3d: received cell #=%d\n",my_MPI_rank,counter);fflush(stdout);
    }

  } /* end of non-zero process */

#else
  /***********************************************************/
  /* this is when we have only one process *******************/
  Image_cell = Image;
  counter = 0;
  do {

  /*printf("Image cell #%d\n",counter);*/

    if (Image_cell->Intensity_em == NULL ) {
      /* these are cells with (Image_cell->r1 > Source_Radius) */
      Image_cell->Intensity_em   = dvector (0,lambda_Max-1);
      Image_cell->Intensity_sca  = dvector (0,lambda_Max-1);
      Image_cell->Intensity_star = NULL;
      Image_cell->tau = dvector (0,lambda_Max-1);
      /* a point on the source surface */
      r_polar_Int = sqrt(Image_cell->r1 * Image_cell->r2);
      phi_polar_Int = 0.5*(Image_cell->phi1+Image_cell->phi2);
      dummy_point.x = r_polar_Int*cos(phi_polar_Int);
      dummy_point.y = r_polar_Int*sin(phi_polar_Int);
      dummy_point.z =sqrt(Rout*Rout-r_polar_Int*r_polar_Int);
      /* coordinate system transformation */ 
      start_point.x = observer_direction.z*dummy_point.x +
                      observer_direction.x*dummy_point.z;
      start_point.y = dummy_point.y;
      start_point.z = -observer_direction.x*dummy_point.x +
                      observer_direction.z*dummy_point.z;
      /* direction of integration */
      integral_direction.x = - observer_direction.x;
      integral_direction.y = - observer_direction.y;
      integral_direction.z = - observer_direction.z;
      /* intensities */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
        if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
                                 Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k){
	  Image_cell->tau[lambda_k] = _VILOZMAJ_Intensity_Integral(lambda_k,
                           start_point,integral_direction,
                           &(Image_cell->Intensity_em[lambda_k]),
                           &(Image_cell->Intensity_sca[lambda_k]) ,FALSE);
        }
    }
    /* scale the polar radius with Rout and check the phi angle */
    Image_cell->r1 /= Rout; Image_cell->r2 /= Rout;
    if (Image_cell->phi1 < 0.0) Image_cell->phi1 = 0.0;
    if (Image_cell->phi2 > PI)  Image_cell->phi2 = PI;


    /*printf("Iem=%E  Isca=%E  \n",Image_cell->Intensity_em[16],
             Image_cell->Intensity_sca[16]);*/

    counter++;

    Image_cell = Image_cell->next;
  } while(Image_cell!=Image);

#endif

  /**************************************************************************/
  /**** this business is finalized by the zero process **********************/
  if (my_MPI_rank==0) {

  printf("IMAGE: total number of image cells = %d\n",
	 (*TOTAL_NUMBER_OF_IMAGE_CELLS));
  fflush(stdout);
  err = _VODENJAK_image_vrml (theta_view,Image);

  /* calculate the total flux */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    total_flux_em[lambda_k]=0.0;
    total_flux_sca[lambda_k]=0.0;
    total_flux_star[lambda_k]=0.0;
  }

  Image_cell = Image;
  do {

    /* dummy = 0.5*(Image_cell->r2 + Image_cell->r1); */
    /* we loose 0.5 because of the factor 2 from symmetry */
    dummy = (Image_cell->r2 + Image_cell->r1);
    dummy *= (Image_cell->r2 - Image_cell->r1);
    dummy *= (Image_cell->phi2 - Image_cell->phi1);
    dummy /= (4.0*PI);

    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
      if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
                                 Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k){
      total_flux_em[lambda_k] +=(Image_cell->Intensity_em[lambda_k]*dummy);
      total_flux_sca[lambda_k]+=(Image_cell->Intensity_sca[lambda_k]*dummy);
      if (Image_cell->Intensity_star!=NULL)
       total_flux_star[lambda_k]+=(Image_cell->Intensity_star[lambda_k]*dummy);
    }

    Image_cell = Image_cell->next;
  } while(Image_cell!=Image);


  /* save the total flux into a file */
if (Output_Lambda_Image<=0.0) {

  dummy = theta_view*180.0/PI;
#ifdef _2DD_MPI_
  sprintf(filename,"%s/Flux-total-%-4.1lfdeg.dat",_LELUYA_MPIOUT_PATH,dummy);
  /*sprintf(filename,"/tmp/Flux-total-%-4.1lfdeg.dat",dummy);*/
#else
  sprintf(filename,"%s/Flux-total-%-4.1lfdeg.dat",_LELUYA_OUTPUT_PATH,dummy);
  /*sprintf(filename,"/tmp/Flux-total-%-4.1lfdeg.dat",dummy);*/
#endif

  flux_file=fopen(filename,"w");

  fprintf(flux_file,"lambda[mic]  lambda*Fem  lambda*Fsca lambda*Fstar  lambda*Ftotal\n");

  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) {
    dummy = total_flux_em[lambda_k] + total_flux_sca[lambda_k] +
            total_flux_star[lambda_k];
    fprintf(flux_file,"%-10.4lf  %E  %E  %E  %E\n",Lambda_Grid[lambda_k],
	    Lambda_Grid[lambda_k]*total_flux_em[lambda_k],
	    Lambda_Grid[lambda_k]*total_flux_sca[lambda_k],
	    Lambda_Grid[lambda_k]*total_flux_star[lambda_k],
            Lambda_Grid[lambda_k]*dummy);
  }

  fclose(flux_file);
} /* end of Output_Lambda_Image<=0.0 */

  /* save the image */
  dummy = theta_view*180.0/PI;
#ifdef _2DD_MPI_
  if (Output_Lambda_Image<=0.0)
    sprintf(filename,"%s/Image-cells-%-4.1lfdeg.dat",_LELUYA_MPIOUT_PATH,dummy);
  /*sprintf(filename,"/tmp/Image-cells-%-4.1lfdeg.dat",dummy);*/
  else
    sprintf(filename,"%s/Image-cells-%-10.4lfmic-%-4.1lfdeg.dat",_LELUYA_MPIOUT_PATH,
            Output_Lambda_Image,dummy);
  /*sprintf(filename,"/tmp/Image-cells-%-10.4lfmic-%-4.1lfdeg.dat",Output_Lambda_Image,dummy);*/
#else
  if (Output_Lambda_Image<=0.0)
    sprintf(filename,"%s/Image-cells-%-4.1lfdeg.dat",_LELUYA_OUTPUT_PATH,dummy);
  /*sprintf(filename,"/tmp/Image-cells-%-4.1lfdeg.dat",dummy);*/
  else
    sprintf(filename,"%s/Image-cells-%-10.4lfmic-%-4.1lfdeg.dat",_LELUYA_OUTPUT_PATH,
            Output_Lambda_Image,dummy);
  /*sprintf(filename,"/tmp/Image-cells-%-10.4lfmic-%-4.1lfdeg.dat",Output_Lambda_Image,dummy);*/
#endif
  flux_file=fopen(filename,"w");

  fprintf(flux_file,"%d\n",(*TOTAL_NUMBER_OF_IMAGE_CELLS));
  if (Output_Lambda_Image>0.0) {
   /* fprintf(flux_file,"lambda[mic]  lambda*Fem  lambda*Fsca lambda*Fstar  lambda*Ftotal\n");*/
    if (Out_lambda_k>=0) {
      dummy = total_flux_em[Out_lambda_k] + total_flux_sca[Out_lambda_k] +
              total_flux_star[Out_lambda_k];
      fprintf(flux_file,"%-10.4lf  %E  %E  %E  %E\n",Output_Lambda_Image,
              Output_Lambda_Image*total_flux_em[Out_lambda_k],
              Output_Lambda_Image*total_flux_sca[Out_lambda_k],
              Output_Lambda_Image*total_flux_star[Out_lambda_k],
              Output_Lambda_Image*dummy);
    }
    else {
	err = _MATH_LIB_log_interpolation (
                   Lambda_Grid[Out_lambda_k1],total_flux_em[Out_lambda_k1],
                   Lambda_Grid[Out_lambda_k2],total_flux_em[Out_lambda_k2],
                   Output_Lambda_Image,&Int_interpolate_em);
	err = _MATH_LIB_log_interpolation (
                   Lambda_Grid[Out_lambda_k1],total_flux_sca[Out_lambda_k1],
                   Lambda_Grid[Out_lambda_k2],total_flux_sca[Out_lambda_k2],
                   Output_Lambda_Image,&Int_interpolate_sca);
	err = _MATH_LIB_log_interpolation (
                   Lambda_Grid[Out_lambda_k1],total_flux_star[Out_lambda_k1],
                   Lambda_Grid[Out_lambda_k2],total_flux_star[Out_lambda_k2],
                   Output_Lambda_Image,&Int_interpolate_star);
	dummy = Int_interpolate_em + Int_interpolate_sca + Int_interpolate_star;
        fprintf(flux_file,"%-10.4lf  %E  %E  %E  %E\n",Output_Lambda_Image,
              Output_Lambda_Image*Int_interpolate_em,
              Output_Lambda_Image*Int_interpolate_sca,
              Output_Lambda_Image*Int_interpolate_em,
              Output_Lambda_Image*dummy);
    }
  }
  Image_cell = Image;
  do {
    fprintf(flux_file,"%E %E %E %E ",Image_cell->r1,Image_cell->r2,
	                              Image_cell->phi1,Image_cell->phi2);
    if (Image_cell->Intensity_star!=NULL) fprintf(flux_file,"1\n");
    else fprintf(flux_file,"0\n");
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
     if (Output_Lambda_Image<0.0 || Out_lambda_k==lambda_k || 
                                 Out_lambda_k1==lambda_k || Out_lambda_k2==lambda_k) {
      fprintf(flux_file,"%E ",Image_cell->tau[lambda_k]);
      fprintf(flux_file,"%E ",Image_cell->Intensity_em[lambda_k]);
      fprintf(flux_file,"%E ",Image_cell->Intensity_sca[lambda_k]);
      if (Image_cell->Intensity_star!=NULL)
	fprintf(flux_file,"%E ",Image_cell->Intensity_star[lambda_k]);
    }
    fprintf(flux_file,"\n");
    Image_cell = Image_cell->next;
  } while(Image_cell!=Image);
  fclose(flux_file);

  free_dvector(total_flux_em,0,lambda_Max-1);
  free_dvector(total_flux_sca,0,lambda_Max-1);
  free_dvector(total_flux_star,0,lambda_Max-1);

  /**** this business was finalized by the zero process **********************/
  /**************************************************************************/
  } /* end of zero process */

#ifdef _2DD_MPI_
  free_ivector(Process_Status,0,total_MPI_process-1);
  free_dvector(Iem11,0,lambda_Max-1);
  free_dvector(Iem12,0,lambda_Max-1);
  free_dvector(Iem13,0,lambda_Max-1);
  free_dvector(Iem21,0,lambda_Max-1);
  free_dvector(Iem22,0,lambda_Max-1);
  free_dvector(Iem23,0,lambda_Max-1);
  free_dvector(Iem31,0,lambda_Max-1);
  free_dvector(Iem32,0,lambda_Max-1);
  free_dvector(Iem33,0,lambda_Max-1);
  free_dvector(Isca11,0,lambda_Max-1);
  free_dvector(Isca12,0,lambda_Max-1);
  free_dvector(Isca13,0,lambda_Max-1);
  free_dvector(Isca21,0,lambda_Max-1);
  free_dvector(Isca22,0,lambda_Max-1);
  free_dvector(Isca23,0,lambda_Max-1);
  free_dvector(Isca31,0,lambda_Max-1);
  free_dvector(Isca32,0,lambda_Max-1);
  free_dvector(Isca33,0,lambda_Max-1);
  free_dvector(Tau11,0,lambda_Max-1);
  free_dvector(Tau12,0,lambda_Max-1);
  free_dvector(Tau13,0,lambda_Max-1);
  free_dvector(Tau21,0,lambda_Max-1);
  free_dvector(Tau22,0,lambda_Max-1);
  free_dvector(Tau23,0,lambda_Max-1);
  free_dvector(Tau31,0,lambda_Max-1);
  free_dvector(Tau32,0,lambda_Max-1);
  free_dvector(Tau33,0,lambda_Max-1);
  free_dvector(Average_I_total,0,lambda_Max-1);
  MPI_Barrier(MPI_communicator);
#endif

  *Image_ = Image;

  return 0;

} /* end of the _VILOZMAJ_Output_Image_Flux */


/****************************************************************************/
/****************************************************************************/
/* subroutine: split a cell into 9 sub-cells ********************************/
struct _image_cell_structure *_VILOZMAJ_Output_Image_Flux_Split_Cell (
                              struct _image_cell_structure *Image_cell,
                              struct _image_cell_structure *Image,
                              int *TOTAL_NUMBER_OF_IMAGE_CELLS,
                              int Refinement_matrix[][3]) {

	/* we need to create 9 sub-cells 

 d*d*d*r1=r2 -------------------------
             |  C31  |  C32  | C33   |
             |       |       |       |
      d*d*r1 -------------------------
             |  C21  |  C22  | C23   |
             |       |       |       |
        d*r1 -------------------------
             |  C11  |  C12  | C13   |
             |       |       |       |
         r1  -------------------------
            phi1  D+phi1  2D+phi1  3D+phi1=phi2
	*/
  /* variables for creation of the sub-cells */
  double r_subcell_step, phi_subcell_step;
  struct _image_cell_structure *Image_subcell;

  r_subcell_step = exp(log(Image_cell->r2/Image_cell->r1) / 3.0);
  phi_subcell_step = (Image_cell->phi2 - Image_cell->phi1)/3.0;

  Refinement_matrix[0][0] = FALSE;
  Refinement_matrix[0][1] = FALSE;
  Refinement_matrix[0][2] = FALSE;
  Refinement_matrix[1][0] = FALSE;
  Refinement_matrix[1][1] = FALSE;
  Refinement_matrix[1][2] = FALSE;
  Refinement_matrix[2][0] = FALSE;
  Refinement_matrix[2][1] = FALSE;
  Refinement_matrix[2][2] = FALSE;

  /* sub-cell C11 (r1,d*r1, phi1, D+phi1) */
  if (Image_cell->phi1 > 0.0) {
     Refinement_matrix[0][0] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                                     (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = Image_cell->r1;
     Image_subcell->r2 = r_subcell_step*Image_cell->r1;
     Image_subcell->phi1 = Image_cell->phi1; 
     Image_subcell->phi2 = phi_subcell_step+Image_cell->phi1;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                           - Image_cell->number_of_visible_points;
     (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
  }
  /* sub-cell C21 (d*r1,d*d*r1, phi1, D+phi1) */
  if (Image_cell->phi1 > 0.0) {
     Refinement_matrix[1][0] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                                   (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = r_subcell_step*Image_cell->r1;
     Image_subcell->r2 = r_subcell_step*r_subcell_step*Image_cell->r1;
     Image_subcell->phi1 = Image_cell->phi1; 
     Image_subcell->phi2 = phi_subcell_step+Image_cell->phi1;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                          - Image_cell->number_of_visible_points;
     (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
  }
  /* sub-cell C31 (d*d*r1,r2, phi1, D+phi1) */
  if (Image_cell->phi1 > 0.0) {
     Refinement_matrix[2][0] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = r_subcell_step*r_subcell_step*Image_cell->r1;
     Image_subcell->r2 = Image_cell->r2;
     Image_subcell->phi1 = Image_cell->phi1; 
     Image_subcell->phi2 = phi_subcell_step+Image_cell->phi1;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                      - Image_cell->number_of_visible_points;
     (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   }
   /* sub-cell C12 (r1,d*r1, D+phi1, 2D+phi1) */
   Refinement_matrix[0][1] = TRUE;
   Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
   Image->prev->next = Image_subcell;
   Image_subcell->prev = Image->prev;
   Image_subcell->next = Image;
   Image->prev = Image_subcell;

   Image_subcell->r1 = Image_cell->r1;
   Image_subcell->r2 = r_subcell_step*Image_cell->r1;
   Image_subcell->phi1 = phi_subcell_step+Image_cell->phi1; 
   Image_subcell->phi2 = 2.0*phi_subcell_step+Image_cell->phi1;
   Image_subcell->Intensity_em   = NULL;
   Image_subcell->Intensity_sca  = NULL;
   Image_subcell->Intensity_star = NULL;
   Image_subcell->tau = NULL;
   Image_subcell->number_of_visible_points = 0;
   Image_subcell->old_number_of_visible_points = 
	            - Image_cell->number_of_visible_points;
   (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   /* sub-cell C22 (d*r1,d*d*r1, D+phi1, 2D+phi1) */
   Refinement_matrix[1][1] = TRUE;
   Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
   Image->prev->next = Image_subcell;
   Image_subcell->prev = Image->prev;
   Image_subcell->next = Image;
   Image->prev = Image_subcell;

   Image_subcell->r1 = r_subcell_step*Image_cell->r1;
   Image_subcell->r2 = r_subcell_step*r_subcell_step*Image_cell->r1;
   Image_subcell->phi1 = phi_subcell_step+Image_cell->phi1; 
   Image_subcell->phi2 = 2.0*phi_subcell_step+Image_cell->phi1;
   Image_subcell->Intensity_em   = NULL;
   Image_subcell->Intensity_sca  = NULL;
   Image_subcell->Intensity_star = NULL;
   Image_subcell->tau = NULL;
   Image_subcell->number_of_visible_points = 0;
   Image_subcell->old_number_of_visible_points = 
	                      - Image_cell->number_of_visible_points;
   (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   /* sub-cell C32 (d*d*r1,r2, D+phi1, 2D+phi1) */
   Refinement_matrix[2][1] = TRUE;
   Image_subcell = (struct _image_cell_structure *) malloc
                         (sizeof(struct _image_cell_structure));
   Image->prev->next = Image_subcell;
   Image_subcell->prev = Image->prev;
   Image_subcell->next = Image;
   Image->prev = Image_subcell;

   Image_subcell->r1 = r_subcell_step*r_subcell_step*Image_cell->r1;
   Image_subcell->r2 = Image_cell->r2;
   Image_subcell->phi1 = phi_subcell_step+Image_cell->phi1; 
   Image_subcell->phi2 = 2.0*phi_subcell_step+Image_cell->phi1;
   Image_subcell->Intensity_em   = NULL;
   Image_subcell->Intensity_sca  = NULL;
   Image_subcell->Intensity_star = NULL;
   Image_subcell->tau = NULL;
   Image_subcell->number_of_visible_points = 0;
   Image_subcell->old_number_of_visible_points = 
	                 - Image_cell->number_of_visible_points;
   (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   /* sub-cell C13 (r1,d*r1, 2D+phi1, phi2) */
   if (Image_cell->phi2 < PI) {
     Refinement_matrix[0][2] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = Image_cell->r1;
     Image_subcell->r2 = r_subcell_step*Image_cell->r1;
     Image_subcell->phi1 = 2.0*phi_subcell_step+Image_cell->phi1; 
     Image_subcell->phi2 = Image_cell->phi2;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                      - Image_cell->number_of_visible_points;
     (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   }
   /* sub-cell C23 (d*r1,d*d*r1, 2D+phi1,phi2) */
   if (Image_cell->phi2 < PI) {
     Refinement_matrix[1][2] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = r_subcell_step*Image_cell->r1;
     Image_subcell->r2 = r_subcell_step*r_subcell_step*Image_cell->r1;
     Image_subcell->phi1 = 2.0*phi_subcell_step+Image_cell->phi1; 
     Image_subcell->phi2 = Image_cell->phi2;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                 - Image_cell->number_of_visible_points;
	  (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   }
   /* sub-cell C33 (d*d*r1,r2, 2D+phi1, phi2) */
   if (Image_cell->phi2 < PI) {
     Refinement_matrix[2][2] = TRUE;
     Image_subcell = (struct _image_cell_structure *) malloc
                              (sizeof(struct _image_cell_structure));
     Image->prev->next = Image_subcell;
     Image_subcell->prev = Image->prev;
     Image_subcell->next = Image;
     Image->prev = Image_subcell;

     Image_subcell->r1 = r_subcell_step*r_subcell_step*Image_cell->r1;
     Image_subcell->r2 = Image_cell->r2;
     Image_subcell->phi1 = 2.0*phi_subcell_step+Image_cell->phi1; 
     Image_subcell->phi2 = Image_cell->phi2;
     Image_subcell->Intensity_em   = NULL;
     Image_subcell->Intensity_sca  = NULL;
     Image_subcell->Intensity_star = NULL;
     Image_subcell->tau = NULL;
     Image_subcell->number_of_visible_points = 0;
     Image_subcell->old_number_of_visible_points = 
	                     -  Image_cell->number_of_visible_points;
     (*TOTAL_NUMBER_OF_IMAGE_CELLS)++;
   }

   /* destroy the parent image cell */
   if (Image_cell->Intensity_em!=NULL) 
       free_dvector(Image_cell->Intensity_em,0,lambda_Max-1);
   if (Image_cell->Intensity_sca!=NULL) 
       free_dvector(Image_cell->Intensity_sca,0,lambda_Max-1);
   if (Image_cell->Intensity_star!=NULL) 
       free_dvector(Image_cell->Intensity_star,0,lambda_Max-1);
   if (Image_cell->tau!=NULL) 
       free_dvector(Image_cell->tau,0,lambda_Max-1);

   Image_subcell = Image_cell->next;
   Image_cell->prev->next = Image_cell->next;
   Image_cell->next->prev = Image_cell->prev;
   free(Image_cell);
   (*TOTAL_NUMBER_OF_IMAGE_CELLS)--;

   return Image_subcell;
}
