/* This function calculates radial profiles of various physical parameters
   on the radial lines with a different theta angle */

#include "VILOZMAJ.h"

int _VILOZMAJ_Radial_Profiles (void){

  int dust_i, lambda_k; /* various counters */

  double theta_line;     /* angle of a radial line */
  int Out_lambda_k1,Out_lambda_k2,Out_lambda_k;/* wavelengths for this radial 
                                                  profile */
  double *temperature;   /* dust temperature at current_point */
  double *old_temperature;   /* dust temperature at previous current_point */
  double u_star, u_scatter, u_emit, u_total; /* energy density */ 
  double u_star1, u_scatter1, u_emit1, u_total1; /* energy density (2nd wavelength) */ 
  int lambda_max_N;    /* grid with the largest number of points */
  struct _3DVector startpoint;/* the first point of a radial line */
  struct _3DVector endpoint;  /* the last point on a radial line */
  struct _3DVector direction; /* direction of a radial line */


  struct _3DVector current_point;/* a point on a radial line during the search for
                                    the physical parameters */
  struct _3DVector old_current_point;
  struct _3DVector current_point_xz;/* ..... cylindrical coordinates */
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _triangle_structure *old_current_triangle;
  struct _triangle_structure *triangle_next; /* next triangle in angular loop */

  double distance_total; /* distance between startpoint and endpoint */
  double current_distance;/* distance between startpoint and current_point */
  double distance_step;  /* a spatial step on a radial line */
  double distance_step_refinement; /* refinement of the integral step */
  int step_function; /* flag which shows if we are crossing a step in Eta */ 

  struct _3DVector next_point; /* point after current_point */
  struct _triangle_structure *next_triangle; /* location of next_point */

  struct _3DVector step_direction; /* direction toward the next angular point */
  double step_distance; /* step to te next angular point */ 

  int err;  /* error flag */

  /* output files */
  FILE *radial_file;
  char filename[256];

  double dummy,dummy1,dummy2;

  temperature = dvector(0,Max_Dust_Type-1);  
  old_temperature = dvector(0,Max_Dust_Type-1);  

  /* find the wavelength for this radial profile */
  Out_lambda_k1=Out_lambda_k2=Out_lambda_k=-1;
  if (Output_Lambda_Image>0.0) {
    /* find the lambda grid number for Output_Lambda_Image */
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
      if (Lambda_Grid[lambda_k]==Output_Lambda_Image) {
        Out_lambda_k = lambda_k;
        break;
      }
    if (Out_lambda_k <0) {
      /* find the lambda grid numbers for interpolation */
      Out_lambda_k2 = 0; 
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
	if (Lambda_Grid[lambda_k]<Output_Lambda_Image) {
	  Out_lambda_k1++;
	  Out_lambda_k2++;
	}
      if (Out_lambda_k1<0 || Out_lambda_k2>=lambda_Max) {
printf("ERROR: The required wavelength (%E) is not part of the wavelength-grid\n",
        Output_Lambda_Image);
        return 1;
      }
      Out_lambda_k = Out_lambda_k1; 
      Out_lambda_k1 = Out_lambda_k2;
      Out_lambda_k2 = -1;
    }
  }

  /* find the most detailed grid */
  lambda_max_N = 0; 
  err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES > err) {
      lambda_max_N = lambda_k;
      err = grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
    }

  /* set the global grid (necessary for some subroutines) */
  TOTAL_NUMBER_OF_VERTICES = 
    grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_VERTICES;
  start_vertex_list = grid_per_lambda[lambda_max_N].start_vertex;
  TOTAL_NUMBER_OF_TRIANGLES = 
    grid_per_lambda[lambda_max_N].TOTAL_NUMBER_OF_TRIANGLES;
  start_triangle_list = grid_per_lambda[lambda_max_N].start_triangle;


  /* the direction and the first point of the radial line (at r=0.5) */
  /*  if (Output_Theta_Image<0.0)
            theta_line = 0.0;
    else
            theta_line = Output_Theta_Image*RAD1; */

  /**************************************************************************/
  /**************************************************************************/
  /***************** radial properties **************************************/
  if (Output_Theta_Image>=0.0) {

  theta_line = Output_Theta_Image*RAD1;

  direction.x = sin(theta_line); direction.y = 0.0; direction.z = cos(theta_line);
  startpoint.x=0.5*direction.x; startpoint.y=0.0; startpoint.z=0.5*direction.z;

  /* find the last point on the line (a point on the Rout sphere)*/
  /* (recall that startpoint.y=0) */
  distance_total=sqrt(startpoint.x*startpoint.x+startpoint.z*startpoint.z);
  dummy = startpoint.x*direction.x+startpoint.z*direction.z;
  dummy = - dummy/distance_total;
  distance_total=distance_total*dummy+
                 sqrt(Rout*Rout-distance_total*distance_total*(1-dummy*dummy));
  endpoint.x=startpoint.x+distance_total*direction.x;
  endpoint.y=             distance_total*direction.y;
  endpoint.z=startpoint.z+distance_total*direction.z;

  /* open the file */
#ifdef _2DD_MPI_
    sprintf(filename,"%s/Radial_Profile_at_theta=%6.4E.dat",_LELUYA_MPIOUT_PATH,(theta_line/RAD1));
#else
    sprintf(filename,"%s/Radial_Profile_at_theta=%6.4E.dat",_LELUYA_OUTPUT_PATH,(theta_line/RAD1));
#endif

  radial_file = fopen(filename,"w");
  fprintf(radial_file,"Leluya: radial profile of the physical properties at %lf mic\n",
                      Output_Lambda_Image);
  fprintf(radial_file,"     R             x             z      ");
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
      fprintf(radial_file,"Temperature_%d[K]   ",dust_i);
  fprintf(radial_file,"Opt.depth   ");
  fprintf(radial_file,"scattering     emission      stellar       total");
  fprintf(radial_file,"\n");

  /* the initial values */
  current_distance = 0.0;
  current_triangle=NULL;
  current_point = startpoint;
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i] = -1.0;

  /* loop over the line of interest */
  while (current_distance<distance_total){

    old_current_point=current_point;
    old_current_triangle=current_triangle; 
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) old_temperature[dust_i]=temperature[dust_i];
    distance_step_refinement = 1; 

    do{ /* loop for detecting dust edges */

	/* find the next  point */ 
	distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,
			     direction,distance_step_refinement,&next_point,&next_triangle);

      /* check if we reached the end_point */
      if (current_distance + distance_step > distance_total) {
        distance_step = distance_total - current_distance;
        current_distance = distance_total;
        current_point = endpoint;
        current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
		                               current_point.y);
        current_point_xz.y = 0.0;
        if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
        else current_point_xz.z = - current_point.z;
        err=_ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
					&current_triangle);
      }
	else {
 	  current_distance += distance_step;
	  current_point = next_point;
	  current_triangle = next_triangle;
	  current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
					      current_point.y);
	  current_point_xz.y = 0.0;
	  if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
	  else current_point_xz.z = - current_point.z;
	}

	/* find the temperature at current_point */
	if (current_triangle==NULL) 
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	    temperature[dust_i]=u_star=u_scatter=u_emit=u_total=-1.0;
	else {
          err = FALSE;
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
	    if ((current_triangle->v1->T_dust[dust_i]>0.0 && 
	        current_triangle->v2->T_dust[dust_i]>0.0 &&
	        current_triangle->v3->T_dust[dust_i]>0.0) &&
                INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
	      temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point_xz.x,current_point_xz.z);
	      if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
              err = TRUE;
	    }
	    else temperature[dust_i]=-1.0;
	  }
	}/* end of temperature inteprolations */
   
	/* check edge crossing */
        step_function=FALSE;
        for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
          if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
            if (temperature[dust_i]*old_temperature[dust_i]<0.0) step_function=TRUE;

        if (step_function && distance_step>1E-10) {
	  current_distance -= distance_step;
	  current_point=old_current_point;
	  current_triangle=old_current_triangle;
	  distance_step_refinement *= 2; 
        }

    } while (step_function && distance_step>1E-10);  /* end of the loop for checking 
                                                        distance step over an edge */

    err = FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        if (temperature[dust_i]>0.0) {err = TRUE;break;}

    /* interpolate the energy density */
    if (err) {
            /* interpolate scattering */
            u_scatter = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v3->identification_number],
                         current_point_xz.x,current_point_xz.z);
            /* interpolate dust emission */
            u_emit = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
            /* interpolate stellar emission */
            u_star = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);

	    if (Out_lambda_k2>=0) {
              /* interpolate scattering1 */
              u_scatter1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v3->identification_number],
                         current_point_xz.x,current_point_xz.z);
              /* interpolate dust emission1 */
              u_emit1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
              /* interpolate stellar emission1 */
              u_star1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
	      /* interpolate the mid value */
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_scatter,
                                          Lambda_Grid[Out_lambda_k1],u_scatter1,
                                          Output_Lambda_Image,&dummy);
              u_scatter=dummy;
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_emit,
                                          Lambda_Grid[Out_lambda_k1],u_emit1,
                                          Output_Lambda_Image,&dummy);
              u_emit=dummy;
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_star,
                                          Lambda_Grid[Out_lambda_k1],u_star1,
                                          Output_Lambda_Image,&dummy);
              u_star=dummy;
	    }
            /* total */
            u_total = u_scatter+u_emit+u_star;
    }
    else u_star=u_scatter=u_emit=u_total=-1.0;

    /* print into the file */
    dummy = sqrt(current_point_xz.x*current_point_xz.x +
	                    current_point_xz.z*current_point_xz.z);
    if (DEBUG.SCREEN_MESSAGE >= 1 ) 
      printf("---radial profile: r=%12.6E\n",dummy);

    fprintf(radial_file,"%12.6E  %12.6E  %12.6E  ",dummy,
                          current_point_xz.x,current_point_xz.z);
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        fprintf(radial_file,"%12.6E  ", temperature[dust_i]);

    /* radial optical depth at Out_lambda_k */
    dummy = _ZELENJAK_tau_over_grid (current_point_xz,startpoint,Out_lambda_k,FALSE);
    if (Out_lambda_k2>=0) {
      dummy1 = _ZELENJAK_tau_over_grid (current_point_xz,startpoint,Out_lambda_k1,FALSE);
      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],dummy,
                                  Lambda_Grid[Out_lambda_k1],dummy1,
                                  Output_Lambda_Image,&dummy2);
      dummy = dummy2;
    }
    fprintf(radial_file,"%12.6E  ", dummy);
    /* print energy */
    fprintf(radial_file,"%12.6E  ", u_scatter);
    fprintf(radial_file,"%12.6E  ", u_emit);
    fprintf(radial_file,"%12.6E  ", u_star);
    fprintf(radial_file,"%12.6E  ", u_total);
    fprintf(radial_file,"\n");

  } /* end of the loop over the radial line */

  fclose(radial_file);

  }
  /**************************************************************************/
  /**************************************************************************/
  /***************** angular properties **************************************/
  else {

    distance_total = -Output_Theta_Image;

  /* open the file */
#ifdef _2DD_MPI_
    sprintf(filename,"%s/Angular_Profile_at_R=%6.4E.dat",_LELUYA_MPIOUT_PATH,
            distance_total);
#else
    sprintf(filename,"%s/Angular_Profile_at_R=%6.4E.dat",_LELUYA_OUTPUT_PATH,
            distance_total);
#endif

  radial_file = fopen(filename,"w");
  fprintf(radial_file,"Leluya: angular profile of the physical properties  at %lf mic\n",
                      Output_Lambda_Image );
  fprintf(radial_file,"  Theta[deg]         x           z       ");
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
      fprintf(radial_file,"Temperature_%d[K]   ",dust_i);
  fprintf(radial_file,"Opt.depth   ");
  fprintf(radial_file,"scattering     emission      stellar       total");
  fprintf(radial_file,"\n");

  /* prepare for the loop */
  theta_line = 0.0;
  current_triangle=NULL;
  do {

    direction.x = sin(theta_line); direction.y=0.0; direction.z=cos(theta_line);

    startpoint.x=0.5*direction.x; startpoint.y=0.0; startpoint.z=0.5*direction.z;

    current_point_xz.x=distance_total*direction.x;
    current_point_xz.y=0.0;
    current_point_xz.z=distance_total*direction.z;

    /* locate the triangle */
    err=_ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
				    &current_triangle);

    /* find the temperature at current_point */
	if (current_triangle==NULL) 
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++)
	    temperature[dust_i]=u_star=u_scatter=u_emit=u_total=-1.0;
	else {
          err = FALSE;
	  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
	    if ((current_triangle->v1->T_dust[dust_i]>0.0 && 
	        current_triangle->v2->T_dust[dust_i]>0.0 &&
	        current_triangle->v3->T_dust[dust_i]>0.0) &&
                INPUT.DUST_DENSITY_TYPE[dust_i]!=99999) {
	      temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point_xz.x,current_point_xz.z);
	      if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
              err = TRUE;
	    }
	    else temperature[dust_i]=-1.0;
	  }
	  /* interpolate the energy density */
	  if (err) {
            /* interpolate scattering */
            u_scatter = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Usca[current_triangle->v3->identification_number],
                         current_point_xz.x,current_point_xz.z);
            /* interpolate dust emission */
            u_emit = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Uem[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
            /* interpolate stellar emission */
            u_star = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k].Ustar[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);

	    if (Out_lambda_k2>=0) {
              /* interpolate scattering1 */
              u_scatter1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Usca[current_triangle->v3->identification_number],
                         current_point_xz.x,current_point_xz.z);
              /* interpolate dust emission1 */
              u_emit1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Uem[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
              /* interpolate stellar emission1 */
              u_star1 = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[Out_lambda_k1].Ustar[current_triangle->v3->identification_number],
			 current_point_xz.x,current_point_xz.z);
	      /* interpolate the mid value */
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_scatter,
                                          Lambda_Grid[Out_lambda_k1],u_scatter1,
                                          Output_Lambda_Image,&dummy);
              u_scatter=dummy;
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_emit,
                                          Lambda_Grid[Out_lambda_k1],u_emit1,
                                          Output_Lambda_Image,&dummy);
              u_emit=dummy;
	      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],u_star,
                                          Lambda_Grid[Out_lambda_k1],u_star1,
                                          Output_Lambda_Image,&dummy);
              u_star=dummy;
	    }
            /* total */
            u_total = u_scatter+u_emit+u_star;
	  }
	  else u_star=u_scatter=u_emit=u_total=-1.0;
	}/* end of temperature and energy inteprolations */

    /* print to file */ 
    if (DEBUG.SCREEN_MESSAGE >= 1 ) 
      printf("---angular profile: theta=%12.6E\n",theta_line/RAD1);

    fprintf(radial_file,"%12.6E  %12.6E  %12.6E  ",theta_line/RAD1,
                          current_point_xz.x,current_point_xz.z);
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
        fprintf(radial_file,"%12.6E  ", temperature[dust_i]);

    /* radial optical depth at Out_lambda_k */
    dummy = _ZELENJAK_tau_over_grid (current_point_xz,startpoint,Out_lambda_k,FALSE);
    if (Out_lambda_k2>=0) {
      dummy1 = _ZELENJAK_tau_over_grid (current_point_xz,startpoint,Out_lambda_k1,FALSE);
      _MATH_LIB_log_interpolation(Lambda_Grid[Out_lambda_k],dummy,
                                  Lambda_Grid[Out_lambda_k1],dummy1,
                                  Output_Lambda_Image,&dummy2);
      dummy = dummy2;
    }
    fprintf(radial_file,"%12.6E  ", dummy);
    /* print energy */
    fprintf(radial_file,"%12.6E  ", u_scatter);
    fprintf(radial_file,"%12.6E  ", u_emit);
    fprintf(radial_file,"%12.6E  ", u_star);
    fprintf(radial_file,"%12.6E  ", u_total);
    fprintf(radial_file,"\n");

    /* increase theta_line */
    if (theta_line < 0.5*PI) {
      /* find step to the next point */
      endpoint.x = endpoint.z = 0.0;
      endpoint.y = 1.0;
      step_direction = _MATH_LIB_3D_vector_product (endpoint,direction);

      step_distance = _ZELENJAK_3Dstep (current_point_xz,current_triangle,
                                       step_direction,1.0,&endpoint,
                                       &triangle_next);
      current_triangle = triangle_next;
 
      step_distance = 1.0 - step_distance*step_distance /
                     (2.0*distance_total*distance_total);

      step_distance = acos(step_distance);

      /* be sure that the step is less than 10deg */
      if (step_distance > 0.174533) step_distance = 0.174533;

      /* increase the angle */ 
      theta_line += step_distance;
      if (theta_line > 0.5*PI) theta_line = 0.5*PI;
    }
    else break;

  } while (theta_line <= 0.5*PI);

  fclose(radial_file);
  }

  free_dvector(old_temperature,0,Max_Dust_Type-1);
  free_dvector(temperature,0,Max_Dust_Type-1);
  return 0;
}
