/* This function is called by the flag -c 403
   It calculates an array from a given window (x1,z1,x2,z2) with 
   a given resolution NxXNz. 

   varaibles: x1 == X1_403, x2 == X2_403, z1 == Z1_403, z2 == Z2_403
              Nx == Nx_403, Nz == Nz_403

   It creates a data file with the following structure:
   
   x[0] z[0]    Temperature
   x[0] z[1]    Temperature
   .......
   x[0] z[Nz]   Temperature
   x[1] z[0]    Temperature
   x[1] z[1]    Temperature
   .......
   x[1] z[Nz]   Temperature
   etc...

   It does the same (in separate files) for temperature errors and 
   energy densities (dust scattering and dust emission).
*/

#include "VILOZMAJ.h"



int _VILOZMAJ_Temperature_Array (int dust_i){

  FILE *ARRAY_FILE;          /* temperature: the flag -c 403 data file */
  FILE *ARRAY_FILE_ERR;      /* temperature error output file */
  FILE *ARRAY_FILE_SCA;      /* dust scattering output file */
  FILE *ARRAY_FILE_EM;       /* dust emission output file */
  char filename[256];
  int plot_type;
  int lambda_k;
  int lambda_max_tau;    /* wavelength with the largest optical depth */
  int err;

  int count_x, count_z;
  double Step_x, Step_z; 
  double x_coord, z_coord;
  double temperature;
  double temp_error;
  double u_sca, u_em, u_star, u_tot;
  struct _triangle_structure *current_triangle; /* location of current_point */

  struct _3DVector startpoint;
  struct _3DVector endpoint;
  double dummy;

  if (INPUT.DUST_DENSITY_TYPE[dust_i]==99999) return 0;

  plot_type = TEMPERATURE_ARRAY_TYPE; /* flag showing: 1=Cartesian 2=Polar coordinates
                                      3=Polar with log(r)*/


  /* find the grid with the largest optical depth (the most detailed grid) */
  lambda_max_tau = 0; 
  dummy = grid_per_lambda[lambda_max_tau].tau;
  for (lambda_k=1; lambda_k<lambda_Max; lambda_k++) 
    if (grid_per_lambda[lambda_k].tau > dummy) {
      lambda_max_tau = lambda_k;
      dummy = grid_per_lambda[lambda_max_tau].tau;
    }

  /* set the global grid (necessary for some subroutines) */
  TOTAL_NUMBER_OF_VERTICES = 
    grid_per_lambda[lambda_max_tau].TOTAL_NUMBER_OF_VERTICES;
  start_vertex_list = grid_per_lambda[lambda_max_tau].start_vertex;
  TOTAL_NUMBER_OF_TRIANGLES = 
    grid_per_lambda[lambda_max_tau].TOTAL_NUMBER_OF_TRIANGLES;
  start_triangle_list = grid_per_lambda[lambda_max_tau].start_triangle;

  /* set the wavelength of output energy densities */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
    if(Output_Lambda_Image<=Lambda_Grid[lambda_k]) break;
  if (lambda_k==lambda_Max) lambda_k=5; /* default */
  printf("OUTPUT AT LAMBDA=%lf\n",Lambda_Grid[lambda_k]);

  /* find the steps in array */
  Step_x = (X2_403  - X1_403) / (Nx_403 );
  Step_z = (Z2_403  - Z1_403) / (Nz_403 );

  /* open the file for temperature */
#ifdef _2DD_MPI_
    sprintf(filename,"%s/Temperature_Array_dust%.2d.dat",_LELUYA_MPIOUT_PATH,
                                                         dust_i);
#else
    sprintf(filename,"%s/Temperature_Array_dust%.2d.dat",_LELUYA_OUTPUT_PATH,
                                                         dust_i);
#endif
   ARRAY_FILE = fopen(filename,"w");
   fprintf(ARRAY_FILE,"X1  X2   Z1  Z2  Nx  Nz\n");
   fprintf(ARRAY_FILE,"%E  %E  %E  %E  %d  %d\n",X1_403,X2_403,Z1_403,Z2_403,
	   Nx_403+1,Nz_403+1);
   fprintf(ARRAY_FILE," X        Z         Temperature [K] \n");

  /* open the file for temperature error*/
#ifdef _2DD_MPI_
    sprintf(filename,"%s/Temperature_Error_Array_dust%.2d.dat",_LELUYA_MPIOUT_PATH,
                                                         dust_i);
#else
    sprintf(filename,"%s/Temperature_Error_Array_dust%.2d.dat",_LELUYA_OUTPUT_PATH,
                                                         dust_i);
#endif
   ARRAY_FILE_ERR = fopen(filename,"w");
   fprintf(ARRAY_FILE_ERR,"X1  X2   Z1  Z2  Nx  Nz\n");
   fprintf(ARRAY_FILE_ERR,"%E  %E  %E  %E  %d  %d\n",X1_403,X2_403,Z1_403,Z2_403,
	   Nx_403+1,Nz_403+1);
   fprintf(ARRAY_FILE_ERR," X        Z         Temperature error[\%] \n");

  /* open the file for scattering energy density */
#ifdef _2DD_MPI_
    sprintf(filename,"%s/U_scattering_Array_dust%.2d.dat",_LELUYA_MPIOUT_PATH,
                                                         dust_i);
#else
    sprintf(filename,"%s/U_scattering_Array_dust%.2d.dat",_LELUYA_OUTPUT_PATH,
                                                         dust_i);
#endif
   ARRAY_FILE_SCA = fopen(filename,"w");
   fprintf(ARRAY_FILE_SCA,"X1  X2   Z1  Z2  Nx  Nz\n");
   fprintf(ARRAY_FILE_SCA,"%E  %E  %E  %E  %d  %d\n",X1_403,X2_403,Z1_403,Z2_403,
	   Nx_403+1,Nz_403+1);
   fprintf(ARRAY_FILE_SCA," X          Z           U_scattering  U_scattering/U_total\n");

  /* open the file for emission energy density */
#ifdef _2DD_MPI_
    sprintf(filename,"%s/U_emission_Array_dust%.2d.dat",_LELUYA_MPIOUT_PATH,
                                                         dust_i);
#else
    sprintf(filename,"%s/U_emission_Array_dust%.2d.dat",_LELUYA_OUTPUT_PATH,
                                                         dust_i);
#endif
   ARRAY_FILE_EM = fopen(filename,"w");
   fprintf(ARRAY_FILE_EM,"X1  X2   Z1  Z2  Nx  Nz\n");
   fprintf(ARRAY_FILE_EM,"%E  %E  %E  %E  %d  %d\n",X1_403,X2_403,Z1_403,Z2_403,
	   Nx_403+1,Nz_403+1);
   fprintf(ARRAY_FILE_EM," X          Z           U_emission  U_emission/U_total \n");

   /* FIND tau=1 SURFACE AND DUST EDGES */
   printf("Skipping the search for tau=zero surface!! ");
   if (plot_type==-3 && dust_i==0) {
     for (count_z=0;count_z<=5*Nz_403; count_z++){
       dummy = RAD1*(Step_z*count_z  + 5*Z1_403);
       startpoint.x = Source_Radius*sin(dummy);
       startpoint.y = 0.0;
       startpoint.z = Source_Radius*cos(dummy);
       endpoint.x = pow(10,X2_403)*sin(dummy);
       endpoint.y = 0.0;
       endpoint.z = pow(10,X2_403)*cos(dummy);
       dummy = _ZELENJAK_tau_over_grid (startpoint,endpoint,Lambda_Total,0);
     }
   }

   /* loop over the array */
   current_triangle=NULL;
   if (plot_type==1) x_coord = X1_403;
   else if (plot_type==2) x_coord = X1_403*sin(RAD1*Z1_403);
   else if (plot_type==3) x_coord = pow(10,X1_403)*sin(RAD1*Z1_403);
   else x_coord = pow(10,X1_403);

   count_x=0;
   do {
     /* printf("TEMPERATURE ARRAY: x=%E\n",x_coord); */
     if (plot_type==1) z_coord = Z1_403;
     else if (plot_type==2) z_coord = X1_403*cos(RAD1*Z1_403);
     else if (plot_type==3) z_coord = pow(10,X1_403)*cos(RAD1*Z1_403);
     else z_coord = Z1_403;

     count_z=0;
     do {
       err=_ZELENJAK_locate_point_in_grid(x_coord,z_coord,&current_triangle);
       if (current_triangle==NULL) { 
	 temperature = 0.0;
         temp_error = 0.0;
       }
       else if (current_triangle->v1->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v1->T_dust[dust_i] > 0.0 &&
                current_triangle->v2->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v2->T_dust[dust_i] > 0.0 &&
		current_triangle->v3->T_dust[dust_i]!=_T99999_ &&
                current_triangle->v3->T_dust[dust_i] > 0.0) {
         /* interpolate the temperature */
	 temperature = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
                                     current_triangle->v1->T_dust[dust_i],
			 current_triangle->v2->x, current_triangle->v2->z,
	                             current_triangle->v2->T_dust[dust_i],
		 current_triangle->v3->x, current_triangle->v3->z,
                                     current_triangle->v3->T_dust[dust_i],
			 x_coord,z_coord);
         /* interpolate the temperature error */
         temp_error = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
                                     current_triangle->v1->T_error[dust_i],
			 current_triangle->v2->x, current_triangle->v2->z,
	                             current_triangle->v2->T_error[dust_i],
		 current_triangle->v3->x, current_triangle->v3->z,
                                     current_triangle->v3->T_error[dust_i],
			 x_coord,z_coord);
         /* interpolate scattering */
         u_sca = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[lambda_k].Usca[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[lambda_k].Usca[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[lambda_k].Usca[current_triangle->v3->identification_number],
  x_coord,z_coord);
  /*
  u_sca = grid_per_lambda[lambda_k].Usca[current_triangle->v1->identification_number]*
          grid_per_lambda[lambda_k].Usca[current_triangle->v2->identification_number]*
          grid_per_lambda[lambda_k].Usca[current_triangle->v3->identification_number];
  u_sca = pow(u_sca,0.33333333);*/


         /* interpolate dust emission */
         u_em = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[lambda_k].Uem[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[lambda_k].Uem[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[lambda_k].Uem[current_triangle->v3->identification_number],
			 x_coord,z_coord);
         /* interpolate stellar emission */
         u_star = _MATH_LIB_plane_interpolation_logZ (
                         current_triangle->v1->x, current_triangle->v1->z,
  grid_per_lambda[lambda_k].Ustar[current_triangle->v1->identification_number],
			 current_triangle->v2->x, current_triangle->v2->z,
  grid_per_lambda[lambda_k].Ustar[current_triangle->v2->identification_number],
		 current_triangle->v3->x, current_triangle->v3->z,
  grid_per_lambda[lambda_k].Ustar[current_triangle->v3->identification_number],
			 x_coord,z_coord);
	 /* total */
         u_tot =  u_sca+u_em+u_star;
       }
       else { temperature = temp_error = u_sca = u_em = u_star = 0.0;u_tot=1.0E30;}
       fprintf(ARRAY_FILE    ,"%E  %E  %E\n",x_coord,z_coord,temperature);
       fprintf(ARRAY_FILE_ERR,"%E  %E  %E\n",x_coord,z_coord,temp_error*100.0);
       fprintf(ARRAY_FILE_SCA,"%E  %E  %E  %E\n",x_coord,z_coord,u_sca,u_sca/u_tot);
       fprintf(ARRAY_FILE_EM ,"%E  %E  %E  %E\n",x_coord,z_coord,u_em,u_em/u_tot);
       count_z++;
       if (plot_type==1) z_coord += Step_z;
       else if (plot_type==2) {
         z_coord = (X1_403+count_x*Step_x)*cos(RAD1*(Z1_403+count_z*Step_z)); 
         x_coord = (X1_403+count_x*Step_x)*sin(RAD1*(Z1_403+count_z*Step_z)); 
       }
       else if (plot_type==3) {
	 z_coord = pow(10,(X1_403+count_x*Step_x))*cos(RAD1*(Z1_403+count_z*Step_z));
         x_coord = pow(10,(X1_403+count_x*Step_x))*sin(RAD1*(Z1_403+count_z*Step_z));
       }
       else {
	 z_coord = z_coord += Step_z;
         x_coord = pow(10,(X1_403+count_x*Step_x));
       }
     } while (count_z <=Nz_403 );
     count_x++;
     if (plot_type==1) x_coord += Step_x;
   } while (count_x <=Nx_403);

  fclose(ARRAY_FILE);
  fclose(ARRAY_FILE_ERR);
  fclose(ARRAY_FILE_SCA);
  fclose(ARRAY_FILE_EM);

  return 0;
}
