/* LELUYA - VILOZMAJ program block */

/* VILOZMAJ block is dealing with the radiative transfer calculation. 
   The grids are given at each wavelength in "grid_per_lambda".
*/

#ifdef _2DD_MPI_
/* There are two types of parallelization implemeted:
     - parallelization over the lambda-grid: the radiative transfer for
       a given wavelength is solved by a single process 
#define _2DD_MPI_lambda_grid 
     - parallelizaton over the spatial-grid: the radiative transfer
       matrix elements are calculated by different processes, where one
       process works on one spatial grid point 
#define _2DD_MPI_spatial_grid
*/
#endif

#include "VILOZMAJ.h"
#include <time.h>

#ifdef _2DD_MPI_
  /* flags that show which processors are available for rad. trans. */
  int *_MPI_Process_Available;
#endif

  /* lambda_jobs */ 
  struct _lambda_job_struct *lambda_jobs; 
  int total_number_of_lambda_jobs;
  int current_lambda_job; /* ID of the lambda_job that has solved==2 */


int _VILOZMAJ_main (void){

  /* One run of _VILOZMAJ_Radiative_Transfer_Solver will solve several wavelengths
     at the same time. Try to put together wavelengths with the same grid 
     and mark them as one job. 
     It may happen that this will be too much for one radiative transfer run. 
     To avoid core dumping, limit the maximum wavelengths per process with
     MAX_LAMBDA_PER_PROCESS (input flag -M , default=5)*/


  struct _lambda_job_struct *current_job; /* a job to solve */
  struct _lambda_job_struct *new_job; /* a new job */
  int job_ID; /* the ID number of a job */
  int NUMBER_OF_VERTICES;

  int lambda_k,lambda_kk; /* counter over wavelengths */
  int p ;       /* counter over the wavelengths in job.run[] */
  int dust_i;

  int total_send, total_receive;

  int err; /* error flag */
  char filename[255];

#ifdef _2DD_MPI_
  int tag = 0;
  int receive_flag;
  MPI_Status Receive_Status, Status;
  int rank_counter;
  int job_counter;
#endif

  /* time */
  time_t sec;
  struct tm *t;
  
#ifdef _2DD_MPI_
  if (my_MPI_rank==0)  _MPI_Process_Available=ivector(0,total_MPI_process-1);
#endif

  /*****************************************************************
   Create a list of lambda_jobs - wavelengths packed together into jobs */

  STEP_FUNCTION_HAZARD_AVOIDANCE = TRUE;
  lambda_jobs = NULL;
  current_job = NULL;
  total_number_of_lambda_jobs = 0;

  /* create the job list */
  do {

    new_job = (struct  _lambda_job_struct *) malloc 
                                   (sizeof(struct _lambda_job_struct));
    new_job->next =NULL;
    new_job->ID=-1;
    new_job->N=0;
    new_job->run=NULL;
    new_job->solved=FALSE; /* flag that shows "not solved yet" */

    p=0;

    /* check if there is a necessity for a job at all */ 
    for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
      if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES>0) {
	p++; 
	lambda_kk=lambda_k;
	break;
      }

    if (p==0) free(new_job);
    else {

      /* find how many wavelengths will belong to this job */
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
	if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES>0 &&
	    grid_per_lambda[lambda_k].start_vertex == 
                        	    grid_per_lambda[lambda_kk].start_vertex && 
	    new_job->N < MAX_LAMBDA_PER_PROCESS) 
	  (new_job->N)++;

      /* create the list of wavelengths */
      new_job->run = ivector(0,new_job->N - 1);
      p=0;
      for (lambda_k=0; lambda_k<lambda_Max; lambda_k++)
	if (grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES>0 &&
	    grid_per_lambda[lambda_k].start_vertex == 
                        	    grid_per_lambda[lambda_kk].start_vertex && 
	    p < MAX_LAMBDA_PER_PROCESS) {

	  new_job->run[p] = lambda_k;
	  grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1;
	  p++;
	}

      /* put the job into the job list */
      new_job->solved=-new_job->N; 
      new_job->ID=total_number_of_lambda_jobs; 
      if (lambda_jobs==NULL) lambda_jobs=new_job;
      else current_job->next = new_job;
      current_job = new_job;
      total_number_of_lambda_jobs++; 
    }

  } while (p>0);

  /* remove the markers from grid_per_lambda */
  for (lambda_k=0; lambda_k<lambda_Max; lambda_k++) 
    grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES *= -1; 

  /*****************************************************************/
  /*****************************************************************/
  /* solve the radiative transfer */
  current_lambda_job =0;
  current_job = lambda_jobs;
#ifdef _2DD_MPI_
  /* mark all processors as available (except #0, of coruse)*/
  if (my_MPI_rank==0) {
    _MPI_Process_Available[0]=FALSE;
    for (p=1;p<total_MPI_process;p++) _MPI_Process_Available[p]=TRUE;
  }
#endif

  /* loop over the lambda_jobs*/ 
  while(current_job!=NULL) {
    if ( current_lambda_job==current_job->ID) {
      /* run the job */
      _VILOZMAJ_Radiative_Transfer_Solver(current_job->N,current_job->run);
    } 
    current_job = current_job->next;
    if (current_lambda_job>=0) current_lambda_job++;
    else current_lambda_job *= -1; /* this means that the current process 
                                      was inverting a matrix and received 
                                      this flag from the zero-process */  
  }

#ifdef _2DD_MPI_
if (my_MPI_rank==0) {
    /* check if all matrix inversions are finished - if not, 
       receive the data */
  current_job = lambda_jobs;
  while(current_job!=NULL) {
    if (current_job->solved<0) {
      /* receive a signal from any process */
      MPI_Recv(&job_ID,1,MPI_INT,MPI_ANY_SOURCE,tag,MPI_communicator,
	       &Receive_Status);
      /* this job_ID flag should be -(p+1) where p=wavelength*/
      p=-job_ID-1;
      /* we need the lambda_job ID number to locate the wavelengths */
      MPI_Recv(&job_ID,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,
	       MPI_communicator,&Status);
      /* locate the job to receive the data */
      current_job = lambda_jobs;
      while(current_job!=NULL && current_job->ID!=job_ID)
	current_job = current_job->next;
      /* receive the data */ 
      for (lambda_kk=0; lambda_kk<current_job->N; lambda_kk++)
        if(lambda_kk==p) {
	  lambda_k = current_job->run[lambda_kk];
	  NUMBER_OF_VERTICES=grid_per_lambda[lambda_k].TOTAL_NUMBER_OF_VERTICES;
	  MPI_Recv(grid_per_lambda[lambda_k].Utot_new,NUMBER_OF_VERTICES,
	    MPI_DOUBLE,Receive_Status.MPI_SOURCE,tag,MPI_communicator,&Status);
        } /* loop over wavelengths in the job */ 
      /* mark this process as availabe */
      _MPI_Process_Available[Receive_Status.MPI_SOURCE]=TRUE;
      /* mark this lambda_job as partially solved */
      current_job->solved += 1;
      /* mark this lambda_job as solved */
      if (current_job->solved==0) current_job->solved==TRUE;
      /* send a signal for the end of lambda-job */
      job_ID = total_number_of_lambda_jobs+1; /* kick that processor out */
      MPI_Send(&job_ID,1,MPI_INT,Receive_Status.MPI_SOURCE,tag,
	       MPI_communicator);
      /* start checking from the beginning */ 
      current_job = lambda_jobs; 
    }
    else 
      current_job = current_job->next;
  }

}
#endif

  /* destroy the job list */
  current_job = lambda_jobs;
  while (current_job!=NULL){
    new_job = current_job->next;
    free_ivector(current_job->run,0,current_job->N - 1);
    free(current_job); 
    current_job=new_job;
  }

  /*****************************************************************/

  if (my_MPI_rank==0) {

    err = _ZELENJAK_Eta_Denominator();
    err = _GOROSTAS_Find_Psi ();
    if (err>0) {STEP_FUNCTION_HAZARD_AVOIDANCE=FALSE;return 1;}
    err = _VILOZMAJ_Find_New_Temperature();
    if (err>0) {STEP_FUNCTION_HAZARD_AVOIDANCE=FALSE;return 1;}
    /* find Source_Radius */
    err = _ZELENJAK_Find_Source_Radius();
    if (err>0) {STEP_FUNCTION_HAZARD_AVOIDANCE=FALSE;return 1;}

#ifdef _2DD_MPI_
   free_ivector(_MPI_Process_Available,0,total_MPI_process-1);
#endif
  }

  STEP_FUNCTION_HAZARD_AVOIDANCE=FALSE;
  return 0;
}

