/* LELUYA - VODENJAK program block */
/* Prototypes of functions for plotting GIF images, VRML worlds, etc.. */

/* This function creates a GIF file */

#include "VODENJAK.h"


int _VODENJAK_plot_image (char image_name[255],
                          double (*func)(double,double,int,int),
                          int image_type,double minX, double maxX, double minY,
                          double maxY, int dust_i, int lambda_k){
  /* input variables:
  image_name   = name of the file without the suffix ".gif" 
  func()       = function which gives the data for plotting
  image_type   = plotting type: 0 = rectangular plot
                                1 = polar plot
                 rectangular plot is an x-y image with the real axes x and z
                 polar plot is an image with the real axes theta and log10(r)
  minX, maxX   = if (image_type=0) then minX and maxX are minimal and maximal
                                        value of the x coordinate
                 if (image_type=1) then minX and maxX are minimal and maximal
                                        value of the theta angle
  minY, maxY   = if (image_type=0) then minY and maxY are minimal and maximal
                                        value of the z coordinate
                 if (image_type=1) then minY and maxY are minimal and maximal
                                        value of the log10(r) value
  dust_i       = dust component
  lambda_k     = wavelength's index
  */

  int gifXsize, gifYsize; /* the real GIF size (resolution) */

  int XedgeL=10, YedgeL=10; /*window of the data image(size:PLOT_IMAGE_SIZE_X */
  int XedgeR, YedgeR;       /* PLOT_IMAGE_SIZE_Y) within the GIF image */

  int GIFx, GIFy;          /* coordinates of the GIF image pixels */
  int yGIF;                /* image is rotated by 180deg: yGIF=gifYsize-GIFy */

  unsigned char color[256][3];   /* colors (RGB values: 0=R 1=G 2=B) */
  int R_value, G_value, B_value; /* RGB color values */

  double point_x,point_z; /* coordinates of a dust point when image_type=0 */
  double point_theta,point_log10r; /* coordinates of a dust point 
                             when image_type=1 */
  double point_value;     /* value at a dust point which will be plotted */

  unsigned char *Line; /* line of numbers sent to the GIF image */
  FILE *PALETTE, *GIF_IMAGE;
  char filename[259];  /* name of a file (either color palette or image) */
  int err;             /* error flag */
  int dummy;           /* dummy variable, no meaning */
  int counter;         /* counter in various loops */

  if (my_MPI_rank!=0) return 0;

  /* check the image type */
  if (image_type<0 || image_type>1) {
    printf("WARNING !!! The GIF image type %d does not exist!!\n",image_type);
    return 1;
  }

  /* set the image parameters */
  XedgeR = XedgeL + PLOT_IMAGE_SIZE_X;
  YedgeR = YedgeL + PLOT_IMAGE_SIZE_Y;
  gifXsize = XedgeR + 3*XedgeL; /* factor 3 comes from the color bar */
  gifYsize = YedgeR + YedgeL;
  Line = cvector(0,gifXsize-1);

  /* open the color palette */
  sprintf(filename,"%s/MALIK/palette.col",_LELUYA_SOURCE_PATH);
  PALETTE = fopen(filename,"rt");
  if (PALETTE == NULL ) {
    printf("WARNING !!! Can not open the color palette file:\n");
    printf("%s    ....continue\n",filename);
    fclose(PALETTE);
    return 1;
  }
  err = fscanf(PALETTE,"%d\n",&dummy);
  /* read the RGB color indeces */
  for (counter=0; counter<256; counter++) {
    err = fscanf(PALETTE,"%d %d %d %d\n",&dummy,&R_value,&G_value,&B_value);
    if (err != 4) {
      printf("WARNING !!! Can not read the color palette file:\n");
      printf("%s    ....continue\n",filename);
      return 1;
    }
    color[counter][0]=R_value;
    color[counter][1]=G_value;
    color[counter][2]=B_value;
  }
  fclose(PALETTE);

  /* open the GIF file */
#ifdef _2DD_MPI_
  sprintf(filename,"%s/%s.gif",_LELUYA_MPIOUT_PATH,image_name);
#else
  sprintf(filename,"%s/%s.gif",_LELUYA_OUTPUT_PATH,image_name);
#endif

  err = EGifOpenFileName(filename);
  if (err != 0) {
    printf("WARNING !!! Can not open the GIF file:\n");
    printf("%s    ....continue\n",filename);
    EGifCloseFile();
    return 1;
  }

  if (DEBUG.SCREEN_MESSAGE >= 1 )
    printf("VODENJAK block: GIF file: %s\n",filename); 

  EGifPutScreenDesc(gifXsize,gifYsize,6,0,8,color);
  EGifPutImageDesc(0,0,gifXsize,gifYsize,8);

  /* go over all GIF image points */
  for (GIFy=0; GIFy<gifYsize; GIFy++){
    yGIF=gifYsize-GIFy-1;
    if (yGIF<YedgeL || yGIF>YedgeR) 
      for (GIFx=0; GIFx<gifXsize; GIFx++) {
        /*if (GIFx>=XedgeL && GIFx<=XedgeR && 
            (yGIF==YedgeL-1 || yGIF==YedgeR+1) ) Line[GIFx]=2;
	    else Line[GIFx]=7; */
        if (yGIF==YedgeL-1 || yGIF==YedgeR+1) Line[GIFx]=2;
        else if (GIFx==XedgeL || GIFx==XedgeR) Line[GIFx]=2;
        else Line[GIFx]=7;
      }
    else
      for (GIFx=0; GIFx<gifXsize; GIFx++) {
        /* black edge */
        if (GIFx < XedgeL-1 ) Line[GIFx]=7;
        else if (GIFx == XedgeL-1 ) Line[GIFx]=2;
        else if (GIFx > XedgeR+1) {
          /* black edges around the color bar */
          if (GIFx < XedgeR+XedgeL || GIFx > gifXsize-XedgeL) Line[GIFx]=7;
          /* color bar */
          else Line[GIFx]=(int)
                          (8.0+172.*(double)(yGIF-YedgeL)/(gifYsize-2*YedgeL));
	}
        else if (GIFx == XedgeR+1) Line[GIFx]=2;
        /* the window for data presentation */
        else {
	  /* if the coordinate system is (x,z) */
          if (image_type==0) {  
            point_x=minX+(double)(GIFx-XedgeL)*(maxX-minX)/(PLOT_IMAGE_SIZE_X);
            point_z=minY+(double)(yGIF-YedgeL)*(maxY-minY)/(PLOT_IMAGE_SIZE_Y);
            point_theta=sqrt(point_x*point_x+point_z*point_z);
	    if (point_theta <= 0.0) { point_x = 0.0; point_z = 0.0;}
	    else {
              point_log10r=pow(10.0,point_theta);
	      point_x = point_log10r*point_x/point_theta;
              point_z = point_log10r*point_z/point_theta;
	    }
	  } /* end of image_type=0 */

	  /* if the coordinate system is (theta, log10(r)) */
          if (image_type==1) {
            point_theta=minX+
                         (double)(GIFx-XedgeL)*(maxX-minX)/(PLOT_IMAGE_SIZE_X);
            point_log10r=minY+
                         (double)(yGIF-YedgeL)*(maxY-minY)/(PLOT_IMAGE_SIZE_Y);
            point_log10r=pow(10.0,point_log10r);
            point_x = point_log10r*sin(point_theta);
            point_z = point_log10r*cos(point_theta);
	  } /* end of image_type=1 */

	  /* find the point value and transform it into a color code */
          point_value=(*func)(point_x,point_z,dust_i,lambda_k);
          if (point_value < 0.0) Line[GIFx]=7;
          else {
            if (point_value > 1.0) point_value = 1.0;
            Line[GIFx]=(int)(8.+172.*point_value);
	  }
	}
      }
     EGifPutLine(Line,gifXsize);
  } /* end of the loop over GIFy */
  EGifCloseFile();

  free_cvector(Line,0,gifXsize-1);
  return 0;
}
