/* LELUYA - VODENJAK program block */
/* Prototypes of functions for plotting GIF images, VRML worlds, etc.. */

/* This function creates GIF and HTML files for visualization of the radiative 
   transfer matrix */

#include "VODENJAK.h"


int _VODENJAK_plot_radiative_transfer_matrix (double **matrixA, double *vectorB,
                                              int lambda_k){
  /* input variables:
  matrixA      = radiative transfer matrix
  vectorB      = radiative transfer emission vector
                 (in the VILOZMAJ block we are solving  A*u=B, where 'u' is the
                  energy density vector)
  lambda_k     = wavelength's index
  */

  int m,n;            /* indexes of the matrixA */
  double point_value; /* a value that is going to be plotted on the image */

  char image_name[300]; /* name of the output files */

  int gifXsize, gifYsize; /* the real GIF size (resolution) */
  double minX, maxX;    /* minimal and maximal value of the GIF's x coordinate */
  double minY, maxY;    /* minimal and maximal value of the GIF's z coordinate */

  int XedgeL=10, YedgeL=10; /*window of the data image(size:PLOT_IMAGE_SIZE_X */
  int XedgeR, YedgeR;       /* PLOT_IMAGE_SIZE_Y) within the GIF image */

  int GIFx, GIFy;          /* coordinates of the GIF image pixels */
  int yGIF;                /* image is rotated by 180deg: yGIF=gifYsize-GIFy */

  unsigned char color[256][3];   /* colors (RGB values: 0=R 1=G 2=B) */
  int R_value, G_value, B_value; /* RGB color values */

  unsigned char *Line; /* line of numbers sent to the GIF image */
  FILE *PALETTE, *GIF_IMAGE;
  char filename[259];  /* name of a file (either color palette or image) */
  int err;             /* error flag */
  int dummy;           /* dummy variable, no meaning */
  int counter;         /* counter in various loops */

  /* set the image parameters */
  minX=0; maxX=TOTAL_NUMBER_OF_VERTICES-1;
  minY=0; maxY=TOTAL_NUMBER_OF_VERTICES-1;
  XedgeR = XedgeL + TOTAL_NUMBER_OF_VERTICES-1;
  YedgeR = YedgeL + TOTAL_NUMBER_OF_VERTICES-1;
  gifXsize = XedgeR + 3*XedgeL; /* factor 3 comes from the color bar */
  gifYsize = YedgeR + YedgeL;
  Line = cvector(0,gifXsize-1);

  if (my_MPI_rank!=0) return 0;

  /* open the color palette */
  sprintf(filename,"%s/MALIK/palette.col",_LELUYA_SOURCE_PATH);
  PALETTE = fopen(filename,"rt");
  if (PALETTE == NULL ) {
    printf("WARNING !!! Can not open the color palette file:\n");
    printf("%s    ....continue\n",filename);
    fclose(PALETTE);
    return 1;
  }
  err = fscanf(PALETTE,"%d\n",&dummy);
  /* read the RGB color indeces */
  for (counter=0; counter<256; counter++) {
    err = fscanf(PALETTE,"%d %d %d %d\n",&dummy,&R_value,&G_value,&B_value);
    if (err != 4) {
      printf("WARNING !!! Can not read the color palette file:\n");
      printf("%s    ....continue\n",filename);
      return 1;
    }
    color[counter][0]=R_value;
    color[counter][1]=G_value;
    color[counter][2]=B_value;
  }
  fclose(PALETTE);

  /* open the GIF file */
 sprintf(image_name,"2D_dusty.rad_trans_matrix.%-6.2lfmic",Lambda_Grid[lambda_k]);
  sprintf(filename,"%s.gif",image_name);
  err = EGifOpenFileName(filename);
  if (err != 0) {
    printf("WARNING !!! Can not open the GIF file:\n");
    printf("%s    ....continue\n",filename);
    EGifCloseFile();
    return 1;
  }

  if (DEBUG.SCREEN_MESSAGE >= 1 )
    printf("VODENJAK block: GIF file: %s\n",filename); 

  EGifPutScreenDesc(gifXsize,gifYsize,6,0,8,color);
  EGifPutImageDesc(0,0,gifXsize,gifYsize,8);

  /* go over all GIF image points */
  for (GIFy=0; GIFy<gifYsize; GIFy++){
    yGIF=gifYsize-GIFy-1;
    if (yGIF<YedgeL || yGIF>YedgeR) 
      for (GIFx=0; GIFx<gifXsize; GIFx++) {
        /*if (GIFx>=XedgeL && GIFx<=XedgeR && 
            (yGIF==YedgeL-1 || yGIF==YedgeR+1) ) Line[GIFx]=2;
	    else Line[GIFx]=7; */
        if (yGIF==YedgeL-1 || yGIF==YedgeR+1) Line[GIFx]=2;
        else if (GIFx==XedgeL || GIFx==XedgeR) Line[GIFx]=2;
        else Line[GIFx]=7;
      }
    else
      for (GIFx=0; GIFx<gifXsize; GIFx++) {
        /* black edge */
        if (GIFx < XedgeL-1 ) Line[GIFx]=7;
        else if (GIFx == XedgeL-1 ) Line[GIFx]=2;
        else if (GIFx > XedgeR+1) {
          /* black edges around the color bar */
          if (GIFx < XedgeR+XedgeL || GIFx > gifXsize-XedgeL) Line[GIFx]=7;
          /* color bar */
          else Line[GIFx]=(int)
                          (8.0+172.*(double)(yGIF-YedgeL)/(gifYsize-2*YedgeL));
	}
        else if (GIFx == XedgeR+1) Line[GIFx]=2;
        /* the window for data presentation */
        else {
	  /* the coordinate system is (n,m) (that is, x<->n, y<->m)*/
          n=(int)(minX+
		  (double)(GIFx-XedgeL)*(maxX-minX)/(TOTAL_NUMBER_OF_VERTICES-1));
          m=gifYsize-yGIF-1;
          m=(int)(minY+
		  (double)(m-YedgeL)*(maxY-minY)/(TOTAL_NUMBER_OF_VERTICES-1));

	  if (matrixA[m][n]==0.0) point_value=0.0;
	  else point_value=log10(matrixA[m][n]*100.0)/8.0 + 1.0;

          if (point_value < 0.0) Line[GIFx]=7;
          else {
            if (point_value > 1.0) point_value = 1.0;
            Line[GIFx]=(int)(8.+172.*point_value);
	  }
	}
      }
     EGifPutLine(Line,gifXsize);
  } /* end of the loop over GIFy */
  EGifCloseFile();

  free_cvector(Line,0,gifXsize-1);
  return 0;
}
