/* LELUYA - ZELENJAK program block */
/* it finds the next integral point (and triangle) from a given current point*/
/* IMPORTANT: this function DOES NOT normalize the direction vector. That is, 
              IT IS ASSUMED THAT the direction vector IS NORMALIZED! */
/* IMPORTANT: this function DOES NOT RECOGNIZE THE SOURCE SURFACE! */
/* IMPORTANT: IF (refinement<=0) THEN this subroutine searches for the triangle
              that contains "startpoint" */

#include "ZELENJAK.h"

double _ZELENJAK_3Dstep (struct _3DVector           startpoint, 
                        struct _triangle_structure *starttriangle,
                        struct _3DVector           direction,
			double                     refinement,
			struct _3DVector           *finalpoint,
			struct _triangle_structure **finaltriangle){

  /* startpoint    = a point in 3D space
     starttriangle = the triangle to which startpoint belongs
     direction     = (normalized) vector which points toward the finalpoint
     refinement    = how many times the min_step should be smaller than default
     finalpoint    = a point that we have to find
     finaltriangle = the triangle to which finalpoint belongs
  */

  double step; /* distance between the startpoint and finalpoint points */
  double min_step; /* minimal step possible */
  double step3D;   /* step that we will return as result */
  struct _3DVector finalpoint_xz; /* cylindrical coordinates */
  double nu1,nu2,nu3; /* these variables help us to find finaltriangle */
  struct _triangle_structure *triangle; /* used in search for finaltriangle */
  int triangle_option; /* this will show which triangles are checked in search
                          for finaltriangle. Options:
                          0 = starttriangle;  
                          1 = starttriangle->v1->triangles;
                          2 = starttriangle->v2->triangles;
                          3 = starttriangle->v3->triangles;
		       */
  struct _vertex_triangles *vertex_triangle;/* used in search for finaltriangle
                              among the triangles adjacent to starttriangle */
  double dust_density_flag; /* this flag shows if the dust exists or not 
                               at a given point */
  int dust_i;   /* dust types counter */
  int err;      /* error flag */  

  /* it looks like this:  finalpoint  =  startpoint+step*direction

                direction
   finalpoint O<----------O startpoint
               \         /
                \       /
                 \     /
                  \   /
                   \ /
                    O
                   origin of the coordinate system

  */   

  *finaltriangle = starttriangle;
  (*finalpoint).x = startpoint.x;
  (*finalpoint).y = startpoint.y;
  (*finalpoint).z = startpoint.z;

  if (refinement>0) {
    /* let's see if we can find anything at all! */
    if (direction.x==0.0 && direction.y==0.0 && direction.z==0.0) return 0.0;

    /* start with the search for finalpoint */

    min_step = _MATH_LIB_distance(0.0,0.0,startpoint.x,startpoint.y);
    min_step = _ZELENJAK_size_limit(min_step,startpoint.z)*_MIN_VERTEX_DISTANCE_;

    if (starttriangle!=NULL) step = starttriangle->integral_step / refinement;
    else step = min_step;

    min_step /= refinement;
  }

  do {

    if (refinement>0) {
      /* calculate finalpoint */
      (*finalpoint).x = startpoint.x + step * direction.x;
      (*finalpoint).y = startpoint.y + step * direction.y;
      (*finalpoint).z = startpoint.z + step * direction.z;
    }

    /* finalpoint in cylindrical coordinates */
    finalpoint_xz.x = _MATH_LIB_distance(0.0,0.0,(*finalpoint).x,
					 (*finalpoint).y);
    finalpoint_xz.y = 0.0;
    if ((*finalpoint).z >= 0.0) finalpoint_xz.z = (*finalpoint).z;
    else finalpoint_xz.z = - (*finalpoint).z;

    /* if the dust density at finalpoint_xz is zero then we know that there
       is no a grid triangle at that point */
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      dust_density_flag = Dust_Density (finalpoint_xz.x,finalpoint_xz.z,
					dust_i,NULL,&err);
      if (dust_density_flag!=0.0) break;
    }

    /* ------ find finaltriangle ------ */
    *finaltriangle = NULL;
    if (starttriangle!=NULL && dust_density_flag!=0.0) {

      /* check the starttriangle first */
      triangle = starttriangle;
      if (finalpoint_xz.x >= triangle->bound_box_down_left_x &&
          finalpoint_xz.x <= triangle->bound_box_up_right_x  &&
          finalpoint_xz.z >= triangle->bound_box_down_left_z &&
          finalpoint_xz.z <= triangle->bound_box_up_right_z) {
        nu1=_MATH_LIB_nu_value(triangle->v1->x,triangle->v1->z,
	   		     triangle->v2->x,triangle->v2->z,
		  	     finalpoint_xz.x,finalpoint_xz.z);
        nu2=_MATH_LIB_nu_value(triangle->v2->x,triangle->v2->z,
			     triangle->v3->x,triangle->v3->z,
			     finalpoint_xz.x,finalpoint_xz.z);
        nu3=_MATH_LIB_nu_value(triangle->v3->x,triangle->v3->z,
			     triangle->v1->x,triangle->v1->z,
			     finalpoint_xz.x,finalpoint_xz.z);
	if (nu1==-2.0) nu1 *= -1;
	if (nu2==-2.0) nu2 *= -1;
	if (nu3==-2.0) nu3 *= -1;
	if (0.0<=nu1 && 0.0<=nu2 && 0.0<=nu3) {
	/* bingo! That's the triangle! */
          *finaltriangle = triangle;
        }
      }

      /* check the triangles around starttriangle */
      if (*finaltriangle==NULL) {

	triangle_option = 1;
	do {
	  /* select the triangle list at one of the vertices */ 
	  switch (triangle_option) {
	  case 1: vertex_triangle = starttriangle->v1->triangles; break;
	  case 2: vertex_triangle = starttriangle->v2->triangles; break;
	  case 3: vertex_triangle = starttriangle->v3->triangles; break;
	  default: break;
	  }

	  /* look for finaltriangle */ 
	  do {
	    triangle = vertex_triangle->triangle;
	    if (triangle!=starttriangle)
              if (finalpoint_xz.x >= triangle->bound_box_down_left_x &&
                  finalpoint_xz.x <= triangle->bound_box_up_right_x  &&
                  finalpoint_xz.z >= triangle->bound_box_down_left_z &&
                  finalpoint_xz.z <= triangle->bound_box_up_right_z) {
		nu1=_MATH_LIB_nu_value(triangle->v1->x,triangle->v1->z,
				       triangle->v2->x,triangle->v2->z,
				       finalpoint_xz.x,finalpoint_xz.z);
		nu2=_MATH_LIB_nu_value(triangle->v2->x,triangle->v2->z,
				       triangle->v3->x,triangle->v3->z,
				       finalpoint_xz.x,finalpoint_xz.z);
		nu3=_MATH_LIB_nu_value(triangle->v3->x,triangle->v3->z,
				       triangle->v1->x,triangle->v1->z,
				       finalpoint_xz.x,finalpoint_xz.z);
		if (0.0<=nu1 && 0.0<=nu2 && 0.0<=nu3) {
		  /* bingo! That's the triangle! */
		  *finaltriangle = triangle;
		}
	      }

	    vertex_triangle=vertex_triangle->next;
	  } while (vertex_triangle!=NULL && *finaltriangle==NULL);

	  triangle_option++;
	} while (triangle_option<4 && *finaltriangle==NULL);

      }/* end of checking the triangles around starttriangle */
    } /* end of "if (starttriangle!=NULL)" */

    /* if we can not find finaltriangle among starttriangle and startriangle's
       neighbors then make a general search through the grid */
    if (*finaltriangle==NULL && dust_density_flag!=0.0) {
      triangle_option = _ZELENJAK_locate_point_in_grid(finalpoint_xz.x,
	  					    finalpoint_xz.z,&triangle);
      *finaltriangle = triangle;
    }
    /* end of ------ find *finaltriangle ------ */

    if (refinement>0) {
      /* memorize the step */
      step3D = step;
      /* next step */
      step /= 2.0;
    }
    else {
      /* this will kick us from the loop */
      step=0.0; step3D=0.0; min_step=1.0;
    }

  } while (
       (*finaltriangle!=NULL && starttriangle==NULL && step>0.01*min_step) ||
       (*finaltriangle==NULL && starttriangle!=NULL && step>0.01*min_step));

  return step3D;

} /* end of _ZELENJAK_step3D */
