/* LELUYA - ZELENJAK program block */
/* calculation of the dust density at a given point 
   ERROR FLAG CODES: 
     err = 1  : x or z or sqrt(x*x+z*z) is out of range 
                (this is a message, not an error!!)
     err = 2  : dust compontent 'dust_counter' doesn't exist
     err = 3  : dust density type INPUT.DUST_DENSITY_TYPE[dust_counter]
                doesn't exist
     err = 4  : density < 0.0 
*/

#include "ZELENJAK.h"

int _ZELENJAK_Dust_Density_point_in_triangle (double x, double z, 
                     struct _triangle_structure *interpol_triangle);

double Dust_Density (double x, double z, int dust_counter, 
                     struct _triangle_structure *interpol_triangle, int *err){

  /* x, y         = coordinates of the point where we need the dust density 
     dust_counter = type of the dust 
     interpol_triangle = interpolation triangle when -v flag (log density) is 
                         active (this triangle has to be from the vertex grid
                         used in Density_2D_Grid)
     err          = error flag 
     informations about the parameters used for calculation of the dust density 
     are stored in the variable Dust_Density_Param[] */

  double density,density_disk;  /* this is the final result which is then sent by 'return'*/
  double param1,param2,param3,param4,param5,param6,param7,param8,param9;
                    /* various parameters used for calculation
                      of the dust density */
  struct _triangle_structure *temp_triangle,*temp2_triangle;
  int dust_i, flag,triangle_option;
  struct _vertex_triangles *vertex_triangle;

  /* if nothing happen, the error flag should be zero */
  *err = 0;

  /* check if the dust component counter is OK */
  if (dust_counter < 0 || dust_counter > Max_Dust_Type) {
    if (DEBUG.SCREEN_MESSAGE >= 1 )
      printf("DUST DENSITY ERROR: dust #%d doesn't exist\n",dust_counter);
    *err = 2;
    return 0.0;
  }

  /* check if x or z are out of the computational domain */
  /* this is not an error, but only a message */
  if ( x<0 ) { *err = 1; x *= -1.0;}
  if ( z<0 ) { *err = 1; z *= -1.0;}
  /* check if r>Rout+0.0001 (this number +0.0001 is just a numerical error) */
  if ( x*x+z*z> (Rout+0.0001)*(Rout+0.0001) ) { *err = 1; return 0.0;}

  /* there are different density models */
  /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==99999) {
    /* Empty Space: this is just empty space used for the grid generation */
    param1 = sqrt(x*x+z*z);
    if (param1>=Dust_Density_Param[dust_counter].param) 
      density = _Density99999_;
    else 
      density = 0.0; 
    *err = 0;
    return density;
  }
  /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==1) {
    /* Welcome to Trinity model ! 
       This model is used by Dejan Vinkovic for testing the code
       and it doesn't have a fixed structure. DO NOT USE THIS MODEL !! */
   
    /* currently I'm using: D*(1+sin(A*r)*sin(b*theta))/2 
       where: r=sqrt(x*x+z*z), cos(theta)=z/r, A=a*pi/Rout
       parameters: 
             D = Dust_Density_Param[dust_counter].param
             a = Dust_Density_Param[dust_counter].next->param
             b = Dust_Density_Param[dust_counter].next->next->param
    */

    /* param1 = D; param2 = A; param3 = b*theta;  */

/*    param4 = sqrt(x*x+z*z);
    param1 = Dust_Density_Param[dust_counter].param;
    param2 = PI*Dust_Density_Param[dust_counter].next->param / Rout;
    param3 = acos(z/param4);
    param3 *= Dust_Density_Param[dust_counter].next->next->param;
    
    density = param1 *(1.0+ sin(param2 * param4) * sin(param3))*0.5;
    if (density < 0.0) {*err=4; return 0.0;} 

    if (x>0.6*Rout && x<0.9*Rout && z>0.1*Rout && z<0.4*Rout) density = 0.0;
*/

    /* this one is a power law 
    param4 = x*x+z*z;
    density = 1/param4; */

    /* this one is a disk */
    /*param4 = sqrt(x*x+z*z);
    param3 = PI/2.0 - acos(z/param4);
    density = exp( - 4.0*param3*param3)/param4;
    if (density<1.0E-5) density = 1.0E-5;
    if (param3 > (PI/18)*(4.0*param4+Rout-5.0)/(Rout-1.0))  density = 0.0;
    */

   /*  umbrella-model */
/*
   param4 = sqrt(x*x+z*z);
   param3 = acos(z/param4)+(0.25*PI*param4/Rout);
   if (param3<PI/3.0) density = 0.0;
   else {
     param3 = PI/2.0 - param3;
     density = cos(2.0*param3)*exp(-param3)/param4;
   }
   if (density < 0.0) density *= -1.0;
*/

   /* trinity - model */
/*
   param4 = sqrt(x*x+z*z);
   param3 = PI/2.0 - acos(z/param4);
   density = cos(2.0*param3)*exp(-param3);
   if (density < 0.0) density *= -1.0;

*/


    /* A simple disk model defined with the angle of cutoff cone and
       outer radius

       the dust exists for Theta>=Theta_cone and R<Rdisk
       dust density = a*rho^(-p)

       parameters:  
       cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       a     = Dust_Density_Param[dust_counter].next->param
       -p    = Dust_Density_Param[dust_counter].next->next->param
       Rdisk = Dust_Density_Param[dust_counter].next->next->next->param
       Rin   = Dust_Density_Param[dust_counter].next->next->next->next->param
    */

    /* flaring  */
    param7 = 2.121320344*pow((x-0.01),1.25)/sqrt(612.);

    param1 = sqrt(x*x+z*z);
    param2 = cos(RAD1*Dust_Density_Param[dust_counter].param);
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = - Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;

    /* use the ordinary density profile */
    if (z>param7 && param1 < param5) {
      if ( param1*param2>=z )
        density = param3*pow(param1,param4);
      else density = 0.0;
    }
    else density = 0.0;

    *err = 0;
    return density;
  }

   /***********************************************************************/
  /***********************************************************************/  
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==2) {
    /* Welcome to Smooth_Sphere model! */
    /* density \propto a*r^(-p)
       where: r=sqrt(x*x+z*z)
       parameters: a = Dust_Density_Param[dust_counter].param
       parameters: p = Dust_Density_Param[dust_counter].next->param
    */

    param1 = sqrt(x*x+z*z);
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = - Dust_Density_Param[dust_counter].next->param;
    density = param2*pow(param1,param3);
    if (density < 0.0) density = 0.0;

    *err = 0;
    return density;
  }

  /***********************************************************************/
  /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==3) {
    /* Welcome to Ellipsoid model !
       Introduced by Zeljko Ivezic & Dejan Vinkovic
                                   Princeton, September 21, 2001

       dust density = 1 / (X*X + e*Z*Z)^(n/2)
       parameters: n    = Dust_Density_Param[dust_counter].param
                   e    = Dust_Density_Param[dust_counter].next->param
    */

    param1 = 0.5*Dust_Density_Param[dust_counter].param;
    param2 = Dust_Density_Param[dust_counter].next->param;

    density = x*x + param2*z*z;
    density = 1.0 / pow(density,param1);

    *err = 0;
    return density;
  }

  /***********************************************************************/
  /***********************************************************************/

  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==4) {
    /* Welcome to the flared disk model with sharp surface*/
    /* this is a flared disk with the density:
         Rho = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m)
       and the surface z=1.5*h0*x^m

       parameters:  
       param2= p  = Dust_Density_Param[dust_counter].param
       param3= m  = Dust_Density_Param[dust_counter].next->param
       param4= h0 = Dust_Density_Param[dust_counter].next->next->param
       param5=Rho0= Dust_Density_Param[dust_counter].next->next->next->param
    */

    param2 = - Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;

    /* flaring  */
    param6 = pow((x-0.01),param3);
    /* density */ 
    if (z<=2.121320344*param6/sqrt(param4)){
      param6 = pow(x,(2*param3));
      density=param5*pow(x,(0.5*param2))*exp(-param4*z*z/param6);
    }
    else density=0.0;

    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==5) {
    /* A spherical shell without a cone: "Cone model"
       if Theta_cone>=0 then the dust exists for Theta>=Theta_cone
       else the dust exists for Theta<Theta_cone

       dust density = a*rho^(-p)

       parameters:  
       param2=cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       param3=    a     = Dust_Density_Param[dust_counter].next->param
       param4=    -p    = Dust_Density_Param[dust_counter].next->next->param
    */
    
    param1 = sqrt(x*x+z*z);
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = - Dust_Density_Param[dust_counter].next->next->param;

/*    if (param2>0.0 && param2<0.9998) param2 += 0.0002; */
/*    else if (param2<0.0 && param2>-0.9998) param2 -= 0.0002; */ 
    density = 0.0;
    if (param2 >= 0.0){
      if ( param1*param2>=z )
          density = param3*pow(param1,param4);
      /*else density = 1.0E-30; */
      else density = 0.0;
    }
    if (param2 < 0.0) {
      if ( param1*(-param2)<=z )
          density = param3*pow(param1,param4);
      else density = 0.0;
    }
    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==6) {
    /* A spherical shell without a cone: "Finite Cone model"
       if Theta_cone<0 then the dust exists for Theta<Theta_cone AND 
       radius < Rcone, else it exists outside this region

       dust density = a*rho^(-p)

       parameters:  
       param2=cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       param3=    a     = Dust_Density_Param[dust_counter].next->param
       param4=    -p    = Dust_Density_Param[dust_counter].next->next->param
       param5= Rcone = Dust_Density_Param[dust_counter].next->next->next->param
    */
    
    param1 = sqrt(x*x+z*z);
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = - Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;

    density = 0.0;
    if (param2 >= 0.0){
      if ( param1*param2>=z || param1>=param5)
          density = param3*pow(param1,param4);
      else density = 1.0E-30;
    }
    if (param2 < 0.0) {
      if ( param1*(-param2)<=z && param1<param5)
          density = param3*pow(param1,param4);
      else density = 0.0;
    }
    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==7) {
    /* A simple disk model defined with the angle of cutoff cone and
       outer radius

       the dust exists for Theta>=Theta_cone and R<Rdisk
       dust density = a*rho^(-p)

       parameters:  
       cos(Theta_cone)= Dust_Density_Param[dust_counter].param
       a     = Dust_Density_Param[dust_counter].next->param
       -p    = Dust_Density_Param[dust_counter].next->next->param
       Rdisk = Dust_Density_Param[dust_counter].next->next->next->param
       Rin   = Dust_Density_Param[dust_counter].next->next->next->next->param
    */
   
    param1 = sqrt(x*x+z*z);
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = - Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;
    param6 = Dust_Density_Param[dust_counter].next->next->next->next->param;

    if (STEP_FUNCTION_HAZARD_AVOIDANCE) {
      /* be careful about the step in the density profile */
      if (param1 < param5*1.0001 && param1*1.0001 > param6) {
        if ( param1*param2>=z-0.0001 )
           density = param3*pow(param1,param4);
        else density = 0.0;
      }
      else density = 0.0;
    }
    else {
      /* use the ordinary density profile */
      if (param1 < param5 && param1 >= param6) {
        if ( param1*param2>=z )
           density = param3*pow(param1,param4);
        else density = 0.0;
      }
      else density = 0.0;
    }

    *err = 0;
    return density;
  }

  /***********************************************************************/
  /***********************************************************************/

  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==8) {
    /* Welcome to the flared disk model */
    if (DEBUG.DENSITY_UPDATE>=0 && Density_2D_Grid.triangle!=NULL && 
        DEBUG.INPUT_DENSITY_UPDATE>=0) {
      /* to speed up, use empty space for R<1 */
      if (x*x+z*z<0.9999) {*err = 0;return 0.0;}
      /* interpolate from the triangle */
      if (interpol_triangle!=NULL) 
        if (_ZELENJAK_Dust_Density_point_in_triangle(x,z,interpol_triangle)){
          if (interpol_triangle->v1->T_dust[dust_counter]>0.0 &&
              interpol_triangle->v2->T_dust[dust_counter]>0.0 &&
              interpol_triangle->v3->T_dust[dust_counter]>0.0) {
	    for (dust_i=0; dust_i<Density_2D_Grid.dust; dust_i++)
	      if (Density_2D_Grid.dust_types[dust_i]==dust_counter) {
                /* interpolate */
                density = _MATH_LIB_plane_interpolation(
                            interpol_triangle->v1->x,interpol_triangle->v1->z,
       Density_2D_Grid.density[dust_i][interpol_triangle->v1->identification_number],
                            interpol_triangle->v2->x,interpol_triangle->v2->z,
       Density_2D_Grid.density[dust_i][interpol_triangle->v2->identification_number],
                            interpol_triangle->v3->x,interpol_triangle->v3->z,
       Density_2D_Grid.density[dust_i][interpol_triangle->v3->identification_number],
                            x,z);
                density = exp(density);
                return density;
		break;
	      }
	  }
	  else {*err = 0;return 0.0;}
	}
      /* forget the input triangle, use Density_2D_Grid.triangle */
      /* locate the triangle first */

      temp2_triangle = Density_2D_Grid.triangle;
      flag = _ZELENJAK_point_in_triangle(x,z,Density_2D_Grid.triangle);
      if (!flag){
	/* check the first neighbors */
	triangle_option = 1;
	do {
	  /* select the triangle list at one of the vertices */ 
	  switch (triangle_option) {
	    case 1: vertex_triangle = Density_2D_Grid.triangle->v1->triangles; break;
	    case 2: vertex_triangle = Density_2D_Grid.triangle->v2->triangles; break;
	    case 3: vertex_triangle = Density_2D_Grid.triangle->v3->triangles; break;
	    default: break;
	  }
	  do {
            /* check the neighboring triangle */ 
	    temp_triangle = vertex_triangle->triangle;
	    flag = _ZELENJAK_point_in_triangle(x,z,temp_triangle);
	    vertex_triangle=vertex_triangle->next;
	  } while (vertex_triangle!=NULL && (!flag));
	  triangle_option++;
	} while (triangle_option<4 && (!flag));

	if (flag) {
	  Density_2D_Grid.triangle = temp_triangle;
	} else {
	  /* no way out: check all triangles */
          temp_triangle = start_triangle_list;
          start_triangle_list = Density_2D_Grid.triangle->next; 
          _ZELENJAK_locate_point_in_grid(x,z,&(Density_2D_Grid.triangle));
          start_triangle_list = temp_triangle;
        }
      }

      if (Density_2D_Grid.triangle==NULL) {
        *err = 0;
        Density_2D_Grid.triangle=temp2_triangle;
        return 0.0;
      }
      else {
        temp_triangle = Density_2D_Grid.triangle;
        if (temp_triangle->v1->T_dust[dust_counter]>0.0 &&
            temp_triangle->v2->T_dust[dust_counter]>0.0 &&
            temp_triangle->v3->T_dust[dust_counter]>0.0) {
          for (dust_i=0; dust_i<Density_2D_Grid.dust; dust_i++)
	    if (Density_2D_Grid.dust_types[dust_i]==dust_counter) {
              /* interpolate */
              density = _MATH_LIB_plane_interpolation(
                            temp_triangle->v1->x,temp_triangle->v1->z,
       Density_2D_Grid.density[dust_i][temp_triangle->v1->identification_number],
                            temp_triangle->v2->x,temp_triangle->v2->z,
       Density_2D_Grid.density[dust_i][temp_triangle->v2->identification_number],
                            temp_triangle->v3->x,temp_triangle->v3->z,
       Density_2D_Grid.density[dust_i][temp_triangle->v3->identification_number],
                            x,z);
              density = exp(density);
              return density;
	      break;
	    }
	}
        else {
          *err = 0;  
          return 0.0;
        }
      }

    }

    /* this is a flared disk with the density:
         Rho = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))
        (NO SHARP STEP-FUNCTION DISK SURFACE!)

       parameters:  
       param2= p  = Dust_Density_Param[dust_counter].param
       param3= m  = Dust_Density_Param[dust_counter].next->param
       param4= h0 = Dust_Density_Param[dust_counter].next->next->param
       param5=Rho0= Dust_Density_Param[dust_counter].next->next->next->param
    */

    param2 = - Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;

    /* density */ 
    if (x<1.0) density = 0.0;
    else {
      param6 = pow(x,2.0*param3);
      density=param5*pow(x,param2)*exp(-param4*z*z/param6);
    }
    if (density<1E-30*param5) density = 0.0;    
    /* if (z>x*0.57737) density = 0.0; 
    if (density>0.0 && density<1.0E-50) density=1.0E-50; */


    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==9) {
    /* A flat disk:

         Rho = Rho0 * x^(-p) for Xmin<x<Xmax and z<Zmax

       parameters:  
         param2 = p  = Dust_Density_Param[dust_counter].param
         param3 = Xmin = Dust_Density_Param[dust_counter].next->param
         param4 = Xmax = Dust_Density_Param[dust_counter].next->next->param
         param5 = Zmax = Dust_Density_Param[dust_counter].next->next->next->param
         param6 = Rho0 = Dust_Density_Param[dust_counter].next->next->next->next->param

    */
   
    /* param1 = sqrt(x*x+z*z); */
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;
    param6 = Dust_Density_Param[dust_counter].next->next->next->next->param;

    if ( x<param3 ) density = 0.0;
    else if ( x>param4 ) density = 0.0;
    /* else if ( z>(x+param5)*param5/(1.0+param5) ) density = 0.0; */
    else if ( z> param5) density = 0.0;
    else {
      density = param6*pow(x,-param2);
    }

    *err = 0;
    return density;
  }

  /***********************************************************************/
  /***********************************************************************/

  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==10) {
    /* Welcome to the flared disk + spherical halo model */
    /* this is a combination of flared disk and spherical halo with the density:

     Halo = halo dust density profile = Rho1 * r^(-s)
     Disk = disk dust density profile = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))

     if (Halo>=Disk) Rho = Halo
     else Rho = Disk

       parameters:  
       param2= -p = Dust_Density_Param[dust_counter].param
       param3= m  = Dust_Density_Param[dust_counter].next->param
       param4= h0 = Dust_Density_Param[dust_counter].next->next->param
       param5=Rho0= Dust_Density_Param[dust_counter].next->next->next->param
       param6= -s = Dust_Density_Param[dust_counter].next->next->next->next->param
       param7=Rho1= Dust_Density_Param[dust_counter].next->next->next->next->next->param
    */

    param2 = - Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;
    param6 = - Dust_Density_Param[dust_counter].next->next->next->next->param;
    param7 = Dust_Density_Param[dust_counter].next->next->next->next->next->param;

    /* disk density */ 
    density = x*x + z*z;
    if (density<1.0 || x<0.5) density_disk = 0.0;
    else {
      density_disk = pow(x,2.0*param3);
      density_disk=param5*pow(x,param2)*exp(-param4*z*z/density_disk);
    }

    /* halo density */
    if (density<Source_Radius) density = 0.0;
    else density = param7 * pow(density,0.5*param6);

    /* result */
    *err = 0;
    if (density_disk > density) return density_disk;
    else return density;

  }

  /***********************************************************************/
  /***********************************************************************/

  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==11) {
    /* Welcome to the flared disk + flared halo model */
    /* this is a combination of flared disk and flared halo with the density:

     Halo = halo dust density profile = Rho1 * x^(-s)*exp(-h1*z^2/x^(2k))
     Disk = disk dust density profile = Rho0 * x^(-p)*exp(-h0*z^2/x^(2m))

     if (Halo>=Disk) Rho = Halo
     else Rho = Disk

       parameters:  
       param2= -p = Dust_Density_Param[dust_counter].param
       param3= m  = Dust_Density_Param[dust_counter].next->param
       param4= h0 = Dust_Density_Param[dust_counter].next->next->param
       param5=Rho0= Dust_Density_Param[dust_counter].next->next->next->param
       param6= -s = Dust_Density_Param[dust_counter].next->next->next->next->param
       param7= k  = Dust_Density_Param[dust_counter].next->next->next->next->next->param
       param8= h1 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param
       param9=Rho1= Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param
    */

    param2 = - Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;
    param4 = Dust_Density_Param[dust_counter].next->next->param;
    param5 = Dust_Density_Param[dust_counter].next->next->next->param;
    param6 = - Dust_Density_Param[dust_counter].next->next->next->next->param;
    param7 = Dust_Density_Param[dust_counter].next->next->next->next->next->param;
    param8 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param;
    param9 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param;

    /* disk density */ 
    density = x*x + z*z;
    if (density<1.0 || x<0.5) density_disk = 0.0;
    else {
      density_disk = pow(x,2.0*param3);
      density_disk=param5*pow(x,param2)*exp(-param4*z*z/density_disk);
    }

    /* halo density */
    if (density<1.0 || x<0.01) density = 0.0;
    else {
      density = pow(x,2.0*param7);
      density = param9*pow(x,param6)*exp(-param8*z*z/density);
    }

    /* result */
    *err = 0;
    if (density_disk > density) return density_disk;
    else return density;

  }

  /***********************************************************************/
  /***********************************************************************/


  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==102) {
    /* Welcome to the CIT3 (Bonn2001) model: CIT3_Jan17_01 */
    
    param1 = sqrt(x*x+z*z);
    param2 = Dust_Density_Param[dust_counter].param;
    param3 = Dust_Density_Param[dust_counter].next->param;

    /* flat sides of the cone (arccos(param2) opening) */
    if (z/param1 > param2) 
      density=param3/sqrt(param1);
    else density=1.0/(param1*param1);
    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/
  if (INPUT.DUST_DENSITY_TYPE[dust_counter]==103) {
    /* A flat disk with a step in density along the x-axis:

         Rho = Rho1 * x^(-p1) for Xmin<x<Xmid and z<Zmax
         Rho = Rho2 * x^(-p2) for Xmid<x<Xmax and z<Zmax

       parameters:
         p1  = Dust_Density_Param[dust_counter].param
         p2  = Dust_Density_Param[dust_counter].next->param
         Xmin = Dust_Density_Param[dust_counter].next->next->param
         Xmid = Dust_Density_Param[dust_counter].next->next->next->param
         Xmax = Dust_Density_Param[dust_counter].next->next->next->next->param
         Zmax = Dust_Density_Param[dust_counter].next->next->next->next->next->param
         Rho1 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param
         Rho2 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param

    */

    /* param1 = sqrt(x*x+z*z); */
    param2 = Dust_Density_Param[dust_counter].param;  /* p1 */
    param3 = Dust_Density_Param[dust_counter].next->param; /* p2 */

    param4 = Dust_Density_Param[dust_counter].next->next->param; /* Xmin */
    param5 = Dust_Density_Param[dust_counter].next->next->next->param; /* Xmid */
    param6 = Dust_Density_Param[dust_counter].next->next->next->next->param; /* Xmax */
    param7 = Dust_Density_Param[dust_counter].next->next->next->next->next->param; /* Xmax */

    if ( x<param4 ) density = 0.0;
    else if ( x>param6 ) density = 0.0;
    else if ( z> param7) density = 0.0;
    else {
      if ( x<param5) {
        param2 = Dust_Density_Param[dust_counter].param;  /* p1 */
        param3 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->param; /* Rho1 */
        density = param3*pow(x,-param2);
      }
      else {
        param2 = Dust_Density_Param[dust_counter].next->param;  /* p2 */
        param3 = Dust_Density_Param[dust_counter].next->next->next->next->next->next->next->param; /* Rho2 */
        density = param3*pow(x,-param2);
      }
    }

    *err = 0;
    return density;
  }

 /***********************************************************************/
 /***********************************************************************/

  *err = 3;
  return 0.0;

} /* end of the function Dust_Density() */



 /***********************************************************************/
 /***********************************************************************/
 /************** checks if a point is in the given triangle *************/
 /***********************************************************************/
int _ZELENJAK_Dust_Density_point_in_triangle (double x, double z, 
                           struct _triangle_structure *triangle){
  double nu1,nu2,nu3;

  /* check if the point (x,z) is inside of the bounding box */
  if (x >= 0.9999*triangle->bound_box_down_left_x)
   if (z >= 0.9999*triangle->bound_box_down_left_z)
    if (x <= 1.0001*triangle->bound_box_up_right_x)
     if (z <= 1.0001*triangle->bound_box_up_right_z) {
      /* check if the point (x,z) is inside of the triangle */
      /* This is the case if the point (x,z) is on the left side relative
         to all triangle's sides:

                    (x,z)
                     /
                    /    LEFT
                   /
                vertex(i)-------vertex(i+1)

                        RIGHT
      */

      nu1=_MATH_LIB_nu_value(triangle->v1->x,triangle->v1->z,
                       triangle->v2->x,triangle->v2->z,x,z);
      nu2=_MATH_LIB_nu_value(triangle->v2->x,triangle->v2->z,
                       triangle->v3->x,triangle->v3->z,x,z);
      nu3=_MATH_LIB_nu_value(triangle->v3->x,triangle->v3->z,
                       triangle->v1->x,triangle->v1->z,x,z);
      if (0.0<=nu1 && 0.0<=nu2 && 0.0<=nu3) {
        /* bingo! That's the triangle! */
        return 1;
      }
    }

  return 0;
}
