/* LELUYA - ZELENJAK program block */
/* This function calculates the denominator of eta-function (see equations)
   over the computational grid for the wavelength Lambda_Total_Value */

#include "ZELENJAK.h"

int _ZELENJAK_Eta_Denominator(void){

  /* calculation requires the integral over radial ray in direction of 
  Theta_Total_Value. This ray starts at 'startpoint' and ends at 'endpoint' */ 
  struct _3DVector startpoint, endpoint;
  struct _3DVector direction; /* direction of integration */
  struct _3DVector current_point; /* a point during integration */
  struct _3DVector next_point; /* point after current_point */
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _triangle_structure *next_triangle; /* location of next_point */
  double distance_total; /* distance between startpoint and endpoint */
  double current_distance; /* distance between startpoint and current_point */
  double distance_step; /* integration spatial step */
  double current_eta_denom; /* eta denominator during integration */
  double eta_denom_step; /* integration step */
  double Denom0, Denom1; /* Eta values at current_point and one point back */
  double *temperature; /* dust temperature at current_point */
  int dust_i; /* dust component counter */
  int err; /* error flag */

  temperature = dvector(0,Max_Dust_Type-1);  
  /* initial values */
  direction.x = sin(Theta_Total_Value);
  direction.y = 0.0;
  direction.z = cos(Theta_Total_Value);

  startpoint.x = 0.1*direction.x; /* we assume that there is no dust
                                     at the distance of 0.1 from the center */
  startpoint.z = 0.1*direction.z;
  startpoint.y = 0.0;

  endpoint.x = Rout*direction.x;
  endpoint.y = 0.0;
  endpoint.z = Rout*direction.z;

  distance_total=_MATH_LIB_3Ddistance(startpoint,endpoint);
  current_point = startpoint;

  start_vertex_list = grid_per_lambda[Lambda_Total].start_vertex;
  TOTAL_NUMBER_OF_VERTICES = 
                       grid_per_lambda[Lambda_Total].TOTAL_NUMBER_OF_VERTICES;
  start_triangle_list = grid_per_lambda[Lambda_Total].start_triangle;
  TOTAL_NUMBER_OF_TRIANGLES = 
                       grid_per_lambda[Lambda_Total].TOTAL_NUMBER_OF_TRIANGLES;

  /* find initial values of the integral variables */
  err = _ZELENJAK_locate_point_in_grid(current_point.x,current_point.z,
				      &current_triangle);
  current_eta_denom = 0.0;
  current_distance = 0.0;

  if (current_triangle==NULL) 
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
  else for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    if (current_triangle->v1->T_dust[dust_i]>0.0 && 
        current_triangle->v2->T_dust[dust_i]>0.0 && 
        current_triangle->v3->T_dust[dust_i]>0.0){
      temperature[dust_i]= _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
                           current_triangle->v1->T_dust[dust_i],
			   current_triangle->v2->x,current_triangle->v2->z,
                           current_triangle->v2->T_dust[dust_i],
			   current_triangle->v3->x,current_triangle->v3->z,
                           current_triangle->v3->T_dust[dust_i],
			   current_point.x,current_point.z);
      if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
    }
    else temperature[dust_i]=-1.0;
  };

  Denom0 = 0.0;
  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
    if (temperature[dust_i]>0.0) 
       Denom0 += (  (Sigma_Dust_Abs[dust_i][Lambda_Total] + 
                     Sigma_Dust_Sca[dust_i][Lambda_Total]  )  *
                 Dust_Density(current_point.x,current_point.z,dust_i,NULL,&err) );
  };

  /* now we can start with the integral */
  while (current_distance<distance_total){

    distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,direction,
				     1.0,&next_point,&next_triangle);

    /* check if we reached the endpoint */
    if (current_distance + distance_step > distance_total) {
      distance_step = distance_total - current_distance;
      current_distance = distance_total;
      current_point = endpoint;
      err=_ZELENJAK_locate_point_in_grid(current_point.x,current_point.z,
					&current_triangle);
    }
    else { /* ...or we have not reached the endpoint */ 
      current_distance += distance_step;
      current_point = next_point;
      current_triangle = next_triangle;
    }

    if (current_triangle==NULL) 
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
    else for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (current_triangle->v1->T_dust[dust_i]>0.0 && 
          current_triangle->v2->T_dust[dust_i]>0.0 && 
          current_triangle->v3->T_dust[dust_i]>0.0){
        temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point.x,current_point.z);
        if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
      }
      else temperature[dust_i]=-1.0;
    }

    Denom1 = 0.0;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) {
      if (temperature[dust_i]>0.0) {
       Denom1 += (  (Sigma_Dust_Abs[dust_i][Lambda_Total] + 
                     Sigma_Dust_Sca[dust_i][Lambda_Total]  )  *
                 Dust_Density(current_point.x,current_point.z,dust_i,NULL,&err) );
      }
    };

    eta_denom_step = (Denom0+Denom1)*distance_step*0.5;
    current_eta_denom += eta_denom_step;

    /* prepare Eta for the next step */
    Denom0 = Denom1;
  }


  free_dvector(temperature,0,Max_Dust_Type-1);

  printf("SEARCH FOR A NEW ETA-DENOMINATOR:\n");
  printf("Old Eta_Denominator=%E, new=%E\n",Eta_Denominator,current_eta_denom);

  Eta_Denominator = current_eta_denom;

  return 0;

} /* end of _ZELENJAK_Eta_Denominator */
