/* LELUYA - ZELENJAK program block */
/* Calculation of the source flux correction due to finite source size,
   including the direction vector. Notice that it differs from 
   Finite_Source_Flux_Correction only in the scalar product between 
   the solid angle and the given vectors Flux_Direction1 and Flux_Direction2 */

/* IMPORTNAT:  Flux_Direction1 and  Flux_Direction2 have to be normalized!! */ 

/* IMPORTANT: this correction function is multiplied by exp(-Tau_center). 
   That is, it is not like in the notes where it is scaled to return 1.0
   in case of a point source. This will return exp(-Tau_center) instead!! 
   KEEP THIS IN MIND WHEN YOU USE THIS SUBROUTINE! */

#include "ZELENJAK.h"


double Finite_Source_Flux_Correction_with_direction (double x, double z,
                                 double Tau_center, int lambda_k,
                                 struct _3DVector Flux_Direction1,
                                 struct _3DVector Flux_Direction2,
                                 struct _vertex_structure *Start_Vertex,
                                 struct _triangle_structure *Start_Triangle,
                                 double *correction2){
  /* 
  (x,z)      = point at which we need this correction
  Tau_center = optical depth toward (0,0)
  lambda_k   = wavelength
  Start_Vertex,Start_Triangle = grid
  */

  struct _3DVector StartPoint; /* 3D vector of the (x,z) point, which is 
			the beginning for the optical depth calculations */
  struct _3DVector EndPoint; /* 3D vector of a point on the stellar surface, 
                     which is the end for the optical depth calculations */
  struct _3DVector ray_direction; /* |StartPoint-EndPoint| */
  double norm_ray_direction; /* ||StartPoint-EndPoint|| */
  double product1, product2; /* ray_direction*Flux_Direction1 and 
                                ray_direction*Flux_Direction2 */ 

  double distance;     /* = sqrt(x^2+z^2) */
  double cos_distance; /* = x/distance */
  double sin_distance; /* = z/distance */
  double apparent_radius;/* = Source_Radius/distance */
  double cos_star;     /* = sqrt(1-apparent_radius^2) */
  double sin_star;     /* = apparent_radius */
  
  int counter_theta, counter_phi; /* counters in the loop over cos(theta)
                                     and phi angle */
  double cos_theta[_FINITE_SOURCE_N_theta+1]; /* grid values of cos(theta) */
  double sin_theta[_FINITE_SOURCE_N_theta+1]; /* grid values of sin(theta) */
  double cos_phi[ _FINITE_SOURCE_N_phi+1]; /* grid values of cos(phi) */
  double sin_phi[ _FINITE_SOURCE_N_phi+1]; /* grid values of sin(phi) */

  struct _3DVector Phi0_Vector; /* surface vector for phi=0 */
  double cos_beta,sin_beta; /* intermediate anlge used for  Phi0_Vector */

  double Rotation_Matrix[3][3]; /* rotation around the (x,0,z) direction */

  double Tau; /* optical depth for a ray (n_theta,n_phi) */
  double correction; /* the correction: for direction Flux_Direction1 */

  distance = sqrt(x*x+z*z);
  apparent_radius = Source_Radius/distance;
  if (apparent_radius < INPUT.POINT_SOURCE_LIMIT) {
    /*if (apparent_radius>=0.0) {*/
    correction = exp(-Tau_center);
    (*correction2) = 0.0;
    return correction;
  }

  /* set some constant variables */
  cos_distance = x/distance;
  sin_distance = z/distance;
  cos_star = sqrt(1.0-apparent_radius*apparent_radius);
  sin_star = apparent_radius;
  StartPoint.x = x; StartPoint.y = 0.0; StartPoint.z = z;
  start_vertex_list = Start_Vertex;
  start_triangle_list = Start_Triangle;

  cos_theta[0]=1.0;
  sin_theta[0]=0.0;
  for (counter_theta=1;counter_theta<_FINITE_SOURCE_N_theta;counter_theta++){
    cos_theta[counter_theta] =
                  1.0-(1.0-cos_star)*counter_theta/_FINITE_SOURCE_N_theta;
    sin_theta[counter_theta] =
                  sqrt(1.0-cos_theta[counter_theta]*cos_theta[counter_theta]);
  }
  cos_theta[_FINITE_SOURCE_N_theta]=cos_star;
  sin_theta[_FINITE_SOURCE_N_theta]=sin_star;

  cos_phi[0]=1.0;
  sin_phi[0]=0.0; 
  for (counter_phi=1;counter_phi<_FINITE_SOURCE_N_phi;counter_phi++){
    cos_phi[counter_phi]=cos(counter_phi*PI/_FINITE_SOURCE_N_phi);
    sin_phi[counter_phi]=sqrt(1.0-cos_phi[counter_phi]*cos_phi[counter_phi]);
  }
  cos_phi[_FINITE_SOURCE_N_phi]=-1.0;
  sin_phi[_FINITE_SOURCE_N_phi]=0.0; 

  /* start loops over the stellar surface */
  correction = 0.0;
  (*correction2) = 0.0;
  for (counter_theta=1;counter_theta<=_FINITE_SOURCE_N_theta;counter_theta++){
    if (counter_theta!=_FINITE_SOURCE_N_theta) {
      cos_beta = sin_theta[counter_theta]/apparent_radius;
      cos_beta = cos_theta[counter_theta]*sqrt(1.0-cos_beta*cos_beta);
      cos_beta += 
          (sin_theta[counter_theta]*sin_theta[counter_theta]/apparent_radius);
      sin_beta = sqrt(1.0-cos_beta*cos_beta);
    }
    else {
      cos_beta = sin_star;
      sin_beta = cos_star;
    }
    /* surface vector for Phi=0 */ 
    Phi0_Vector.x = apparent_radius*(x*cos_beta - z*sin_beta);
    Phi0_Vector.y=0.0;
    Phi0_Vector.z = apparent_radius*(z*cos_beta + x*sin_beta);
    /* cosine of the ray direction and Flux_Direction */
    ray_direction.x = StartPoint.x - Phi0_Vector.x;
    ray_direction.y = StartPoint.y - Phi0_Vector.y;
    ray_direction.z = StartPoint.z - Phi0_Vector.z;
    norm_ray_direction = sqrt(ray_direction.x*ray_direction.x + 
           ray_direction.y*ray_direction.y + ray_direction.z*ray_direction.z);
    ray_direction.x /= norm_ray_direction;
    ray_direction.y /= norm_ray_direction;
    ray_direction.z /= norm_ray_direction;
    product1 = ray_direction.x*Flux_Direction1.x + 
               ray_direction.y*Flux_Direction1.y +
               ray_direction.z*Flux_Direction1.z;
    product2 = ray_direction.x*Flux_Direction2.x + 
               ray_direction.y*Flux_Direction2.y +
               ray_direction.z*Flux_Direction2.z;
    /* correction */ 
    Tau = _ZELENJAK_tau_over_grid(StartPoint,Phi0_Vector,lambda_k,TRUE);
    if (INPUT.TAU_LIMIT > 0.0) /* check for the -E flag on the tau limit */
      if (Tau>=INPUT.TAU_LIMIT) if (Tau<Tau_center) Tau=Tau_center;

    if (counter_theta<_FINITE_SOURCE_N_theta) {
      correction += (0.5*exp(-Tau)*product1);
      (*correction2) += (0.5*exp(-Tau)*product2);
    }
    else {
      correction += (0.5*0.5*exp(-Tau)*product1);
      (*correction2) += (0.5*0.5*exp(-Tau)*product2);
    }
    for (counter_phi=1;counter_phi<=_FINITE_SOURCE_N_phi;counter_phi++){
      /* rotation matrix */
      Rotation_Matrix[0][0] = cos_distance*cos_distance +
                              cos_phi[counter_phi]*sin_distance*sin_distance;
      Rotation_Matrix[0][1] = -sin_phi[counter_phi]*sin_distance;
      Rotation_Matrix[0][2] = cos_distance*sin_distance*
                                                 (1.0-cos_phi[counter_phi]);
      Rotation_Matrix[1][0] = -Rotation_Matrix[0][1];
      Rotation_Matrix[1][1] = cos_phi[counter_phi];
      Rotation_Matrix[1][2] = -sin_phi[counter_phi]*cos_distance;

      Rotation_Matrix[2][0] = Rotation_Matrix[0][2];
      Rotation_Matrix[2][1] = - Rotation_Matrix[1][2];
      Rotation_Matrix[2][2] = cos_distance*cos_distance*cos_phi[counter_phi]
                              +sin_distance*sin_distance;
      /* find the surface vector */
      EndPoint.x = Rotation_Matrix[0][0]*Phi0_Vector.x +
                   Rotation_Matrix[0][2]*Phi0_Vector.z;
      EndPoint.y = Rotation_Matrix[1][0]*Phi0_Vector.x +
                   Rotation_Matrix[1][2]*Phi0_Vector.z;
      EndPoint.z = Rotation_Matrix[2][0]*Phi0_Vector.x +
                   Rotation_Matrix[2][2]*Phi0_Vector.z;
      /* find the optical depth */
      Tau = _ZELENJAK_tau_over_grid(StartPoint,EndPoint,lambda_k,TRUE);
      if (INPUT.TAU_LIMIT > 0.0) /* check for the -E flag on the tau limit */
        if (Tau>=INPUT.TAU_LIMIT) if (Tau<Tau_center) Tau=Tau_center;

      /* cosine of the ray direction and Flux_Direction */
      ray_direction.x = StartPoint.x - EndPoint.x;
      ray_direction.y = StartPoint.y - EndPoint.y;
      ray_direction.z = StartPoint.z - EndPoint.z;
      norm_ray_direction = sqrt(ray_direction.x*ray_direction.x + 
           ray_direction.y*ray_direction.y + ray_direction.z*ray_direction.z);
      ray_direction.x /= norm_ray_direction;
      ray_direction.y /= norm_ray_direction;
      ray_direction.z /= norm_ray_direction;
      product1 = ray_direction.x*Flux_Direction1.x + 
                 ray_direction.y*Flux_Direction1.y +
                 ray_direction.z*Flux_Direction1.z;
      product2 = ray_direction.x*Flux_Direction2.x + 
                 ray_direction.y*Flux_Direction2.y +
                 ray_direction.z*Flux_Direction2.z;
      /* add to the total result */ 
      Tau = exp(-Tau); 
      if (counter_theta==_FINITE_SOURCE_N_theta) Tau *= 0.5;
      if (counter_phi==_FINITE_SOURCE_N_phi) Tau*= 0.5;
      correction += (Tau*product1); 
      (*correction2) += (Tau*product2); 
    }
  }

  /* cosine of the ray direction and Flux_Direction */
  ray_direction.x = StartPoint.x;
  ray_direction.y = StartPoint.y;
  ray_direction.z = StartPoint.z;
  norm_ray_direction = sqrt(ray_direction.x*ray_direction.x + 
        ray_direction.y*ray_direction.y + ray_direction.z*ray_direction.z);
  ray_direction.x /= norm_ray_direction;
  ray_direction.y /= norm_ray_direction;
  ray_direction.z /= norm_ray_direction;
  product1 = ray_direction.x*Flux_Direction1.x + 
             ray_direction.y*Flux_Direction1.y +
             ray_direction.z*Flux_Direction1.z;
  product2 = ray_direction.x*Flux_Direction2.x + 
             ray_direction.y*Flux_Direction2.y +
             ray_direction.z*Flux_Direction2.z;
  /* send the final result */ 
  correction *= (2.0/(_FINITE_SOURCE_N_phi));
  correction += (exp(-Tau_center)*product1);
  correction *= ((1.0-cos_star)/(_FINITE_SOURCE_N_theta)/
                                            apparent_radius/apparent_radius);

  (*correction2) *= (2.0/(_FINITE_SOURCE_N_phi));
  (*correction2) += (exp(-Tau_center)*product2);
  (*correction2) *= ((1.0-cos_star)/(_FINITE_SOURCE_N_theta)/
                                            apparent_radius/apparent_radius);

  return correction; 

}
