/* LELUYA - ZELENJAK program block */
/* This function calculates optical depth over the computational 
   grid in 3D space */
/* IMPORTANT: this function DOES NOT RECOGNIZE THE SOURCE SURFACE! */

#include "ZELENJAK.h"

double _ZELENJAK_tau_over_grid (struct _3DVector startpoint, 
			       struct _3DVector endpoint, int lambda_k,
                               int tau_limit_flag){
  /* tau_limit_flag=TRUE/FALSE : if TRUE then use the optical depth limit
     from the -E input flag (if the flag is activated) */ 

  /* calculate the optical depth between startpoint and endpoint over the
     computational grid */
  struct _3DVector direction; /* direction of integration */
  struct _3DVector current_point; /* a point in 3D space during integration */
  struct _3DVector old_current_point; /* current_point from previous step */
  struct _3DVector next_point; /* point after current_point */
  struct _3DVector current_point_xz; /* cylindrical coordinates */
  struct _triangle_structure *current_triangle; /* location of current_point */
  struct _triangle_structure *old_current_triangle; /* from previous step */
  struct _triangle_structure *next_triangle; /* location of next_point */
  double distance_total; /* distance between startpoint and endpoint */
  double current_distance; /* distance between startpoint and current_point */
  double distance_step; /* integration spatial step */
  double distance_step_refinement; /* refinement of the integral step */
  int step_function; /* flag which shows if we are crossing a step in Eta */ 
  double tau; /* optical depth distance between startpoint and current_point */
  double tau_step; /* integration tau step */
  double Eta0, Eta1; /* Eta values at current_point and one point back */
  double *temperature; /* dust temperature at old current_point */
  double *new_temperature; /* dust temperature at current_point */
  int dust_i; /* dust component counter */
  int err; /* error flag */
  double A,B,C,distance_old,Eta_old; /* quadratic optical depth interpolation*/
  int quadratic; /* flag which shows if the quadratic interpolation is used */

  double dummy; 

  temperature = dvector(0,Max_Dust_Type-1);  
  new_temperature = dvector(0,Max_Dust_Type-1);  

  /* initial values */
  distance_total=_MATH_LIB_3Ddistance(startpoint,endpoint);
  if (distance_total == 0.0) return 0.0;

  direction.x = (endpoint.x - startpoint.x)/distance_total;
  direction.y = (endpoint.y - startpoint.y)/distance_total;
  direction.z = (endpoint.z - startpoint.z)/distance_total;
  
  current_point = startpoint;

  current_point_xz.x = _MATH_LIB_distance(0.0,0.0,startpoint.x,startpoint.y);
  current_point_xz.y = 0.0;
  if (startpoint.z >= 0.0) current_point_xz.z = startpoint.z;
  else current_point_xz.z = - startpoint.z;

  err = _ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
				      &current_triangle);
  tau = 0.0;
  current_distance = 0.0;

  for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) temperature[dust_i]=-1.0;
  if (current_triangle!=NULL)  
   for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
    if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
      if (current_triangle->v1->T_dust[dust_i]>0.0 && 
          current_triangle->v2->T_dust[dust_i]>0.0 && 
          current_triangle->v3->T_dust[dust_i]>0.0){
        temperature[dust_i]= _MATH_LIB_plane_interpolation 
                          (current_triangle->v1->x,current_triangle->v1->z,
                           current_triangle->v1->T_dust[dust_i],
			   current_triangle->v2->x,current_triangle->v2->z,
                           current_triangle->v2->T_dust[dust_i],
			   current_triangle->v3->x,current_triangle->v3->z,
                           current_triangle->v3->T_dust[dust_i],
			   current_point_xz.x,current_point_xz.z);
        if (temperature[dust_i]==0.0) temperature[dust_i]=-1.0;
      }
      else temperature[dust_i]=-1.0;
    }

  Eta0 = Eta(current_point_xz.x,current_point_xz.z,lambda_k,temperature,&err);

  /* quadratic interpolation */
  distance_old = 0.0;
  Eta_old = 0.0;
  C = 0.0;
  quadratic=0;
  Eta1=0.0;

  /* now we can start with the integral */
  while (current_distance<distance_total){
    old_current_point=current_point;
    old_current_triangle=current_triangle; 
    distance_step_refinement = 1.0; 

    do { /* loop for the case of crossing a dust edge (step function) with a too big tau */
    do { /* loop for the case of crossing a dust edge with a too big space step */

    distance_step = _ZELENJAK_3Dstep (current_point,current_triangle,direction,
                          distance_step_refinement,&next_point,&next_triangle);

    /* check if we reached the endpoint */
    if (current_distance + distance_step > distance_total) {
      distance_step = distance_total - current_distance;
      current_distance = distance_total;
      current_point = endpoint;
      current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
					    current_point.y);
      current_point_xz.y = 0.0;
      if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
      else current_point_xz.z = - current_point.z;
      err=_ZELENJAK_locate_point_in_grid(current_point_xz.x,current_point_xz.z,
					&current_triangle);
    }
    else { /* ...or we have not reached the endpoint */ 
      current_distance += distance_step;
      current_point = next_point;
      current_point_xz.x=_MATH_LIB_distance(0.0,0.0,current_point.x,
					    current_point.y);
      current_point_xz.y = 0.0;
      if (current_point.z >= 0.0) current_point_xz.z = current_point.z;
      else current_point_xz.z = - current_point.z;
      current_triangle = next_triangle;
    }


    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
                                               new_temperature[dust_i]=-1.0;
    quadratic = 0;
    if (current_triangle!=NULL){
      for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
	if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
	  quadratic = 1;
          if (current_triangle->v1->T_dust[dust_i]>0.0 && 
              current_triangle->v2->T_dust[dust_i]>0.0 && 
              current_triangle->v3->T_dust[dust_i]>0.0){
            new_temperature[dust_i]= _MATH_LIB_plane_interpolation 
                            (current_triangle->v1->x,current_triangle->v1->z,
                             current_triangle->v1->T_dust[dust_i],
		             current_triangle->v2->x,current_triangle->v2->z,
                             current_triangle->v2->T_dust[dust_i],
			     current_triangle->v3->x,current_triangle->v3->z,
                             current_triangle->v3->T_dust[dust_i],
			     current_point_xz.x,current_point_xz.z);
	    if (new_temperature[dust_i]==0.0) new_temperature[dust_i]=-1.0;
	  }
          else new_temperature[dust_i]=-1.0;
	}
    }


    Eta1=Eta(current_point_xz.x,current_point_xz.z,lambda_k,new_temperature,&err);

    /* check edge crossing */
    step_function=FALSE;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999)
      if (new_temperature[dust_i]*temperature[dust_i]<0.0) {
        quadratic = 0;
        step_function=TRUE;
      }

    if (step_function && distance_step>1E-10) {
	current_distance -= distance_step;
	current_point=old_current_point;
	current_triangle=old_current_triangle;
	distance_step_refinement *= 2.0; 
    }

    } while (step_function && distance_step>1E-10);  /* end of the loop for checking 
                                                        distance step over an edge */

    /* optical depth interpolation*/
    if (quadratic == 0 || C==0.0) {
      /* linear interpolation */ 
      tau_step = (Eta0+Eta1)*distance_step*0.5;
      if (quadratic == 1 && C==0.0) C=1.0; /* this is a special case when we
        use linear interpolation only because it was also used in previous step */
      else C=0.0;
    }
    else {
      /* quadratic interpolation */
      if (Eta0==0.0 && Eta1==0.0) tau_step = 0.0;
      else {
        B = (Eta0 - Eta_old)/(current_distance - distance_step - distance_old);
        C = (Eta1 - Eta_old)/(current_distance - distance_old) - B;
        C /= distance_step;
        B = B - C*(current_distance - distance_step + distance_old);
        A = Eta_old - distance_old*(B + C*distance_old);
        tau_step = distance_step*(A + B*(current_distance - 0.5*distance_step)+
	           C*(current_distance*current_distance - 
                      current_distance*distance_step + 
                      distance_step*distance_step/3.0));
        /* failed quadrature */
	if (tau_step<0.0) tau_step = (Eta0+Eta1)*distance_step*0.5;
      }
    }

    /* this "if" detects a step function with a significant influence 
       on the optical depth */
    if (step_function) {
      if (tau_step>0.01) {
	current_distance -= distance_step;
	current_point=old_current_point;
	current_triangle=old_current_triangle;
	distance_step_refinement *= 2; 
      }
      else {

        /***********************************************/
        /* search for the tau=1 surface and dust edges */
	if (DEBUG.CONTINUE == 403 && TEMPERATURE_ARRAY_TYPE == 4) {
          printf("%13.11E %13.11E  ",current_point_xz.x,current_point_xz.z);
          for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
            if (INPUT.DUST_DENSITY_TYPE[dust_i]!=99999){
               if (new_temperature[dust_i]*temperature[dust_i]<0.0)
		 printf("%d  ",dust_i);
	       else 
		 printf("-1 ");
	    }
	  printf("\n");fflush(stdout);
	}
        /***********************************************/
        step_function=FALSE;
      }
    }

    /***********************************************/
    /* search for the tau=1 surface and dust edges */
    if (DEBUG.CONTINUE == 403 && TEMPERATURE_ARRAY_TYPE == 3) {
      if (tau<1.0 && tau+tau_step>1.0){
	if (tau_step>0.001) {
          current_distance -= distance_step;
	  current_point=old_current_point;
	  current_triangle=old_current_triangle;
	  distance_step_refinement *= 2; 
	  step_function=TRUE;
	}
	else {
	  printf("%13.11E %13.11E %13.11E\n",current_point_xz.x,current_point_xz.z,tau + tau_step);
	  distance_step_refinement = 1;
	}
      }
    }
    /***********************************************/

    } while (step_function); /* end of the loop for checking tau step over an edge */

    tau += tau_step;

    /* check for the -E input flag: limit on the tau integral */
    if (tau_limit_flag) 
      if (INPUT.TAU_LIMIT > 0.0)
        if (tau > INPUT.TAU_LIMIT) break;

    /* prepare Eta for the next step */
    Eta_old = Eta0;
    Eta0 = Eta1;
    distance_old = current_distance - distance_step;
    for (dust_i=0; dust_i<Max_Dust_Type; dust_i++) 
      temperature[dust_i]=new_temperature[dust_i];
  }

  free_dvector(temperature,0,Max_Dust_Type-1);
  free_dvector(new_temperature,0,Max_Dust_Type-1);

  /* debugger option *
  tau_step = _VODENJAK_tau_over_grid (startpoint, endpoint, lambda_k);
  *****************/

  return tau;

} /* end of _ZELENJAK_tau_over_grid */
